package appl.data;

import java.io.File;
import java.io.Serializable;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import appl.util.GeneralUtil;

/**
 * Loads an object on demand with the help of an <code>DataLoader</code>. The
 * unload() method unloads a serializable {@link Object} into a temporary folder
 * on the hard disk. Unloading can be disabled via the
 * {@link #setUnloading(boolean)} method.
 * 
 * @see DataLoader
 * @author Dominik Appl
 */
public class LateLoadingProxy implements DataProxy, Serializable {

	private File unloadFile;

	protected transient final Logger LOG = LogManager.getLogger(this.getClass()
			.getName());

	/** Object loaded? */
	protected boolean loaded = false;

	protected transient DataLoader dataLoader;

	/**
	 * The loader which constructs the object
	 */
	protected transient DataLoader intialDataLoader;

	protected Object baseObject;

	private boolean unloadingEnabled;

	/**
	 * Creates a new instance with the specified data loader
	 * 
	 * @param loader
	 *            the load method of the {@link DataLoader} is called for
	 *            loading
	 */
	public LateLoadingProxy(DataLoader loader) {
		if (loader == null) {
			throw new UnsupportedOperationException(
					"Error: the dataloader may not be null");
		}
		setUnloadDir("Temp");
		this.intialDataLoader = loader;
		this.dataLoader = loader;
	}

	private LateLoadingProxy() {
	};

	/**
	 * returns the File, in which the object is/will be unloaded
	 */
	protected File getUnloadFile() {
		return unloadFile;
	}

	/**
	 * Loads the data into memory (if not already loaded).<br>
	 * This method is thread-safe
	 * 
	 * @throws LoadingException
	 */
	public synchronized void loadData() throws LoadingException {
		if (loaded)
			return;
		try {
			if (dataLoader == null)
				throw new UnsupportedOperationException(
						"A dataloader was found to be null!");
			baseObject = dataLoader.load();
			if (LOG.isDebugEnabled())
				LOG.debug("Successfully loaded " + baseObject + " with "
						+ dataLoader.getLoadInfo());
			loaded = true;

		} catch (LoadingException e) {
			throw e;
		}
	}

	/***************************************************************************
	 * If no Data is loaded or unloading is disabled this method does nothing,
	 * else it saves the Data into a temporary File. <br>
	 * Warning: Should be Thread-Safe with loadData, but was not tested
	 **************************************************************************/
	public synchronized void unloadData() {
		// only do something if the Data is loaded and unloading is enabled:
		if (!loaded)
			return;
		if (!unloadingEnabled)
			return;
		// there is no way to save the Grid to a File, when its not serializible
		if (!(baseObject instanceof java.io.Serializable)) {
			LOG.warn("Could not unload the Data because the class "
					+ baseObject.getClass().getName() + "is not serializable");
			return;
		}

		// unload the Data
		try {
			GeneralUtil.SerializeToFile(baseObject, unloadFile);
			loaded = false;
			baseObject = null;
			dataLoader = new SerializedDataLoader(unloadFile, true);
			LOG.debug("Successfully unloaded Grid into File: "
					+ unloadFile.getPath());
		} catch (Exception e) {
			LOG.error("Error while trying to unload a grid into file '"
					+ unloadFile.getAbsolutePath() + "'");
			e.printStackTrace();
		}

	}

	/**
	 * This class supports late loading, of course.
	 * 
	 * @see appl.data.LateLoadable#unloadData()
	 * @return true;
	 * @see appl.data.LateLoadable#isLateLoadable()
	 */
	public boolean isLateLoadable() {
		return true;
	}

	/**
	 * returns the baseObject
	 * 
	 * @see appl.data.DataProxy#getProxiedObject()
	 */
	public Object getProxiedObject() {
		try {
			loadData();
		} catch (LoadingException e) {
			System.err.println("Error while loading Griddata ... ");
			e.printStackTrace();
		}
		return baseObject;
	}

	/** returns true, if the baseObject is loaded */
	public boolean isLoaded() {
		return loaded;
	}

	public void setUnloadDir(String newUnloadDirectory) {
		unloadFile = new File(newUnloadDirectory + File.separatorChar
				+ this.toString() + ".tmp");
		// the unload file will be deleted after the JVM shuts down (only after
		// "normal" shutdown)
		unloadFile.deleteOnExit();
	}

	/**
	 * Set to false, if you want to disable the unloading in a temporary file
	 */
	public void setUnloading(boolean unloadingEnabled) {
		this.unloadingEnabled = unloadingEnabled;
	}

	/**
	 * @return true, if unloading is enabled
	 */
	public boolean isUnloadingEnabled() {
		return unloadingEnabled;
	}
}
