package appl.parallel;

import java.io.Serializable;
import java.lang.management.ManagementFactory;
import java.lang.management.MemoryPoolMXBean;
import java.lang.management.MemoryUsage;
import java.lang.management.ThreadMXBean;
import java.net.UnknownHostException;
import java.util.List;
import java.util.Properties;

/**
 * Sets the following properties: <br>
 * <br>
 * Maximum Java Memory <br>
 * Free Java Memory<br>
 * System Information<br>
 * System Architecture<br>
 * System Version<br>
 * Hostname<br>
 * IP<br>
 * 
 * @see ComputingResourceProperties
 * @author Dominik Appl
 */
public class SimpleResourceProperties implements ComputingResourceProperties,
		Serializable {

	protected Properties properties = new Properties();

	public SimpleResourceProperties() {
		java.net.InetAddress i;
		try {
			i = java.net.InetAddress.getLocalHost();
			properties.setProperty("Hostname", i.getHostName());

			// also the DefaultName of the machine, if not specified
			properties.setProperty("Name", i.getHostName());
			properties.setProperty("IP", i.getHostAddress());
		} catch (UnknownHostException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		properties.setProperty("Memory Maximum", ""
				+ getFormatedMemoryString(Runtime.getRuntime().maxMemory()));
		properties.setProperty("Memory Free", ""
				+ getFormatedMemoryString(Runtime.getRuntime().freeMemory()));
		properties.setProperty("Memory Used", ""
				+ getFormatedMemoryString(getUsedMemory()));
		properties.setProperty("Operating System", System
				.getProperty("os.name"));
		properties.setProperty("System Architecture", System
				.getProperty("os.arch"));
		properties.setProperty("System Version", System
				.getProperty("os.version"));
		properties.setProperty("Available Processors", Runtime.getRuntime()
				.availableProcessors()
				+ "");

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.ComputingResourceProperties#getProperties()
	 */
	public Properties getProperties() {
		return properties;
	}

	protected long getUsedMemory() {
		ThreadMXBean tb = ManagementFactory.getThreadMXBean();
		List<MemoryPoolMXBean> pools = ManagementFactory.getMemoryPoolMXBeans();
		long sum = 0;
		for (MemoryPoolMXBean pool : pools) {
			MemoryUsage peak = pool.getPeakUsage();
			sum += peak.getUsed();
		}
		return sum;
	}

	protected String getFormatedMemoryString(long bytes) {
		long mb = bytes / (1024 * 1024);
		long kb = bytes / 1024 - (mb * 1024);
		// only the first char of kb
		kb = kb / 100;
		return mb + "." + kb + " MB";

	}

	/*
	 * (non-Javadoc) only indirection
	 * 
	 * @see appl.parallel.ComputingResourceProperties#getProperty(java.lang.String)
	 */
	public String getProperty(String key) {
		return properties.getProperty(key);
	}

	/*
	 * (non-Javadoc) only indirection
	 * 
	 * @see appl.parallel.ComputingResourceProperties#setProperty(java.lang.String,
	 *      java.lang.String)
	 */
	public void setProperty(String key, String value) {
		properties.setProperty(key, value);
	}

	public String toString() {
		return getProperty("Name");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.ComputingResourceProperties#getName()
	 */
	public String getName() {
		return getProperty("Name");
	}

	public boolean isAvailable() {
		return getProperty("isAvailable").equals("true");
	}

	public String getIP() {
		return getProperty("IP");
	}

	public String getPort() {
		return getProperty("Port");
	}
}
