package appl.parallel.client;

import java.awt.Rectangle;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.rmi.AccessException;
import java.rmi.NotBoundException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.ServerNotActiveException;
import java.rmi.server.UnicastRemoteObject;
import java.util.HashMap;
import java.util.Hashtable;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import appl.ext.XuluConfig;
import appl.parallel.event.CommEvent;
import appl.parallel.event.CommEventSink;
import appl.parallel.event.TransferEvent;
import appl.parallel.event.CommEvent.CommType;
import appl.parallel.spmd.split.DataPartition;
import appl.parallel.util.Helper;
// fuer Doku
import javax.activation.DataHandler;
import appl.parallel.server.PartitionDataServer;
import appl.parallel.server.XuluServer;
import appl.parallel.spmd.split.SplittableResource;

/**
 * This class is used for storing all data which is needed for parallelization
 * and may be requested remotely. On {@link DataPartition DataPartitions} the
 * request of partitions is supported. {@link PartitionDataServer}s running on
 * {@link XuluServer}s will typically use {@link DataHandler DataHandlers} for
 * retrieving the partitions.
 *
 * @see PartitionDataServer
 * @author Dominik Appl
 */
public class ClientDataServer extends UnicastRemoteObject implements DataServer {

	protected final Logger LOG = LogManager
			.getLogger(this.getClass().getName());

	private final String bindingName = "DataServer";

	// for retrieval with the baseID
	private Hashtable<Integer, DataPartition> dataByID;

	HashMap<String, String> hostnames = new HashMap<String, String>(10);

	private int registryPort = 1099;

	private final CommEventSink eventSink;

	/**
	 * The standard Constructor (binds itself to the running registry)
	 *
	 * @param eventSink
	 *            the sink is used for generating events
	 * @throws RemoteException
	 *             if the connection fails
	 */
	public ClientDataServer(CommEventSink eventSink) throws RemoteException {
		this.eventSink = eventSink;
		dataByID = new Hashtable<Integer, DataPartition>();

		// bind to the registry - create a new reg, if no reg is found
		try {
			// try to get port from registry
			int port = XuluConfig.getXuluConfig().getIntProperty(
					"XuluClient.registryport");
			if (port != 0)
				registryPort = port;
			Helper.bind(bindingName, this, registryPort);
		} catch (Exception e) {
			LOG.error("Error while binding SPMDClient to the registry."
					+ e.getMessage());
		}
		LOG.info("DataServer up and running....");
	}

	/**
	 * Returns the data identified by the given id
	 *
	 * @param baseID
	 *            Id identifiying the data
	 * @see SplittableResource#getRootID()
	 */
	public DataPartition getData(int baseID) throws RemoteException {
		DataPartition result = dataByID.get(baseID);
		if (LOG.isDebugEnabled())
			LOG.debug("Data retrieved from SPMDClient: ID:" + baseID);

		// generate transfer event. Notice that this is very slow and
		// will therefore only be created if a listener has registered
		if (eventSink.isTransferMonitoringEnabled()) {
			try {
				// try to get Hostname (for performance reasons this is done
				// only one time)
				String hostIP = getClientHost();
				// check if already looked up;
				String hostname = hostnames.get(hostIP);
				if (hostname == null) {
					hostname = InetAddress.getByName(hostIP).getHostName();
					hostnames.put(hostIP, hostname);
				}
				eventSink.fireRemoteEvent(new TransferEvent("DataServer",
						hostname, CommType.TRANSFER_DATA, result));
			} catch (ServerNotActiveException e) {
				// if this is a purely local call no event
				// should be generated.
			} catch (UnknownHostException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		return result;
	}

	/**
	 * DataPartitions can be added
	 *
	 * @see appl.parallel.client.DataServer#addData(appl.parallel.spmd.split.DataPartition)
	 */
	public void addData(DataPartition splittable) {
		this.dataByID.put(splittable.getRootID(), splittable);
	}

	/**
	 * stops the client and removes it from the registry
	 */
	public void close() {
		Helper.unbind(bindingName);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see appl.parallel.server.PartitionDataServer#getPartition(int,
	 *      java.awt.Rectangle)
	 */
	public DataPartition getPartition(int id, Rectangle bounds)
			throws RemoteException {
		DataPartition data = (DataPartition) dataByID.get(id);
		if (data == null && LOG.isDebugEnabled()) {
			LOG.error("Partition with id " + id + " and bounds " + bounds
					+ "was requested but not found on SPMDClient");
			return null;
		}
		DataPartition result = data.getPartition(bounds);
		if (eventSink.isTransferMonitoringEnabled()) {
			try {
				eventSink.fireRemoteEvent(new TransferEvent("DataServer",
						getClientHost(), CommType.TRANSFER_DATA, result));
			} catch (ServerNotActiveException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		if (LOG.isDebugEnabled())
			LOG.debug("Data retrieved from SPMDClient: ID:" + id
					+ " partition: " + bounds);
		return result;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see appl.parallel.server.PartitionDataServer#setPartition(int,
	 *      appl.parallel.spmd.split.DataPartition, java.awt.Rectangle)
	 */
	public void updatePartition(int id, DataPartition partition,
			Rectangle bounds) throws RemoteException {
		DataPartition data = (DataPartition) dataByID.get(id);
		if (data == null && LOG.isDebugEnabled()) {
			LOG.error("Partition of data with id " + id + " and bounds "
					+ bounds + "should be set but was not found on SPMDClient");
			return;
		}
		data.setPartition(partition, bounds);
		if (LOG.isDebugEnabled())
			LOG.debug("Partition set on SPMDClient with id " + id
					+ " and partition bounds: " + bounds);

	}

	/*
	 * (non-Javadoc)
	 *
	 * @see appl.parallel.client.DataServer#removeData(int)
	 */
	public void removeData(int id) {
		dataByID.remove(id);
	}

}
