package appl.parallel.client;

import java.io.IOException;
import java.rmi.RemoteException;
import java.util.Iterator;
import java.util.Vector;

import javax.swing.JOptionPane;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.apache.log4j.lf5.StartLogFactor5;

import com.sun.jini.tool.ClassServer;

import appl.ext.XuluConfig;
import appl.parallel.ComputingResource;
import appl.parallel.ComputingResourceContainer;
import appl.parallel.ComputingResourceProperties;
import appl.parallel.model.ParallelStepModel;
import appl.parallel.server.XuluServer;
import appl.parallel.services.DiscoveryService;
import appl.parallel.services.GlobalDiscoveryService;
import appl.parallel.spmd.AdvancedSPMDClientController;
import appl.parallel.spmd.SPMDClientController;
import appl.parallel.spmd.SPMDClientInterface;
import appl.parallel.spmd.split.SplitMap;
import appl.parallel.util.Helper;
import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.appl.XuluPlugin;
import edu.bonn.xulu.model.AbstractStepModel;
import edu.bonn.xulu.model.XuluModel;
import schmitzm.lang.AbstractNamedObject;

/**
 * This plugin is responsible that all services required for remote
 * execution are properly started. It first starts the
 * {@link GlobalDiscoveryService}. Then it uses the following entries of
 * {@link XuluConfig} for configuration: <br>
 * <br>
 * <code>RemoteExecutionController.spmd.start</code> - if, <code>true</code>
 * the {@link ClientDataServer} is started. <br><br>
 * <code>RemoteExecutionController.http.start</code> - if, <code>true</code>
 * the {@link ClassServer HTTP-Class server} is started. <br><br>
 * <code>RemoteExecutionController.startLocalXuluServer</code> - if,
 * <code>true</code> a local {@link XuluServer} is started. <br>.
 * 
 * @author Dominik Appl
 */
public class RemoteExecutionController extends AbstractNamedObject implements
		XuluPlugin {

	private final Logger LOG = LogManager.getLogger(this.getClass().getName());

	private ClientDataServer spmdClient;

	private boolean started = false;

	private XuluServer localServer;

	private ClassServer httpServer;

	private RemoteEventHandler eventHandlerProxy;

	private GlobalDiscoveryService discoveryService;

	private boolean startHTTP = false;

	private boolean startSPMD = false;

	private boolean startXuluServer = false;

	private boolean httpVerbose = false;

	private int httpPort = 80;

	private String httpBaseDir = ".\\classes"; //default: will be overwritten with the configuration
	
	private Vector<ResourceChangeListener> activeChangeListeners = new Vector<ResourceChangeListener>();

	private Vector<ComputingResourceContainer> lastKnownResources;
	
	public RemoteExecutionController() {
		startDiscovery();
	}

	/**
	 * Starts the Plugin
	 * 
	 * @see edu.bonn.xulu.appl.XuluPlugin#execute(edu.bonn.xulu.XuluModellingPlatform)
	 */
	public void execute(XuluModellingPlatform appl) {
		started = true;
		loadConfig();
		startEventSystem();
		startSPMDClient();
		startXuluServer();
		startDiscovery();
		startHTTPServer();
		//discover own server
		refreshResources();
	}

	private void startEventSystem() {
		// start EventProxy
		try {
			eventHandlerProxy = new RemoteEventHandler();
			eventHandlerProxy.startService();
		} catch (RemoteException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
	}

	private void startDiscovery() {
		if (discoveryService != null)
			discoveryService.stopService();
		discoveryService = new GlobalDiscoveryService();
	}

	/**
	 * Load the configuration form the {@link XuluConfig}
	 */
	private void loadConfig() {
		XuluConfig config = XuluConfig.getXuluConfig();
		startSPMD = config
				.getBooleanProperty("RemoteExecutionController.spmd.start");
		startXuluServer = config
				.getBooleanProperty("RemoteExecutionController.startlocalxuluserver");
		startHTTP = config
				.getBooleanProperty("RemoteExecutionController.http.start");
		httpVerbose = config
				.getBooleanProperty("RemoteExecutionController.http.verbose");
		if (config.getIntProperty("RemoteExecutionController.http.port") != 0)
			httpPort = config
					.getIntProperty("RemoteExecutionController.http.port");
		if (config.getProperty("RemoteExecutionController.http.basedir") != null)
			httpBaseDir = config
					.getProperty("RemoteExecutionController.http.basedir");
	}

	private void startHTTPServer() {
		// start the http server
		if (startHTTP) {
			try {
				httpServer = new ClassServer(httpPort, httpBaseDir, true,
						httpVerbose);
				httpServer.start();
			} catch (IOException e) {
				LOG.error("Could not create HTTP-Server!", e);
				e.printStackTrace();
			}
		}
	}

	private void startSPMDClient() {
		try {
			if (startSPMD)
				spmdClient = new ClientDataServer(eventHandlerProxy);
		} catch (RemoteException e) {
			LOG.error("SPMD Client could NOT be started. " + e.getMessage());
			e.printStackTrace();
		}
	}

	private void startXuluServer() {
		try {
			if (startXuluServer)
				localServer = new XuluServer(false, spmdClient);
		} catch (RemoteException e) {
			LOG.error("XuluServer could NOT be started. " + e.getMessage());
			e.printStackTrace();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.bonn.xulu.appl.XuluPlugin#isStarted()
	 */
	public boolean isStarted() {
		return started;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.bonn.xulu.appl.XuluPlugin#isVisible()
	 */
	public boolean isVisible() {
		// TODO Auto-generated method stub
		return false;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.bonn.xulu.appl.XuluPlugin#setVisible(boolean)
	 */
	public void setVisible(boolean visible) {
		// TODO Auto-generated method stub

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.bonn.xulu.appl.XuluPlugin#stop()
	 */
	public void stop() {
		started = false;
		if (httpServer != null)
			httpServer.stop();
		if (spmdClient != null)
			spmdClient.close();
		spmdClient = null;
		if (localServer != null)
			localServer.stopServer();
		spmdClient = null;
		eventHandlerProxy.stopService();
	}

	/**
	 * Gets active Resources and triggers the discovery-process. After
	 * the resources are found an update is called on all registered 
	 * {@link ResourceChangeListener}s.
	 * 	 * 
	 */
	public void refreshResources() {
		Vector<ComputingResourceContainer> remoteResources = discoveryService
				.getRemoteResources();
		
		lastKnownResources = remoteResources;
		//notify listeners about the resource update
		for (ResourceChangeListener listener : activeChangeListeners) {
			listener.updateResources(remoteResources);
		}

		// The advanced Server still makes some problems and therefore is here
		// deactivated. May be fixed later.
		// try {
		// if(localServer!=null){
		// ComputingResourceProperties localInfo =
		// localServer.getResourceInformation();
		// localInfo.setProperty("Name", "Advanced Internal Server");
		// remoteResources.add(new ComputingResourceContainer(localServer,
		// localInfo));
		// }
		//				
		// } catch (RemoteException e) {
		// // TODO Auto-generated catch block
		// e.printStackTrace();
		// }
	}
	
	
	/**
	 * adds the change lister
	 * 
	 * @param r the change listener
	 */
	public void addResourceChangeListener(ResourceChangeListener r){
		activeChangeListeners.add(r);
	}
	
	/**
	 * Removes the change listener
	 * @param r the change listener to remove
	 */
	public void removeResourceChangeListener(ResourceChangeListener r){
		activeChangeListeners.remove(r);
	}

	

	/**
	 * Calculates weight for the given resources for the use in {@link SplitMap}
	 * splitmaps. For this the PerfomanceRating read out of the
	 * {@link XuluConfig} property <br>
	 * "PerformanceRating." + name of the machine <br>
	 * If the rating is 0 an average rating is assumed.
	 * 
	 * @param computingResources
	 * @return the weights
	 */
	private double[] calculateWeights(
			Vector<ComputingResourceContainer> computingResources) {
		// first get the rating. Try to find the value in the
		// (user defined)configuration first, if not there try resource info.
		int ratings[] = new int[computingResources.size()];
		for (int i = 0; i < ratings.length; i++) {
			ComputingResourceContainer container = computingResources.get(i);
			String resName = container.getInformation().getName();
			String rating = XuluConfig.getXuluConfig().getProperty(
					"PerformanceRating." + resName, false);
			// if not found take the benchmark value
			if (rating == null) {
				rating = container.getInformation().getProperty("Rating");
			}
			if (rating == null)
				rating = "0";
			ratings[i] = Integer.valueOf(rating);
		}
		return Helper.calculateWeights(ratings);
	}

	/**
	 * Gives a new {@link SPMDClientController}.
	 * 
	 * @param computingResources
	 *            the resources to be used by the new controller
	 * @return a new {@link SPMDClientController} for the given resources
	 */
	public SPMDClientController getNewSPMDClientController(
			ComputingResourceContainer[] computingResources) {
		Vector<ComputingResourceContainer> resvec = new Vector<ComputingResourceContainer>(
				computingResources.length);
		for (ComputingResourceContainer container : computingResources) {
			resvec.add(container);
		}
		return getNewSPMDClientController(resvec);
	}

	/**
	 * Gives a new {@link SPMDClientController}.
	 * 
	 * @param computingResources
	 *            the resources to be used by the new controller
	 * @return a new {@link SPMDClientController} for the given resources
	 */
	public SPMDClientController getNewSPMDClientController(
			Vector<ComputingResourceContainer> computingResources) {
		double[] weights = calculateWeights(computingResources);
		return new AdvancedSPMDClientController(computingResources, weights,
				spmdClient, eventHandlerProxy);
	}

	/**
	 * @return the execution controller of the given
	 *         {@link XuluModellingPlatform} or null, if not found
	 */
	public static RemoteExecutionController getRemoteExecutionController(
			XuluModellingPlatform modellingPlatform) {
		XuluPlugin[] plugins = modellingPlatform.getRegistry().getPlugins();
		for (XuluPlugin plugin : plugins) {
			if (plugin instanceof RemoteExecutionController)
				return (RemoteExecutionController) plugin;
		}
		return null;
	}

	public RemoteEventHandler getEventProxy() {
		return eventHandlerProxy;
	}

	/**
	 * Prepares a Model for execution, which means it creates a
	 * {@link SPMDClientController} with the given Resources
	 * 
	 * @param model
	 *            the model which is to be executed parallel
	 * @param selectedResourceContainers
	 *            the resources on which the model is executed
	 */
	public void prepareModelForSPMDExecution(ParallelStepModel model,
			ComputingResourceContainer[] selectedResourceContainers) {
		SPMDClientInterface controller = getNewSPMDClientController(selectedResourceContainers);
		model.setSPMDController(controller);
	}
}
