/*
 *    Geotools2 - OpenSource mapping toolkit
 *    http://geotools.org
 *    (C) 2002, Geotools Project Managment Committee (PMC)
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package appl.parallel.data.xulugridfile;
import org.geotools.gce.arcgrid.ArcGridRaster;
import org.geotools.resources.NIOUtilities;

import appl.util.RasterMetaData;

import java.awt.image.DataBuffer;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.RandomAccessFile;
import java.io.Reader;
import java.io.StreamTokenizer;
import java.net.URL;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.channels.FileChannel;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import javax.media.jai.RasterFactory;


/**
 * This class converts a ArcGridInfo_ASCII file to a {@link XuluGridFile}. It directly writes from
 * the the ArcGridInfo to the given file. It does not use the memory!
 *
 * It is a modified geotools import class. Buffering was added.
 *
 * @author Dominik Appl <br>
 * For reading this class is based partly on the Geotools class {@link ArcGridRaster}. This was
 * done to make sure that EVERY ArcInfoASCII can be read. The used {@link ArcGridRaster} was created by <br>
 * @author <a href="mailto:ckl@dacelo.nl">Christiaan ten Klooster</a>
 * @author <a href="mailto:aaime@users.sf.net">Andrea Aime</a>
 * @author <a href="mailto:simboss_ml@tiscali.it">Simone Giannecchini
 *         (simboss)</a>
 * @see XuluGridFile
 * @see http://svn.geotools.org/geotools/tags/2.2.0/plugin/arcgrid/src/org/geotools/gce/arcgrid/ArcGridRaster.java
 */

public class XuluGridFileConverter extends ArcGridRaster {

    protected final File inputArcInfoGrid;
	protected final File outputXuluGridFile;

	/**
	 * @throws FileNotFoundException
	 */
	public XuluGridFileConverter(File inputArcInfoGrid, File outputXuluGridFile) throws FileNotFoundException {
		super(new BufferedReader(new InputStreamReader(new FileInputStream(inputArcInfoGrid))), false);
		this.inputArcInfoGrid = inputArcInfoGrid;
		this.outputXuluGridFile = outputXuluGridFile;
	}


   /**
     * partly copied from Geotools. www.geotools.org.
     *
     * @param type the type of the values. Use the
     * {@link DataBuffer} constants to identify the type
     *
     * @return the new XuluGridFile
     *
     * @throws IOException
     * @throws XuluGridFileException
     */
    public XuluGridFile convertArcInfoToXuluGridFile(int type) throws IOException, XuluGridFileException {



    	// open reader and make tokenizer
        Reader reader = openReader();
        StreamTokenizer st = new StreamTokenizer(reader);

        // parse header
        parseHeader(st);

        //create metadata and outputFile
        RasterMetaData meta = new RasterMetaData(type,getNCols(),getNRows(),0,0,getXlCorner(),getYlCorner(),getCellSize(),null);
        RandomAccessFile outputFile = new RandomAccessFile(outputXuluGridFile,"rw");

        // reconfigure tokenizer
        st.resetSyntax();
        st.parseNumbers();
        st.whitespaceChars(' ', ' ');
        st.whitespaceChars(' ', '\t');
        st.whitespaceChars('\n', '\n');
        st.whitespaceChars('\r', '\r'); //linefeed (on windows only?)
        st.whitespaceChars('\f', '\f'); //form feed (on printers????)
        st.eolIsSignificant(false);
        st.ordinaryChars('E', 'E');



        // Read values from grid and put into raster.
        // Values must be numbers, which may be simple <num>, or expressed
        // in scientific notation <num>E<exp>.
        // The following loop can read both, even if mixed.
        // The loop expects a token to be read already
        st.nextToken();


        for (int y = 0; y < getNRows(); y++) {
        	ByteArrayOutputStream baos;
        	if(type == DataBuffer.TYPE_DOUBLE)
        		baos = new ByteArrayOutputStream(getNCols()*8);
        	else
        		baos = new ByteArrayOutputStream(getNCols()*4);

        	for (int x = 0; x < getNCols(); x++) {
                // this call always reads the next token
                double d = readCell(st, x, y);

                // mask no data values with NaN
                if (d == getNoData()) {
                    d = Double.NaN;
                } else {
                    minValue = Math.min(minValue, d);
                    maxValue = Math.max(maxValue, d);
                }
                switch(type){
                case DataBuffer.TYPE_FLOAT:
                	BufferedHelper.writeInt(Float.floatToIntBits((float)d),baos);
                break;
                case DataBuffer.TYPE_DOUBLE:
                	BufferedHelper.writeLong(Double.doubleToLongBits(d),baos);
                case DataBuffer.TYPE_INT:
                	BufferedHelper.writeInt((int)d,baos);
                break;
                default:
                	throw new UnsupportedOperationException("Unsupported datatype");
            }
                outputFile.write(baos.toByteArray());
                baos.reset();
        }
        }
        reader.close();
        try {
			XuluGridFile.writeWorldFileForMetaData(outputXuluGridFile, meta);
		} catch (XuluGridFileException e) {
			throw new IOException ("Could not write WorldFile");
		}
        return new XuluGridFile(outputXuluGridFile,"rw");
    }

    /**
     *copied unmodified form superclass (had to be copied, because the geotools method was private)
     */
    private double readCell(StreamTokenizer st, int x, int y)
        throws IOException {
        double d = 0;

        // read a token, expected: a number
        switch (st.ttype) {
        case StreamTokenizer.TT_NUMBER:
            d = (float) st.nval;

            break;

        case StreamTokenizer.TT_EOF:
            throw new IOException("Unexpected EOF at " + x + "," + y);

        default:
            throw new IOException("Unknown token " + st.ttype);
        }

        // read another. May be an exponent of this number.
        // If its not an exponent, its the next number. Fall through
        // and token is prefetched for next loop...
        switch (st.nextToken()) {
        case 'e':
        case 'E':

            // now read the exponent
            st.nextToken();

            if (st.ttype != StreamTokenizer.TT_NUMBER) {
                throw new IOException("Expected exponent at " + x + "," + y);
            }

            // calculate
            d = d * Math.pow(10.0, st.nval);

            // prefetch for next loop
            st.nextToken();

            break;

        case StreamTokenizer.TT_NUMBER:
        case StreamTokenizer.TT_EOF:
            break;

        default:
            throw new IOException("Expected Number or EOF");
        }

        return d;
    }

   /** You can use this method to manually convert datatypes
 * @param args
 */
public static void main(String[] args) {
	   int type=DataBuffer.TYPE_FLOAT;
	   if((args.length!=2 && args.length!=3)){
		   parameterMessage();
	   return;
	   }
	   if(!args[1].endsWith(".xgrid")){
		   parameterMessage();
		   return;
	   }
	   if(args.length==3){
			if(args[2].equals("INT"))
				type=DataBuffer.TYPE_INT;
	   		if(args[2].equals("DOUBLE"))
	   			type = DataBuffer.TYPE_DOUBLE;
   		}
		XuluGridFileConverter converter;
		try {
			converter = new XuluGridFileConverter(new File(args[0]),new File(args[1]));
			converter.convertArcInfoToXuluGridFile(type);
		} catch (FileNotFoundException e) {
			System.out.println("Could not find file " + args[0]);
		} catch (IOException e) {
			System.out.println("IOException while converting: " + e.getMessage());
		} catch (XuluGridFileException e) {
			System.out.println("XuluGridFileException while converting: " + e.getMessage());
		}


   }
   public static void parameterMessage(){
   System.out.println("Syntax: XuluGridFileConverter Source Target [type], where \n" +
   		"Source = a raster in the ArcInfo_ASCII format \n" +
   		"Target = the new XuluGridFile with the extension .xgrid \n" +
   		"[type] = (optional) the type of the XuluGridFile. Possible Values: INT,FLOAT,DOUBLE\n" +
   		" default type is FLOAT");
   }
}

