package appl.parallel.data.xulugridfile;

import java.awt.Rectangle;
import java.io.File;
import java.rmi.RemoteException;

import appl.data.LoadingException;
import appl.parallel.client.ClientDataServer;
import appl.parallel.data.AbstractDataHandler;
import appl.parallel.data.PartitionDataHandler;
import appl.parallel.spmd.split.DataPartition;

/**
 * This is a loader for use with shared file systems like ocfs2 or gfs. It only
 * works in conjunction with the {@link XuluWritableGridFile} and supports
 * loading data from and into this file format. That means reading and writing
 * is done directly into the file. The Xulu GUI won't notice changes if you do not
 * reload the XuluWritableGrid!
 * 
 * @see XuluGridFile
 * @see XuluWritableGridFile
 * @see PartitionDataHandler
 * 
 * @author Dominik Appl
 */
public class XuluGridSharedFileSystemLoader extends AbstractDataHandler{
	
	private String gridFileName;
	private transient XuluWritableGridFile xuluWritableGridFile;
	
	/**
	 * constructs a new {@link XuluGridSharedFileSystemLoader}. 
	 * 
	 * @param rootID
	 *            the id of the data
	 * @param client
	 *            a spmd client which may be used to get metainformation on
	 *            client side before transfering the loader to its destination
	 * @param partitionBounds
	 *            the bounds of the partition to be retrieved on server side
	 * @param unloadBounds
	 *            the bounds of the partition which is to be uploaded to the
	 *            client after calculation (may only be the calculation area)
	 */
	public XuluGridSharedFileSystemLoader(int rootID, ClientDataServer client, Rectangle partitionBounds, Rectangle unloadBounds) {
		super(rootID, client, partitionBounds, unloadBounds);
		DataPartition partition;
		try {
			partition = client.getData(rootID);
			if (partition instanceof XuluWritableGridFile) {
				XuluWritableGridFile xwg = (XuluWritableGridFile) partition;
				gridFileName = xwg.getGridFileName();
			}
			else throw new UnsupportedOperationException("This Dataloader can only be used XuluGridFiles");
		} catch (RemoteException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	
	/**
	 * only used by {@link #clone()}
	 */
	private XuluGridSharedFileSystemLoader(int rootID, ClientDataServer spmdClient, Rectangle partitionBounds, Rectangle unloadBounds, String gridFileName) {
		super(rootID, null, partitionBounds, unloadBounds);
		this.gridFileName=gridFileName;
	}

	/**
	 * Used for deserialization. Do not use for other purposes
	 */
	public XuluGridSharedFileSystemLoader() {
		super();
	}

	/* (non-Javadoc)
	 * @see appl.data.AbstractDataHandler#newInstance(int, appl.parallel.client.SPMDClient, java.awt.Rectangle, java.awt.Rectangle)
	 */
	public PartitionDataHandler newInstance(int rootID, ClientDataServer client, Rectangle partitionBounds, Rectangle unloadBounds) {
		// TODO Auto-generated method stub
		return new XuluGridSharedFileSystemLoader(rootID, client, partitionBounds, unloadBounds);
	}
	
	/* (non-Javadoc)
	 * @see appl.data.AbstractDataHandler#getLoadInfo()
	 */
	@Override
	public String getLoadInfo() {
		return "Trying to load " + gridFileName;
	}

	/* (non-Javadoc)
	 * @see appl.data.AbstractDataHandler#getUnloadInfo()
	 */
	@Override
	public String getUnloadInfo() {
		return "Trying to unload " + gridFileName;
	}

	/** loads the data directly from the gridfile 
	 * @see appl.parallel.data.AbstractDataHandler#load()
	 */
	@Override
	public DataPartition load() throws LoadingException {
		//load from the gridfile
		if(xuluWritableGridFile==null)
			xuluWritableGridFile = new XuluWritableGridFile(new File(gridFileName));
		data =  xuluWritableGridFile.getPartition(partitionBounds);
		return data;
	}

	/** unloads into the file! not into the Xulu-Client or anything!
	 */
	@Override
	public void unload() {
		//unload to the local file
		if(xuluWritableGridFile==null)
			xuluWritableGridFile = new XuluWritableGridFile(new File(gridFileName));
//		 construct return grid only if necessary
		DataPartition returnPartition = null;
		if (unloadBounds.equals(data.getPartitionBounds()))
			returnPartition = data;
		else
			returnPartition = data.getPartition(unloadBounds);
		xuluWritableGridFile.setPartition(returnPartition, unloadBounds);
	}
	
	public PartitionDataHandler clone() {
		return new XuluGridSharedFileSystemLoader(rootID, spmdClient, this.partitionBounds,
				this.unloadBounds, this.gridFileName);
	}
}
