package appl.parallel.data.xulugridfile.factories;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;

import appl.data.DataProxy;
import appl.data.WritableGridLLProxy;
import appl.ext.XuluConfig;
import appl.parallel.data.xulugridfile.XuluGridFile;
import appl.parallel.data.xulugridfile.XuluGridFileConverter;
import appl.parallel.data.xulugridfile.XuluWritableGridFile;
import appl.util.RasterMetaData;
import appl.util.RasterUtil;

import schmitzm.data.WritableGrid;
import schmitzm.data.WritableGridArray;

import edu.bonn.xulu.io.ImportFactory;
import edu.bonn.xulu.io.InstantiationFactory;
import edu.bonn.xulu.io.ExportFactory;
import edu.bonn.xulu.io.AbstractFactory;
import edu.bonn.xulu.appl.XuluRegistry;

import edu.bonn.xulu.plugin.data.grid.SingleGrid;
import edu.bonn.xulu.plugin.io.grid.array.WritableGridArrayFactory_ArcInfoAsciiGrid;

// nur fuer Doku

/**
 * Diese Factory importiert und exportiert Instanzen des Datentyps
 * {@link SingleGrid} aus/in das ArcInfo-ASCII-Grid-Format.

 * @author Dominik Appl
 * @version 1.0
 */
public class SingleGridFactory_ArcInfoAsciiGrid extends AbstractFactory implements ImportFactory, ExportFactory {
  private static final ImportFactory importFac = new WritableGridArrayFactory_ArcInfoAsciiGrid();
  private static final ExportFactory exportFac = new WritableGridArrayFactory_ArcInfoAsciiGrid();
  ////////////////////////////////////////////////////////////////////////
  ///////////////     Implementierung von ImportFactory     //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Liefert den Datentyp, den die Factory erzeugt.
   * @return immer {@link SingleGrid SingleGrid.class}
   */
  public Class getImportType() {
    return SingleGrid.class;
  }

  /**
   * Liefert den Datentyp, den die Factory als Import-Quelle benoetigt.
   * @return immer {@link File File.class}
   */
  public Class getImportSourceType() {
    return File.class;
  }

  /**
   * Importiert eine Instanz von {@link SingleGrid} aus einer Datei im
   * ArcInfoAsciiGrid-Format.
   * @param input Eingabe-Quelle (muss ein {@link File} sein!)
   * @param reg   Instanz der Xulu-Registry, ueber die eine Factory ermittelt
   *              wird, die eine Standard-Instanz von {@link SingleGrid} erzeugt, in
   *              die das Raster importiert wird
   * @exception UnsupportedOperationException falls als Eingabe-Quelle keine
   *            Datei angegeben wird
   */
  public SingleGrid importObject(Object input, XuluRegistry reg) throws Exception {
    AbstractFactory.checkImportSourceObject(this,input,getImportSourceType(),true);
    // Aus Registry die passende Factory fuer SingleGrid suchen
    InstantiationFactory instFac = AbstractFactory.getInstantiationFactoryFromRegistry(reg,getImportType(),true);
    // Raster importieren und in ein SingleGrid einfuegen
    SingleGrid sg = (SingleGrid)instFac.newInstance(false);
    File inputFile = (File) input;
    File outputFile = new File(inputFile.getAbsoluteFile()+".xgrid");
    //check if already a gridFile (should not be..)
    XuluGridFile xuluGridFile;
    if(((File)input).getAbsolutePath().endsWith(".xgrid"))
    	xuluGridFile = new XuluGridFile((File)input);
    else
    {	
    //check if a the import should be converted even if a xulugridfile with the same name exists
    boolean alwaysreload = XuluConfig.getXuluConfig().getBooleanProperty("Datatypes.xulugrid.alwaysconvert");		
    //only convert if nessecary
    if(alwaysreload || (!outputFile.exists())){
    	 //the ArcInfoGrid must be converted to XuluGridFile
        XuluGridFileConverter converter = new XuluGridFileConverter(inputFile,outputFile);
        int dataType = RasterUtil.getRasterMetaData_from_ArcGridASCII_File((File) input).getDataType();
        xuluGridFile = converter.convertArcInfoToXuluGridFile(dataType);
    }
    else
    	xuluGridFile = new XuluGridFile(outputFile);
    }
    WritableGrid grid = new XuluWritableGridFile(xuluGridFile);
    sg.setGrid(grid);
    // Dateinamen in die Bezeichnung aufnehmen
    sg.setDescription(sg.getDescription().concat(" [").concat(((File)input).getName()).concat("]"));
    return sg;
  }

  ////////////////////////////////////////////////////////////////////////
  ///////////////     Implementierung von ExportFactory     //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Prueft, ob ein Objekt exportiert werden kann. Dies ist der
   * Fall, wenn es sich um ein {@link SingleGrid} handelt, das ein
   * {@link WritableGridArray} in seiner Raster-Eigenschaft besitzt.
   * @param obj zu pruefendes Objekt
   */
  public boolean isExportable(Object obj) {
    return obj!=null &&
           obj instanceof SingleGrid &&
           ((SingleGrid)obj).getGrid() instanceof XuluWritableGridFile;
  }

  /**
   * Prueft, ob ein Objekttyp exportiert werden kann. Dies ist der
   * Fall, wenn es sich um {@link SingleGrid} (oder eine Unterklasse)
   * handelt.
   * @param c zu pruefender Objekttyp
   */
  public boolean isExportable(Class c) {
    return SingleGrid.class.isAssignableFrom(c);
  }

  /**
   * Liefert den Objekt-Typ, den die Factory als Ziel zum Exportieren
   * benoetigt.
   * @return immer {@link File File.class}
   */
  public Class getExportDestinationType() {
    return File.class;
  }

  /**
   * Exportiert eine Instanz von {@link SingleGrid} in eine ArcInfoAsciiGrid-Datei.
   * @param grid zu exportiertendes Raster (muss ein {@link SingleGrid} sein
   * @param output Export-Ziel (muss ein {@link File} sein!)
   * @exception UnsupportedOperationException falls als Export-Ziel keine
   *            Datei angegeben wird
   * @exception IllegalArgumentException falls es sich bei dem angegeben Raster
   *            nicht um ein {@link SingleGrid} handelt
   */
  public void exportObject(Object grid, Object output) throws Exception {
    AbstractFactory.checkExportDestinationObject(this,output,getExportDestinationType(),true);
    if ( !isExportable(grid) )
      throw new IllegalArgumentException(getClass().getName().concat(" can not export instances of ").concat(grid.getClass().getName()));
    exportFac.exportObject( ((DataProxy)((SingleGrid)grid).getGrid()).getProxiedObject(), new FileOutputStream((File)output) );
  }

}
