/**
 * 
 */
package appl.parallel.event;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.Serializable;

/**
 * A TransferEvent is event which is associated with the size of a data transfer
 * in bytes. It is measured by serializing the data into byte stream. This is
 * very slow! So use this events carefully. You should check if monitoring for
 * this event type is enabled, before generating it! See {@link CommEventSink}
 * for details.
 * 
 * @author Dominik Appl
 */
public class TransferEvent extends CommEvent {

	int size = 0;

	/**
	 * @param src
	 *            the source of the event
	 * @param target
	 *            the target of the event
	 * @param type
	 *            the type of the event
	 * @param toTransferObjects
	 *            the objects transfered (Warning: the size will be measured by
	 *            serializing the object. This is slow!)
	 * @see #getObjectSize(Object)
	 */
	public TransferEvent(String src, String target, CommType type,
			Object... toTransferObjects) {
		super(src, target, type);
		for (Object object : toTransferObjects) {
			size += getObjectSize(object);
		}
	}

	/**
	 * Determines the object size in a very primitive way (serializing to a byte
	 * array and returning the byteSize) I do not have to mention here
	 * explicitly that this is slow, do I? There is no real alternative, because
	 * in java exists. no function like <code>sizeOf</code> in C++
	 * 
	 * @param object
	 *            a {@link Serializable} object
	 * @return the size in bytes
	 */
	protected int getObjectSize(Object object) {
		if (object == null)
			return 0;

		byte[] byteArray = null;
		try {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			ObjectOutputStream oos = new ObjectOutputStream(baos);
			oos.writeObject(object);
			oos.close();
			byteArray = baos.toByteArray();
			baos.close();
		} catch (IOException e) {
			e.printStackTrace();
			return 0;
		}
		return byteArray.length;
	}

	public int getSize() {
		return size;
	}

	public String toString() {
		StringBuffer returnString = new StringBuffer();
		returnString.append(type.toString());
		returnString.append(" Transfer size: ").append(size).append(" bytes");
		if (!src.equals(""))
			returnString.append(" source: ").append(src);
		if (!target.equals(""))
			returnString.append(" target: ").append(target);

		return returnString.toString();
	}

}
