package appl.parallel.services;

import java.util.Iterator;
import java.util.Vector;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import appl.ext.XuluConfig;
import appl.parallel.ComputingResource;
import appl.parallel.ComputingResourceContainer;

/**
 * This class is responsible for discovering Resources. For this it first starts
 * all known {@link DiscoveryService DiscoveryServices} and queries them for
 * known {@link ComputingResource Computing Resources}. <br>
 * <br>
 * This class starts all {@link DiscoveryService}s registered with the following
 * property in the {@link XuluConfig}: <br>
 * <br>
 * DiscoveryServices.activeServices<br>
 * <br>
 * Register {@link DiscoveryService DiscoveryServices} with their fully
 * qualified java name because
 * {@link java.lang.Class#forName(String)  Class.forName} will be called on that
 * String.
 * 
 * @author Dominik Appl
 */
public class GlobalDiscoveryService implements Service {

	private final Logger LOG = LogManager.getLogger(this.getClass().getName());

	private static GlobalDiscoveryService instance;

	Vector<DiscoveryService> services = new Vector<DiscoveryService>();

	private boolean isRunning;

	/**
	 * @return all active {@link ComputingResource RemoteResources} in a
	 *          container class with meta in the a ComputingResourceContainer
	 *          Object. They are queried from all known Discovery Services. Do
	 *          not query more than necessary!
	 */
	public Vector<ComputingResourceContainer> getRemoteResources() {
		Vector<ComputingResourceContainer> discResources = new Vector<ComputingResourceContainer>();
		// query all services for RemoteResources
		for (DiscoveryService disc : services) {
			Vector<ComputingResourceContainer> newResources = disc
					.getResources();
			// do not add double entrys
			for (ComputingResourceContainer container : newResources) {
				if (!(checkAlreadyInside(container, discResources)))
					discResources.add(container);
			}
		}
		return discResources;
	}

	/**
	 * checks if the a remoteResource of a given container is in a Vector of
	 * RemoteResourceContainers
	 *
	 * @param container
	 * @return
	 */
	private boolean checkAlreadyInside(ComputingResourceContainer container,
			Vector<ComputingResourceContainer> containers) {
		for (ComputingResourceContainer originalcontainer : containers) {
			if (originalcontainer.getResource().equals(container.getResource()))
				return true;
		}
		return false;
	}

	/**
	 * Looks up all services from the config file, loads them and also starts
	 * them.
	 */
	public GlobalDiscoveryService() {
		String[] classnames = XuluConfig.getXuluConfig().getMultiProperty(
				"DiscoveryServices.activeServices");
		for (String classname : classnames) {
			Class serviceclass;
			try {
				serviceclass = Class.forName(classname);
				DiscoveryService service = (DiscoveryService) serviceclass
						.newInstance();
				services.add(service);
				service.startService();

			} catch (ClassNotFoundException e) {
				LOG.error("Class was not found for string" + classname);
				LOG.error(e.getMessage());
			} catch (InstantiationException e) {
				LOG.error("Could not load class " + classname);
				LOG.error(e.getMessage());
			} catch (IllegalAccessException e) {
				LOG.error("Could not load class " + classname);
				LOG.error(e.getMessage());
			}
		}
		// allow some time for the services to discover Resources
		try {
			Thread.currentThread().sleep(200);
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	
	/*
	 * (non-Javadoc)
	 *
	 * @see appl.parallel.services.Service#isRunning()
	 */
	public boolean isRunning() {
		return isRunning;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see appl.parallel.services.Service#startService()
	 */
	public void startService() {
		isRunning = true;

	}

	/*
	 * (non-Javadoc)
	 *
	 * @see appl.parallel.services.Service#stopService()
	 */
	public void stopService() {
		for (DiscoveryService s : services) {
			s.stopService();
		}
		isRunning = false;

	}

}
