package appl.parallel.services;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.MulticastSocket;
import java.net.UnknownHostException;
import java.rmi.Naming;
import java.rmi.RemoteException;
import java.util.Iterator;
import java.util.Vector;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import appl.ext.XuluConfig;
import appl.parallel.ComputingResource;
import appl.parallel.ComputingResourceContainer;
import appl.parallel.server.XuluServer;
import appl.parallel.test.PingTestObject;
import appl.parallel.util.Helper;

/**
 * Responsible for discovery of RemoteResources on the network, especially but
 * not only {@link XuluServer XuluServers}. It listens for the following
 * messages: <br>
 * <br>
 * "Hello Xulu from XuluServer" - sent by RemoteResources like
 * {@link XuluServer} when starting <br>
 * "Goodbye Xulu" - sent by RemoteResources like {@link XuluServer} when
 * stopping <br>
 * <br>
 * and sends: <br>
 * "Hello Servers" - when started If multicasting does not work try to disable
 * your firewall or better configure it to allow multicast packages. Notice
 * also, that multicasting may not work in virtual machines like MS Virtual PC
 * 2004. <Warning><b>THIS CLASS IS STILL FOR DEMONSTRATION ONLY</b></Warning>
 * 
 * @author Dominik Appl
 * @see XuluServer
 */
public class MulticastDiscoveryService implements DiscoveryService {

	// the data to send
	final String sendHello = "Hello Servers";

	// data to receive
	final String helloFromXuluServers = "Hello Xulu from XuluServer";

	final String recGoodBye = "Goodbye Xulu";

	// the Thread which waits on Server hello/goodbyes
	volatile ReceiveThread receiveThread;

	// volatile RenewalThread renewalThread;

	// the thread that checks if servers are alive
	// volatile CheckThread checkThread;

	// the data is read from XuluConfig
	InetAddress group;

	int port;

	int lease;

	MulticastSocket socket;

	boolean isRunning = false;

	private final Logger LOG = LogManager.getLogger(this.getClass().getName());

	/**
	 * Constructs a new service reading port and lease time from
	 * {@link appl.ext.XuluConfig}
	 */
	public MulticastDiscoveryService() {
		XuluConfig config = XuluConfig.getXuluConfig();
		port = Integer.valueOf(config
				.getProperty("DiscoveryServices.multicast.port"));
		// lease = Integer.valueOf(config
		// .getProperty("MulticastDiscoveryService.lease"));
		try {
			group = InetAddress.getByName(config
					.getProperty("DiscoveryServices.multicast.group"));
		} catch (UnknownHostException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/**
	 * Constructs a new service
	 * 
	 * @param port
	 *            port where the multicasting happens
	 * @param lease
	 *            timeintervall (in ms) after whi
	 */
	public MulticastDiscoveryService(int port, int lease) {
		this();
		this.port = port;
		this.lease = lease;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.services.Service#startService()
	 */
	public void startService() {
		try {
			LOG.info("Discoveryservice started");
			// All Servers must listen on the port specified, in order fo
			// the socket to receive
			if (socket == null)
				socket = new MulticastSocket(port);
			// open a multicast group and join it

			socket.joinGroup(group);

			socket.setSoTimeout(5000);

			// create and send Hello world
			DatagramPacket datagram = new DatagramPacket(sendHello.getBytes(),
					sendHello.length(), group, port);
			System.out.println(group);
			LOG.info("Send hello to all listening servers with broadcast IP "
					+ group);

			socket.send(datagram);

			if (receiveThread == null) {
				receiveThread = new ReceiveThread(socket, helloFromXuluServers,
						recGoodBye);
			}
			if (!(receiveThread.isAlive()))
				receiveThread.start();

		} catch (UnknownHostException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.services.DiscoveryService#getResources()
	 */
	public Vector<ComputingResourceContainer> getResources() {
		return receiveThread.getResources();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.services.Service#stopService()
	 */
	public void stopService() {
		receiveThread.stopThread();
		// renewalThread.stopThread();
		try {
			Thread.currentThread().sleep(100);
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		socket.close();

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.services.Service#isRunning()
	 */
	public boolean isRunning() {
		return isRunning;
	}

	/** waits for incoming calls */

	class ReceiveThread extends Thread {

		// Resources discovered so far
		Vector<ComputingResource> discoveredResources;

		private volatile Thread receiveThread;

		boolean exit = false;

		MulticastSocket socket;

		private final String recHello2;

		private final String recGoodBye2;

		private boolean discoveryInProgress = false;

		/**
		 * @return a Vector of discovered Ressources before return, all values
		 *          are pinged and only returned if alive
		 */
		public Vector<ComputingResourceContainer> getResources() {
			return Helper.getResourceContainersForVector(discoveredResources);
		}

		/**
		 * @param socket2
		 *            Socket which is used for listening
		 * @param helloFromXuluServers
		 *            String that equals the hello message of the server
		 * @param recGoodBye
		 *            String that equals the Goordbye message of the server
		 */
		public ReceiveThread(MulticastSocket socket, String recHello,
				String recGoodBye) {
			recHello2 = recHello;
			recGoodBye2 = recGoodBye;
			this.socket = socket;
			discoveredResources = new Vector<ComputingResource>();
		}

		/**
		 * Safe method to stop the Thread. Use this method instead of
		 * {@link Thread}
		 */
		public void stopThread() {
			if (receiveThread != null) {
				exit = true;
				// the only way to stop the Thread is to send a package so that
				// the receive operation no longer blocks
				DatagramPacket datagram = new DatagramPacket(
						"Stopping receive Thread".getBytes(),
						"Stopping receive Thread".length(), group, port);
				try {
					socket.send(datagram);
					// wait on stop
					Thread.currentThread().sleep(100);
				} catch (IOException e) {
					e.printStackTrace();
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				} finally {
					exit = false;
				}
				exit = false;

			}

		}

		public void run() {
			receiveThread = Thread.currentThread();
			while (!exit) {
				try {
					// receive buffer

					byte[] buffer = new byte[256];
					DatagramPacket datagram = new DatagramPacket(buffer,
							buffer.length);
					socket.receive(datagram);
					// get the message in the right
					byte[] result = new byte[datagram.getLength()];
					System.arraycopy(datagram.getData(), 0, result, 0, datagram
							.getLength());
					// check if message is from XuluServer
					if (new String(result).equals(helloFromXuluServers)) {
						LOG.debug("received hello from server "
								+ datagram.getAddress());

						String newname = "rmi:/" + datagram.getAddress()
								+ "/XuluServer";

						ComputingResource server = (ComputingResource) Naming
								.lookup(newname);
						addRessource(server, datagram.getAddress().toString());

					} else if (LOG.isDebugEnabled())
						LOG.debug("received message |" + new String(result)
								+ "| from " + datagram.getAddress());

				} catch (java.net.SocketTimeoutException e) {
					// catch timeout exception
				}

				catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				} catch (Exception e) {
					if (!(e instanceof InterruptedException))
						e.printStackTrace();
				}
			}
			System.out.println("exit Thread " + this.toString());
		}

		/**
		 * adds ressource to vector but first looks if its not already added
		 * 
		 * @param server
		 */
		private void addRessource(ComputingResource server, String IP) {
			// check for double entries
			for (ComputingResource s : discoveredResources) {
				if (s.equals(server)) {
					LOG.debug("Not adding " + server.toString()
							+ " to discovered Ressources again!");
					return;
				}
			}
			LOG.info("Adding Xulu Server: " + IP + "to discovered Ressources");
			// try {
			// Thread.currentThread().sleep(2000);
			// } catch (InterruptedException e) {
			// // TODO Auto-generated catch block
			// e.printStackTrace();
			// }
			discoveredResources.add(server);
		}

		/**
		 * Gets the {@link Resource
		 */
		public void pingResources() {
			PingTestObject po = new PingTestObject(32);
			// cannot make a for each loop because of
			// java.util.ConcurrentModificationException
			int i = discoveredResources.size() - 1;
			while (i >= 0) {
				boolean error = true;
				ComputingResource r = discoveredResources.get(i);
				if (!(r == null)) {

					try {
						r.ping();
						error = false;
						LOG.debug("Ping to " + r + " was successfull!");
					} catch (Exception e) {
						e.printStackTrace();
					}
					// if not successfull remove object
					if (error) {
						LOG.debug("Ping to " + r
								+ " was NOT successfull! Removing ressource");

						discoveredResources.remove(r);
					}
				}
				i--;
			}
		}

	}

	/**
	 * simple Thread that caused rediscovery of Ressources at a given time
	 * intervall **
	 */

	class RenewalThread extends Thread {

		private volatile Thread thisThread;

		boolean exit = false;

		private final long timeintervall;

		private volatile ReceiveThread discoveryToRenew;

		/**
		 * invokes rediscovery on the given thread at the given time intervall
		 */
		public RenewalThread(ReceiveThread discoveryToRenew, long timeintervall) {
			this.discoveryToRenew = discoveryToRenew;
			this.timeintervall = timeintervall;

		}

		/**
		 * Safe method to stop this Thread. Use this method instead of
		 * {@link Thread}
		 */
		public void stopThread() {
			if (thisThread != null)
				exit = true;
		}

		public void run() {
			thisThread = Thread.currentThread();
			while (!exit) {
				discoveryToRenew.pingResources();
				try {
					Thread.currentThread().sleep(timeintervall);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
			System.out.println("exit renewal Thread");
		}

	}

	/**
	 * just for testing..edit as you like
	 */
	public static void main(String[] args) {
		BasicConfigurator.configure();
		MulticastDiscoveryService bro = new MulticastDiscoveryService();
		bro.startService();
		try {
			Thread.currentThread().sleep(5000);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		// bro.stopService();
		while (true)
			try {
				Vector<ComputingResourceContainer> v = bro.getResources();
				for (ComputingResourceContainer container : v) {
					System.out.print("Active Object: ");
					System.out.println(container.getInformation().getProperty(
							"name")
							+ " with IP "
							+ container.getInformation().getProperty("IP"));
				}
				Thread.currentThread().sleep(5000);
			} catch (InterruptedException e) {

			}
	}

}
