package appl.parallel.services;

import java.rmi.RemoteException;
import java.util.Vector;
import java.util.concurrent.ConcurrentLinkedQueue;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import appl.parallel.event.CommEventSink;
import appl.parallel.event.RemoteEvent;
import appl.parallel.event.RemoteEventSink;
import appl.parallel.event.TimeEvent;
import appl.parallel.event.TransferEvent;

/**
 * Delays events and fires them after the delay all <b>at once</b> in one
 * object (this will save communication time)
 * 
 * @author Dominik Appl
 */
public class RemoteEventProxy implements CommEventSink, Service {

	/**
	 * This Thread submits the events after the delay
	 * 
	 * @author Dominik Appl
	 */
	private class SubmitThread extends Thread {

		private boolean exit = false;

		private final ConcurrentLinkedQueue<RemoteEvent> queue2;

		private final CommEventSink sink;

		private final long delay;

		SubmitThread(ConcurrentLinkedQueue<RemoteEvent> queue,
				CommEventSink sink, long delay) {
			queue2 = queue;
			this.sink = sink;
			this.delay = Math.max(100, delay);
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Thread#run()
		 */
		public void run() {
			while (!exit) {
				try {
					this.sleep(delay);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				// add all new events to a vector submit them at once
				Vector<RemoteEvent> events = new Vector<RemoteEvent>();
				while (queue.peek() != null)
					events.add(queue.poll());
				try {
					if (events.size() > 0)
						sink.fireRemoteEvents((RemoteEvent[]) events
								.toArray(new RemoteEvent[events.size()]));
				} catch (RemoteException e) {
					// TODO Auto-generated catch block
					LOG.warn("Could not transmit events." + e.getMessage(), e);
					events.clear();
				}
			}
		}

		public void startThread() {
			exit = false;
			this.start();
		}

		public void stopThread() {
			exit = true;
		}
	}

	private final Logger LOG = LogManager.getLogger(this.getClass().getName());

	private boolean timeMonitoring = false;

	private boolean transferMonitoring = false;

	private boolean isRunning;

	SubmitThread submitThread;

	ConcurrentLinkedQueue<RemoteEvent> queue;

	private final CommEventSink eventSink;

	private final long delay;

	/**
	 * Creates a new proxy
	 * 
	 * @param targetSink
	 *            the sink where the events should be submitted to
	 * @param delay
	 *            the delay (in milliseconds)
	 */
	public RemoteEventProxy(CommEventSink targetSink, long delay) {
		this.eventSink = targetSink;
		this.delay = delay;
		queue = new ConcurrentLinkedQueue<RemoteEvent>();

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.event.RemoteEventSink#fireRemoteEvent(appl.parallel.event.RemoteEvent)
	 */
	public void fireRemoteEvent(RemoteEvent e) throws RemoteException {
		queue.add(e);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.event.RemoteEventSink#fireRemoteEvents(appl.parallel.event.RemoteEvent[])
	 */
	public void fireRemoteEvents(RemoteEvent[] e) throws RemoteException {
		for (RemoteEvent event : e) {
			fireRemoteEvent(event);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.services.Service#isRunning()
	 */
	public boolean isRunning() {
		return isRunning;
	}

	/**
	 * @return true if the service is running and time monitoring is enabled
	 */
	public boolean isTimeMonitoringEnabled() throws RemoteException {
		return timeMonitoring;
	}

	/**
	 * @return true if the service is running and tansfer monitoring is enabled
	 */
	public boolean isTransferMonitoringEnabled() throws RemoteException {
		return transferMonitoring;

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.services.Service#startService()
	 */
	public void startService() {
		isRunning = true;
		if (submitThread != null)
			submitThread.stopThread();
		submitThread = new SubmitThread(queue, eventSink, delay);
		submitThread.startThread();
		try {
			this.timeMonitoring = eventSink.isTimeMonitoringEnabled();
			this.transferMonitoring = eventSink.isTransferMonitoringEnabled();
		} catch (RemoteException e) {
			LOG.error("Could not connect to CommEventSink!!!");
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.services.Service#stopService()
	 */
	public void stopService() {
		isRunning = false;
		if (submitThread != null)
			submitThread.stopThread();
	}
}
