package appl.parallel.spmd;

import java.io.PrintStream;

import appl.parallel.server.PartitionDataManager;
import appl.parallel.spmd.split.DataPartition;

/**
 * This class extends the {@link SPMDClientInterface} with additional features for performance tuning. You can merge
 * partitions in background (so that the not very cpu-intensive communication can
 * happen during extensive calculations). <br>
 * 
 * @author Dominik Appl
 */
public interface AdvancedSPMDClientInterface extends SPMDClientInterface {

	/**
	 * Merges like {@link SPMDClientController#mergePartition(int)}, but in a
	 * separate thread, so that communication does not block computation. See
	 * also {@link SyncPoint synchronization points}
	 * 
	 * @param partition
	 *            the partition to be merged (must be splittable)
	 * @param s
	 *            a syncpoint
	 */
	public void mergePartition(Object partition, SyncPoint s);

	/**
	 * Merges like
	 * {@link SPMDClientController#mergeMultiData(MultiDataObject, int)}. But
	 * allows also {@link SyncPoint synchronization points}
	 * 
	 * @param multidata
	 *            the multidataobject
	 * @param idx
	 *            the partition of the {@link MultiDataObject} to be merged
	 * @param s
	 *            a {@link SyncPoint}
	 */
	public void mergeMultiData(MultiDataObject multidata, int idx, SyncPoint s);

	/**
	 * Same functionality as {@link #mergePartition(Object, SyncPoint)}, gives
	 * a message to the given {@link PrintStream}
	 * 
	 * @param partition
	 *            the partition to merge
	 * @param s
	 *            a {@link SyncPoint}
	 * @param stream
	 *            the message is given out to this stream (or null for no
	 *            message)
	 * @param message
	 *            the message to be displayed when finished (or null for no
	 *            message)
	 * @see #mergeMultiData(MultiDataObject, int, SyncPoint)
	 */
	public void mergePartition(Object partition, SyncPoint s,
			PrintStream stream, String message);

	/**
	 * Merges like
	 * {@link SPMDClientController#mergeMultiData(MultiDataObject, int)}. But
	 * allows also {@link SyncPoint synchronization points}. At the end of the
	 * merge a message is given to the provided {@link PrintStream}.
	 * 
	 * @param multidata
	 *            the multidataobject
	 * @param idx
	 *            the partition of the {@link MultiDataObject} to be merged
	 * @param s
	 *            a {@link SyncPoint}
	 * @param stream
	 *            the message is given out to this stream (or null for no
	 *            message)
	 * @param message
	 *            the message to be diplayed when finished (or null for no
	 *            message)
	 */
	public void mergeMultiData(MultiDataObject multidata, int idx, SyncPoint s,
			PrintStream stream, String message);

	/**
	 * Waits until the Thread with the specified SyncPoint finishes. After that
	 * the Thread is removed. A second thread which synchronizes to the same
	 * Point will directly continue.
	 * 
	 * @param s
	 *            the {@link SyncPoint}
	 * @throws UnsupportedOperationException
	 *             if the SyncPoint does not exist and has never existed
	 */
	public void synchronizeToSyncPoint(SyncPoint s);

}