package appl.parallel.spmd;

import java.rmi.RemoteException;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import schmitzm.data.WritableGrid;

import edu.bonn.xulu.plugin.data.grid.MultiGrid;

import appl.parallel.client.DataServer;
import appl.parallel.server.PartitionDataServer;
import appl.parallel.spmd.split.DataPartition;

/**
 * A multi data object stores multiple elements of the same type.
 * While the related {@link MultiDataInfo} stores only Information about 
 * multi-data and is used by Xulu / V for transfer of metadata, this class
 * is a actual user object. The user can retrieve a multidata object e.g. via the
 * {@link SPMDServerController} and work with it (e.g. add elements)
 * 
 * @see SPMDServerInterface#getMultiData(String)
 * @see MultiDataInfo
 * 
 * @author Dominik Appl
 **/

public class MultiDataObject {

	private final Logger LOG = LogManager.getLogger(this.getClass().getName());

	private final MultiDataInfo multiDataInfo;

	private final DataServer dataServer;

	private MultiGrid managedGrid = null;

	/**
	 * Construct a new Multidata object
	 * 
	 * @param info a {@link MultiDataInfo} object with the identification of the objects
	 * @param dataServer a dataServer which is responsible for the actual loading of the data
	 */
	public MultiDataObject(MultiDataInfo info, DataServer dataServer) {
		this.multiDataInfo = info;
		this.dataServer = dataServer;
	}

	/**
	 * @param idx the position of the element
	 * @return the element at the indicated position
	 */
	public Object getElement(int idx) {
		try {
			checkIndex(idx);
			return dataServer.getData(multiDataInfo.getMultiID(idx));
		} catch (RemoteException e) {
			LOG.error("RemoteExeception while retrieving data with ID "
					+ multiDataInfo.getMultiID(idx)
					+ " from PartitionDataServer");
		}
		return null;
	}

	/**
	 * checks whether the indicated index is valid
	 * @throws ArrayIndexOutOfBoundsException if idx is below zero or higher than number of elements 
	 */
	private void checkIndex(int idx) {
		if ((idx > multiDataInfo.getCount() - 1) || idx <0 )
			throw new ArrayIndexOutOfBoundsException("The Element with index "
					+ idx + " was requested, but the MultiDataObject has only "
					+ multiDataInfo.getCount() + " elements.");

	}

	/**
	 * @return the number of elements in the multidataobject
	 */
	public int getCount() {
		return multiDataInfo.getCount();
	}

	/**
	 * Adds an element to MultiDataObjects. Method should be thread safe.
	 * 
	 * @return the index of the new element
	 */
	public synchronized int addElement() {
		//the new ID of the MultiDataObject is generated by making
		//the first id negative and than subtracting the index from 
		//the ID
		int idx = multiDataInfo.getCount() - 1;
		try {
			if (idx == -1)
				throw new UnsupportedOperationException(
						"The MultiDataInfo must have at least one element!");
			//simply get the first element and create a empty one.
			int firstElementID = multiDataInfo.getMultiID(0);
			DataPartition firstElement = dataServer.getData(firstElementID);
			//generate a new id by making the first ID negative and subtracting the index value
			int newID = multiDataInfo.getNewIDForIndex(idx + 1);
			//there is a rare chance of colliding with existing ids:
			while (dataServer.getData(newID) != null) {
				newID--;
			}
			//create new element 
			DataPartition newElement = firstElement.getEmpty(newID);
			multiDataInfo.addElement(newID);
			//add the element to the dataServer and to the managed MultiGrid (if any)
			dataServer.addData(newElement);
			if (managedGrid != null)
				managedGrid.addGrid((WritableGrid) newElement);
			return idx + 1;
		} catch (RemoteException e) {
			LOG
					.error("RemoteExeception while trying to add a new element with idx "
							+ idx + " to MultiData");
		}
		return 0;
	}

	/**
	 * Adds an element with the specified ID.
	 * 
	 * @param id the id of the element to add
	 * @return the index of the new element or 0 if an error occurred
	 */
	public int addElement(int id) {
		//actual maxIndex
		int idx = multiDataInfo.getCount() - 1;
		try {
			if (idx == -1)
				throw new UnsupportedOperationException(
						"The MultiDataInfo must have at least one element!");
			//simply get the first element and create a empty one.
			int firstElementID = multiDataInfo.getMultiID(0);
			DataPartition firstElement = dataServer.getData(firstElementID);

			//create new element 
			DataPartition newElement = firstElement.getEmpty(id);
			multiDataInfo.addElement(id);
			//add the element to the dataServer and to the managed MultiGrid (if any)
			dataServer.addData(newElement);
			if (managedGrid != null)
				managedGrid.addGrid((WritableGrid) newElement);
			return idx + 1;
		} catch (RemoteException e) {
			LOG
					.error("RemoteExeception while trying to add a new element with idx "
							+ idx + " to MultiData");
		}
		return 0;
	}

	/**
	 * Manages a multigrid, which means that when a new element is added 
	 * to the multiObject, then the same element is also added to 
	 * the managed multiGrid.
	 * 
	 * @param toManageMultiGrid the grid to be managed. Only one grid will be 
	 * managed at a time.
	 *  
	 */
	public void setManagedGrid(MultiGrid toManageMultiGrid) {
		this.managedGrid = toManageMultiGrid;
	}

	/**
	 * @return the encapsulted {@link MultiDataInfo}
	 */
	public MultiDataInfo getMultiInfo() {
		return this.multiDataInfo;
	}

	public String getName() {
		return multiDataInfo.getName();
	}

	/**
	 * removes the element with the specified index from the dataServer
	 * 
	 * @param idx the element
	 */
	public void deleteElement(int idx) {
		checkIndex(idx);
		try {
			dataServer.removeData(multiDataInfo.getMultiID(idx));
		} catch (RemoteException e) {
			LOG.error("Could not reach DataServer" , e);
			e.printStackTrace();
		}	
	}
}
