package appl.parallel.spmd;

import java.awt.Rectangle;
import java.util.HashMap;

import edu.bonn.xulu.plugin.data.grid.GridList;
import edu.bonn.xulu.plugin.data.grid.MultiGrid;

import appl.parallel.ComputingResource;
import appl.parallel.spmd.split.DataPartition;
import appl.parallel.spmd.split.SplittableResource;

/**
 * Using this interface the programmer can access the parallel functionality on server side.
 * The programmer should subclass {@link AbstractSPMDTask} and use its method {@link AbstractSPMDTask#getSPMDServerController()} to
 * access the parallel control.
 * @author Dominik Appl
 */
public interface SPMDServerInterface {

	/**
	 * Returns the base parameter with the specified id.
	 *
	 * @param parameterName the name of the parameter specified on client side
	 * @return the base parameter
	 *
	 * @see SPMDClientInterface#addBaseParameter(Object, String)
	 */
	public Object getBaseParameter(String parameterName);

	/**
	 * Gives back the area which is actively used for read/write (in contrast
	 * to neighborhood area, which is often accessed will be accessed read only).
	 *
	 * @param partition
	 *            a partition for which the the bounds are requested
	 * @return the calculation bounds in local coordinates (local means relative
	 *         to the upper left corner of all available data)
	 */
	public Rectangle getLocalCalculationBounds(DataPartition partition);

	/**
	 * Gives back the area which is available for local access (including
	 * Neighborhood)
	 *
	 * @param partition
	 *            a partition for which the the bounds are requested
	 * @return the calculation bounds in local coordinates - the upper left
	 *         corner of the {@link Rectangle} will be (0,0)
	 */
	public Rectangle getLocalBounds(DataPartition partition);

	/**
	 * Gives back the global bounds, which are the bounds of all partitions over all servers.
	 * To be clear: This are the bounds of the unsplitted data!
	 * @param partition partition for which the bounds are queried
	 * @return the global bounds
	 */
	public Rectangle getGlobalBounds(DataPartition partition);

	/**
	 * Gives back the area which is actively used for read/write (in contrast
	 * to neighborhood area, which is often accessed will be accessed read only).
	 *
	 * @return the calculation bounds in local coordinates (local means relative
	 *         to the upper left corner of all available data) of the reference
	 *         resource (which is the first resource added to split-control on
	 *         client side)
	 *
	 * @see SPMDClientController#addToSplitControl(Object, String)
	 * @see SPMDClientController#setReferenceResource(Object)
	 */
	public Rectangle getLocalCalculationBounds();

	/**
	 * Gives back the area which is available for local access (including
	 * Neighborhood)
	 *
	 * @return the calculation bounds in local coordinates of the reference
	 *         resource (which is the first resource added to split-control on
	 *         client side) - the upper left corner of the {@link Rectangle}
	 *         will be (0,0)
	 * @see SPMDClientController#addToSplitControl(Object, String)
	 * @see SPMDClientController#setReferenceResource(Object)
	 */
	public Rectangle getLocalBounds();

	/**
	 * The whole {@link DataPartition} available for this task. In contrast to
	 * {@link #getLocalCalculationBounds()} the neighborhood is included!
	 *
	 * @param name
	 *            the name of the splittable
	 * @return the partition of the resource which is available for local
	 *         calculation
	 * @see SPMDClientController#addToSplitControl(Object, String)
	 */
	public DataPartition getPartition(String name);

	/**
	 * Gets the array of {@link DataPartition DataPartitions} which were
	 * submitted on client side via
	 * {@link SPMDClientController#addToMultiDataSplitControl(Object[], String)}
	 * All resources are retrieved from the data source and given back. If you
	 * only need a specific resource you can use
	 * {@link #getMultiPartition(String, int)} to avoid loading all data.
	 *
	 * @param name
	 *            the name of the splittable
	 * @return the partition of the resource which is available for local
	 *         calculation
	 * @see SPMDClientController#addToMultiDataSplitControl(Object[], String)
	 * @see #getMultiPartition(String, int)
	 */
	public DataPartition[] getMultiPartition(String name);

	/**
	 * Gives back a MultiDataObject. A MultiDataObject can encapsulate multiple
	 * objects and is especially intended for use with {@link GridList} and
	 * {@link MultiGrid} types. See {@link MultiDataObject} for more details.
	 *
	 * @param name the name associated with the object on client side
	 * @return the MDO
	 */
	public MultiDataObject getMultiData(String name);

	/**
	 * Gets a {@link DataPartition} which was submitted on client side via
	 * {@link SPMDClientController#addToMultiDataSplitControl(Object[], String)}.
	 * Only the grid with the index is returned. All other partitions are NOT
	 * retrieved.
	 *
	 * @param name
	 *            the name of the splittable
	 * @param idx
	 * @return the partition of the resource which is available for local
	 *         calculation
	 * @see SPMDClientController#addToMultiDataSplitControl(Object[], String)
	 * @see #getMultiPartition(String)
	 */
	public DataPartition getMultiPartition(String name, int idx);

	/**
	 * Convenience method. Gets the coordinate out from the Rectangle which the
	 * {@link #getLocalCalculationBounds()} returns
	 *
	 * @return the x-coordinate upperleft corner of the calcarea
	 * @see SPMDServerController#getLocalCalculationBounds()
	 * @see SPMDServerController#getLocalCalcMinX()
	 * @see SPMDServerController#getLocalCalcMaxX()
	 * @see SPMDServerController#getLocalCalcMinY()
	 * @see SPMDServerController#getLocalCalcMaxY()
	 */
	public int getLocalCalcMinX();

	/**
	 * Convenience method. Gets the coordinate out from the Rectangle which the
	 * {@link #getLocalCalculationBounds()} returns
	 *
	 * @return the y-coordinate upper left corner of the calculation area
	 * @see SPMDServerController#getLocalCalculationBounds()
	 * @see SPMDServerController#getLocalCalcMinX()
	 * @see SPMDServerController#getLocalCalcMaxX()
	 * @see SPMDServerController#getLocalCalcMinY()
	 * @see SPMDServerController#getLocalCalcMaxY()
	 */
	public int getLocalCalcMinY();

	/**
	 * Convenience method. Gets the coordinate out from the Rectangle which the
	 * {@link #getLocalCalculationBounds()} returns
	 *
	 * @return the x-coordinate lower right corner of the calculation area
	 * @see SPMDServerController#getLocalCalculationBounds()
	 * @see SPMDServerController#getLocalCalcMinX()
	 * @see SPMDServerController#getLocalCalcMaxX()
	 * @see SPMDServerController#getLocalCalcMinY()
	 * @see SPMDServerController#getLocalCalcMaxY()
	 */
	public int getLocalCalcMaxX();

	/**
	 * Convenience method. Gets the coordinate out from the Rectangle which the
	 * {@link #getLocalCalculationBounds()} returns
	 *
	 * @return the y-coordinate lower right corner of the calculation area
	 * @see SPMDServerController#getLocalCalculationBounds()
	 * @see SPMDServerController#getLocalCalcMinX()
	 * @see SPMDServerController#getLocalCalcMaxX()
	 * @see SPMDServerController#getLocalCalcMinY()
	 * @see SPMDServerController#getLocalCalcMaxY()
	 */
	public int getLocalCalcMaxY();

	/**
	 * Removes the partition from memory. You may want to call {@link System#gc()} afterwards
	 * @param name the name of the partition to remove
	 */
	public void destroyPartition(String name);

	/**
	 * Removes the partition from memory. You may want to call {@link System#gc()} afterwards
	 * @param name the name of the multidata
	 * @param idx the index of the partition to remove
	 */
	public void destroyMultiPartition(String name,int idx);
}
