package appl.parallel.spmd;

import java.awt.Rectangle;
import java.io.Serializable;

import appl.parallel.server.XuluServer;

import schmitzm.data.WritableGrid;

/**
 * A SPMDTask should be executed on servers using the SPMD-paradigm. 
 * <br>
 * The Task provides access to the {@link SPMDClientInterface} and the
 * {@link AdvancedSPMDClientInterface}. <br>
 * <br>
 * @see AbstractSPMDTask for more details
 * 
 * @author Dominik Appl
 */
public interface SPMDTask extends Serializable {

	/**
	 * Gives access the {@link AdvancedSPMDServerInterface} for advanced parallel
	 * programming, including performance optimizations like preloading.
	 * @return the interface
	 */
	public AdvancedSPMDServerInterface getAdvancedSPMDServerController();

	/**
	 * Gives access the {@link SPMDServerInterface} for parallel programming.
	 * 
	 * @return the interface
	 */
	public SPMDServerInterface getSPMDServerController();

	/**
	 * Associates the given SPMDController with the task. Will be called e.g. by
	 * the {@link XuluServer}.
	 * 
	 * @param controller
	 */
	public void setSPMDServerController(AdvancedSPMDServerInterface controller);

	/**
	 * Currently not used. May be used later when introducing write access on
	 * neighborhood regions.
	 * 
	 * @param incomingGrid
	 * @param location
	 */
	public void outgoingUpdate(WritableGrid incomingGrid, Rectangle location);

	/**
	 * Currently not used. May be used later when introducing write access on
	 * neighborhood regions.
	 * 
	 * @param outgoingGrid
	 * @param location
	 */
	public void incomingUpdate(WritableGrid outgoingGrid, Rectangle location);

	/**
	 * Starts the task with the given parameters. Overwrite this method to
	 * implement the task.
	 * 
	 * @param parameters
	 * @return the result of the task computation (if any)
	 */
	public Object run(Object... parameters);

	/**
	 * Initializes the task. Initializing is only done once!
	 */
	public void initialize();

	/**
	 * @return whether the task is initialized or {@link #initialize()} must be
	 *         called
	 */
	public boolean isInitialized();

	/**
	 * If you want to support multiple processors, this method must return true.
	 * 
	 * @return whether multithreading can be used by this task
	 */
	public boolean supportsMultiThreading();

}
