package appl.parallel.spmd;

/**
 * Specifies details about a synchronization point. These points are used in
 * the {@link AdvancedSPMDClientController} or {@link AdvancedSPMDServerController}
 * as preloading-Points or multithread-barriers. They are associated with a priority. 
 * 
 * @author Dominik Appl
 */
public class SyncPoint {
	
	/**
	 * A high priority means a high priority on operating system level. It should
	 * be used by implementations of {@link AdvancedSPMDClientController} and {@link AdvancedSPMDServerController}
	 * 
	 * @author Dominik Appl
	 */
	public enum Priority
	{
		LOW(Thread.MIN_PRIORITY),
		BELOW_NORMAL(Thread.NORM_PRIORITY-1),
		NORMAL(Thread.NORM_PRIORITY),
		ABOVE_NORMAL(Thread.NORM_PRIORITY+1),
		HIGH(Thread.MAX_PRIORITY);
		
		private final int realPriorityNo;
		
		Priority(int realPriorityNo)
		{
		   this.realPriorityNo=realPriorityNo;
		}
		
		/**
		 * @return the priority
		 */
		public int getPriority(){
			return realPriorityNo;
		}
		
	}
	private final int id;
	private final Priority priority;

	/**
	 * Creates a new Sync-Point with normal priority
	 * 
	 * @param pointID a id for this syncPoint
	 */
	public SyncPoint(int pointID) {
		this(pointID,Priority.NORMAL);
	}
	
	/**
	 * Creates a new Sync-Point with normal {@link Priority} (and with hashcode as id)
	 */
	public SyncPoint() {
		this.id = this.hashCode();
		this.priority = Priority.NORMAL;
	}
	
	/**
	 * @param id a id for the 
	 * @param priority 
	 */
	public SyncPoint(int id, Priority priority ) {
		this.id = id;
		this.priority = priority;
	}

	/**
	 * Creates a new Sync-Point with the given {@link Priority}
	 */
	public SyncPoint(Priority priority) {
		id = this.hashCode();
		this.priority = priority;
	}

	/**
	 * @return the id
	 */
	public int getId() {
		return id;
	}

	/**
	 * @return the priority of the Thread (as system dependent for direct use in {@link Thread Threads})
	 */
	public int getPriority() {
		return priority.getPriority();
	}
}
