package appl.parallel.spmd.split;

import java.io.Serializable;

import appl.data.DataLoader;
import appl.parallel.data.PartitionDataHandler;

// fuer Doku
import javax.activation.DataHandler;
import appl.parallel.client.DataServer;
import appl.parallel.data.XuluClientLoader;

/**
 * This class collects the metadata of a partitioning
 * The following information is encapsulated in a SinglePartitionInfo Object:
 * <ul>
 * <li>
 * The <b>ID</b>of the {@link SplittableResource SplittableResource}
 * </li>
 * <li>The <b>baseResourceName</b> of the resource (given by the parallel programmer)</li>
 * <li>The <b>{@link DataHandler}</b>, which loads the data of the partition. This is a very important aspect.
 * The handler defines how the data is actually loaded. It may for example be loaded
 * from a {@link DataServer} using a {@link XuluClientLoader}, but it may also be used
 * to load data from local disk, databases etc. without communication with the client.
 * The {@link DataHandler} is also responsible for writing the the data back to the source
 * </li>
 * <li>
 * The <b>{@link SplitMap}</b> which represents the current partitioning.
 * </li>
 * <li>
 * A <b>splitmap position</b>, which identifies the partition number in the split.
 * </li>
 *
 * @author Dominik Appl
 *
 *
 */
public class SinglePartitionInfo implements Serializable, PartitionInfo{

	private final int baseResourceID;
    private final PartitionDataHandler handler;
    private final SplitMap splitMap;
    private final int splitMapPos;
    private final String baseResourceName;

    /**
     * The constructor.
     *
     * @param baseResourceID the root id (see {@link DataPartition#getRootID()})
     * @param baseResourceName the name of the resource (given by parallel programmer)
     * @param handler {@link PartitionDataHandler}
     * 			which is responsible for loading and unloading
     * @param splitMap splitmap describing the partitioning
     * @param splitMapPos the position inside the splitmap
     */
    public SinglePartitionInfo(int baseResourceID, String baseResourceName, PartitionDataHandler handler, SplitMap splitMap, int splitMapPos){
        this.baseResourceID = baseResourceID;
        this.baseResourceName = baseResourceName;
        this.handler = handler;
        this.splitMap = splitMap;
        this.splitMapPos = splitMapPos;
    }

    /**
	 * @return the baseResourceID
	 */
	public int getBaseResourceID() {
		return baseResourceID;
	}
    /**
	 * @return the handler
	 */
	public PartitionDataHandler getPartitionDataHandler() {
		return handler;
	}
    /* (non-Javadoc)
	 * @see appl.parallel.spmd.split.PartitionInfo#getSplitMap()
	 */
	public SplitMap getSplitMap() {
		return splitMap;
	}
    /* (non-Javadoc)
	 * @see appl.parallel.spmd.split.PartitionInfo#getSplitMapPos()
	 */
	public int getSplitMapPos() {
		return splitMapPos;
	}

    /* (non-Javadoc)
	 * @see appl.parallel.spmd.split.PartitionInfo#getBaseResourceName()
	 */
	public String getBaseResourceName() {
		return baseResourceName;
	}

	public PartitionInfo clone(int newRootID){
		PartitionDataHandler clonedHandler = handler.clone();
		clonedHandler.setRootID(newRootID);
		return new SinglePartitionInfo(newRootID,
				this.baseResourceName,clonedHandler,this.splitMap,
				this.splitMapPos);
	}

}
