package appl.parallel.spmd.split;

import java.awt.Rectangle;
import java.awt.geom.Rectangle2D;

import appl.parallel.spmd.split.AbstractSplitMap.NeighborhoodBoxingMode;
import appl.util.RasterMetaData;
import schmitzm.data.WritableGrid;

/**
 * Responsible for splitting a 2D Area (e.g a {@link WritableGrid}) in a 
 * 1D fashion, which means in this case horizontal (by rows). This is only a
 * a virtual split (a map of a split). 
 * 
 * A neighborhood range can be specified to create partitions that
 * overlap each other. This overlapping can be done in two different
 * ways: <br><br>
 * 
 * <b>Inboxing</b>(default):<br><br>
 * Inboxing means, that the neighborhood area is not part of the calculation
 * area. 
 * <br> <br>
 * <b>Outboxing:</b><br>
 * Outboxing means that the neighborhood area is part of the calculation area.
 *  <br> <br>
 * 
 *
 * Note that this class is flexible. It may also be applied 
 * to one dimensional data structures like for example arrays. 
 * If u want to do this, simply choose one of the dimensions as 1. <br>
 * 
 * @author Dominik Appl
 */

public class SplitMap1DHorizontal extends AbstractSplitMap implements SplitMap {



	public SplitMap1DHorizontal(int width, int height, int neighborhoodRange, int noOfPartitions,
			NeighborhoodBoxingMode boxingMode) {
		
		super(width,height,neighborhoodRange, noOfPartitions, boxingMode);
	}
	

	/**
	 * needed for serialization
	 */
	public SplitMap1DHorizontal(){
		
	}
	
	/* (non-Javadoc)
	 * @see appl.parallel.spmd.split.DataSplitter#getDescription()
	 */
	public String getDescription() {
		return "1D-GridSplitter";
	}

	/* (non-Javadoc)
	 * @see appl.parallel.spmd.split.DataSplitter#split(appl.parallel.spmd.split.SplittableRessource)
	 */
	public void makeMap() {
			//the local calculation area is partitioned
			int nextUpperLeftCorner = (int) globalBounds.getY(); //next y-value (will be 0)
			for (int i = 0; i < noOfPartitions; i++) {
				
				//calculate length based on weights
				int pHeight = (int) (globalBounds.getHeight() * weights[i]);
				
				//for the last position make sure, that ALL data is used 
				//and no line is lost due to rounding errors in weight calculation
				if(i==noOfPartitions-1)
					pHeight = (int) globalBounds.getHeight() - nextUpperLeftCorner;
				
				//create inboxing(!) partition 
				partitionCalculationBounds[i] = new Rectangle(
						(int) globalBounds.getX(), nextUpperLeftCorner, (int)globalBounds.getWidth(),
						pHeight);
				
				//if there is only one partition there are no explicit neighborhood bounds
				if(noOfPartitions==1)
					partitionNeighborhoodBounds[i]=partitionCalculationBounds[i];
				//else simply calculate the neighborhood bounds out of the previosly 
				//created calculation Area (@see AbstractSplitMap)
				else 
					partitionNeighborhoodBounds[i] = new Rectangle(
							(int) globalBounds.getX(),
							(int) partitionCalculationBounds[i].getY() - neighborhoodRange,
							(int)(partitionCalculationBounds[i].getWidth()),
						    (int)(partitionCalculationBounds[i].getHeight()+ 2 * neighborhoodRange));
				
			    //the first and the last partition have a neighborhood area only on one side of 
			    //the calculation area:
			    partitionNeighborhoodBounds[i] = partitionNeighborhoodBounds[i].intersection(globalBounds);
				nextUpperLeftCorner += pHeight; //+1 not required - the partitions will not overlap
				
				//the partitionCalculation bounds were created for inboxing:
				if(boxingMode==NeighborhoodBoxingMode.outBoxing)
					partitionCalculationBounds[i]=partitionNeighborhoodBounds[i];			
		}
	}
	
	/* (non-Javadoc)
	 * @see appl.parallel.spmd.split.SplitMap#getNeighborsForPosition(int)
	 */
	public int[] getNeighborsForPosition(int pos) {
		int[] oneReturn = new int[1];
		int[] twoReturns = new int[2];

		//this is very simple in one 1D case, of course:
		//if only one Partition return null
		if (noOfPartitions == 1)
			return new int[0];
		
		//for first partition
		if ((pos == 0)) {
			oneReturn[0] = 1;
			return oneReturn;
		}
		//for last partition
		if (pos == noOfPartitions - 1) {
			oneReturn[0] = noOfPartitions - 2;
			return oneReturn;
		}
		//for every other partition
		twoReturns[0] = pos - 1;
		twoReturns[1] = pos + 1;
		return twoReturns;
	}

}
