package appl.parallel.thread;

import java.nio.channels.Pipe.SinkChannel;
import java.rmi.RemoteException;
import java.util.concurrent.Callable;

import appl.parallel.ComputingResourceProperties;
import appl.parallel.client.RemoteEventHandler;
import appl.parallel.event.CommEvent;
import appl.parallel.event.CommEventSink;
import appl.parallel.event.TimeEvent;
import appl.parallel.event.TransferEvent;
import appl.parallel.event.CommEvent.CommType;

/**
 * A simple helper class for more easy Thread handling with anonymous classes.
 *
 * @see DataServerThread
 * @see ComputingResourceThread
 *
 * @author Dominik Appl
 */
public abstract class ExecutionThread implements Callable {

	protected final Object server;

	protected final Object argument;

	protected final CommType commType;

	protected final ComputingResourceProperties serverInfos;

	protected final CommEventSink eventSink;

	protected boolean disableTransferEvents=false;

	/**
	 * creates a new thread
	 *
	 * @param server
	 *            the server which is accessed with this thread
	 * @param serverInfos the serverinfos (used for event info)
	 * @param argument
	 *            an argument which can be used inside the anonymous class
	 * @param type the type of the communication
	 * @param sink events are submitted to this sink
	 * @param disableTransferEvents says whether to ignore transfer events (if the transfer volume is to low to be counted))
	 */
	public ExecutionThread(Object server,
			ComputingResourceProperties serverInfos, Object argument,
			CommType type, CommEventSink sink, boolean disableTransferEvents) {
		this.server = server;
		this.serverInfos = serverInfos;
		this.argument = argument;
		this.commType = type;
		this.eventSink = sink;
		this.disableTransferEvents = disableTransferEvents;
	}

	/**
	 * @return the server given the constructor (should be overwrited by a
	 *         subclass)
	 */
	protected Object getServer() {
		return server;
	}

	/**
	 * @return the argument given in the constructor as int value
	 */
	protected int getIntArgument() {
		return (Integer) argument;
	}

	/**
	 * @return the argument given in the constructor as object value
	 */
	protected Object getObjectArgument() {
		return argument;
	}

	/**
	 * @return the argument given in the constructor as array value
	 */
	protected Object[] getObjectArrayArgument() {
		return (Object[]) argument;
	}

	/**
	 * Sends standard {@link CommEvent CommEvents} to the {@link CommEventSink} given with the
	 * constructor. Executes the {@link #run()} method and returns its result.
	 *
	 * @return the result of {@link #run()}.
	 *
	 * @see java.util.concurrent.Callable#call()
	 */
	public Object call() throws Exception {
		long l = System.nanoTime();
		Object result = run();
		if (eventSink.isTimeMonitoringEnabled())
			fireTimeEvents((System.nanoTime() - l), result);
		if (eventSink.isTransferMonitoringEnabled())
			fireTransferEvent(result);
		return result;

	}

	/**
	 * Sends the time events for this execution. Method generates only the event of
	 * this execution. If you want more events (e.g. details from a remote
	 * execution) you can overwrite this method. This method is only called if time
	 * monitoring is enabled
	 *
	 * @param execTime
	 *            the time of THIS execution
	 * @param result the result of the computation (may be of use when overwriting)
	 */
	protected void fireTimeEvents(long execTime, Object result) {
		// localhost as String for performance reasons
		try {
			eventSink.fireRemoteEvent(new TimeEvent(execTime, "localhost",
					serverInfos.getName(), commType));
		} catch (RemoteException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/**
	 * The TransferEvents for this execution. Method generates only the event
	 * out of the size of {@link #getObjectArgument()}.If you want more events
	 * (e.g. details from a remote execution) you can overwrite this method.
	 * This method is only called if time monitoring is enabled.
	 *  Notice that this will be very slow for large Data!!
	 *
	 * @param result
	 *            the result of the execution (per default not needed)
	 */
	protected void fireTransferEvent(Object result) {
		if(!disableTransferEvents)
		// client as String for performance reasons
		try {
			eventSink.fireRemoteEvent(new TransferEvent("Xulu/V-Client",
					serverInfos.getName(), commType, argument));
		} catch (RemoteException e) {
			e.printStackTrace();
		}

	}

	/**
	 * This method should implement the real running code.
	 * @return the result you want to retrieve later
	 * @throws Exception the exceptions thrown by your code
	 */
	protected abstract Object run() throws Exception;

}
