package appl.util;

import java.awt.image.DataBuffer;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;

import schmitzm.data.WritableGrid;
import schmitzm.geotools.io.GeoImportUtil;

import org.geotools.gce.arcgrid.ArcGridRaster;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

/** 
 * This Class provides some simple mostly independent functions 
 * related to Geo-Rasters and {@link WritableGrid WritableGrids}.
 * 
 * @author Dominik Appl
 *
 */

public class RasterUtil {

		/**
		 * Extracts metadata out of a ArcGridASCII File without reading the
		 * whole Grid (only the Header is parsed).
		 * 
		 * @param src the ArcInfoASCII file
		 * @return the meta data
		 */
	public static RasterMetaData getRasterMetaData_from_ArcGridASCII_File(
			File src) {

		try {
			ArcGridRaster raster = new ArcGridRaster(new InputStreamReader(
					new FileInputStream(src)), false);
			raster.parseHeader();
			  CoordinateReferenceSystem crs = GeoImportUtil.determineProjection(src);
			return new RasterMetaData(DataBuffer.TYPE_FLOAT, raster.getNCols(),
					raster.getNRows(), 0, 0, raster.getXlCorner(), raster
							.getYlCorner(), raster.getCellSize(),crs);
		    
		   } catch (FileNotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			return new RasterMetaData(DataBuffer.TYPE_UNDEFINED, 0, 0, 0, 0, 0, 0,
					0,null);
	}

	/**
	 * Runs through every grid cell and checks if the two Grids are equal (based
	 * on the {@link RasterMetaData} of the first Grid. If the difference is
	 * smaller than 0.0001 only conversion Warning is given. Output to console
	 * may be enabled. Writes out warning to sysout, if metaData is not equal,
	 * but continues checking
	 * 
	 * @param a
	 *            1st grid
	 * @param b
	 *            2nd grid
	 * @param outputToConsole
	 *            if true the results and some infos are written on System.out
	 * @return true, if the grids match
	 */
	public static boolean checkEqual(WritableGrid a, WritableGrid b,
			boolean outputToConsole) {

		if (a == null && b == null)
			return true;
		if (a == null || b == null)
			return false;

		//for errors which are probably caused by rounding  
		float diffTolerance = 1.0e-5f;
		float maxdiff = 0; // maximum difference found
		int maxdiffX = 0, maxdiffY = 0; // coordinates of maxdiff

		RasterMetaData metaData = new RasterMetaData(a);
		if (!metaData.equals(new RasterMetaData(b))) {
			if (outputToConsole)
				System.out
						.println("Warning! Metadata of the two Grids is not equal!");
		}

		// run through the whole grid
		float aValue, bValue;
		float currentdiff;
		int counter = 0;
		for (int y = metaData.getMinY(); y < metaData.getHeight()
				- metaData.getMinY(); y++)
			for (int x = metaData.getMinX(); x < metaData.getWidth()
					- metaData.getMinY(); x++) {
				counter++;
				aValue = a.getRasterSampleAsFloat(x, y);
				bValue = b.getRasterSampleAsFloat(x, y);
				//check equal
				if (aValue == bValue)
					continue;
				if (Float.isNaN(aValue) && Float.isNaN(bValue))
					continue;

				//check for light differences which might indicate conversionErrors
				currentdiff = Math.abs(aValue - bValue);
				if (currentdiff < diffTolerance) {
					if (currentdiff > maxdiff) {
						maxdiff = currentdiff;
						maxdiffX = x;
						maxdiffY = y;
					}
					continue;
				}
				// else not equal:
				if (outputToConsole)
					System.out
							.println("Warning: Two Grids were not equal at pos ("
									+ x
									+ ","
									+ y
									+ ") --- first value "
									+ a.getRasterSampleAsFloat(x, y)
									+ " and second value "
									+ b.getRasterSampleAsFloat(x, y));
				return false;
			}
		if (outputToConsole && !(maxdiff == 0))
			System.out
					.println("The Gridvalues differ slightly. The maximum Difference is: "
							+ maxdiff
							+ " at ("
							+ maxdiffX
							+ ","
							+ maxdiffY
							+ "). Finished grid compare after "
							+ counter
							+ " checks");
		else if (outputToConsole)
			System.out
					.println("Grids found equal after " + counter + " checks");
		return true;
	}


	/**
	 * Copies all values from a {@link WritableGrid} into another
	 * {@link WritableGrid}. Of course the grids must have the same dimensions.
	 * 
	 * @param src
	 * @param target
	 * @throws UnsupportedOperationException
	 *             if the dimensions of the grids does not match or one of the
	 *             grids is null
	 */
	public static void copyInto(WritableGrid src, WritableGrid target){
		if(src==null || target==null)
			throw new UnsupportedOperationException("Error: One of the grids was null");
		//check if width and height are matching:
		if((src.getHeight()!=target.getHeight()) || (src.getWidth()!=target.getWidth()))
				throw new UnsupportedOperationException("The bounds of the two rasters don't not match! Copy failed!");
		for(int y=0; y < src.getHeight();y++)
			for(int x=0; x < src.getWidth(); x++)
				target.setRasterSample(src.getRasterSample(x,y), x,y);
		
	}
	
	/**
	 * Prints the grid to the console (useful for testing).
	 * 
	 * @param grid the grid to be printed
	 * @param spacing the space to be reserved for one cell-value
	 * @param precision the decimal places to be printed out
	 * @param name the name will be shown in the heading of the output
	 */
	public static void printGrid(WritableGrid grid, int spacing, int precision, String name){
		//generate format String
		System.out.println("****************************** " + name +  " ******************************");
		int width =grid.getWidth();
		int noOfCells = grid.getWidth() * grid.getHeight();
		Object[] cellarray = new Object[noOfCells];
		StringBuffer buffer = new StringBuffer(noOfCells * 9);
		String singleCellFormatString = "%" + spacing + "." + precision + "f ";
		buffer.append("%n ");
		for(int y=0; y < grid.getHeight();y++){
			for(int x=0; x < grid.getWidth(); x++){
				buffer.append(singleCellFormatString);
				cellarray[y*width+x]=grid.getRasterSampleAsFloat(x,y);
			}
			buffer.append("%n ");
		}
		
		System.out.printf(buffer.toString(), cellarray);
	}
	
	/**
	 * Prints the grid to the console (useful for testing).
	 * Spacing will be 5 and precision will be 2.
	 * 
	 * @param grid the grid to be printed
	 * @param name the name will be shown in the heading of the output
	 */
	public static void printGrid(WritableGrid grid, String name){
		printGrid(grid,5,2,name);
	}
	
	/**
	 * Prints the grid to the console (useful for testing). 
	 * Spacing will be 5 and precision will be 2.
	 * 
	 * @param grid the grid to be printed
	 */
	public static void printGrid(WritableGrid grid){
		printGrid(grid," ");
	}
}
