/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.appl;

import java.io.*;
import schmitzm.lang.OperationCanceledException;
import edu.bonn.xulu.*;
import edu.bonn.xulu.appl.*;

/**
 * Diese Klasse bildet eine Basis-Implementierung eines {@link ScriptInterpreter}.
 * Sie basiert darauf, dass die {@linkplain Command Befehle} zunaechst alle
 * aus einem InputStream eingelesen und danach hintereinander ausgefuehrt
 * werden. Implementierungen dieser Klasse brauchen lediglich noch das
 * Einlesen der Befehle aus dem Stream implementieren. Das Ausfuehren (mit
 * entsprechenden Status-Meldungen) uebernimmt diese Oberklasse.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class AbstractCommandInterpreter extends AbstractScriptInterpreter {
  /** Speichert die Anzahl an Fehlern, die fuer das Skript aufgetreten sind.*/
  protected int errorCount = 0;
  /** Speichert den Namen des Skript-Interpreters. */
  protected String name = "";
  /** Speichert die Xulu-Instanz fuer die das Skript ausgefuehrt wird. */
  protected XuluModellingPlatform appl = null;

  /**
   * Fuehrt ein Script aus.
   * @param in   Eingabe-Stream in dem das Skript zu finden ist.
   * @param appl Applikation, in der das Skript auszufuehren ist
   */
  public void execute(InputStream in, XuluModellingPlatform appl) {
    this.appl = appl;
    errorCount = 0;
    Command[] command = readCommands(in);
    System.out.println("\nStart executing script...");
    for (int i=0; i<command.length; i++) {
      System.out.print(">>>> "+command[i].getDescription()+"...");
      try {
        command[i].execute(appl);
        System.out.println(" done!");
      } catch (OperationCanceledException err) {
        // Wirft eine Operation eine solche Exception, wird die komplette
        // Script-Ausfuehung abgebrochen
        System.out.println(" failed!");
        System.out.println("Script canceled with "+errorCount+" errors...\n");
        return;
      } catch (Exception err) {
        errorCount++;
        System.out.println(" failed!");
        System.out.println(err);
      }

    }
    System.out.println("Script finished with "+errorCount+" errors...\n");
  }

  /**
   * Liest ein Skript ein und erzeugt daraus eine Reihe von Befehlen.
   * @param in Eingabe-Stream in dem das Skript zu finden ist
   */
  protected abstract Command[] readCommands(InputStream in);

  /**
   * Diese Interface beschreibt einen unter der Xulu-Plattform ausfuehrbaren
   * Befehl.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  protected static interface Command {
    /**
     * Fuehrt den Befehl aus.
     * @param appl Applikation unter der der Befehl auszufuehren ist
     * @exception java.lang.Exception wenn ein Fehler bei der Ausfuehrung aufritt
     */
    public void execute(XuluModellingPlatform appl) throws Exception;
    /**
     * Liefert eine Beschreibung, was der Befehl macht.
     */
    public String getDescription();

    /**
     * Prueft, ob alle Parameter korrekt gesetzt sind. Ggf. wird eine Exception
     * geworfen.
     */
    public void check() throws RuntimeException;
  }

}
