/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.appl;

import schmitzm.data.event.ObjectEvent;
import edu.bonn.xulu.appl.AbstractManager;
import edu.bonn.xulu.data.XuluObject;
import edu.bonn.xulu.model.XuluModel;
import edu.bonn.xulu.XuluModellingPlatform;

/**
 * Ein EventHandler stellt eine Zuordnung her zwischen
 * <ol>
 *   <li>einem Objekt</li>
 *   <li>einem Ereignis-Typ</li>
 *   <li>einer {@linkplain Handler Ereignis-Routine} (Handler)
 * </ol>
 * Loest das Objekt ein Ereignis vom spezifizierten Typ aus, so wird der
 * Handler ausgefuehrt.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class EventHandler implements AbstractManager.ContentType {
  /** Objekt, auf dessen Ereignisse gelauscht wird. */
  protected Object object = null;
  /** Ereignis, auf das reagiert wird. */
  protected Class<ObjectEvent> eventType = null;
  /** Handler, der ausgefuehrt wird, wenn {@link #object} ein Ereignis
   *  vom Typ {@link #eventType} ausloest. */
  protected Handler handler;

  // Fuer AbstractManager.ContentType
  private XuluModellingPlatform appl    = null;
  private boolean               visible = false;
  private String                title   = "";

  /**
   * Erzeugt einen neuen Event-Handler.
   * @param object    Objekt, das beobachtet wird
   * @param eventType Ereignis-Typ auf den reagiert werden soll (muss eine
   *                  Subklasse von {@link ObjectEvent} sein)
   * @param handler   Handler der ausgefuehrt werden soll
   * @param appl      Xulu-Applikation in der der Handler ausgefuehrt wird
   */
  public EventHandler(Object object, Class eventType, Handler handler, XuluModellingPlatform appl) {
    if ( !ObjectEvent.class.isAssignableFrom(eventType) )
      throw new IllegalArgumentException("Subclass of ObjectEvent expected for eventType!");
    this.object    = object;
    this.eventType = eventType;
    this.handler   = handler;
    this.appl      = appl;
    // Titel initialisieren
    if (object == null)
      title = "*";
    else if (object instanceof XuluObject)
      title = ((XuluObject)object).getDescription();
    else if (object instanceof XuluModel)
      title = ((XuluModel)object).getName();
    else
      title = "untitled";
  }

  /**
   * Erzeugt einen neuen Event-Handler. Dieser beobachtet kein bestimmtes
   * Objekt, sondern ist global.
   * @param eventType Ereignis-Typ auf den reagiert werden soll (muss eine
   *                  Subklasse von {@link ObjectEvent} sein)
   * @param handler   Handler der ausgefuehrt werden soll
   * @param appl      Xulu-Applikation in der der Handler ausgefuehrt wird
   */
  public EventHandler(Class eventType, Handler handler, XuluModellingPlatform appl) {
    this(null,eventType,handler,appl);
  }

  /**
   * Liefert das Objekt, das beobachtet wird.
   */
  public Object getObject() {
    return object;
  }

  /**
   * Liefert den Ereignistyp, auf den der Handler ausgefuehrt werden soll.
   */
  public Class<ObjectEvent> getEventType() {
    return eventType;
  }

  /**
   * Liefert den Handler, der ausgefuehrt werden soll, wenn das beobachtete Objekt
   * das Ereignis ausloest.
   */
  public Handler getHandler() {
    return handler;
  }

  /**
   * Prueft, ob es sich um einen Event-Handler handelt, der fuer <i>alle</i>
   * Objekte zustaendig ist.
   */
  public boolean isGlobal() {
    return object == null;
  }


  ///////////////////////////////////////////////////////////////////
  //////// Implementierung von AbstractManager.ContentType  /////////
  ///////////////////////////////////////////////////////////////////
  /**
   * Fordert die HandlerFactory auf, den Handler zu veraendern.
   */
  public void setVisible(boolean visible) {
    if ( visible && this.getHandler().getFactory()!=null ) {
      this.visible = true;
      this.getHandler().getFactory().modifyHandler( appl, getHandler() );
    }
    this.visible = false;

  }

  /**
   * Liefert <code>true</code> falls der Handler durch die Factory gerade
   * veraendert wird.
   */
  public boolean isVisible() {
    return visible;
  }

  /**
   * Liefert die Bezeichnung des Objekts fuer das der Event-Handler zustaendig
   * ist.
   */
  public String getTitle() {
    return title;
  }

  /**
   * Setzt die Bezeichnung der Komponente. Hiermit wird die Komponente
   * in einer grafischen Darstellung (Tabelle) identifiziert.
   */
  public void setTitle(String newTitle) {
    this.title = newTitle;
  }

  /**
   * Schliesst den EventHandler.
   */
  public void close() {

  }
}
