/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.appl;

import java.io.File;
import org.apache.log4j.Logger;

import appl.ext.XuluConfig;
import schmitzm.lang.LimitedVector;
import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.io.IODefinition;
import edu.bonn.xulu.io.ImportFactory;


/**
 * Diese Klasse stellt eine Liste der letzten in Xulu vorgenommenen Import-Operationen
 * dar.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class RecentImports extends LimitedVector<IODefinition.ImportDefinition> {
  private final Logger LOG = Logger.getLogger(this.getClass().getName());

  private static final String KEY_PREFIX  = "RecentImports.";
  private static final String KEY_MAXLOAD = KEY_PREFIX + "MAXLOAD";
  private static final String KEY_SIZE = KEY_PREFIX + "SIZE";
  private static final String KEY_FACTORY = ".FACTORY";
  private static final String KEY_SOURCE  = ".SOURCE";

  /**
   * Erzeugt eine neue leere Liste, in der letzten Importe abgelegt werden.
   * @param maxLoad maximale Groesse der Liste
   */
  public RecentImports(int maxLoad) {
    super(maxLoad);
  }

  /**
   * Laed die Liste der zuletzt vorgenommenen Import-Operationen aus den
   * Xulu-Properties.
   * @param appl Xulu-Instanz
   * @see XuluConfig
   */
  public void loadFromXuluProperties(XuluModellingPlatform appl) {
    XuluConfig config = XuluConfig.getXuluConfig();
    // MaxLoad-Wert einlesen
    int maxLoad = config.getIntProperty(KEY_MAXLOAD);
    if (maxLoad <= 0) {
      maxLoad = 10;
      config.setIntProperty(KEY_MAXLOAD,maxLoad);
    }
    clear();
    setMaxLoad(maxLoad);

    // Anzahl an Elementen einlesen
    int size = config.getIntProperty(KEY_SIZE);
    if (size <= 0)
      config.setIntProperty(KEY_SIZE,0);

    // Import-Definitionan einlesen
    for (int i=0; i<size; i++) {
      // Name der Factory ermitteln
      String facName = config.getProperty(KEY_PREFIX + i + KEY_FACTORY);
      ImportFactory[] impFac = appl.getRegistry().getImportFactories(facName);
      if ( impFac.length <= 0 ) {
        LOG.error("Warning [RecentImports]: No import factory with name '"+facName+"' registered in Xulu-Registry");
        continue;
      }
      // Source-Objekte einlesen
      String[] impSourceName = config.getMultiProperty(KEY_PREFIX + i + KEY_SOURCE);
      if ( impSourceName.length <= 0 ) {
        LOG.error("Warning [RecentImports]: No source objects found in Xulu properties for recent import "+i);
        continue;
      }

      // Import-Quellen erstellen (ES WIRD ANGENOMMEN, DASS ES SICH BEI
      // DEN QUELL-OBJEKTEN UM DATEIEN HANDELT!!)
      Object impSource = null;
      if ( Object[].class.isAssignableFrom( impFac[0].getImportSourceType() ) ) {
        impSource = new File[ impSourceName.length ];
        for (int j=0; j<impSourceName.length; j++)
          ((File[])impSource)[j] = new File(impSourceName[j]);
      } else
        impSource =  new File(impSourceName[0]);

      // Import-Definition erstellen
      add( new IODefinition.ImportDefinition(impSource,impFac[0],appl.getRegistry()) );
    }
  }

  /**
   * Speichert die Liste der zuletzt vorgenommenen Import-Operationen in den
   * Xulu-Properties.
   * @param appl Xulu-Instanz
   * @see XuluConfig
   */
  public void storeToXuluProperties(XuluModellingPlatform appl) {
    XuluConfig config = XuluConfig.getXuluConfig();
    config.removeAll(KEY_PREFIX);
    // MaxLoad-Wert speichern
    config.setProperty(KEY_MAXLOAD,String.valueOf(getMaxLoad()));
    // Anzahl an Elementen speichern
    config.setProperty(KEY_SIZE,String.valueOf(size()));
    // Import-Definitionan speichern
    for (int i=0; i<size(); i++) {
      IODefinition.ImportDefinition impDef = elementAt(i);
      // Name der Factory
      String facName = impDef.getFactory().getName();
      // Quell-Objekte in String konvertieren (ES WIRD ANGENOMMEN, DASS ES
      // SICH BEI DEN QUELL-OBJEKTEN UM DATEIEN HANDELT!!)
      Object   impSource = impDef.getIOObject();
      String[] impSourceName = null;
      if ( impSource instanceof File )
        impSourceName = new String[] { ((File)impSource).getAbsolutePath() };
      else if ( impSource instanceof File[] ) {
        impSourceName = new String[ ((File[])impSource).length ];
        for (int j=0; j<impSourceName.length; j++)
          impSourceName[j] = ((File[])impSource)[j].getAbsolutePath();
      } else {
        LOG.error("Warning [RecentImports]: Can not store recent import "+i+" in Xulu properties (import source is no file)");
        continue;
      }
      // Import-Definition speichern
      config.setProperty(KEY_PREFIX + i + KEY_FACTORY, facName);
      config.setMultiProperty(KEY_PREFIX + i + KEY_SOURCE, impSourceName);
    }
  }
}
