/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.data;

import schmitzm.data.property.PropertySet;
import schmitzm.data.property.Properties;
import schmitzm.data.property.Property;
import edu.bonn.xulu.io.AbstractFactory;
import edu.bonn.xulu.io.InstantiationFactory;

/**
 * Diese Klasse bildet die Basisimplementierung eines <u>dynamischen</u>
 * Xulu-Datenobjekts. Diesem koennen Eigenschaften
 * ({@link schmitzm.data.property.Property Propertys}) hinzugefuegt oder
 * entfernt werden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class DynamicXuluObject extends AbstractXuluObject {
  /**
   * Factory, die Standard-Instanzen von <code>DynamicXuluObject</code>
   * erzeugt.
   */
  public static final InstantiationFactory DEFAULT_FACTORY = new DefaultFactory();

  /**
   * Erzeugt ein dynamisches Xulu-Objekt.
   * @param desc    Beschreibung fuer den Datenpool
   * @param propSet Eigenschaften aus denen das Objekt zusammengesetzt ist
   */
  public DynamicXuluObject(String desc, PropertySet propSet) {
    super(desc);
    // Struktur kopieren
    // Bemerke: this.propSet = propSet geht nicht mehr, da Variable
    //                         da this.propSet als final deklariert ist!
    Property[] prop = propSet.getProperties();
    for (int i=0; i<prop.length; i++)
      this.propSet.addProperty(prop[i]);

  }

  /**
   * Erzeugt ein dynamisches Xulu-Objekt. Dieses besitzt noch keine
   * Eigenschaften.
   * @param desc    Beschreibung fuer den Datenpool
   */
  public DynamicXuluObject(String desc) {
    this(desc,new PropertySet());
  }

  /**
   * Erzeugt ein dynamisches Xulu-Objekt. Diesem wird eine Default-Beschreibung
   * fuer den Datenpool zugewiesen.
   * @param propSet Eigenschaften aus denen das Objekt zusammengesetzt ist
   */
  public DynamicXuluObject(PropertySet propSet) {
    this(null,propSet);
  }

  /**
   * Erzeugt ein dynamisches Xulu-Objekt. Dieses besitzt noch keine Eigenschaften
   * und ihm wird eine Default-Beschreibung fuer den Datenpool zugewiesen.
   */
  public DynamicXuluObject() {
    this(new PropertySet());
  }


  /**
   * Fuegt dem Xulu-Objekt eine Eigenschaft hinzu. Dabei wird auf Eindeutigkeit
   * des Namens geachtet!
   * @param prop Eigenschaft
   * @exception schmitzm.data.property.PropertyException falls das Xulu-Objekt
   *            bereits eine Eigenschaft dieses Namens besitzt
   */
  public void addProperty(Property prop) {
    propSet.addProperty(prop);
  }

  /**
   * Entfernt eine Eigenschaft aus dem Xulu-Objekt (sofern diese existiert).<br>
   * <b>Die entfernte Eigenschaft wird mittels {@link Property#dispose()}
   * vollstaendig zerstoert.</b>
   * @param name Name der Eigenschaft
   */
  public void removeProperty(String name) {
    Property p = propSet.removeProperty(name);
    if ( p != null ) p.dispose();
  }

  /**
   * Entfernt eine Eigenschaft aus der Menge.
   * Falls die Menge eine solche Eigenschaft nicht besitzt, passiert nichts.<br>
   * <b>Die entfernte Eigenschaft wird mittels {@link Property#dispose()}
   * vollstaendig zerstoert.</b>
   * @param prop zu entfernende Eigenschaft
   */
  public void removeProperty(Property prop) {
    removeProperty(prop.getName());
  }

  /**
   * Entfernt Eigenschaften aus der Menge. Falls die Menge eine dieser
   * Eigenschaften nicht enthaelt, passiert fuer die entsprechende Eigenschaft
   * nichts.<br>
   * <b>Alle entfernten Eigenschaften werden mittels {@link Property#dispose()}
   * vollstaendig zerstoert.</b>
   * @param prop zu entfernende Eigenschaft
   */
  public void removeProperties(Properties prop) {
    Property[] p = propSet.removeProperties(prop).getProperties();
    for (int i=0; i < p.length; i++)
      p[i].dispose();
  }

  ///////////////////////////////////////////////////////////////////
  // Implementierung von InstantiationFactory
  ///////////////////////////////////////////////////////////////////
  /**
   * Diese Klasse implementiert eine Standard-Instanz-Factory fuer
   * {@link DynamicXuluObject}.
   * @see edu.bonn.xulu.data.DynamicXuluObject#DEFAULT_FACTORY
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class DefaultFactory extends AbstractFactory implements InstantiationFactory {
    /**
     * Liefert <code>DynamicXuluObject.class</code> als die Klasse, die
     * von {@link #newInstance(boolean)} instanziiert wird.
     */
    public Class getInstanceType() {
      return DynamicXuluObject.class;
    }

    /**
     * Erzeugt eine neue leere Instanz von {@link DynamicXuluObject}.
     * @param dialog wird nicht beachtet, da kein Dialog zum erzeugen einer
     *               leeren Standard-Instanz benoetigt wird
     */
    public DynamicXuluObject newInstance(boolean dialog) {
      return new DynamicXuluObject();
    }

    /**
     * Erzeugt ein dynamisches Xulu-Objekt, das dieselben Eigenschaften hat,
     * wie die Vorlage. Der Inhalt der Eigenschaften wird jedoch nicht kopiert.
     * @param dialog siehe {@link #newInstance(boolean)}
     * @param sample Vorlage
     * @exception IllegalArgumentException falls es sich bei der Vorlage nicht
     *            um eine <code>Properties</code>-Instanz handelt
     */
    public DynamicXuluObject newInstance(boolean dialog, Object sample) {
      if ( sample == null )
        throw new IllegalArgumentException(this.getClass().getSimpleName().concat(" can not create an instance of ").concat(getInstanceType().getSimpleName()).concat(" from null") );
      if ( !Properties.class.isAssignableFrom(sample.getClass()) )
        throw new IllegalArgumentException(this.getClass().getSimpleName().concat(" can not create an instance of ").concat(getInstanceType().getSimpleName()).concat(" from ").concat(sample.getClass().getName()) );

      Property[] prop = ((Properties)sample).getProperties();
      DynamicXuluObject newObj = newInstance(dialog);
      for (int i=0; i<prop.length; i++)
        newObj.addProperty( prop[i].cloneStructure() );
      return newObj;
    }
  }
}
