/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.gui;

import java.awt.Container;
import java.awt.Color;
import javax.swing.BoxLayout;
import javax.swing.SpringLayout;
import java.util.Enumeration;
import java.util.Collection;
import java.util.Vector;

import schmitzm.swing.SpringUtilities;

import schmitzm.data.property.Properties;
import schmitzm.data.property.Property;
import schmitzm.data.event.ObjectListener;
import schmitzm.data.event.ObjectEvent;
import schmitzm.data.event.NameChangeEvent;
import schmitzm.swing.SelectionInputOption;

import edu.bonn.xulu.appl.DataPool;
import edu.bonn.xulu.data.XuluObject;
import edu.bonn.xulu.model.ModelContentManager;
import edu.bonn.xulu.model.ModelResource;


/**
 * Diese Klasse stellt eine Container dar, ueber den die im Datenpool befindlichen
 * Objekte den benoetigten Ressourcen eines Modells (siehe {@link ModelContentManager})
 * durch den Anwender zugeordnet werden koennen.<br>
 * Manuelle Eingaben (in Form eines Textfeldes) sind nicht moeglich. Fuer jede
 * Ressource gibt es ein Auswahlfeld, in dem ein (passendes) Xulu-Objekt aus
 * dem Datenpool oder eine seiner Propertys selektiert werden koennen.<br>
 * Die Ressourcen werden nach folgender Regel farbig markiert:
 * <ul>
 * <li>Eingabe-Ressource >> GRUEN</li>
 * <li>Ausgabe-Ressource >> BLAU</li>
 * <li>Ein- und Ausgabe-Ressource >> ROT</li>
 * <li>Temporaere Ressource >> ORANGE</li>
 * </ul>
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ModelContentManagerContainer extends Container {
  /** Speichert eine Referenz auf den ContentManager */
  protected ModelContentManager contManager = null;
  /** Speichert eine Referenz auf den Datenpool */
  protected DataPool dataPool = null;

  private SelectionInputOption[] inputOption = null;
  private ObjectListener         objectListener = null;

  /**
   * Erzeugt einen neuen Container.
   * @param contentManager ModelContentManager, dem Daten zugeordnet werden
   * @param dataPool       DatenPool, aus dem die Daten stammen
   */
  public ModelContentManagerContainer(ModelContentManager contentManager, DataPool dataPool) {
    super();
    this.contManager = contentManager;
    this.dataPool    = dataPool;
//    this.setLayout( new BoxLayout(this,BoxLayout.Y_AXIS) );
    this.setLayout( new SpringLayout() );
    // Auswahl-Optionen fuer Ressourcen erzeugen
    this.inputOption = new SelectionInputOption[contManager.getResourceCount()];
    for (int i=0; i<inputOption.length; i++) {
      inputOption[i] = new SelectionInputOption.Combo(
          "["+(i+1)+"] "+contManager.getResource(i).getDesc(),
          !contManager.getResource(i).isNullable()
      );
      // je nach Kategorie erhaelt die Ressource eine andere Farbe
      inputOption[i].getInputComponent().setBackground(
          getColorForResource( contManager.getResource(i) )
      );
      // InputOption hinzufuegen
      this.add(inputOption[i]);
    }
    // Komponenten in Form eines Grids anordnen
    int cols = (inputOption.length-1) / 9   + 1; // max. 9 Zeilen pro Spalte
    int rows = (inputOption.length-1) / cols + 1;
    SpringUtilities.makeCompactGrid(this,rows,cols,0,0,10,10);

    // Auswahl-Optionen fuer Ressourcen befuellen
    reorganize();

    // Listener auf den Datenpool, damit die Auswahlfelder automatisch
    // erweitert/reduziert werden, wenn Objekte in den Datenpool eingefuegt,
    // daraus entfernt werden, oder ihren Namen aendern
    objectListener = new ObjectListener() {
      public void performObjectEvent(ObjectEvent e) {
        if ( e instanceof DataPool.DataPoolChangeEvent ) {
          reorganize();
        }
      }
    };
    dataPool.addObjectListener( objectListener );
  }

  /**
   * Liefert die Farb-Codierung fuer eine Ressource. Sub-Klassen koennen diese
   * Methode ueberschreiben, um die Farb-Codierung zu aendern.
   */
  protected Color getColorForResource(ModelResource resource) {
    if ( resource.isCategory( ModelResource.CATEGORY_TEMP ) )
        return Color.ORANGE;
    if ( resource.isCategoryAll( ModelResource.CATEGORY_INPUT | ModelResource.CATEGORY_OUTPUT) )
        return Color.RED;
    if ( resource.isCategoryAll( ModelResource.CATEGORY_INPUT ) )
      return Color.GREEN;
    if ( resource.isCategoryAll( ModelResource.CATEGORY_OUTPUT ) )
      return Color.BLUE;
    return Color.WHITE;
  }
  /**
   * Zerstoert den Container.
   */
  public void dispose() {
    this.removeAll();
    dataPool.removeObjectListener(objectListener);
  }

  /**
   * Deaktiviert den Container und alle darin enthalten Optionen.
   */
  public void setEnabled(boolean enabled) {
    super.setEnabled(enabled);
    for (int i=0; i<this.getComponentCount(); i++)
      this.getComponent(i).setEnabled(enabled);
  }

  /**
   * Belegt ein Ressourcen-Auswahlfeld mit einem Objekt.
   * @param i          Nummer der Ressource
   * @param objectName Im Auswahlfeld angezeigter Objekt-Name
   */
  public void setResourceByName(int i, String objectName) {
    inputOption[i].setSelectedDisplayItem(objectName);
  }

  /**
   * Belegt ein Ressourcen-Auswahlfeld mit einem Objekt.
   * @param i      Nummer der Ressource
   * @param object Passendes Objekt aus dem Datenpool (muss in der Auswahl
   *               vorhanden sein!)
   */
  public void setResourceByObject(int i, Object object) {
    inputOption[i].setSelectedItem(object);
  }

  /**
   * Aktualisiert die Auswahlfelder.
   */
  public void reorganize() {
    Vector         objects = new Vector<Object>();
    Vector<String> names   = new Vector<String>();

    for (int i=0; i<inputOption.length; i++) {
      ModelResource resource = contManager.getResource(i);

      objects.clear();
      names.clear();
      XuluGUIUtil.determineObjectsFromDataPool(resource,dataPool,objects,names);

      inputOption[i].setSelectionObjects(objects.toArray(),names.toArray(new String[0]));
    }

    this.validate();
  }

  /**
   * Liefert alle in den Eingabe-Optionen ausgewaehlten Objekte des Datenpools.
   */
  public Object[] getSelection() {
    Object[] selection = new Object[contManager.getResourceCount()];

//    //===================================================================
//    // ZUM TESTEN DES CLUE-MODELLS
//    // Ressourcen automatisch zuordnen, wenn erste Option nicht zugeordnet ist
//    if ( contManager instanceof edu.bonn.xulu.plugin.model.ClueModelContentManager &&
////         inputOption != null &&
//         inputOption[0].getSelectedIndex() == -1) {
//      inputOption[ClueModelContentManager.MODELSTEPS].setSelectedItem(      dataPool.getObjects("Clue-Parameter")[0].getProperty("Step Count") );
//      inputOption[ClueModelContentManager.TOLERANCETYPE].setSelectedItem(   dataPool.getObjects("Clue-Parameter")[0].getProperty("Demand Tolerance Type") );
//      inputOption[ClueModelContentManager.TOLERANCESINGLE].setSelectedItem( dataPool.getObjects("Clue-Parameter")[0].getProperty("Single Demand Tolerance") );
//      inputOption[ClueModelContentManager.TOLERANCEAVERAGE].setSelectedItem(dataPool.getObjects("Clue-Parameter")[0].getProperty("Average Demand Tolerance") );
//      inputOption[ClueModelContentManager.CONVFACTQM].setSelectedItem(      dataPool.getObjects("Clue-Parameter")[0].getProperty("Demand Conversion") );
//      inputOption[ClueModelContentManager.BASESCENARIO].setSelectedItem(    dataPool.getObjects("Base LU Scenario")[0].getProperty("Grid") );
//      inputOption[ClueModelContentManager.LUHISTORY].setSelectedItem(       dataPool.getObjects("Base LU History")[0].getProperty("Grid") );
//      inputOption[ClueModelContentManager.LUCONV].setSelectedItem(          dataPool.getObjects("Clue-Parameter")[0].getProperty("LU Conversion Matrix") );
//      inputOption[ClueModelContentManager.ELASTICITY].setSelectedItem(      dataPool.getObjects("Clue-Parameter")[0].getProperty("Conversion Elasticity") );
//      inputOption[ClueModelContentManager.AREARESTR].setSelectedItem(       dataPool.getObjects("Area Restrictions")[0].getProperty("Grid") );
//      inputOption[ClueModelContentManager.DEMAND].setSelectedItem(          dataPool.getObjects("Clue-Parameter")[0].getProperty("Demand Scenario") );
//      inputOption[ClueModelContentManager.STATDF].setSelectedItem(          dataPool.getObjects("Static Driving Forces")[0] );
//      inputOption[ClueModelContentManager.DYNDFTYPES].setSelectedItem(      dataPool.getObjects("Clue-Parameter")[0].getProperty("Dynamic Driving Forces") );
//      inputOption[ClueModelContentManager.DYNDF].setSelectedItem(           dataPool.getObjects("Dynamic Driving Forces")[0] );
//      inputOption[ClueModelContentManager.REGRESSION].setSelectedItem(      dataPool.getObjects("Clue-Parameter")[0].getProperty("Regression Results") );
//      inputOption[ClueModelContentManager.NBHWEIGHTS].setSelectedItem(      dataPool.getObjects("Clue-Parameter")[0].getProperty("Neighborhood Weights") );
//      inputOption[ClueModelContentManager.NBHSETTINGS].setSelectedItem(     dataPool.getObjects("Clue-Parameter")[0].getProperty("Neighborhood Settings") );
//      inputOption[ClueModelContentManager.NBHREGRESSION].setSelectedItem(   dataPool.getObjects("Clue-Parameter")[0].getProperty("Neighborhood Regression") );
//      inputOption[ClueModelContentManager.ACTLU].setSelectedItem(           dataPool.getObjects("Out: Actual LU")[0].getProperty("Grid") );
//      inputOption[ClueModelContentManager.ACTLUHISTORY].setSelectedItem(    dataPool.getObjects("Out: Actual LU History")[0].getProperty("Grid") );
//      inputOption[ClueModelContentManager.LUPROB].setSelectedItem(          dataPool.getObjects("Out: LU Probabilities")[0] );
//      inputOption[ClueModelContentManager.NBHPROB].setSelectedItem(         dataPool.getObjects("Out: Neighborhood Probabilities")[0] );
//      inputOption[ClueModelContentManager.ITERVAR].setSelectedItem(         dataPool.getObjects("Clue-Parameter")[0].getProperty("Iteration Variables") );
//      inputOption[ClueModelContentManager.STEPRESULTS].setSelectedItem(     dataPool.getObjects("Out: Step Results")[0] );
//    }
    //===================================================================
    for (int i=0; i<contManager.getResourceCount(); i++)
      selection[i] = ((SelectionInputOption)getComponent(i)).getValue();
    return selection;
  }
}
