/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.gui;

import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.MutableTreeNode;

import schmitzm.swing.TreeSelectionDialog;
import schmitzm.swing.tree.EmptyNode;

import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.appl.DataPool;
import edu.bonn.xulu.appl.ModelControlManager;
import edu.bonn.xulu.appl.XuluConstants;
import edu.bonn.xulu.model.XuluModel;

// nur fuer Doku
import schmitzm.swing.MultipleOptionPane;

/**
 * Diese Klasse stellt eine Eingabe-Option fuer einen {@link MultipleOptionPane}-Dialog
 * dar, in der ein Daten-Objekt aus dem {@linkplain DataPool Xulu-Datenpool}
 * oder ein Modell aus dem {@linkplain ModelControlManager Xulu-Modell-Manager}
 * eingegeben werden kann. Hierzu besteht fuer den Anwender die Moeglichkeit,
 * ueber einen Browse-Dialog aus allen vorhandenen Datenpool-Objekten -und
 * Modellen auszuwaehlen.<br>
 * <br>
 * Zur Zeit ist die manuelle Eingabe eines Objekts in das Textfeld der Eingabe-Option
 * gesperrt (es kann nur gebrowsed werden). Die Methode {@link #convertFromString(String)}
 * muss noch implementiert werden.
 * @todo Manuelle Eingabe eines Objekts implementieren
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ObjectOrModelInputOption extends DataPoolInputOption {
  /** Speichert die Xulu-Applikation. */
  protected XuluModellingPlatform appl = null;

  /**
   * Erzeugt einen neu Input-Option
   * @param appl Xulu-Applikation die die Datenbasis (Xulu-Objekte und Modelle) bereitstellt
   * @param label Beschreibung der Eingabe-Option
   * @param inputNeeded bestimmt, ob eine Eingabe erforderlich (oder optional) ist
   * @param defValue Standardwert fuer die Eingabe-Option
   */
  public ObjectOrModelInputOption(XuluModellingPlatform appl, String label, boolean inputNeeded, Object defValue) {
    super(appl.getDataPool(), label, inputNeeded, defValue);
    this.appl = appl;

    // Eingabe-Feld sperren (bis Objekterkennung aus manuellem Textstring
    // implementiert ist!)
    inpComp.setEnabled(false);
  }

  /**
   * Fuehrt die Browse-Aktion aus, wenn der Anwender auf den entsprechenden
   * Button der Eingabe-Option klickt. Es wird ein Anwender-Dialog angezeigt,
   * der in einer Baumstruktur alle Datenpool-Objekte und Modelle zur Auswahl stellt.
   * @param defValue Standard-Auswahl fuer den Browse-Dialog (wird <b>nicht</b> genutzt!)
   * @return <code>null</code> wenn der Browse-Dialog abgebrochen wurde
   * @see TreeSelectionDialog
   */
  public Object performBrowse(Object defValue) {
    TreeSelectionDialog dialog = new TreeSelectionDialog(new ObjectOrModellChoseTreeModel(appl.getDataPool(), appl.getModelControlManager()));
    // Objekt von Anwender abfragen
    Object[] o = dialog.show(this,XuluConstants.XULUGUI_RES.getString("EventObject"),false);
    if ( o == null || o.length < 1 )
      return null; // Dialog abgebrochen
    return o[0];
  }

  /**
   * <b>Noch nicht implementiert</b><br>
   * Parst eine textuellen Angabe und liefert ein entsprechendes Objekt aus dem
   * Datenpool oder Modell-Manager.
   * @param objectStr String, der das Datenpool-Objekt (bzw. Modell) identifiziert.
   * @exception UnsupportedOperationException bei <b>jedem</b> Aufruf
   */
  public Object convertFromString(String objectStr) {
    throw new UnsupportedOperationException(XuluConstants.SWING_RES.getString("InvalidInputMess"));
  }

  /**
   * Liefert die textuelle Bezeichnung zu einem Datenpool-Objekt, bzw. Modell.
   * @param object Datenpool-Objekt oder Xulu-Modell
   */
  public String convertToString(Object object) {
    if ( object instanceof XuluModel )
      return ((XuluModel)object).getName();
    return super.convertToString(object);
  }

  /**
   * Diese Klasse stellt ein TreeModel dar, in dem es zwei Root-Knoten gibt.
   * Unterhalb des ersten werden alle Datenpool-Objekte aufgefuehrt und
   * unterhalb des anderen alle geladenen Modelle.
   */
  private static class ObjectOrModellChoseTreeModel extends DefaultTreeModel {
    public ObjectOrModellChoseTreeModel(DataPool dp, ModelControlManager mm) {
      super(new EmptyNode("ROOT"));
      ((EmptyNode)root).add( (MutableTreeNode)new DataPoolTreeModel(dp, false, XuluConstants.XULUGUI_RES.getString("Comp_DataPool")).getRoot());
      ((EmptyNode)root).add( (MutableTreeNode)new ModelManagerTreeModel(mm, XuluConstants.XULUGUI_RES.getString("Comp_Models")).getRoot());
    }
  }
}
