/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.gui;

import java.awt.Component;
import javax.swing.JOptionPane;
import java.util.ResourceBundle;
import java.io.File;


import schmitzm.lang.LangUtil;
import schmitzm.swing.ExceptionDialog;
import schmitzm.data.property.AccessViolationException;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.PropertyWriteAccess;

import edu.bonn.xulu.appl.XuluConstants;
import edu.bonn.xulu.data.XuluObject;
import edu.bonn.xulu.data.XuluDataException;
import schmitzm.swing.MultipleOptionPane;
import javax.swing.JCheckBox;


/**
 * Diese Klasse enthaelt statische Methoden zur Anzeige von GUI-Messages
 * ({@link javax.swing.JOptionPane}.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class XuluGUIMessages implements XuluConstants {
  /** Auswahl von {@link #checkFileOverwrite(Component,File,boolean)} und
   *  {@link #approveFileOverwrite(Component,String,boolean)} */
  public static enum FileOverrideOption {
    /** Dialog wurde abgebrochen. */
    CANCEL,
    /** Auswahl "Ueberschreiben". */
    OVERWRITE,
    /** Auswahl "Immer ueberschreiben". */
    OVERWRITE_ALL
  }

  /**
   * Prueft, ob eine Datei existiert und fordert gegenfalls zum Bestaetigen
   * des Ueberschreiben auf.
   * @param parent Uebergeordnete GUI-Komponente fuer die Meldung
   * @param file zu ueberpruefende Datei
   * @param owAllOption Flag, ob die Option "Alle ueberschreiben" zur Auswahlt
   *                    gestellt wird
   * @return {@link FileOverrideOption#CANCEL} falls der Dialog abgebrochen wurde;
   *         {@link FileOverrideOption#OVERWRITE} falls das Ueberschreiben
   *         bestaetigt wurde; {@link FileOverrideOption#OVERWRITE_ALL} falls
   *         das Ueberschreiben generell bestaetigt wurde
   */
  public static FileOverrideOption checkFileOverwrite(Component parent, File file, boolean owAllOption) {
    if ( !file.exists() )
      return FileOverrideOption.OVERWRITE;
    return approveFileOverwrite(parent, file.getName(), owAllOption);
  }

  /**
   * Zeigt einen Dialog an, in dem das Ueberschreiben einer Datei bestaetigt
   * werden muss.
   * @param parent Uebergeordnete GUI-Komponente fuer die Meldung
   * @param filename Bezeichnung der Datei
   * @param owAllOption Flag, ob die Option "Alle ueberschreiben" zur Auswahlt
   *                    gestellt wird
   * @return {@link FileOverrideOption#CANCEL} falls der Dialog abgebrochen wurde;
   *         {@link FileOverrideOption#OVERWRITE} falls das Ueberschreiben
   *         bestaetigt wurde; {@link FileOverrideOption#OVERWRITE_ALL} falls
   *         das Ueberschreiben generell bestaetigt wurde
   */
  public static FileOverrideOption approveFileOverwrite(Component parent, String filename, boolean owAllOption) {
    // Fehler-Dialog
    // --> Abfrage Abbrechen / Ueberschreiben / Immer ueberschreiben
    String cancelOption = SWING_RES.getString("Cancel");
    String overwriteOption = SWING_RES.getString("Overwrite");
    String overwriteAllOption = SWING_RES.getString("OverwriteAll");
    Object selection = MultipleOptionPane.showMultipleOptionDialog(
        parent,
        JOptionPane.WARNING_MESSAGE,
        SWING_RES.getString("Warning")+"...",
        SWING_RES.getString("FileExists")+": "+filename,
        owAllOption ? new String[] {cancelOption, overwriteOption, overwriteAllOption} : new String[] {cancelOption, overwriteOption},
        cancelOption
        );
    // Ueberschreiben
    if (overwriteOption.equals(selection))
      return FileOverrideOption.OVERWRITE;
    // Immer ueberschreiben
    if (overwriteAllOption.equals(selection))
      return FileOverrideOption.OVERWRITE_ALL;
    // Abbrechen
    return FileOverrideOption.CANCEL;
  }

  /**
   * Zeigt eine allgemeine Fehlermeldung (fuer alle "unbekannten" Fehler) an.
   * @param parent Uebergeordnete GUI-Komponente fuer die Meldung
   * @param err    Fehler der aufgetreten ist
   */
  public static void showError(Component parent, Exception err) {
    showError(parent,err,null);
  }

  /**
   * Zeigt eine allgemeine Fehlermeldung (fuer alle "unbekannten" Fehler) an.
   * @param parent Uebergeordnete GUI-Komponente fuer die Meldung
   * @param mess   Fehlerbeschreibung
   */
  public static void showError(Component parent, String mess) {
    showError(parent,null,mess);
  }

  /**
   * Zeigt eine allgemeine Fehlermeldung (fuer alle "unbekannten" Fehler) an.
   * @param parent  Uebergeordnete GUI-Komponente fuer die Meldung
   * @param err     Fehler der aufgetreten ist
   * @param addMess Zusaetzliche Fehlermeldung
   */
  public static void showError(Component parent, Exception err, String addMess) {
    String mess = (addMess==null) ? "" : addMess;
    if ( err!=null && err.getMessage()!=null && !err.getMessage().equals("") ) {
      if (!mess.equals(""))
        mess = mess.concat(":  ");
      mess = mess.concat(err.getMessage());
    }

    String title = (err!=null) ? err.getClass().getSimpleName() : SWING_RES.getString("Error");
    ExceptionDialog.show(parent,err,title,addMess);
  }

  /**
   * Zeigt eine Fehlermeldung fuer Zugriffsfehler an.
   * @param parent Uebergeordnete GUI-Komponente fuer die Meldung
   * @param err    Fehler der aufgetreten ist
   */
    public static void showAccessError(Component parent, AccessViolationException err) {
    String desc = XULUERR_RES.getString("Access");
    if ( err.getAccess() instanceof PropertyWriteAccess )
      desc = XULUERR_RES.getString("WriteAccess");
    if ( err.getAccess() instanceof PropertyReadAccess )
      desc = XULUERR_RES.getString("ReadAccess");

    ExceptionDialog.show(
      parent,
      err,
      XULUERR_RES.getString("AccessError"),
      LangUtil.stringConcatWithSep(" ",XULUERR_RES.getString("AccessErrMess_1"),desc,XULUERR_RES.getString("AccessErrMess_2"),"!")
    );

  }

  /**
   * Zeigt eine Fehlermeldung fuer Datentyp-Konflikte an.
   * @param parent Uebergeordnete GUI-Komponente fuer die Meldung
   * @param err    Fehler der aufgetreten ist
   */
  public static void showBaseTypeError(Component parent, NumberFormatException err) {
    ExceptionDialog.show(
      parent,
      err,
      XULUERR_RES.getString("DataTypeError"),
      XULUERR_RES.getString("TypeConvErrMess")+"!"
    );
  }

  /**
   * Zeigt eine Fehlermeldung fuer Instanziierungs-Konflikte an.
   * @param parent Uebergeordnete GUI-Komponente fuer die Meldung
   * @param err    Fehler der aufgetreten ist
   */
  public static void showInstantiationError(Component parent, InstantiationException err) {
    ExceptionDialog.show(
      parent,
      err,
      XULUERR_RES.getString("InstantiationError"),
      XULUERR_RES.getString("InstantiationErrMess")
    );
  }

  /**
   * Zeigt eine Fehlermeldung fuer Cast-Konflikte an.
   * @param parent   Uebergeordnete GUI-Komponente fuer die Meldung
   * @param err      Fehler der aufgetreten ist
   * @param expClass Erwartete Klasse
   */
  public static void showClassCastError(Component parent, ClassCastException err, Class expClass) {
    ExceptionDialog.show(
      parent,
      err,
      XULUERR_RES.getString("ClassCastError"),
      LangUtil.stringConcatWithSep(" ",XULUERR_RES.getString("InstExpMess_1"),expClass.getName(),XULUERR_RES.getString("InstExpMess_2"))
    );
  }

  /**
   * Zeigt eine Fehlermeldung fuer ClassNotFound-Konflikte an.
   * @param parent   Uebergeordnete GUI-Komponente fuer die Meldung
   * @param err      Fehler der aufgetreten ist
   * @param expClass Erwartete Klasse
   */
  public static void showClassNotFoundError(Component parent, ClassNotFoundException err, String expClass) {
    ExceptionDialog.show(
      parent,
      err,
      XULUERR_RES.getString("ClassError"),
      LangUtil.stringConcat(XULUERR_RES.getString("ClassNotFoundErrMess"),":  ",expClass)
    );
  }

  /**
   * Zeigt eine Fehlermeldung fuer Daten-Konflikte an.
   * @param parent   Uebergeordnete GUI-Komponente fuer die Meldung
   * @param err      Fehler der aufgetreten ist
   */
  public static void showDataError(Component parent, XuluDataException err) {
    String mess = "";
    XuluObject data = err.getXuluData();
    if ( data != null && !data.getDescription().trim().equals("") )
      mess = mess.concat(data.getDescription()).concat(":  ");
    mess = mess.concat(err.getMessage());
    ExceptionDialog.show(parent,err,XULUERR_RES.getString("DataError"),mess);
  }

  /**
   * Zeigt eine Info-Meldung an.
   * @param parent   Uebergeordnete GUI-Komponente fuer die Meldung
   * @param mess     Meldung
   */
  public static void showInfoMessage(Component parent, String mess) {
    JOptionPane.showMessageDialog(
      parent,
      mess,
      SWING_RES.getString("Information")+"...",
      JOptionPane.INFORMATION_MESSAGE
    );
  }

}
