/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.gui;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

import appl.ext.XuluConfig;

import java.util.Vector;

import schmitzm.lang.AlreadyHandledException;
import schmitzm.swing.MultipleOptionPane;
import schmitzm.swing.SelectionInputOption;
import schmitzm.swing.InputOption;
import schmitzm.swing.ExceptionDialog;
import schmitzm.swing.SwingUtil;
import schmitzm.swing.menu.ObjectMenuItem;
import schmitzm.data.event.ObjectListener;
import schmitzm.data.event.ObjectEvent;

import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.appl.ScriptInterpreter;
import edu.bonn.xulu.appl.XuluComponent;
import edu.bonn.xulu.appl.XuluComponentUtil;
import edu.bonn.xulu.appl.DataPool;
import edu.bonn.xulu.appl.EventManager;
import edu.bonn.xulu.appl.XuluRegistry;
import edu.bonn.xulu.appl.VisualisationManager;
import edu.bonn.xulu.appl.ModelControlManager;

import edu.bonn.xulu.gui.XuluGUIMessages;
import edu.bonn.xulu.gui.DataPoolFrame;
import edu.bonn.xulu.plugin.gui.DataPoolFrame_Tree;
import edu.bonn.xulu.plugin.gui.DataPoolFrame_BasicTable;
import edu.bonn.xulu.gui.StatusFrame;
import edu.bonn.xulu.plugin.gui.StatusFrame_BasicTextArea;
import edu.bonn.xulu.gui.RegistryFrame;
import edu.bonn.xulu.plugin.gui.RegistryFrame_Tree;
import edu.bonn.xulu.gui.AbstractManagerFrame;
import edu.bonn.xulu.plugin.gui.AbstractManagerFrame_BasicTable;
import edu.bonn.xulu.plugin.gui.EventManagerFrame_BasicTable;
import edu.bonn.xulu.plugin.gui.PluginManagerFrame_BasicTable;
import edu.bonn.xulu.gui.ModelControlFrame;

/**
 * Diese Klasse stellt das Hauptfenster der Xulu-Modelling-Platform dar.
 * Die einzelenen Komponenten werden in Unterfenstern dargestellt, welche
 * wahlweise angezeigt oder verborgen werden koennen:
 * <ul>
 * <li>Ein Fenster fuer den {@link DataPoolFrame Datenpool}</li>
 * <li>Ein Fenster fuer die {@link AbstractManagerFrame Visualisierungsfenster}</li>
 * <li>Ein Fenster fuer die {@link AbstractManagerFrame aktiven Modelle/Modellfenster}</li>
 * <li>Ein Fenster fuer den {@link AbstractManagerFrame Event-Manager}</li>
 * <li>Ein Fenster fuer die {@link RegistryFrame Registry}</li>
 * <li>Ein Fenster fuer die {@link StatusFrame Statusausgaben}</li>
 * </ul>
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class XuluMainFrame extends JFrame implements XuluComponent {
  //########## Konstanten fuer Labels ##########
  private final String VISTOOLS_DESC     = XULUGUI_RES.getString("Comp_Visualisation");
  private final String MODELS_DESC       = XULUGUI_RES.getString("Comp_Models");
  private final String EVENTMANAGER_DESC = XULUGUI_RES.getString("Comp_EventManager");
  private final String DATAPOOL_DESC     = XULUGUI_RES.getString("Comp_DataPool");
  private final String REGISTRY_DESC     = XULUGUI_RES.getString("Comp_Registry");
  private final String STATUS_DESC       = XULUGUI_RES.getString("Comp_Log");
  private final String PLUGINS_DESC      = XULUGUI_RES.getString("Comp_PluginManager");

  //########## Komponenten ##########
  private XuluModellingPlatform appl = null;
  private DataPool              dataPool = null;
  private XuluRegistry          registry = null;
  private VisualisationManager  visManager = null;
  private ModelControlManager   modelManager = null;
  private EventManager          eventManager = null;
  private XuluMainFrame         mainFrame = this;

  //########## GUI ##########
  private JDesktopPane               desktopPane = new JDesktopPane();
  private LayoutManager              layout      = new GridBagLayout();
  private DataPoolFrame              dataPoolFrame;
  private AbstractManagerFrame       visManagerFrame;
  private AbstractManagerFrame       modelManagerFrame;
  private AbstractManagerFrame       eventManagerFrame;
  private RegistryFrame              registryFrame;
  private XuluInternalFrame          pluginManagerFrame;
  private StatusFrame                statusFrame;
  private XuluLanguageManagerFrame   languageDlg = new XuluLanguageManagerFrame(this,true);
  private XuluMainFrameInfo          infoDlg     = new XuluMainFrameInfo(this,true);

  private Cursor                     waitCursor = new Cursor(Cursor.WAIT_CURSOR);
  private Cursor                     activeCursor = new Cursor(Cursor.DEFAULT_CURSOR);

  /**
   * Einheitlicher <@linkplain JFileChooser FileChooser> fuer die Xulu-Applikation.
   * Dieser kann in allen Untermodulen verwendet werden. Somit kann auf ein
   * einheitliches Ausgangsverzeichnis zurueckgegriffen werden.<br>
   * <b>Bemerke:</b><br>
   * Der File-Chooser ueberschreibt die Methoden <code>showOpenDialog(..)</code>
   * und <code>showSaveDialog(..)</code>, so dass die Datei-Auswahl bei
   * jedem Anzeigen initialisiert wird!
   */
  public final JFileChooser FILE_CHOOSER = new JFileChooser() {
    public int showOpenDialog(Component parent) {
      this.setSelectedFile( new java.io.File("") );
      this.setSelectedFiles( new java.io.File[0] );
      return super.showOpenDialog(parent);
    }
    public int showSaveDialog(Component parent) {
      this.setSelectedFile( new java.io.File("") );
      this.setSelectedFiles( new java.io.File[0] );
      return super.showSaveDialog(parent);
    }
  };

  //########## Menue-Struktur ##########
  private JMenuBar  menuBar = new JMenuBar();
  private JMenu     menuFile = new JMenu();
  private JMenuItem menuFile_Exit = new JMenuItem();
  private JMenu     menuView = new JMenu();
  private JMenuItem menuView_DataPool      = new JCheckBoxMenuItem();
  private JMenuItem menuView_Registry      = new JCheckBoxMenuItem();
  private JMenuItem menuView_Status        = new JCheckBoxMenuItem();
  private JMenuItem menuView_VisManager    = new JCheckBoxMenuItem();
  private JMenuItem menuView_ModelManager  = new JCheckBoxMenuItem();
  private JMenuItem menuView_EventManager  = new JCheckBoxMenuItem();
  private JMenuItem menuView_PluginManager = new JCheckBoxMenuItem();
  private JMenuItem menuView_Arange        = new JMenuItem();
  private JMenuItem menuView_ArangeAuto    = new JCheckBoxMenuItem();
  private JMenu     menuAdvanced           = new JMenu();
  private JMenuItem menuAdvanced_Languages = new JMenuItem();
  private JMenu     menuHelp               = new JMenu();
  private JMenuItem menuHelp_About         = new JMenuItem();
  private JMenu     menuScript             = new JMenu();
  private JMenu     menuScript_Recent      = new JMenu();
  private JMenu     menuModel = new JMenu();
  private JMenu     menuModel_New = new JMenu();
  private JMenuItem menuModel_New_Model = new JMenuItem();
  private JMenuItem menuModel_New_DynModel = new JMenuItem();
  private JMenuItem menuModel_Reload = new JMenuItem();

  /**
   * Erzeugt ein neues Haupt-Fenster des Xulu-Applikation.
   * @param appl Xulu-Applikation fuer die dieses Fenster die GUI bildet
   */
  public XuluMainFrame(XuluModellingPlatform appl) {
    this.appl         = appl;
    this.dataPool     = appl.getDataPool();
    this.registry     = appl.getRegistry();
    this.modelManager = appl.getModelControlManager();
    this.FILE_CHOOSER.setCurrentDirectory( appl.getWorkingDirectory() );

    enableEvents(AWTEvent.WINDOW_EVENT_MASK);
    try {
      guiInit();    // GUI initialisieren
      menuInit();   // Menue initialisieren
      menuUpdate(); // dynamische Menue-Eintraege erzeugen
      popupMenuInit(); // Popup-Menues initialisieren
    }
    catch(Exception e) {
      ExceptionDialog.show(e);
    }
  }

  /**
   * Zerstoert das Fenster. Ruft <code>super.dispose()</code> auf.
   */
  public void dispose() {
    super.dispose();
    appl = null;
  }

  /**
   * Prueft, ob das Fenster zerstoert ist.
   */
  public boolean isDisposed() {
    return appl == null;
  }

  /**
   * Liefert das {@linkplain DataPool Datenpool}-Fenster der Xulu-Anwendung.
   */
  public DataPoolFrame getDataPoolFrame() {
    return this.dataPoolFrame;
  }

  /**
   * Liefert das Fenster der Xulu-Anwendung, in dem die
   * {@linkplain VisualisationManager Visualisierungstools} verwaltet werden.
   */
  public AbstractManagerFrame getVisualisationManagerFrame() {
    return this.visManagerFrame;
  }

  /**
   * Liefert das Fenster der Xulu-Anwendung, in dem die Modelle / Modellfenster
   * verwaltet werden.
   * @see ModelControlManager
   */
  public AbstractManagerFrame getModelManagerFrame() {
    return this.modelManagerFrame;
  }

  /**
   * Liefert das {@linkplain EventManager EventManager}-Fenster der Xulu-Anwendung.
   */
  public AbstractManagerFrame getEventManagerFrame() {
    return this.eventManagerFrame;
  }

  /**
   * Liefert das Fenster der Xulu-Anwendung, in dem die {@linkplain XuluRegistry Xulu-Registry}
   * verwaltet wird.
   */
  public RegistryFrame getRegistryFrame() {
    return this.registryFrame;
  }

  /**
   * Liefert das Fenster der Xulu-Anwendung, in dem die Xulu-Plugins
   * verwaltet wird.
   */
  public XuluInternalFrame getPluginManagerFrame() {
    return this.pluginManagerFrame;
  }

  /**
   * Liefert das Status-Fenster der Xulu-Anwendung.
   */
  public StatusFrame getStatusFrame() {
    return this.statusFrame;
  }

  /**
   * Initialisierung der GUI.
   */
  private void guiInit() throws Exception  {
    //setIconImage(Toolkit.getDefaultToolkit().createImage(XuluMainFrame.class.getResource("[Ihr Symbol]")));
    this.setIconImage(XuluComponent.XULU_ICON.getImage());
    this.setSize(new Dimension(800, 600));
    this.setTitle(XULUGUI_RES.getString("XuluTitle"));
//    this.setContentPane( new JScrollPane(desktopPane) );
    this.layout = new GridBagLayout();
    this.setContentPane( desktopPane );
    this.setJMenuBar(menuBar);
    desktopPane.setLayout( null );

    // Listener auf Registry, um Menues anzupassen
    registry.addObjectListener( new ObjectListener() {
      public void performObjectEvent(ObjectEvent e) {
        if ( !e.getSource().contains( registry ) )
          return;
        // dynamische Menues neu organisieren
        menuUpdate();
      }
    });

    // Frame hinzufuegen
//    dataPoolFrame     = (DataPoolFrame)initInternalFrame( new DataPoolFrame_BasicTable(appl), true, getWidth()/3, 2*(getHeight()-50)/3, menuView_DataPool );
    dataPoolFrame      = (DataPoolFrame)initInternalFrame( new DataPoolFrame_Tree(appl), true, getWidth()/3, 2*(getHeight()-50)/3, menuView_DataPool );
    visManagerFrame    = (AbstractManagerFrame)initInternalFrame( new AbstractManagerFrame_BasicTable(VISTOOLS_DESC,appl.getVisualisationManager()), true, getWidth()/3, 2*(getHeight()-50)/3, menuView_VisManager );
    modelManagerFrame  = (AbstractManagerFrame)initInternalFrame( new AbstractManagerFrame_BasicTable(MODELS_DESC, new AbstractManagerFrame_BasicTable.AbstractManagerTableModel(appl.getModelControlManager()) {
                                                                         // TableModel fuer Modell-Fenster soll als Typ nicht
                                                                         // die Klasse des Modell-Fensters, sondern den Typ
                                                                         // des Modells anzeigen
                                                                         public Object getValueAt(int row, int col) {
                                                                           if ( col == 1 )
                                                                             return ((ModelControlFrame)manager.get(row)).getModel().getClass().getName();
                                                                           return super.getValueAt(row,col);
                                                                         }
                                                                  }), true, getWidth()/3, 2*(getHeight()-50)/3, menuView_ModelManager );
    eventManagerFrame  = (AbstractManagerFrame)initInternalFrame( new EventManagerFrame_BasicTable(appl,EVENTMANAGER_DESC), true, getWidth()/3, 2*(getHeight()-50)/3, menuView_EventManager );
    registryFrame      = (RegistryFrame)initInternalFrame( new RegistryFrame_Tree(registry), false, getWidth()/3, 2*(getHeight()-50)/3, menuView_Registry );
    statusFrame        = (StatusFrame)initInternalFrame( new StatusFrame_BasicTextArea(appl), true, getWidth(), getHeight()/3,menuView_Status);
    pluginManagerFrame = (XuluInternalFrame)initInternalFrame( new PluginManagerFrame_BasicTable(PLUGINS_DESC,appl), false, getWidth()/3, 2*(getHeight()-50)/3, menuView_PluginManager );

    // (SK) Die Standard-Ausgaben wird immer mit dem Statusfenster verbunden,
    // ausser wenn ein Eintrag "Console.Logging" in der XuluProperties Datei existiert,
    // welcher nicht "false" lautet.
    XuluConfig config = XuluConfig.getXuluConfig();
    String consoleLogging = config.getProperty("Console.Logging");
    if (consoleLogging == null) {
    	consoleLogging = "false";
    	config.setProperty("Console.Logging",consoleLogging);
    }
    if 	((!consoleLogging.toLowerCase().equals("true"))
    	   && (!consoleLogging.equals("1"))
    	) {
    	System.setOut( statusFrame.out );
    	System.setErr( statusFrame.out );
    }
  }

  /**
   * Initialisiert einen internen Frame und fuegt ihn dem DesktopPane hinzu.
   * @param frame  neue Frame
   * @param width  Breite des Frames
   * @param height Hoehe des Frames
   * @param connectedMenu MenuItem, mit dem der Frame un/sichtbar gemacht wird
   * @return der uebergebene Frame, damit Zuweiseung "schoener" moeglich
   */
  private JInternalFrame initInternalFrame(JInternalFrame frame, boolean visible, int width, int height, final JMenuItem connectedMenu) {
    frame.setSize(new Dimension(width,height));
    frame.setClosable(true);
    frame.setVisible(visible);
    frame.setDefaultCloseOperation( JInternalFrame.HIDE_ON_CLOSE );
    frame.addComponentListener( new ComponentAdapter() {
      public void componentHidden(ComponentEvent e) {
        if ( connectedMenu != null )
          connectedMenu.setSelected(false);
      }
    });
    desktopPane.add( frame );
    return frame;
  }

  /**
   * Ordnet die Fenster (in einer festen Anordnung) an.
   */
  public void arrangeInnerFrames() {
    final int COL_COUNT = 3;
    final int ROW_COUNT = 4;

    // Alle Fenster entfernen und dabei alle "nicht-standard" Fenster merken (die
    // u.U. von Plugins erzeugt wurden). Diese duerfen nicht verloren gehen!
//    Vector<JInternalFrame> others = new Vector<JInternalFrame>(); // Probleme mit Stefans PLAF; da ist eine Komponente dabei, die nicht JInternalFrame ist!
    Vector<Component> others = new Vector<Component>();
    for ( ; desktopPane.getComponentCount() > 0; ) {
      Component frame = (Component)desktopPane.getComponent(0);
      if ( frame != dataPoolFrame &&
           frame != visManagerFrame &&
           frame != modelManagerFrame &&
           frame != eventManagerFrame &&
           frame != registryFrame &&
           frame != statusFrame &&
           frame != pluginManagerFrame )
        others.add( frame );
      desktopPane.remove(frame);
    }

    desktopPane.setLayout( layout );
    desktopPane.add( dataPoolFrame, new GridBagConstraints(0, 0, 1, 3, 0.3, 0.6,
        GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(0,0,0,0),0,0) );
    desktopPane.add( visManagerFrame, new GridBagConstraints(1, 0, 1, 1, 0.3, 0.2,
        GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(0,0,0,0),0,0) );
    desktopPane.add( modelManagerFrame, new GridBagConstraints(1, 1, 1, 1, 0.3, 0.2,
        GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(0,0,0,0),0,0) );
    desktopPane.add( eventManagerFrame, new GridBagConstraints(1, 2, 1, 1, 0.3, 0.2,
        GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(0,0,0,0),0,0) );
    desktopPane.add( registryFrame, new GridBagConstraints(2, 1, 1, 2, 0.4, 0.3,
        GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(0,0,0,0),0,0) );
    desktopPane.add( statusFrame, new GridBagConstraints(0, ROW_COUNT-1, COL_COUNT, 1, 1.0, 0.4,
        GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(0,0,0,0),0,0) );
    desktopPane.add( pluginManagerFrame, new GridBagConstraints(2, 0, 1, 1, 0.4, 0.3,
        GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(0,0,0,0),0,0) );
    desktopPane.doLayout();
    if ( !menuView_ArangeAuto.isSelected() )
      desktopPane.setLayout( null );

    // "nicht-standard" Fenster wieder einfuegen
    for ( int i=0; i<others.size(); i++ ) {
      desktopPane.add(others.elementAt(i), new GridBagConstraints());
      if ( others.elementAt(i) instanceof JInternalFrame )
        ((JInternalFrame)others.elementAt(i)).toFront();
    }
  }

  /**
   * Initialisierung der Menues.
   */
  private void menuInit() {
    // Menue "Datei"
    menuBar.add(menuFile);
    menuFile.setText( XULUGUI_RES.getString("Menu_File") );
    menuFile.setMnemonic('d');
//    menuFile.addSeparator();
    menuFile.add(menuFile_Exit);
    menuFile_Exit.setText( XULUGUI_RES.getString("Menu_File_Exit") );
    menuFile_Exit.addActionListener(new ActionListener()  {
      public void actionPerformed(ActionEvent e) {
        menuFile_Exit_actionPerformed(e);
      }
    });

    // Menue "Modell"
    menuBar.add(menuModel);
    menuModel.setText( XULUGUI_RES.getString("Menu_Model") );
    menuModel.setMnemonic('m');
    menuModel.add(menuModel_New);
    menuModel.addSeparator();
    menuModel.add(menuModel_Reload);
    menuModel_Reload.setText( XULUGUI_RES.getString("Menu_Model_Reload") );
    menuModel_Reload.addActionListener(new ActionListener()  {
      public void actionPerformed(ActionEvent e) {
        menuModel_Reload_actionPerformed(e);
      }
    });

    // Menue "Modell -> Neu"
    menuModel_New.setText( XULUGUI_RES.getString("Menu_Model_New") );
    menuModel_New.add(menuModel_New_Model);
    menuModel_New_Model.setText( XULUGUI_RES.getString("Model") );
    menuModel_New_Model.addActionListener(new ActionListener()  {
      public void actionPerformed(ActionEvent e) {
        menuModel_New_Model_actionPerformed(e);
      }
    });
//    menuModel_New.add(menuModel_New_DynModel);
    menuModel_New_DynModel.setText( XULUGUI_RES.getString("DynModel") );
    menuModel_New_DynModel.addActionListener(new ActionListener()  {
      public void actionPerformed(ActionEvent e) {
        menuModel_New_DynamicModel_actionPerformed(e);
      }
    });

    // Menue "Ansicht"
    menuBar.add(menuView);
    menuView.setText( XULUGUI_RES.getString("Menu_View") );
    menuView.setMnemonic('a');
    menuView.add(menuView_DataPool);
    menuView_DataPool.setText( DATAPOOL_DESC );
    menuView_DataPool.setSelected( dataPoolFrame.isVisible() );
    menuView_DataPool.addActionListener(new ActionListener()  {
      public void actionPerformed(ActionEvent e) {
        menuView_DataPool_actionPerformed(e);
      }
    });
    menuView.add(menuView_VisManager);
    menuView_VisManager.setText( VISTOOLS_DESC );
    menuView_VisManager.setSelected( visManagerFrame.isVisible() );
    menuView_VisManager.addActionListener(new ActionListener()  {
      public void actionPerformed(ActionEvent e) {
        menuView_VisManager_actionPerformed(e);
      }
    });
    menuView.add(menuView_ModelManager);
    menuView_ModelManager.setText( MODELS_DESC );
    menuView_ModelManager.setSelected( modelManagerFrame.isVisible() );
    menuView_ModelManager.addActionListener(new ActionListener()  {
      public void actionPerformed(ActionEvent e) {
        menuView_ModelManager_actionPerformed(e);
      }
    });
    menuView.add(menuView_EventManager);
    menuView_EventManager.setText( EVENTMANAGER_DESC );
    menuView_EventManager.setSelected( eventManagerFrame.isVisible() );
    menuView_EventManager.addActionListener(new ActionListener()  {
      public void actionPerformed(ActionEvent e) {
        menuView_EventManager_actionPerformed(e);
      }
    });
    menuView.add(menuView_Registry);
    menuView_Registry.setText( REGISTRY_DESC );
    menuView_Registry.setSelected( registryFrame.isVisible() );
    menuView_Registry.addActionListener(new ActionListener()  {
      public void actionPerformed(ActionEvent e) {
        menuView_Registry_actionPerformed(e);
      }
    });
    menuView.add(menuView_PluginManager);
    menuView_PluginManager.setText( PLUGINS_DESC );
    menuView_PluginManager.setSelected( pluginManagerFrame.isVisible() );
    menuView_PluginManager.addActionListener(new ActionListener()  {
      public void actionPerformed(ActionEvent e) {
        menuView_PluginManager_actionPerformed(e);
      }
    });
    menuView.add(menuView_Status);
    menuView_Status.setText( STATUS_DESC );
    menuView_Status.setSelected( statusFrame.isVisible() );
    menuView_Status.addActionListener(new ActionListener()  {
      public void actionPerformed(ActionEvent e) {
        menuView_Status_actionPerformed(e);
      }
    });
    menuView.addSeparator();
    menuView.add(menuView_Arange);
    menuView_Arange.setText( XULUGUI_RES.getString("Menu_View_Arrange") );
    menuView_Arange.addActionListener(new ActionListener()  {
      public void actionPerformed(ActionEvent e) {
        menuView_Arange_actionPerformed(e);
      }
    });
    menuView.add(menuView_ArangeAuto);
    menuView_ArangeAuto.setText( XULUGUI_RES.getString("Menu_View_AutoArrange") );
    menuView_ArangeAuto.addActionListener(new ActionListener()  {
      public void actionPerformed(ActionEvent e) {
        menuView_ArangeAuto_actionPerformed(e);
      }
    });

    // Menue "Skript"
    menuBar.add(menuScript);
    menuScript.setText( XULUGUI_RES.getString("Menu_Scripts") );
    menuScript.setMnemonic('s');

    // Menue "Skript -> Skript erneut ausfuehren"
    menuScript_Recent.setText( XULUGUI_RES.getString("Menu_RecentScripts") );

    // menu advanced
    menuAdvanced.setText( XULUGUI_RES.getString("Menu_Advanced") );
    menuAdvanced.setMnemonic('a');
    menuBar.add(menuAdvanced);
    menuAdvanced.add(menuAdvanced_Languages);
    menuAdvanced_Languages.setText( XULUGUI_RES.getString("Menu_Advanced_Languages") );
    menuAdvanced_Languages.addActionListener(new ActionListener()  {
        public void actionPerformed(ActionEvent e) {
        menuAdvanced_Languages_actionPerformed(e);
      }
    });

    // Menue "Hilfe"
    menuHelp.setText( XULUGUI_RES.getString("Menu_Help") );
    menuHelp.setMnemonic('h');
    menuBar.add(menuHelp);
    menuHelp.add(menuHelp_About);
    menuHelp_About.setText( XULUGUI_RES.getString("Menu_Help_Info") );
    menuHelp_About.addActionListener(new ActionListener()  {
       	public void actionPerformed(ActionEvent e) {
        menuHelp_About_actionPerformed(e);
      }
    });

  }

  /**
   * Nachtraegliches Update der Menues.
   */
  private void menuUpdate() {
    // Menue "Datei"
    // ...

    // Menue "Datei -> Neu"
    // ...

    // Menue "Ansicht"
    // ...

    // Menue "Skript"
    menuScript.removeAll();
    menuScript.add(menuScript_Recent);
    menuScript.addSeparator();
    ScriptInterpreter[] si = registry.getScriptInterpreter();
    for (int i=0; i<si.length; i++) {
      String name = si[i].getClass().getSimpleName();
      if ( si[i].getName() != null && !si[i].getName().trim().equals("") )
        name = si[i].getName();
      JMenuItem item = new ObjectMenuItem(name,si[i],null);
      item.addActionListener(new ActionListener()  {
        public void actionPerformed(ActionEvent e) {
          menuScript_actionPerformed(e);
        }
      });
      menuScript.add(item);
    }

    // Menue "Skript -> Erneut ausfuehren"
    menuScript_Recent.removeAll();
    for (ScriptInterpreter.ScriptExecutionDefinition sed : registry.getRecentScripts()) {
      JMenuItem item = new ObjectMenuItem(sed.getName(),sed,null);
      item.addActionListener(new ActionListener()  {
        public void actionPerformed(ActionEvent e) {
          menuRecentScript_actionPerformed(e);
        }
      });
      menuScript_Recent.add(item,0);
    }


    // Menue "Hilfe"
    // ...
  }

  /**
   * Initialisierung der Popup-Menues.
   */
  private void popupMenuInit() {
    // Menue fuer Datenpool
  }

  /**
   * �berschrieben, so dass eine Beendigung beim Schlie�en des Fensters moeglich ist.
   */
  protected void processWindowEvent(WindowEvent e) {
    super.processWindowEvent(e);
    if (e.getID() == WindowEvent.WINDOW_CLOSING) {
      menuFile_Exit_actionPerformed(null);
    }
  }

  ///////////////////////////////////////////////////////////////////
  // Menue-Aktionen
  ///////////////////////////////////////////////////////////////////
  /**
   * Aktion Datei | Beenden durchfuehren.
   */
  public void menuFile_Exit_actionPerformed(ActionEvent e) {
    appl.dispose();
  }

  /**
   * Aktion Modell | Neu | Modell durchfuehren.
   */
  public void menuModel_New_Model_actionPerformed(ActionEvent e) {
    XuluGUIUtil.createNewModel(this,appl);
  }

  /**
   * Aktion Modell | Neu | Dynamisches Modell durchfuehren.
   */
  public void menuModel_New_DynamicModel_actionPerformed(ActionEvent e) {
    XuluGUIUtil.createNewDynamicModel(this,appl);
  }

  /**
   * Aktion Modell | Neu laden durchfuehren.
   */
  public void menuModel_Reload_actionPerformed(ActionEvent e) {
    XuluGUIUtil.reloadModelClass(appl);
  }

  /**
   * Aktion Ansicht | Daten-Pool durchfuehren.
   */
  public void menuView_DataPool_actionPerformed(ActionEvent e) {
    dataPoolFrame.setVisible( menuView_DataPool.isSelected() );
  }

  /**
   * Aktion Ansicht | Visualisierungstools durchfuehren.
   */
  public void menuView_VisManager_actionPerformed(ActionEvent e) {
    visManagerFrame.setVisible( menuView_VisManager.isSelected() );
  }

  /**
   * Aktion Ansicht | Aktive Modelle durchfuehren.
   */
  public void menuView_ModelManager_actionPerformed(ActionEvent e) {
    modelManagerFrame.setVisible( menuView_ModelManager.isSelected() );
  }

  /**
   * Aktion Ansicht | Event-Manager durchfuehren.
   */
  public void menuView_EventManager_actionPerformed(ActionEvent e) {
    eventManagerFrame.setVisible( menuView_EventManager.isSelected() );
  }

  /**
   * Aktion Ansicht | Registrierung durchfuehren.
   */
  public void menuView_Registry_actionPerformed(ActionEvent e) {
    registryFrame.setVisible( menuView_Registry.isSelected() );
  }

  /**
   * Aktion Ansicht | Plugin-Manager durchfuehren.
   */
  public void menuView_PluginManager_actionPerformed(ActionEvent e) {
    pluginManagerFrame.setVisible( menuView_PluginManager.isSelected() );
  }

  /**
   * Aktion Ansicht | Status-Ausgaben durchfuehren.
   */
  public void menuView_Status_actionPerformed(ActionEvent e) {
    statusFrame.setVisible( menuView_Status.isSelected() );
  }

  /**
   * Aktion Ansicht | Fenster anordnen durchfuehren.
   */
  public void menuView_Arange_actionPerformed(ActionEvent e) {
    arrangeInnerFrames();
  }

  /**
   * Aktion Ansicht | Automatische Fensteranordnung durchfuehren.
   */
  public void menuView_ArangeAuto_actionPerformed(ActionEvent e) {
//    menuView_Arange.setEnabled( !menuView_ArangeAuto.isSelected() );
    if ( menuView_ArangeAuto.isSelected() )
      menuView_Arange_actionPerformed(null);
    else {
      desktopPane.setLayout(null);
      desktopPane.doLayout();
    }
  }

  /**
   * Aktion Skript | ... durchfuehren.
   */
  public void menuScript_actionPerformed(ActionEvent e) {
    if ( !(e.getSource() instanceof ObjectMenuItem) )
      return;
    XuluGUIUtil.performScriptInThread(
      appl,
      (ScriptInterpreter) ((ObjectMenuItem)e.getSource()).getObject()
    );
  }

  /**
   * Aktion Skript | Erneut ausfuehren | ... durchfuehren.
   */
  public void menuRecentScript_actionPerformed(ActionEvent e) {
    if ( !(e.getSource() instanceof ObjectMenuItem) )
      return;
    XuluGUIUtil.performScriptInThread(
      appl,
      (ScriptInterpreter.ScriptExecutionDefinition) ((ObjectMenuItem)e.getSource()).getObject()
    );
  }

  /**
   * Aktion Erweitert | Xulu Sprach-Pakete
   */
  public void menuAdvanced_Languages_actionPerformed(ActionEvent e) {
    SwingUtil.centerFrameOnScreen(languageDlg);
    languageDlg.setVisible(true);
  }

  /**
   * Aktion Hilfe | Info durchfuehren.
   */
  public void menuHelp_About_actionPerformed(ActionEvent e) {
    SwingUtil.centerFrameOnScreen(infoDlg);
    infoDlg.setVisible(true);
//    Dimension dlgSize = dlg.getPreferredSize();
//    Dimension frmSize = getSize();
//    Point loc = getLocation();
//    dlg.setLocation((frmSize.width - dlgSize.width) / 2 + loc.x, (frmSize.height - dlgSize.height) / 2 + loc.y);
//    dlg.setModal(true);
//    dlg.pack();
//    dlg.setVisible(true);
  }

}
