/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.io;

import edu.bonn.xulu.appl.XuluRegistry;
import edu.bonn.xulu.io.Factory;

/**
 * Diese Klasse bildet eine Basisklasse fuer alle Factorys und implementiert
 * die <code>equals</code>-Methode des {@link Factory#equals(Factory) Factory}-Interfaces.
 * Da Factorys nur quasi-statische Methoden besitzen, gelten zwei Factorys
 * als gleich, wenn sie Instanz derselben Klasse sind.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class AbstractFactory implements Factory {
  /** Speichert den Namen, mit der die Factory gegenueber dem Anwender
   *  identifiziert wird. */
  protected String name = "";

  /**
   * Erzeugt eine neue Factory.
   * @param name Bezeichnung/Beschreibung der Factory
   */
  public AbstractFactory(String name) {
    setName(name);
  }

  /**
   * Erzeugt eine neue Factory ohne Namen.
   */
  public AbstractFactory() {
    setName("");
  }

  /**
   * Liefert den Namen der Factory.
   */
  public String getName() {
    return name;
  }

  /**
   * Setzt den Namen der Factory.
   * @param name neuer Name
   */
  public void setName(String name) {
    this.name = name;
  }

  /**
   * Prueft, ob die Factory mit einer anderen uebereinstimmt. Dies ist der Fall
   * wenn sie Instanzen derselben Klasse sind.
   */
  public boolean equals(Factory fac) {
    return fac.getClass().equals( this.getClass() );
  }

  /**
   * Prueft, ob die Factory mit einer anderen uebereinstimmt. Dies ist der Fall
   * wenn sie Instanzen derselben Klasse sind.
   * @return <code>false</code>, wenn das uebergebene Objekt keine Implementierung
   *         von {@link Factory} ist
   * @see #equals(Factory)
   */
  public boolean equals(Object o) {
    if ( !(o instanceof Factory) )
      return false;
    return equals((Factory)o);
  }

  /**
   * Wirft eine Exception oder gibt <code>false</code> zurueck, je nachdem
   * ob das <code>throwException</code>-Flag gesetzt ist oder nicht.
   */
  private static boolean performNoGood(boolean throwException, RuntimeException err) {
    if ( !throwException )
      return false;
    throw err;
  }

  /**
   * Liefert eine InstantiationFactory aus der Xulu-Registry, die einen
   * bestimmten Datentyp erzeugt.
   * @param reg  Xulu-Registry
   * @param need Datentyp zu dem eine Factory gesucht wird
   * @param throwException bestimmt, ob eine Exception geworfen wird, wenn keine
   *        Factory gefunden wird
   * @exception java.lang.UnsupportedOperationException wenn die Suche in der Registry
   *            fehl schlaegt und das <code>throwException</code>-Flag gesetzt ist
   */
  public static InstantiationFactory getInstantiationFactoryFromRegistry(XuluRegistry reg, Class need, boolean throwException) {
    if ( reg == null && throwException )
      throw new UnsupportedOperationException("Can search for a factory in null-registry");

    InstantiationFactory fac = reg.getDefaultFactory(need);
    if ( fac == null && throwException )
      throw new UnsupportedOperationException("No factory found in registry, that creates an instances of ".concat(need.getName()).concat("!"));
    return fac;
  }

  /**
   * Prueft, ob ein Eingabequelle-Objekt fuer eine ImportFactory geeignet ist.
   * @param fac   Factory, fuer die dieser Check aufgerufen wird
   * @param input Eingabequelle, die geprueft wird
   * @param need  Klasse, die fuer die Factory als Eingabequelle benoetigt wird
   * @param throwException bestimmt, ob eine Exception geworfen wird, oder
   *        nicht
   * @exception java.lang.UnsupportedOperationException wenn der Check fehl
   *            schlaegt und das <code>throwException</code>-Flag gesetzt ist
   */
  public static boolean checkImportSourceObject(ImportFactory fac, Object input, Class need, boolean throwException) {
    if ( input == null )
      return performNoGood( throwException, new UnsupportedOperationException(fac.getClass().getName().concat(".importObject(.) can not import from null-input")) );
    if ( !need.isInstance(input) )
      return performNoGood( throwException, new UnsupportedOperationException(fac.getClass().getName().concat(".importObject(.) needs ").concat(need.getName()).concat(" for input-Parameter!")) );
    return true;
  }

  /**
   * Prueft, ob ein Ausgabeziel-Objekt fuer eine ExportFactory geeignet ist.
   * @param fac    Factory, fuer die dieser Check aufgerufen wird
   * @param output Ausgabequelle, die geprueft wird
   * @param need   Klasse, die fuer die Factory als Ausgabeziel benoetigt wird
   * @param throwException bestimmt, ob eine Exception geworfen wird, oder
   *        nicht
   * @exception java.lang.UnsupportedOperationException wenn der Check fehl
   *            schlaegt und das <code>throwException</code>-Flag gesetzt ist
   */
  public static boolean checkExportDestinationObject(ExportFactory fac, Object output, Class need, boolean throwException) {
    if ( output == null )
      return performNoGood( throwException, new UnsupportedOperationException(fac.getClass().getName().concat(".exportObject(.) can not export to null-output")) );
    if ( !need.isInstance(output) )
      return performNoGood( throwException, new UnsupportedOperationException(fac.getClass().getName().concat(".exportObject(.) needs ").concat(need.getName()).concat(" for output-Parameter!")) );
    return true;
  }
}
