/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.model;

import java.io.PrintStream;
import java.util.Map;
import java.util.Vector;
import java.awt.Component;

import schmitzm.data.property.Access;
import schmitzm.lang.AbstractNamedObject;
import schmitzm.lang.LangUtil;
import schmitzm.lang.ResourceProvider;
import schmitzm.swing.CaptionsChangeable;

import edu.bonn.xulu.model.XuluModel;
import edu.bonn.xulu.model.ModelContentManager;
import edu.bonn.xulu.model.ModelCanceledException;
import edu.bonn.xulu.model.event.ModelListener;
import edu.bonn.xulu.model.event.ModelEvent;
import edu.bonn.xulu.model.event.ModelInitialisedEvent;
import edu.bonn.xulu.model.event.ModelStartedEvent;
import edu.bonn.xulu.model.event.ModelStoppedEvent;
import edu.bonn.xulu.model.event.ModelDisposedEvent;

import org.apache.log4j.Logger;
import edu.bonn.xulu.appl.XuluConstants;

/**
 * Diese Klasse stellt eine Basis-Implementierung eines Xulu-Modells dar.
 * Es werden alle Methoden des {@link XuluModel}-Interface implementiert, in dem
 * entsprechende Flags sinnvoll gesetzt werden.<br>
 * Eine konkrete Modell-Implementierung muss nur noch folgende Methoden
 * implementieren:
 * <ul>
 * <li><b>{@link #performModelInit()}</b><br>
 *     Initialisierung/Ruecksetzen aller internen Modell-Ressourcen und
 *     Herstellen evt. fuer den Ablauf benoetigter Lese/Schreib-Rechte.</li>
 * <li><b>{@link #performModelStart()}</b><br>
 *     Implementierung des kompletten Modellablaufs (in einer Schleife, die
 *     endet, wenn das {@link #stopped}-Flag gesetzt ist).</li>
 * <li><b>{@link #performModelDispose()}</b><br>
 *     Loeschen aller internen Ressourcen aus dem Speicher und Freigabe aller
 *     externen Ressourcen/Rechte.</li>
 * </ul>
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class AbstractXuluModel extends AbstractNamedObject implements XuluModel {
  /** ResourceProvider fuer allgemeine Uebersetzungen. */
  private ResourceProvider RESOURCE = XuluConstants.XULUGUI_RES;
  /** ResourceProvider fuer modellspezifische Uebersetzungen. Diese Variable ist inital
   *  {@code null} und muss von der jeweiligen Modell-Implementierung belegt werden!! */
  protected ResourceProvider modelResourceProvider = null;

  private Logger LOGGER = LangUtil.createLogger(this);

  /** Ausgabe-Stream fuer Statusausgaben. Standardmaessig auf
   *  {@link System#out System.out} eingestellt */
  protected PrintStream statusOut = System.out;

  /** Speichert den ContentManager fuer das Modell. Diese Variable ist
   *  initial <code>null</code> und muss von der Implementierung befuellt werden! */
  protected ModelContentManager contManager = null;

  /** Speichert die Listener fuer das Modell. */
  protected Vector listeners = new Vector<ModelListener>();

  /** Dieses Flag speichert, ob das Modell gestoppt ist. Wird von
   *  {@link #startModelling()} und {@link #stopModelling()} gesetzt.
   *  Initial <code>true</code>.<br>
   *  <b>Dieses Flag sollte an geeigneter Stelle im Modellablauf abgefragt
   *  werden, um z.B. einen Zwischenschritt abzubrechen. */
  protected boolean stopped = true;

  /** Dieses Flag speichert, ob die Modellierung gerade laeuft. Wird von
   *  {@link #startModelling()} und {@link #stopModelling()} gesetzt.
   *  Initial <code>false</code>. */
  protected boolean running = false;

  /** Dieses Flag speichert, ob die Modellierung initialisiert ist, also
   *  sich alle internen Ressourcen in einem Ausgangszustand befinden.
   *  Wird von {@link #initModel()} und {@link #startModelling()} gesetzt.
   *  Initial <code>false</code>. */
  protected boolean initialised = false;

  /** Dieses Flag speichert, ob die internen Modell-Ressourcen "zerstoert"
   *  sind.
   *  Wird von {@link #initModel()} und {@link #disposeModel()} gesetzt.
   *  Initial <code>true</code>. */
  protected boolean disposed = true;

  /** Speichert die Ereignis-Typen, die waehrend des Modellablaufs geworfen
   *  werden. Unterklassen koennen diese Liste erweitern.
   *  @see #getEvents() */
  protected Vector events = new Vector<ModelEvent>();

  /**
   * Erzeugt ein neues Modell.
   * @param contManager verwaltet die Modell-Ressourcen
   * @param name        Bezeichnung fuer das Modell
   */
  public AbstractXuluModel(ModelContentManager contManager, String name) {
    this.contManager = contManager;
    // Ereignisse in Liste aufnehmen
    events.add( ModelInitialisedEvent.class );
    events.add( ModelStartedEvent.class );
    events.add( ModelStoppedEvent.class );
    events.add( ModelDisposedEvent.class );
  }

  /**
   * Erzeugt ein neues Modell.
   * @param contManager verwaltet die Modell-Ressourcen
   */
  public AbstractXuluModel(ModelContentManager contManager) {
    this(contManager,"");
  }

  /**
   * Liefert den ContentManager des Modells.
   */
  public ModelContentManager getContentManager() {
    return this.contManager;
  }

  /**
   * Setzt den Stream fuer Status-Ausgaben des Modells.
   */
  public void setStatusOut(PrintStream out) {
    this.statusOut = out;
  }

  /**
   * Liefert den Stream fuer Status-Ausgaben des Modells.
   */
  public PrintStream getStatusOut() {
    return statusOut;
  }

  /**
   * Initialisiert das Model, indem der Modell-Ablauf gestoppt und
   * {@link #performModelInit()} aufgerufen wird.
   * Darueberhinaus wird das {@link #running}-Flag auf <code>false</code>, das
   * {@link #stopped}-Flag auf <code>true</code>, das {@link #initialised}-Flag
   * auf <code>true</code> und das {@link #disposed}-Flag auf <code>false</code>
   * gesetzt.
   */
  public void initModel() {
    if ( isRunning() )
      stopModelling();
    running = false;
    stopped = true;
    performModelInit();
    initialised = true;
    disposed = false;
    fireModelInitialised();
  }

  /**
   * Muss die eigentliche Initialisierung des Modells implementieren und
   * alle <b>internen</b> Modell-Ressourcen auf einen Ausgangszustand
   * zurueck setzen, so dass das Modell komplett neu gestartet werden
   * kann.
   */
  public abstract void performModelInit();

  /**
   * Prueft, ob das Modell initialisiert ist - sich also im Ausgangszustand
   * befindet - und bereit ist gestartet zu werden.
   * @see #initialised
   */
  public boolean isInitialised() {
    return this.initialised;
  }

  /**
   * Startet die Modellierung, indem {@link #performModelStart()} aufgerufen
   * wird. Desweiteren wird das {@link #running}-Flag auf <code>true</code>,
   * das {@link #initialised}-Flag auf <code>false</code> und das
   * {@link #stopped}-Flag auf <code>false</code> gesesetzt.
   * @return {@code false} wenn das Modell auf irgendeine Art (kontrolliert oder
   *         Fehler) abgebrochen wurde; {@code true} nur wenn das Modell komplett
   *         durchgelaufen ist
   */
  public boolean startModelling() {
    running = true;
    initialised = false;
    stopped = false;
    fireModelStarted();
    try {
      performModelStart();
      // Am Ende der Modellierung wird das Modell zerstoert, damit alle
      // Ressourcen wieder freigegeben werden
      disposeModel();
      return true;
    } catch ( ModelCanceledException err ) {
      // erwartete Exception
      // --> wird das Modell gestoppt, wird es zerstoert, damit alle
      //     Ressourcen wieder freigegeben werden
      disposeModel();
    } catch ( Exception err ) {
      getStatusOut().println("\n"+RESOURCE.getString("Model.ErrorOccured")+"\n   "+err+"\n");
      performError();
      err.printStackTrace();
    }
    return false;
  }

  /**
   * Prueft, ob das Modell gestoppt ist. Ist dies der Fall wird eine
   * {@link ModelCanceledException} geworfen.<br>
   * Diese Methode sollte an geeigneten Stellen waehrend des Modellablaufs
   * aufgerufen werden, um diesen aufgrund eines externen Befehls
   * abzubrechen. Die Exception wird in der {@link #startModelling()}-Methode
   * wieder abgefangen.
   * @exception ModelCanceledException falls das Modell gestoppt ist
   */
  protected void checkBreakingCommands() throws ModelCanceledException {
    if ( isStopped() || Thread.interrupted() )
      throw new ModelCanceledException();
  }

  /**
   * Implementiert den eigentlichen Modellablauf. Dies kann z.B. durch einen
   * einfachen Befehlsablauf oder eine Schleife geschehen.
   */
  public abstract void performModelStart();

  /**
   * Prueft, ob das Modell gerade laeuft.
   * @see #running
   */
  public boolean isRunning() {
    return this.running;
  }

  /**
   * Stoppt die Modellierung. Diese Implementierung setzt lediglich das
   * {@link #running}-Flag auf <code>false</code> und das {@link #stopped}-Flag
   * auf <code>true</code>.<br>
   * Die Implementierung muss (in {@link #performModelStart()}) selbst auf
   * {@link #isStopped()} oder das {@link #stopped}-Flag checken, um eine
   * Schleife oder zeitaufwaendige Zwischenschritte abzubrechen.
   */
  public void stopModelling() {
    running = false;
    stopped = true;
    fireModelStopped();
  }

  /**
   * Prueft, ob das Modell gestoppt ist.
   * @see #stopped
   */
  public boolean isStopped() {
    return this.stopped;
  }

  /**
   * Diese Methode "zerstoert" das Modell, in dem es zunaechst gestoppt und
   * dann {@link #performModelDispose()} aufgerufen wird.
   * Desweiteren wird das {@link #disposed}-Flag auf <code>true</code>
   * gesetzt.
   */
  public void disposeModel() {
    if ( isRunning() )
      stopModelling();
    performModelDispose();
    disposed = true;
    fireModelDisposed();
  }

  /**
   * Implementiert die eigentliche Freigabe der Modell-Ressourcen, indem alle
   * <b>internen</b> Modell-Ressourcen aus dem Speicher geloescht und die
   * externen wieder frei gegeben werden.
   */
  public abstract void performModelDispose();

  /**
   * Prueft, ob das Modell (dessen interen Ressourcen) "zerstoert" ist.
   * @see #disposed
   */
  public boolean isDisposed() {
    return this.disposed;
  }

  /**
   * Diese Methode wird aufgrufen, wenn ein unerwarteter Fehler bei der
   * Modellierung auftritt, welcher das Beenden der Modellierung zu Folge
   * hat. Diese Implementierung ruft lediglich {@link #disposeModel()} auf.
   */
  public void performError() {
    disposeModel();
  }

  /**
   * Fuegt einen ModelListener hinzu.
   * @return <code>true</code> wenn dass Hinzufuegen erfolgreich war (siehe
   *         {@link Vector#add(Object)})
   */
  public boolean addModelListener(ModelListener listener) {
    return listeners.add(listener);
  }

  /**
   * Entfernt einen ModelListener von dem Modell.
   * @return <code>true</code> wenn dass Entfernen erfolgreich war (siehe
   *         {@link Vector#remove(Object)})
   */
  public boolean removeModelListener(ModelListener listener) {
    return listeners.remove(listener);
  }

  /**
   * Informiert alle {@link ModelListener} ueber ein Modell-Ereinis.
   */
  protected void fireModelEvent(ModelEvent e) {
    for (int i=0; i<listeners.size(); i++)
      ((ModelListener)listeners.get(i)).performModelEvent(e);
  }

  /**
   * Informiert alle {@link ModelListener} darueber, dass das Modell initialisiert
   * wurde.
   */
  protected void fireModelInitialised() {
    getStatusOut().println( RESOURCE.getString("Model.Init") );
    fireModelEvent(new ModelInitialisedEvent(this));
  }

  /**
   * Informiert alle {@link ModelListener} darueber, dass das Modell gestartet
   * wurde.
   */
  protected void fireModelStarted() {
    getStatusOut().println( RESOURCE.getString("Model.Start") );
    fireModelEvent(new ModelStartedEvent(this));
  }

  /**
   * Informiert alle {@link ModelListener} darueber, dass das Modell
   * gestoppt wurde.
   */
  protected void fireModelStopped() {
    getStatusOut().println( RESOURCE.getString("Model.Stop") );
    fireModelEvent(new ModelStoppedEvent(this));
  }

  /**
   * Informiert alle {@link ModelListener} darueber, dass das Modell
   * "zerstoert" wurde.
   */
  protected void fireModelDisposed() {
    getStatusOut().println( RESOURCE.getString("Model.Dispose") );
    fireModelEvent(new ModelDisposedEvent(this));
  }

  /**
   * Liefert immer <code>null</code>. Sub-Klassen koennen diese Methode
   * ueberschreiben und eine spezielle (zusaetzliche) Modell-GUI
   * zurueck geben.
   */
  public Component getModelSpecificGUI() {
    return null;
  }

  /**
   * Liefert alle Ereignis-Typen, die waehrend des Modellablaufs initiiert werden.
   * @return leerer Array, falls keine Ereignisse initiiert werden
   */
  public Class<ModelEvent>[] getEvents() {
    return (Class<ModelEvent>[])events.toArray( new Class[0] );
  }

  /**
   * Gibt ein Lese- oder Schreibrecht wieder frei. Dabei wird zunaechst
   * geprueft, ob das uebergebene Objekt nicht <code>null</code> ist.
   */
  protected static void releaseAccess(Access a) {
    if ( a!=null )
      a.release();
  }

  /**
   * Liefert den {@link ResourceProvider}, der die Uebersetzungen
   * fuer Modell-Ausgaben und modellspezifische GUI verwaltet.
   * @return {@code null}, wenn das Modell nicht multi-lingual ist
   */
  public ResourceProvider getResourceProvider() {
    return modelResourceProvider;
  }
  
  /**
   * Belegt die sprachspezifischen Elemente des Modells (Status-Ausgaben, GUI)
   * neu mit Elementen aus der Map. Diese Standard-Implementierung macht nichts,
   * ausser die {@code resetCaptions}-Methode des Content-Managers und der
   * modell-spez. GUI aufzurufen, sofern diese {@link CaptionsChangeable} sind.
   * @param captionMap Label-Beschreibungen
   */
  public void resetCaptions(Map<String, Object> captionMap) { 
    if ( this.contManager instanceof CaptionsChangeable )
      ((CaptionsChangeable)contManager).resetCaptions(captionMap);
    Object gui = this.getModelSpecificGUI(); 
    if ( gui != null && gui instanceof CaptionsChangeable )
      ((CaptionsChangeable)gui).resetCaptions(captionMap);
  }

  /**
   * Ermittelt einn String-Wert aus der {@code captionMap}. Wenn dort kein Wert
   * hinterlegt ist, wird auf den ResourceProvider zurueckgegriffen.
   * @param captionMap    Map mit neuen Bezeichnungen
   * @param rp            ResourceProvider
   * @param captionKey    Key unter dem in der Map und im ResourceProvider gesucht wird 
   * @param captionParams Parameter, die durch den ResourceProvider in die Caption "eingebaut" werden
   * @see ResourceProvider#getString(String, Object...)
   */
  public static String determineCaption(Map<String,Object> captionMap, ResourceProvider rp, String captionKey, Object... captionParams) {
    // Beschreibung nach Moeglichkeit aus der Map ziehen
    Object newCaptionObj = (captionMap != null) ? captionMap.get(captionKey) : null;
    if ( newCaptionObj != null )
      return newCaptionObj.toString();
    // Wenn keine Beschreibung in Map hinterlegt, dann die "normale"
    // Lokalisation verwenden
    if ( rp != null )
      return rp.getString(captionKey,captionParams);
    return ( "" ); 
  }
}
