/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/
package edu.bonn.xulu.model;

import edu.bonn.xulu.model.AbstractModelResource;
import edu.bonn.xulu.data.XuluDataException;

/**
 * Standard-Implementierung von {@link AbstractModelResource}. Die Check-Methoden
 * pruefen lediglich auf Typ-Vertraeglichkeit mit dem der Ressource zugeordneten
 * Datentyp, sowie auf die not-null Eigenschaft.
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
 * @version 1.0
 */
public class DefaultModelResource extends AbstractModelResource {

  /**
   * Erzeugt eine leere Ressource.
   * @param category Ordnet die Ressource in eine Kategorie ein (siehe
   *                 {@link #setCategory(int)})
   * @param desc Beschreibung der Ressource
   * @param type Datentyp der Ressource
   * @param nullable gibt an, ob die Ressource unbelegt bleiben darf
   * @exception XuluDataException falls der Datentyp <code>null</code> ist.
   */
  public DefaultModelResource(int category, String desc, Class type, boolean nullable) {
    super(category, desc, type, nullable);
  }

  /**
   * Erzeugt Ressource und ordnet ihr ein konretes Datenobjekt zu. Der Datentyp
   * wird implizit aus dem uebergebenen Objekt ermittelt.
   * @param category Ordnet die Ressource in eine Kategorie ein (siehe
   *                 {@link #setCategory(int)})
   * @param desc Beschreibung der Ressource
   * @param data Daten-Objekt fuer die Ressource
   * @param nullable gibt an, ob die Ressource unbelegt bleiben darf
   * @exception XuluDataException falls das Datenobjekt <code>null</code> ist,
   *           denn aus <code>null</code> kann kein Datentyp ermittelt werden
   */
  public DefaultModelResource(int category, String desc, Object data, boolean nullable) {
    super(category, desc, data, nullable);
  }

  /**
   * Erzeugt eine leere Ressource.
   * @param desc Beschreibung der Ressource
   * @param type Datentyp der Ressource
   * @param nullable gibt an, ob die Ressource unbelegt bleiben darf
   * @exception XuluDataException falls der Datentyp <code>null</code> ist.
   * @exception IllegalArgumentException falls eine unbekannte Kategorie
   *            angegeben wird
   */
  public DefaultModelResource(String desc, Class type, boolean nullable) {
    super(desc, type, nullable);
  }

  /**
   * Erzeugt Ressource und ordnet ihr ein konretes Datenobjekt zu. Der Datentyp
   * wird implizit aus dem uebergebenen Objekt ermittelt.
   * @param desc Beschreibung der Ressource
   * @param data Daten-Objekt fuer die Ressource
   * @param nullable gibt an, ob die Ressource unbelegt bleiben darf
   * @exception XuluDataException falls das Datenobjekt <code>null</code> ist,
   *           denn aus <code>null</code> kann kein Datentyp ermittelt werden
   * @exception IllegalArgumentException falls eine unbekannte Kategorie
   *            angegeben wird
   */
  public DefaultModelResource(String desc, Object data, boolean nullable) {
    super(desc, data, nullable);
  }

  /**
   * Prueft ein beliebiges Objekt auf Typ- und Wert-Konsistenz mit der
   * Ressource. Ruft lediglich {@link #checkTypeAndError(Object)} auf.
   * @throws XuluDataException falls der Typ oder der Wert des zugeordneten
   *         Objekts nicht mit der Ressource vertraeglich ist
   */
  public void checkAndError(Object data) throws XuluDataException {
    checkTypeAndError(data);
  }

  /**
   * Prueft ein beliebiges Objekt nur auf Typ-Vertraeglichkeit mit der
   * Ressource.
   * @throws XuluDataException falls der Objekt-Typ nicht mit der Ressourcen
   *         vertraeglich ist
   */
  public void checkTypeAndError(Object data) throws XuluDataException {
    if ( data!=null && !getType().isInstance(data) ||
         data==null && !nullable )
      throwDataException(getType().getSimpleName());
  }
}
