/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.model;

import edu.bonn.xulu.data.XuluObject;
import edu.bonn.xulu.data.XuluDataException;

/**
 * Eine Ressource beschreibt allgemein ein Datenobjekt, das für die Modellierung
 * benoetigt wird. Im Unterschied zu einem konkreten Daten-Objekt, existiert die
 * Ressource auch, wenn noch gar keine Daten vorhanden sind.<br>
 * Die Ressource beschreibt also vor allem die Struktur des benoetigten
 * Objekts. Gleichzeitig nimmt sie aber auch die konkreten Daten auf und ist
 * fuer deren Typenvertraeglichkeit verantwortlich.<br>
 * Eine Ressource besteht aus 3 Teilen:
 * <ul>
 * <li><b>Typ (Klasse)</b><br>
 *     beschreibt den Datentyp, der für die Ressource benoetigt wird (zulaessig ist)</li>
 * <li><b>Daten</b><br>
 *     nimmt ein (zulaessiges) Datenobjekt fuer die Ressource auf</li>
 * <li><b>Beschreibung</b><br>
 *     macht gegenueber dem Anwender (im UI) kenntlich, zu welchem Zweck die
 *     Ressource benoetigt wird</li>
 * </ul>
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public interface ModelResource {
  /** Definiert die Ressource als "andere" Ressource */
  public static final int CATEGORY_OTHER = 0;
  /** Definiert die Ressource als Eingabe-Ressource */
  public static final int CATEGORY_INPUT = 1;
  /** Definiert die Ressource als Ausgabe-Ressource */
  public static final int CATEGORY_OUTPUT = 2;
  /** Definiert die Ressource als temporaere Ressource */
  public static final int CATEGORY_TEMP = 4;

  /**
   * Liefert den Datentyp der Ressource.
   */
  public Class getType();

  /**
   * Liefert eine Beschreibung, worum es sich bei der Ressource handelt. Diese
   * macht dem Anwender kenntlich, wofuer die Ressource vom Modell benoetigt wird.
   */
  public String getDesc();

  /**
   * Setzt eine Beschreibung, worum es sich bei der Ressource handelt. Diese
   * macht dem Anwender kenntlich, wofuer die Ressource vom Modell benoetigt wird.
   */
  public void setDesc(String desc);

  /**
   * Liefert das Daten-Objekt der Ressource. Initial sollte dieses <code>null</code>
   * sein.
   */
  public Object getData();

  /**
   * Ordnet der Ressource ein bestimmtes Datenobjekt zu. Wird i.d.R. vom
   * XULU-ModelInputManager aufgerufen, nachdem der Anwender Daten aus dem
   * DatenPool zugeordnet hat.
   * @param data  Daten-Objekt
   * @param check bestimmt, ob das Datenobjekt sofort auf Typ-Vertraeglichkeit
   *        ueberprueft werden soll
   */
  public void setData(Object data, boolean check);

  /**
   * Gibt an, ob die Ressource opional ist (also unbelegt bleiben darf).
   */
  public boolean isNullable();

  /**
   * Bestimmt, ob die Ressource opional ist (also unbelegt bleiben darf).
   */
  public void setNullable(boolean nullable);

  /**
   * Ordnet der Ressource ein bestimmtes Datenobjekt zu. Zuvor wird das Objekt
   * auf Korrektheit ueberprueft. Wird i.d.R. vom XULU-ModelInputManager
   * aufgerufen, nachdem der Anwender Daten aus dem DatenPool zugeordnet hat.
   * @param data  Daten-Objekt
   * @exception XuluDataException falls das zugeordnete Objekt <code>null</code>
   *         ist, oder die Klasse nicht mit dem Ressourcen-Typ vertraeglich ist
   * @see #checkAndError()
   */
  public void setData(Object data) throws XuluDataException;

  /**
   * Liefert die Kategorie der Ressource. Dies kann die Konstante
   * {@link #CATEGORY_OTHER} sein, oder eine logische
   * Oder-Verknuepfung der Konstanten {@link #CATEGORY_INPUT},
   * {@link #CATEGORY_OUTPUT} und {@link #CATEGORY_TEMP}.
   */
  public int getCategory();

  /**
   * Setzt die Kategorie der Ressource. Dies kann die Konstante
   * {@link #CATEGORY_OTHER} sein, oder eine logische
   * Oder-Verknuepfung der Konstanten {@link #CATEGORY_INPUT},
   * {@link #CATEGORY_OUTPUT} und {@link #CATEGORY_TEMP}.
   */
  public void setCategory(int category);

  /**
   * Prueft, ob die Ressource einer bestimmten Kategorie
   * zugeordnet ist. Liefert <code>true</code> wenn die Ressource <b>einer</b>
   * der angegebenen Kategorien zugeordnet ist (ODER).
   * @param category Oder-Verknuepfung von Kategorien
   * @return <code>this.getCategory() == CATEGORY_OTHER && category == CATEGORY_OTHER ||
   *         (this.getCategory() & category) > 0</code>
   */
  public boolean isCategory(int category);

  /**
   * Prueft, ob die Ressource einer bestimmten Kategorie-Kombination
   * zugeordnet ist. Liefert <code>true</code> wenn die Ressource <b>allen</b>
   * angegebenen Kategorien zugeordnet ist (UND).
   * @return <code>this.getCategory() == CATEGORY_OTHER && category == CATEGORY_OTHER ||
   *         (this.getCategory() & category) == category</code>
   */
  public boolean isCategoryAll(int category);

  /**
   * Prueft ein beliebiges Objekt auf Typ- und Wert-Konsistenz mit der
   * Ressource.
   */
  public boolean check(Object data);

  /**
   * Prueft ein beliebiges Objekt auf Typ- und Wert-Konsistenz mit der
   * Ressource. Kann manuell aufgerufen werden und wird automatisch von
   * {@link #setData(Object)} aufgerufen.
   * @exception XuluDataException falls der Typ oder der Wert des zugeordneten
   *            Objekts nicht mit der Ressource vertraeglich ist
   */
  public void checkAndError(Object data) throws XuluDataException;

  /**
   * Prueft ein beliebiges Objekt nur auf Typ-Vertraeglichkeit mit der
   * Ressource.
   */
  public boolean checkType(Object data);

  /**
   * Prueft ein beliebiges Objekt nur auf Typ-Vertraeglichkeit mit der
   * Ressource.
   * @exception XuluDataException falls der Objekt-Typ nicht mit der Ressourcen
   *            vertraeglich ist
   */
  public void checkTypeAndError(Object data) throws XuluDataException;

  /**
   * Prueft das der Ressource zugeordnete Objekt auf Typ- und Wert-Konsistenz
   * mit der Ressource.
   */
  public boolean check();

  /**
   * Prueft das der Ressource zugeordnete Objekt auf Typ- und Wert-Konsistenz
   * mit der Ressource.
   * @exception XuluDataException falls der Typ oder der Wert des zugeordneten
   *            Objekts nicht mit der Ressource vertraeglich ist
   */
  public void checkAndError() throws XuluDataException;
}
