/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.model;

import edu.bonn.xulu.data.XuluDataException;
import schmitzm.data.property.ValueProperty;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.ValuePropertyType;
import schmitzm.temp.BaseTypeUtil;

/**
 * Diese Implementierung von <code>ModelResource</code> laesst nur
 * {@linkplain ValueProperty}-Objekte eines bestimmten Datentyps als
 * Datenobjekt zu. Ausserdem darf das Daten-Objekt der Ressource (wahlweise)
 * nicht auf <code>null</code> gesetzt werden.<br>
 * Details zur Funktionalitaet und Bestimmung dieser Klasse sind in
 * {@link ModelResource} zu finden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ValuePropertyResource extends PropertiesResource {
  /** Speichert den Inhaltstyp der {@link ValueProperty}. */
  protected ValuePropertyType propType = null;

  /**
   * Erzeugt eine neue Ressource
   * @param category Ordnet die Ressource in eine Kategorie ein (siehe
   *                 {@link #setCategory(int)})
   * @param desc Beschreibung der Ressource
   * @param prop Beispiel-Property, die die Ressource aufnehmen darf
   * @param nullable gibt an, ob die Ressource unbelegt bleiben darf
   */
  public ValuePropertyResource(int category, String desc, ValueProperty prop, boolean nullable) {
    this(category,desc,prop.getClass(),prop.getPropertyType(),nullable);
  }

  /**
   * Erzeugt eine neue Ressource
   * @param category Ordnet die Ressource in eine Kategorie ein (siehe
   *                 {@link #setCategory(int)})
   * @param desc Beschreibung der Ressource
   * @param propClass (Unter)Klasse von {@link ValueProperty}, die die Art
   *                  der Ressource bestimmt (z.B. Skalar, Matrix, Liste)
   * @param propType  Typ den die Property als Wert aufnehmen/liefern soll
   * @param nullable gibt an, ob die Ressource unbelegt bleiben darf
   */
  public ValuePropertyResource(int category, String desc, Class propClass, Class propType, boolean nullable) {
    this(category,desc,propClass,ValuePropertyType.createValuePropertyType(propType),nullable);
  }

  /**
   * Erzeugt eine neue Ressource
   * @param category Ordnet die Ressource in eine Kategorie ein (siehe
   *                 {@link #setCategory(int)})
   * @param desc Beschreibung der Ressource
   * @param propClass (Unter)Klasse von {@link ValueProperty}, die die Art
   *                  der Ressource bestimmt (z.B. Skalar, Matrix, Liste)
   * @param propType  Typ den die Property als Wert aufnehmen/liefern soll
   * @param nullable gibt an, ob die Ressource unbelegt bleiben darf
   */
  public ValuePropertyResource(int category, String desc, Class propClass, ValuePropertyType propType, boolean nullable) {
    super(category,desc,propClass,nullable);
    if ( !ValueProperty.class.isAssignableFrom(propClass) )
      throw new IllegalArgumentException("Resource '"+desc+": Parameter 'propClass' muss bei a subclass of ValueProperty!");
    this.propType = propType;
  }

  /**
   * Erzeugt eine neue Ressource
   * @param desc Beschreibung der Ressource
   * @param prop Beispiel-Property, die die Ressource aufnehmen darf
   * @param nullable gibt an, ob die Ressource unbelegt bleiben darf
   */
  public ValuePropertyResource(String desc, ValueProperty prop, boolean nullable) {
    this(desc,prop.getClass(),prop.getPropertyType(),nullable);
  }

  /**
   * Erzeugt eine neue Ressource
   * @param desc Beschreibung der Ressource
   * @param propClass (Unter)Klasse von {@link ValueProperty}, die die Art
   *                  der Ressource bestimmt (z.B. Skalar, Matrix, Liste)
   * @param propType  Typ den die Property als Wert aufnehmen/liefern soll
   * @param nullable gibt an, ob die Ressource unbelegt bleiben darf
   */
  public ValuePropertyResource(String desc, Class propClass, Class propType, boolean nullable) {
    this(desc,propClass,ValuePropertyType.createValuePropertyType(propType),nullable);
  }

  /**
   * Erzeugt eine neue Ressource
   * @param desc Beschreibung der Ressource
   * @param propClass (Unter)Klasse von {@link ValueProperty}, die die Art
   *                  der Ressource bestimmt (z.B. Skalar, Matrix, Liste)
   * @param propType  Typ den die Property als Wert aufnehmen/liefern soll
   * @param nullable gibt an, ob die Ressource unbelegt bleiben darf
   */
  public ValuePropertyResource(String desc, Class propClass, ValuePropertyType propType, boolean nullable) {
    this(CATEGORY_OTHER,desc,propClass,propType,nullable);
  }

  /**
   * Liefert den Datentyp, der in der Prop
   * @return ValuePropertyType
   */
  public ValuePropertyType getPropertyType() {
    return this.propType;
  }

  /**
   * Neben den Checks der
   * {@linkplain PropertiesResource#checkTypeAndError(Object) Oberklasse},
   * prueft diese Methode, ob der Inhaltstyp der Property <code>data</code>
   * dem fuer die Ressource verlangten Typ entspricht.
   * @param data Instanz von {@link ValueProperty}
   */
  public void checkTypeAndError(Object data) {
    super.checkTypeAndError(data);

    // null-Daten koennen nicht auf Typ geprueft werden, werden also
    // als korrekt angenommen
    if ( data == null )
      return;

    Class objDataType = ((ValueProperty)data).getType();
    Class resDataType = getPropertyType().getType();

    // Numberische Properties koennen u.U. in Ressourcen-Typ umgewandelt werden
    if ( Number.class.isAssignableFrom( objDataType ) &&
         Number.class.isAssignableFrom( resDataType ) &&
         BaseTypeUtil.isConvertable(objDataType, resDataType) )
      return;

    // Datenobjekt auf den richtigen Property-Inhaltstyp checken
    if ( data!= null && !resDataType.isAssignableFrom(objDataType) )
      throwDataException(resDataType.getSimpleName().concat(" property"));
  }
}
