/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.model.event;

import schmitzm.data.event.ObjectEvent;
import edu.bonn.xulu.model.XuluModel;
import edu.bonn.xulu.model.event.ModelListener;

/**
 * Diese Klasse stellt eine Basis-Implementierung des Interfaces
 * {@link ModelListener} dar, welche nach den Ereignissen...
 * <ul>
 *   <li>{@link ModelInitialisedEvent}</li>
 *   <li>{@link ModelStartedEvent}</li>
 *   <li>{@link ModelStoppedEvent}</li>
 *   <li>{@link ModelDisposedEvent}</li>
 * </ul>
 * ... unterscheidet. Fuer jedes Ereignis wird durch {@link #performModelEvent(ModelEvent)}
 * eine eigene Adapter-Methode aufgerufen. Handelt es sich bei einem Ereignis
 * um keines der oben genannten, wird dies an {@link #performOtherEvent(ModelEvent)}
 * weitergereicht.<br>
 * <b>Die implementierten Methoden {@link #modelInitialised(XuluModel)},
 * {@link #modelStarted(XuluModel)}, {@link #modelStopped(XuluModel)},
 * {@link #modelDisposed(XuluModel)} und {@link #performOtherEvent(ModelEvent)}
 * machen allesamt nichts.</b> Sie dienen lediglich der Vereinfachung,
 * dass nicht benoetigte Methoden auch nicht implementiert werden muessen.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class ModelAdapter implements ModelListener {
  /**
   * Leitet das Ereignis an {@link #performModelEvent(ModelEvent)} weiter.
   * @param e Modell-Ereignis
   * @exception IllegalArgumentException falls das uebergebene Ereignis
   *            kein {@link ModelEvent} ist.
   */
  public void performObjectEvent(ObjectEvent e) {
    if ( !(e instanceof ModelEvent) )
      throw new IllegalArgumentException("ModelListener.performObjectEvent(.) can only be applied for ModelEvents!");
    performModelEvent((ModelEvent)e);
  }

  /**
   * Wird fuer jedes Modell-Ereignis ausgeloest. Je nach Event-Typ wird die
   * entsprechende Adapter-Methode aufgerufen.
   * @param e Modell-Ereignis
   * @see #modelInitialised(XuluModel)
   * @see #modelStarted(XuluModel)
   * @see #modelStopped(XuluModel)
   * @see #modelDisposed(XuluModel)
   * @see #performOtherEvent(ModelEvent e)
   */
  public void performModelEvent(ModelEvent e) {
    if ( e instanceof ModelInitialisedEvent )
      this.modelInitialised(e.getSourceModel());
    else if ( e instanceof ModelStartedEvent )
      this.modelStarted(e.getSourceModel());
    else if ( e instanceof ModelStoppedEvent )
      this.modelStopped(e.getSourceModel());
    else if ( e instanceof ModelDisposedEvent )
      this.modelDisposed(e.getSourceModel());
    else
      this.performOtherEvent(e);
  }


  /**
   * Wird ausgeloest, nachdem ein Modell initialisiert wurde.
   * @param model Xulu-Modell das initialisiert wurde
   */
  public void modelInitialised(XuluModel model) {
  }

  /**
   * Wird ausgeloest, nachdem ein Modell gestartet wurde.
   * @param model Xulu-Modell das gestartet wurde
   */
  public void modelStarted(XuluModel model) {
  }

  /**
   * Wird ausgeloest, nachdem ein Modell komplett gestoppt hat.
   * @param model Xulu-Modell das gestoppt wurde
   */
  public void modelStopped(XuluModel model) {
  }

  /**
   * Wird ausgeloest, nachdem ein Modell "zerstoert" wurde.
   * @param model Xulu-Modell das zerstoert wurde
   */
  public void modelDisposed(XuluModel model) {
  }

  /**
   * Wird von {@link #performModelEvent(ModelEvent)} ausgeloest, wenn ein Ereignis
   * "unbekannt" ist.
   * @param e unbekanntes Modell-Ereignis
   */
  public void performOtherEvent(ModelEvent e) {
  }
}
