/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.appl;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.JMenuItem;

import schmitzm.lang.LangUtil;

import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.appl.AbstractXuluPlugin;
import edu.bonn.xulu.gui.XuluInternalFrame;
import edu.bonn.xulu.gui.XuluGUIMessages;

/**
 * Diese Klasse stellt eine abstrakte Oberklasse fuer Plugins dar, die
 * sich (bzw. ein zum Plugin gehoerendes Anwendungsfenster) in das Xulu-Menue
 * integrieren.<br>
 * Diese Oberklasse sorgt automatisch dafuer, dass beim Starten des Plugins
 * ein Menue-Eintrag erzeugt wird, bei dessen Aktivierung das Plugin-Fenster
 * angezeigt wird. Beim Beenden des Plugins wird das Plugin-Fenster verborgen und
 * der Menue-Eintrag wieder entfernt.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class AbstractMenuPlugin extends AbstractXuluPlugin {
  /** Speichert die Menue-Nummer, in der der Eintrag zum Anzeigen des
   *  des Applikationsfensters eingebunden wird. */
  protected int xuluMenuNo = 0;
  /** Speichert den Menue-Eintrag, der zum Anzeigen des Applikationsfensters
   *  dient wird. */
  protected JMenuItem         menuItem   = null;
  /** Speichert die Instanz des Applikationsfensters (wird beim Starten des
   *  Plugins befuellt). */
  protected XuluInternalFrame pluginAppl = null;

  /**
   * Erzeugt eine neue Plugin-Instanz.
   * @param xuluMenuNo Nummer des Xulu-Menues in das der Menue-Eintrag eingebettet wird
   * @param menuItem   Menue-Eintrag fuer das Plugin-Fenster
   * @param hasGUI     Flag, ob das Plugin (zusaetzlich) eine Konfigurations-GUI besitzt
   */
  public AbstractMenuPlugin(int xuluMenuNo, JMenuItem menuItem, boolean hasGUI) {
    super(hasGUI);
    this.xuluMenuNo  = xuluMenuNo;
    this.menuItem    = menuItem;
    this.menuItem.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if ( pluginAppl != null ) {
          pluginAppl.setVisible(true);
          pluginAppl.toFront();
        }
      }
    });
  }

  /**
   * Erzeugt eine neue Plugin-Instanz ohne Konfigurations-GUI.
   * @param xuluMenuNo Nummer des Xulu-Menues in das der Menue-Eintrag eingebettet wird
   * @param menuItem   Menue-Eintrag fuer das Plugin-Fenster
   */
  public AbstractMenuPlugin(int xuluMenuNo, JMenuItem menuItem) {
    this(xuluMenuNo,menuItem,false);
  }

  /**
   * Erzeugt eine neue Plugin-Instanz ohne Konfigurations-GUI.
   * @param xuluMenuNo   Nummer des Xulu-Menues in das der Menue-Eintrag eingebettet wird
   * @param menuItemName Name fuer den Menue-Eintrag des Plugin-Fensters
   */
  public AbstractMenuPlugin(int xuluMenuNo, String menuItemName) {
    this(xuluMenuNo, new JMenuItem(menuItemName));
  }

  /**
   * Erzeugt das Applikations-Fenster fuer das Plugin. Dieses wird mit dem
   * Menue-Eintrag verbunden.
   * Wird erst aufgerufen, wenn das Plugin gestartet wird.
   * @see #execute(XuluModellingPlatform)
   */
  protected abstract XuluInternalFrame createPluginApplication() throws Exception;


  /**
   * Startet das Plugin. Dabei wird die Plugin-Applikation neu erzeugt und
   * ein Menu-Punkt im Xulu-Menue angelegt.
   * @param appl Xulu-Instanz
   */
  public void execute(XuluModellingPlatform appl) {
    super.execute(appl);
    try {
      // Instanz der Plugin-Applikation erzeugen
      pluginAppl = createPluginApplication();
      // Plugin-Applikation in Menu einbinden
      appl.getMainFrame().getJMenuBar().getMenu(xuluMenuNo).add(menuItem);
      // Plugin-Applikation (InternalFrame) dem Hauptfenster hinzufuegen
      appl.getMainFrame().getContentPane().add(pluginAppl);
      started = true;
    } catch (Exception err) {
      XuluGUIMessages.showError(appl.getMainFrame(),err);
    }
  }

  /**
   * Stoppt das Plugin. Der Menu-Punkt wird wieder entfernt.
   */
  public void stop() {
    super.stop();
    // Plugin-Applikation schliessen
    if ( pluginAppl != null ) {
      pluginAppl.setVisible(false);
      pluginAppl.dispose();
    }
    // Plugin-Applikation (InternalFrame) aus dem Hauptfenster entfernen
    appl.getMainFrame().getContentPane().remove(pluginAppl);
    // Code-Generator aus Menu entfernen
    appl.getMainFrame().getJMenuBar().getMenu(xuluMenuNo).remove(menuItem);
    pluginAppl = null;
  }
}
