/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.appl;

import schmitzm.data.event.ObjectEvent;
import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.appl.AbstractHandler;
import edu.bonn.xulu.appl.VisualisationTool;
import edu.bonn.xulu.gui.XuluGUIMessages;
import edu.bonn.xulu.data.XuluObject;
import edu.bonn.xulu.io.ExportFactory;

/**
 * Diese Klasse stellt einen Handler dar, mit dem ein Objekt des Datenpools
 * (automatisiert) exportiert wird.
 * Instanzen dieses Handlers werden (interaktiv)
 * ueber eine {@link DataExportHandlerFactory} erzeugt, bzw. modifiziert.
 * @author <a href="mailto:schmitzm@bonn.edu">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class DataExportHandler extends AbstractHandler {
  /** Speichert das Objekt, das vom Handler exportiert wird. */
  protected Object object = null;

  /** Speichert die Factory, mit der das Objekt exportiert wird. */
  protected ExportFactory exportFactory = null;

  /** Speichert das Objekt, in das der Export vorgenommen wird (z.B. eine Datei). */
  protected Object destObject = null;

  /** Speichert eine Referenz auf die Xulu-Applikation. */
  protected XuluModellingPlatform appl = null;

  /**
   * Erzeugt einen neuen Handler. Dem Handler ist noch kein konkretes Objekt und
   * auch noch keine Factory (fuer den Export) zugeordnet.
   * @param fac Factory, die fuer den Handler zustaendig ist.
   * @param appl Instanz der Xulu-Applikation
   * @see #setObject(Object)
   * @see #setExportFactory(ExportFactory)
   */
  public DataExportHandler(DataExportHandlerFactory fac, XuluModellingPlatform appl) {
    this(fac,(ExportFactory)null,null,null,appl);
  }

  /**
   * Erzeugt einen neuen Handler.
   * @param obj zu exportierendes Objekt
   * @param expFac Factory Export-Factory, mit der das Objekt exportiert wird
   * @param fac Factory, die fuer den Handler zustaendig ist.
   * @param destObj Objekt in das der Export vorgenommen wird (z.B. eine Datei)
   * @param appl Instanz der Xulu-Applikation
   */
  public DataExportHandler(Object obj, ExportFactory expFac, Object destObj, DataExportHandlerFactory fac, XuluModellingPlatform appl) {
    super(fac);
    this.object  = obj;
    this.exportFactory = expFac;
    this.destObject = destObj;
    this.appl = appl;
  }

  /**
   * Schliesst den Handler. Macht nichts.
   */
  public void close() {
  }

  /**
   * Fuehrt den Handler aus. Das Objekt wird durch die Export-Factory in
   * eine Datei (oder je nach Factory mehrere Dateien) exportiert. Bereits
   * existierede Dateien werden automatisch ueberschrieben.
   * @param e Ereignis fuer das der Handler ausgefuehrt wird (wird nicht verwendet!).
   */
  public void execute(ObjectEvent e) {
    try {
      exportFactory.exportObject( object, destObject );
    } catch (Exception err) {
      XuluGUIMessages.showError(null,err);
    }
  }


  /**
   * Liefert das Objekt, das vom Handler exportiert wird.
   */
  public Object getObject() {
    return object;
  }

  /**
   * Setzt das Objekt, das vom Handler exportiert wird.
   * @param obj zu visualisierendes Objekt
   */
  public void setObject(Object obj) {
    this.object = obj;
  }

  /**
   * Liefert die Factory, mit der das Objekt exportiert wird.
   */
  public ExportFactory getExportFactory() {
    return exportFactory;
  }

  /**
   * Setzt die Factory, mit der das Objekt exportiert wird.
   * @param expFac Factory fuer den Datenexport
   */
  public void setExportFactory(ExportFactory expFac) {
    this.exportFactory = expFac;
  }

  /**
   * Liefert das Objekt, in das der Export vorgenommen wird. z.B. eine Datei.
   */
  public Object getExportDestination() {
    return this.destObject;
  }

  /**
   * Setzt das Objekt, in das der Export vorgenommen wird.
   * @param destObject Objekt in das exportiert wird (z.B. eine Datei)
   */
  public void setExportDestination(Object destObject) {
    this.destObject = destObject;
  }
}
