/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.appl;

import java.text.DateFormat;
import java.util.GregorianCalendar;
import java.awt.BorderLayout;
import javax.swing.JFrame;
import javax.swing.JLabel;

import schmitzm.lang.AbstractNamedObject;
import schmitzm.lang.WorkingThread;
import schmitzm.swing.SwingWorker;

import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.appl.XuluPlugin;


/**
 * Diese Klasse stellt ein Plugin fuer die Xulu-Applikation dar, welches die
 * aktuelle Uhrzeit (und Datum) in der Titel-Leiste des Xulu-Hauptfensters
 * anzeigt.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class DateTimePlugin implements XuluPlugin {
  private String                name        = "";
  private ClockThread           clockThread = null;

  /**
   * Erzeugt ein neues Plugin.
   */
  public DateTimePlugin() {
  }

  /**
   * Startet die Ausfuehrung des Plugins.
   * @param appl Instanz der Xulu-Applikation
   */
  public void execute(XuluModellingPlatform appl) {
    // alten Titel des Fensters merken
    this.clockThread = new ClockThread(appl);
    clockThread.start();
  }

  /**
   * Stoppt die Ausfuehrung des Plugins.
   */
  public void stop() {
    clockThread.terminate();
  }

  /**
   * Prueft, ob das Plugin gestartet ist.
   */
  public boolean isStarted() {
    return clockThread != null && clockThread.isRunning();
  }

  /**
   * Liefert eine Beschreibung des Plugins.
   */
  public String getName() {
    return name;
  }

  /**
   * Setzt die Beschreibung des Plugins.
   * @param name neue Bezeichnung
   */
  public void setName(String name) {
    this.name = name;
  }

  /**
   * Macht nichts, da das Plugin keine GUI besitzt.
   */
  public void setVisible(boolean visible) {
  }

  /**
   * Liefert immer <code>false</code>, da das Plugin keine GUI besitzt.
   */
  public boolean isVisible() {
    return false;
  }

  /**
   * Diese Klasse stellt den Uhrzeit-Thread dar, der die Uhrzeit in der
   * Titelleiste des Xulu-Hauptfensters aktualisiert.
   */
  private class ClockThread extends WorkingThread {
    private int                   lastSec;
    private String                baseTitle;
    private XuluModellingPlatform appl        = null;
    private DateFormat            dateFormat  = DateFormat.getDateTimeInstance();

    /**
     * Erzeugt einen neuen Uhrzeit-Thread.
     * @param appl Instanz der Xulu-Applikation
     */
    public ClockThread(XuluModellingPlatform appl) {
      this.appl = appl;
    }

    /**
     * Initialisiert den Thread.
     */
    public void performInit() {
      this.lastSec  = -1;
      this.baseTitle = appl.getMainFrame().getTitle();
    }

    /**
     * Implementiert die Aktualisierung der Zeit-Anzeige.
     */
    public void performWork() {
      while (true) {
        this.checkBreakingCommands();
        GregorianCalendar calendar = new GregorianCalendar();
        int actSec = calendar.get(GregorianCalendar.SECOND);
        if ( actSec != lastSec && appl.getMainFrame() != null )
          appl.getMainFrame().setTitle( baseTitle + "  [" + dateFormat.format( calendar.getTime() ) + "]");
        lastSec = actSec;
        try {
          this.sleep(100);
        }
        catch (InterruptedException err) {
        }
      }
    }

    /**
     * Wird der Thread gestoppt, wird die Titelleiste des Xulu-Hauptfensters
     * auf ihren urspruenglichen Wert zurueckgesetzt.
     */
    public void performDispose() {
      appl.getMainFrame().setTitle(baseTitle);
    }
  }
}
