/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.appl;

import schmitzm.swing.MultipleOptionPane;
import schmitzm.swing.InputOption;
import schmitzm.swing.SelectionInputOption;

import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.appl.Handler;
import edu.bonn.xulu.appl.HandlerFactory;
import edu.bonn.xulu.appl.VisualisationTool;
import edu.bonn.xulu.appl.XuluConstants;
import edu.bonn.xulu.io.Factory;
import edu.bonn.xulu.io.AbstractFactory;
import edu.bonn.xulu.gui.DataPoolInputOption;
import edu.bonn.xulu.io.ExportFactory;
import schmitzm.swing.FileInputOption;
import java.io.File;
import schmitzm.swing.event.InputOptionAdapter;
import edu.bonn.xulu.gui.XuluGUIMessages;
import java.util.Vector;
import schmitzm.data.event.ObjectEvent;
import schmitzm.io.IOUtil;
import java.text.DecimalFormat;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import edu.bonn.xulu.model.event.AbstractStepModelEvent;
import edu.bonn.xulu.model.event.AbstractIterationStepEvent;

/**
 * Mit dieser Factory werden Instanzen des {@link FileExportHandler}
 * erstellt und modifiziert. Die benoetigten Informationen (zu exportierendes
 * Datenpool-Objekt, Ausgabe-Datei und Export-Factory) werden interaktiv
 * ueber Dialoge vom Anwender abgefragt.
 * @author <a href="mailto:schmitzm@bonn.edu">Martin Schmitz</a> (University of  Bonn/Germany)
 * @version 1.0
 */
public class FileExportHandlerFactory extends DataExportHandlerFactory {
  private File          originalDestFile      = null;
  private Matcher       counterPatternMatcher = null;
  private DecimalFormat counterFormat         = null;

  /**
   * Erzeugt eine neue Handler-Factory.
   */
  public FileExportHandlerFactory() {
    this("");
  }

  /**
   * Erzeugt eine neue Handler-Factory.
   * @param name Beschreibung der Factory
   */
  public FileExportHandlerFactory(String name) {
    super(name);
  }

  /**
   * Erzeugt einen neuen {@link DataExportHandler}.
   * @param appl Instanz der Xulu-Applikation
   */
  public DataExportHandler createHandler(XuluModellingPlatform appl) {
    DataExportHandler handler = super.createHandler(appl);
    if ( handler == null )
      return null; // Dialog abgebrochen

    // neue Handler-Instanz erzeugen, die vor dem Datei-Schreiben den
    // Dateinamen je nach Event um eine ID erweitert, damit die Datei nicht
    // ueberschrieben wird
    handler = new DataExportHandler(handler.getObject(),handler.getExportFactory(),handler.getExportDestination(),(DataExportHandlerFactory)handler.getFactory(),appl) {
                    public void execute(ObjectEvent e) {
                      if ( counterFormat != null ) {
                        String eventID = "";
                        // Bei StepModelEvent wird der Dateiname um die Step-No.
                        // erweitert
                        if ( e instanceof AbstractStepModelEvent )
                          eventID += counterFormat.format( ((AbstractStepModelEvent)e).getStepNo() );
                        // Bei IterationStepEvent wird der Dateiname ZUDEM
                        // um die Iteration-No. erweitert
                        if ( e instanceof AbstractIterationStepEvent )
                          eventID += "." + counterFormat.format( ((AbstractIterationStepEvent)e).getIterNo() );
                        // Export-Ziel durch neue Datei ersetzen
                        destObject = new File(
                            counterPatternMatcher.replaceFirst(eventID)
                        );
                      }
                      // Export durchfuehren
                      super.execute(e);
                      // Export-Ziel wieder durch Original ersetzen, damit
                      // EventHandler unveraendert bleibt
                      destObject = originalDestFile;
                    }
    };

    return handler;
  }

  /**
   * Ermittelt die fuer den {@link DataExportHandler} benoetigten Informationen
   * ueber einen Anwender-Dialog.
   * @param appl Instanz der XuluModellingPlatform
   * @param defObject vorgeblendetes Objekt, welches exportiert wird (kann {@code null} sein)
   * @param defExpFac vorgeblendete Export-Factory (kann {@code null} sein)
   * @param defDestObj vorgeblendete Export-Datei (kann {@code null} sein)
   * @return Array in dem in Element 0 das ausgewaehlte Datenpool-Objekt,
   *         in Element 1 die Export-Factory und in Element 2 die Ziel-Datei
   *         enthalten ist
   * @exception IllegalArgumentException falls {@code defDestObj} kein {@link File} ist
   */
  protected Object[] showDialog(final XuluModellingPlatform appl, Object defObject, ExportFactory defExpFac, Object defDestObj) {
    if ( defDestObj != null && !(defDestObj instanceof File) )
      throw new IllegalArgumentException(getClass().getSimpleName()+" only accepts File as destination Object: "+defDestObj.getClass().getSimpleName());

    // existierende Export-Factories ermitteln
    ExportFactory[] expFactory     = appl.getRegistry().getExportFactories();
    String[]        expFactoryDesc = new String[expFactory.length];
    for (int i=0; i<expFactoryDesc.length; i++)
      expFactoryDesc[i] = expFactory[i].getName();

    final DataPoolInputOption  dataPoolOption = new DataPoolInputOption(appl.getDataPool(),XuluConstants.XULUGUI_RES.getString("ExportSource"),true,null);
    final SelectionInputOption factoryOption  = new SelectionInputOption.Combo(XuluConstants.XULUGUI_RES.getString("ExportFactory")+":",true);
    final FileInputOption      fileOption     = new FileInputOption(XuluConstants.XULUGUI_RES.getString("ExportDest")+":",true);

    // wenn sich das zu Datenpool-Objekt aendert, muss die Auswahl der
    // moeglichen Export-Factories entsprechend aktualisiert werden
    dataPoolOption.addInputOptionListener( new InputOptionAdapter() {
      public void optionChanged(InputOption option, Object oldValue, Object newValue) {
        // Export-Factories ermitteln, die den Datentyp unterstuetzen und
        // in eine Datei exportieren
        ExportFactory[] allExpFactory     = appl.getRegistry().getExportFactories();
        Vector<ExportFactory> expFactoryVec = new Vector();
        for ( ExportFactory expFac : allExpFactory )
          if ( File.class.isAssignableFrom( expFac.getExportDestinationType() ) &&
               expFac.isExportable( newValue ) )
            expFactoryVec.add( expFac );
        ExportFactory[] expFactory = expFactoryVec.toArray( new ExportFactory[0] );
        String[]        expFactoryDesc = new String[expFactory.length];
        for (int i=0; i<expFactoryDesc.length; i++)
          expFactoryDesc[i] = expFactory[i].getName();
        factoryOption.setSelectionObjects(expFactory,expFactoryDesc);
      }
    });

    // etwaigen Default-Werte fuer Datenpool-Objekt erst jetzt setzen, damit
    // Listener aufgerufen wird
    dataPoolOption.setValue( defObject );
    factoryOption.setValue( defExpFac );
    fileOption.setValue( (File)defDestObj );

    // Dialog erzeugen
    Object[] input = MultipleOptionPane.showMultipleInputDialog(null,XuluConstants.XULUGUI_RES.getString("Handler"), new InputOption[] {
        dataPoolOption,
        factoryOption,
        fileOption
    });

    // Dialog abgebrochen
    if ( input == null || input.length < 3 )
      return null;

    // Wildcard "{number_format}" fuer Datei-Nummerierung aus
    // Ziel-Dateinamen parsen und daraus ein NumberFormat erstellen
    originalDestFile      = (File)input[2];
    String exportFileName = originalDestFile.getAbsolutePath();
    counterPatternMatcher = Pattern.compile("\\{.+\\}").matcher(exportFileName);
    if ( counterPatternMatcher.find() ) {
      String matchPattern = counterPatternMatcher.group();   // = "{000...00}"
      String counterPattern = matchPattern.substring(1,matchPattern.length()-1); // = "000...00"
      counterFormat = new DecimalFormat(counterPattern);
    } else
      counterFormat = null; // WICHTIG, da Variable auch als Flag verwendet wird!!

    return input;
  }
}
