/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.appl;

import schmitzm.data.event.ObjectEvent;
import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.appl.AbstractHandler;
import edu.bonn.xulu.appl.VisualisationTool;
import edu.bonn.xulu.gui.XuluGUIMessages;
import edu.bonn.xulu.data.XuluObject;

/**
 * Diese Klasse stellt einen Handler dar, mit dem die Visualisierung eines
 * Objekts aktualisiert wird. Instanzen dieses Handlers werden (interaktiv)
 * ueber eine {@link VisualisationHandlerFactory} erzeugt, bzw. modifiziert.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class VisualisationHandler extends AbstractHandler {
  /** Speichert das Objekt, das vom Handler dargestellt/aktualisiert wird. */
  protected Object object = null;

  /** Speichert das Visualisierungstool, in dem das Objekt dargestellt wird/ist. */
  protected VisualisationTool visTool = null;

  /** Speichert das Klasse des Visualisierungstool, in dem das Objekt dargestellt wird/ist. */
  protected Class<VisualisationTool> visToolClass = null;

  /** Speichert eine Referenz auf die Xulu-Applikation. */
  protected XuluModellingPlatform appl = null;

  /**
   * Erzeugt einen neuen Handler. Dem Handler ist noch kein konkretes Objekt und
   * auch noch kein Visualisierungstool (fuer die Darstellung) zugeordnet.
   * @param fac Factory, die fuer den Handler zustaendig ist.
   * @param appl Instanz der Xulu-Applikation
   * @see #setObject(Object)
   * @see #setVisualisationTool(VisualisationTool)
   */
  public VisualisationHandler(VisualisationHandlerFactory fac, XuluModellingPlatform appl) {
    this(fac,(VisualisationTool)null,null,appl);
  }

  /**
   * Erzeugt einen neuen Handler.
   * @param obj zu visualisierendes Objekt
   * @param visTool Tool, in dem das Objekt vom Handler visualisiert wird
   * @param fac Factory, die fuer den Handler zustaendig ist.
   * @param appl Instanz der Xulu-Applikation
   */
  public VisualisationHandler(Object obj, VisualisationTool visTool, VisualisationHandlerFactory fac, XuluModellingPlatform appl) {
    this(obj,(visTool != null) ? (Class<VisualisationTool>)visTool.getClass() : null,fac,appl);
    this.visTool = visTool;
  }

  /**
   * Erzeugt einen neuen Handler. Beim ersten Aufruf des Handlers wird eine
   * Instanz des Visualisierungstools erstellt.
   * @param obj zu visualisierendes Objekt
   * @param visToolClass Klasse des Tools, in dem das Objekt vom Handler visualisiert wird
   * @param fac Factory, die fuer den Handler zustaendig ist.
   * @param appl Instanz der Xulu-Applikation
   */
  public VisualisationHandler(Object obj, Class<VisualisationTool> visToolClass, VisualisationHandlerFactory fac, XuluModellingPlatform appl) {
    super(fac);
    this.object  = obj;
    this.visTool = null;
    this.visToolClass = visToolClass;
    this.appl = appl;
  }

  /**
   * Schliesst den Handler. Macht nichts.
   */
  public void close() {
  }

  /**
   * Fuehrt den Handler aus. Das Objekt wird im Visualisierungstool aktualisiert,
   * bzw. dem Visualisierungstool hinzugefuegt, falls es noch nicht dargestellt
   * wird.
   * @param e Ereignis fuer das der Handler ausgefuehrt wird (wird nicht verwendet!).
   */
  public void execute(ObjectEvent e) {
    // Wenn das Vis-Tool noch nicht existiert, wird eine Instanz erstellt
    // und dem Vis-Manager hinzugefuegt
    if ( visTool == null )
      try {
        visTool = visToolClass.newInstance();
        appl.getVisualisationManager().add(visTool,true);
      } catch (Exception err) {
          XuluGUIMessages.showError(null,err);
      }

    // Wenn Objekt noch nicht im Vis-Tool dargestellt wird, wird es
    // hinzugefuegt
    if ( !visTool.update(object) ) {
      String desc = "neues Objekt";
      if ( object instanceof XuluObject )
        desc =  ((XuluObject)object).getDescription();
      visTool.add(object, desc);
    }
  }


  /**
   * Liefert das Objekt, das vom Handler dargestellt/aktualisiert wird.
   */
  public Object getObject() {
    return object;
  }

  /**
   * Setzt das Objekt, das vom Handler dargestellt/aktualisiert wird.
   * @param obj zu visualisierendes Objekt
   */
  public void setObject(Object obj) {
    this.object = obj;
  }

  /**
   * Liefert das Visualisierungstool, in dem der Handler das Objekt darstellt.
   */
  public VisualisationTool getVisualisationTool() {
    return visTool;
  }

  /**
   * Setzt das Visualisierungstool, in dem der Handler das Objekt darstellt.
   * @param visTool Tool, in dem das Objekt vom Handler visualisiert wird
   */
  public void setVisualisationTool(VisualisationTool visTool) {
    this.visTool = visTool;
    this.visToolClass = (visTool==null) ? null : (Class<VisualisationTool>)visTool.getClass();
  }

  /**
   * Setzt die Klasse des Visualisierungstools, in dem der Handler das Objekt darstellt.
   * Ist dem Handler bereits ein konkretes Visualisierungstool zugeordnet und
   * entspricht dieses <b>nicht</b> dieser Klasse, so wird das bestehende
   * Visualisierungstool ({@link #visTool}) zurueckgesetzt.
   * @param visToolClass Tool, in dem das Objekt vom Handler visualisiert wird
   */
  public void setVisualisationToolClass(Class<VisualisationTool> visToolClass) {
    this.visToolClass = visToolClass;
    if ( !(visToolClass.isInstance(this.visTool)) )
      this.visTool = null;
  }

  /**
   * Liefert die Klasse des Visualisierungstools, in dem der Handler das Objekt darstellt.
   */
  public Class<VisualisationTool> getVisualisationToolClass() {
    return this.visToolClass;
  }
}
