/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.appl;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.BufferedReader;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.MalformedURLException;

import schmitzm.io.IOUtil;
import schmitzm.lang.NamedObject;
import schmitzm.lang.LangUtil;

import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.io.InstantiationFactory;
import edu.bonn.xulu.io.ImportFactory;
import edu.bonn.xulu.io.ExportFactory;
import edu.bonn.xulu.io.TypeMapping;

import edu.bonn.xulu.appl.XuluPlugin;
import edu.bonn.xulu.appl.ScriptInterpreter;
import edu.bonn.xulu.appl.XuluRegistry;
import edu.bonn.xulu.appl.XuluRegistryReader;


// nur fuer Doku
import edu.bonn.xulu.data.XuluObject;
import edu.bonn.xulu.appl.VisualisationTool;
import edu.bonn.xulu.model.XuluModel;

/**
 * Diese Klasse dient dazu, die Xulu-Registry aus einem ASCII-InputStream
 * (z.B. einer ASCII-Datei) einzulesen.<br>
 * Dieser muss folgendes (zeilenweises) Format haben:<br>
 * <br>
 * <code>[Global]</code><br>
 * <i><small>Alle Klassen in diesem Bereich werden in der Xulu-Registry fuer alles
 *    registriert, wofuer sie geeignet sind (siehe
 *    {@link XuluRegistry#register(Object)}).</small></i><br>
 * <code><i>package.class</i> [<i>name</i>]<br>
 *       <i>package.class</i> [<i>name</i>]<br>
 *       ...<br></code>
 * <code>[DataType]</code><br>
 * <i><small>Alle Klassen in diesem Bereich werden in der Xulu-Registry als
 *    Datentyp registriert. Sie muessen (Sub)Klassen von {@link XuluObject}
 *    darstellen.</small></i><br>
 * <code><i>package.class</i><br>
 *       <i>package.class</i><br>
 *       ...<br></code>
 * <code>[DefaultFactory]</code><br>
 * <i><small>Alle Klassen in diesem Bereich werden in der Xulu-Registry nur als
 *    Default-Factory registriert. Sie muessen (Sub)Klassen von
 *    {@link InstantiationFactory} darstellen und
 *    ueber einen Standard-Konstruktor instanziierbar sein.</small></i><br>
 * <code><i>package.class</i> [<i>name</i>]<br>
 *       <i>package.class</i> [<i>name</i>]<br>
 *       ...<br></code>
 * <code>[ImportFactory]</code><br>
 * <i><small>Alle Klassen in diesem Bereich werden in der Xulu-Registry nur als
 *    Import-Factory registriert. Sie muessen (Sub)Klassen von
 *    {@link ImportFactory} darstellen und
 *    ueber einen Standard-Konstruktor instanziierbar sein.</small></i><br>
 * <code><i>package.class</i> [<i>name</i>]<br>
 *       <i>package.class</i> [<i>name</i>]<br>
 *       ...<br></code>
 * <code>[ExportFactory]</code><br>
 * <i><small>Alle Klassen in diesem Bereich werden in der Xulu-Registry nur als
 *    Export-Factory registriert. Sie muessen (Sub)Klassen von
 *    {@link ExportFactory} darstellen und
 *    ueber einen Standard-Konstruktor instanziierbar sein.</small></i><br>
 * <code><i>package.class</i> [<i>name</i>]<br>
 *       <i>package.class</i> [<i>name</i>]<br>
 *       ...<br></code>
 * <code>[TypeMapping]</code><br>
 * <i><small>Alle Tupel in diesem Bereich werden in der Xulu-Registry nur als
 *    TypeMapping registriert. Die Typ-Klasse muss eine (Sub)Klasse von
 *    {@link XuluObject} sein und die Default-Factory von {@link InstantiationFactory}.
 *    Beides sind Pflichtangaben fuer ein TypeMapping. Alle weiteren Angaben sind optional
 *    und repraesentieren Im- und Export-Factorys. Sie muessen (Sub)Klassen von
 *    {@link ImportFactory} bzw. {@link ExportFactory} darstellen und
 *    ueber einen Standard-Konstruktor instanziierbar sein.</small></i><br>
 * <code><i>package.typeClass package.defaultFacClass [package.factoryClass ...]</i><br>
 *       <i>package.typeClass package.defaultFacClass [package.factoryClass ...]</i><br>
 *       ...<br></code>
 * <code>[Visualisation]</code><br>
 * <i><small>Alle Klassen in diesem Bereich werden in der Xulu-Registry als
 *    Visualisierungstool registriert. Sie muessen (Sub)Klassen von
 *    {@link VisualisationTool} darstellen und
 *    ueber einen Standard-Konstruktor instanziierbar sein.</small></i><br>
 * <code><i>package.class</i><br>
 *       <i>package.class</i><br>
 *       ...<br></code>
 * <code>[Model]</code><br>
 * <i><small>Alle Klassen in diesem Bereich werden in der Xulu-Registry als
 *    Modell-Klasse registriert. Sie muessen (Sub)Klassen von {@link XuluModel}
 *    darstellen und
 *    ueber einen Standard-Konstruktor instanziierbar sein.</small></i><br>
 * <code><i>package.class</i><br>
 *       <i>package.class</i><br>
 *       ...<br></code>
 * <code>[Script]</code><br>
 * <i><small>Alle Klassen in diesem Bereich werden in der Xulu-Registry als
 *    Skript-Interpreter registriert. Sie muessen (Sub)Klassen von {@link ScriptInterpreter}
 *    darstellen und ueber einen Standard-Konstruktor instanziierbar sein.</small></i><br>
 * <code><i>package.class</i> [<i>name</i>]<br>
 *       <i>package.class</i> [<i>name</i>]<br>
 *       ...<br></code>
 * <code>[Plugin]</code><br>
 * <i><small>Alle Klassen in diesem Bereich werden in der Xulu-Registry als
 *    Xulu-Plugin registriert. Sie muessen (Sub)Klassen von {@link XuluPlugin}
 *    darstellen und ueber einen Standard-Konstruktor instanziierbar sein.</small></i><br>
 * <code><i>package.class</i> [AUTOSTART] [<i>name</i>]<br>
 *       <i>package.class</i> [AUTOSTART] [<i>name</i>]<br>
 *       ...<br></code>
 * <br>
 * Leerzeilen und Zeilen, die mit <code>//</code>, <code>|</code> oder <code>#</code>
 * beginnen, werden als Kommentare komplett ignoriert.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class XuluRegistryReader_BasicAscii implements XuluRegistryReader {
  public static final String TAG_GLOBAL            = "[Global]";
  public static final String TAG_DATATYPE          = "[DataType]";
  public static final String TAG_DEFAULTFAC        = "[DefaultFactory]";
  public static final String TAG_IMPORTFAC         = "[ImportFactory]";
  public static final String TAG_EXPORTFAC         = "[ExportFactory]";
  public static final String TAG_TYPEMAPPING       = "[TypeMapping]";
  public static final String TAG_VISUALISATION     = "[Visualisation]";
  public static final String TAG_MODEL             = "[Model]";
  public static final String TAG_SCRIPTINTERPRETER = "[Script]";
  public static final String TAG_PLUGIN            = "[Plugin]";

  private static final int TYPE_UNKNOWN           = -99;
  private static final int TYPE_DATALINE          = -1;
  private static final int TYPE_COMMENT           = 0;
  private static final int TYPE_DATATYPE          = 1;
  private static final int TYPE_DEFAULTFAC        = 2;
  private static final int TYPE_IMPORTFAC         = 3;
  private static final int TYPE_EXPORTFAC         = 4;
  private static final int TYPE_TYPEMAPPING       = 5;
  private static final int TYPE_VISUALISATION     = 6;
  private static final int TYPE_GLOBAL            = 7;
  private static final int TYPE_MODEL             = 8;
  private static final int TYPE_SCRIPTINTERPRETER = 9;
  private static final int TYPE_PLUGIN            = 10;

  private XuluModellingPlatform appl = null;

  /**
   * Erzeugt einen neuen Reader fuer die Xulu-Registry.
   */
  public XuluRegistryReader_BasicAscii() {
  }

  /**
   * Bestimmt die Art, wie die folgenden Zeilen interpretiert werden muessen.
   */
  private int determineLineType(String line) {
    if ( line==null || line.trim().equals("") || IOUtil.isCommentLine(line) )
      return TYPE_COMMENT;

    if ( line.startsWith("[") ) {
      if ( line.equals(TAG_GLOBAL) )            return TYPE_GLOBAL;
      if ( line.equals(TAG_DATATYPE) )          return TYPE_DATATYPE;
      if ( line.equals(TAG_DEFAULTFAC) )        return TYPE_DEFAULTFAC;
      if ( line.equals(TAG_IMPORTFAC) )         return TYPE_IMPORTFAC;
      if ( line.equals(TAG_EXPORTFAC) )         return TYPE_EXPORTFAC;
      if ( line.equals(TAG_TYPEMAPPING) )       return TYPE_TYPEMAPPING;
      if ( line.equals(TAG_VISUALISATION) )     return TYPE_VISUALISATION;
      if ( line.equals(TAG_MODEL) )             return TYPE_MODEL;
      if ( line.equals(TAG_SCRIPTINTERPRETER) ) return TYPE_SCRIPTINTERPRETER;
      if ( line.equals(TAG_PLUGIN) )            return TYPE_PLUGIN;
      return TYPE_UNKNOWN;
    }

    return TYPE_DATALINE;
  }

  /**
   * Interpretiert eine Zeile als TypeMapping.
   */
  private TypeMapping readTypeMapping(String line) throws Exception {
    StringTokenizer st = new StringTokenizer(line);

    try {
      // 1. Token: DatenTyp
      Class dataType = loadClass(st.nextToken(),false);
      // 2. Token: DefaultFactory
      InstantiationFactory defaultFac = (InstantiationFactory)loadClass(st.nextToken(),false).newInstance();
      TypeMapping mapping = new TypeMapping(dataType,defaultFac);
      // alle weiteren Token: Im- oder ExportFactorys
      while ( st.hasMoreTokens() ) {
        Object fac = loadClass(st.nextToken(),false).newInstance();
        if ( fac instanceof ImportFactory )
          mapping.addImportFactory( (ImportFactory)fac );
        if ( fac instanceof ExportFactory )
          mapping.addExportFactory( (ExportFactory)fac );
        if ( !(fac instanceof ImportFactory || fac instanceof ExportFactory) )
          throw new Exception("Invalid TypeMapping: Import or export factory expected.");
      }
      return mapping;
    } catch (NoSuchElementException err) {
      throw new Exception("Invalid TyperMapping: First two parameters are mandatory.");
    }
  }

  /**
   * Schreibt eine Fehlermeldung in die Standard-Fehlerausgabe.
   * @param lineNo Zeilennummer der Datei in der der Fehler auftrat
   * @param line   komplette Zeile, die den Fehler verursacht hat
   * @param mess   zusaetzliche Fehlermeldung
   */
  private void giveErrorMess(int lineNo, String line, String mess) {
    System.err.println( getClass().getSimpleName().concat(" --> ").concat(mess) );
    System.err.println( "     Line ".concat(String.valueOf(lineNo)).concat(": ").concat(line) );
  }

  /**
   * Trennt eine Zeile am ersten Leerzeichen/Tab.
   */
  private String[] divideLine(String line) {
    StringTokenizer st = new StringTokenizer(line,"\n\t ");
    String[] ret = new String[2];
    ret[0] = st.nextToken().trim();
    if ( st.hasMoreTokens() )
      ret[1] = st.nextToken("\n").trim();
    return ret;
  }

  /**
   * Instanziiert eine Klasse und wenn es sich dabei um eine Instanz
   * von {@link NamedObject} handelt, wird es benannt.
   * @param line Zeile aus dem Eingabe-File
   * @return das instanziierte Objekt
   */
  private Object instantiateAndName(String line) throws Exception {
    String[] linePart = divideLine(line);
    Object o = loadClass(linePart[0],false).newInstance();
    if ( o instanceof NamedObject )
      ((NamedObject)o).setName(linePart[1]);
    return o;
  }

  /**
   * Erweitert eine bestehende XuluRegistry. Bestehende Eintraege werden
   * ueberschrieben. Ist die uebergebene Registy <code>null</code> wird
   * eine neue (leere) Instanz erstellt.
   * Kann ein Eintrag aus dem Stream nicht richtig interpretiert oder
   * instanziiert werden, wird die Zeile ignoriert und eine Fehlermeldung
   * in die Standard-Fehlerausgabe geschrieben.
   * @param appl     Instanz der Xulu-Applikation fuer die die Registry eingelesen
   *                 werden soll
   * @param input    Eingabe-Stream aus dem die Registry-Daten gelesen werden
   * @param registry XuluRegistry die erweitert werden soll (kan <code>null</code> sein
   * @return Instanz der XuluRegistry, die erstellt oder erweitert wurde
   * @throws java.io.IOException falls ein unerwarteter Fehler beim Einlesen auftritt.
   */
  public XuluRegistry readRegistry(XuluModellingPlatform appl, InputStream input, XuluRegistry registry) throws IOException {
    this.appl = appl;
    if ( registry==null )
      registry = new XuluRegistry();

    BufferedReader in = new BufferedReader( new InputStreamReader(input) );

    // Es soll nicht fuer jede eingelesene Zeile ein Event geworfen werden.
    // Eine generelle Mitteilung an die Listener am Ende reicht.
    registry.setEventFiringEnabled(false);

    String line = "";
    int    typeOfFollowingData = TYPE_COMMENT;
    for (int lineNo=1; in.ready() && (line = in.readLine()) != null; lineNo++) {
      // Typ der Zeile ermitteln
      int type = determineLineType(line);

      // unbekannte Zeile
      if ( type == TYPE_UNKNOWN ) giveErrorMess(lineNo,line,"Unknown registry tag");
      // Kommentarzeile ignorieren
      if ( type == TYPE_COMMENT ) continue;
      // Tag-Zeile
      if ( type > 0 ) typeOfFollowingData = type;
      // Zeile mit Registry-Data
      Class expClass  = null;
      try {
        if ( type == TYPE_DATALINE )
          switch (typeOfFollowingData) {
            case TYPE_GLOBAL: registry.register( instantiateAndName(line) );
                              break;
            case TYPE_DATATYPE: registry.registerDataType( loadClass(line,false) );
                                break;
            case TYPE_DEFAULTFAC: expClass = InstantiationFactory.class;
                                  registry.registerDefaultFactory( (InstantiationFactory)instantiateAndName(line) );
                                  break;
            case TYPE_IMPORTFAC: expClass = ImportFactory.class;
                                 registry.registerImportFactory( (ImportFactory)instantiateAndName(line) );
                                 break;
            case TYPE_EXPORTFAC: expClass = ExportFactory.class;
                                 registry.registerExportFactory( (ExportFactory)instantiateAndName(line) );
                                 break;
            case TYPE_TYPEMAPPING: registry.registerTypeMapping( (TypeMapping)readTypeMapping(line) );
                                   break;
            case TYPE_VISUALISATION: registry.registerVisualisationTool( loadClass(line,false) );
                                     break;
            case TYPE_MODEL: registry.registerModelClass( loadClass(line,true) );
                             break;
            case TYPE_SCRIPTINTERPRETER: expClass = ScriptInterpreter.class;
                                         registry.registerScriptInterpreter( (ScriptInterpreter)instantiateAndName(line) );
                                         break;
            case TYPE_PLUGIN: expClass = XuluPlugin.class;
                                         String[] linePart = divideLine(line);
                                         XuluPlugin plugin = (XuluPlugin)loadClass(linePart[0],false).newInstance();
                                         boolean autostart = linePart[1].startsWith("AUTOSTART");
                                         String  name      = autostart ? linePart[1].substring(9).trim() : linePart[1];
                                         plugin.setName( name );
                                         if ( autostart )
                                           registry.registerAutostartPlugin( plugin );
                                         else
                                           registry.registerPlugin( plugin );
                                         break;
          }
      } catch (ClassCastException err) {
        giveErrorMess(lineNo,line,"Instance of ".concat(expClass.getName()).concat(" expected"));
      } catch (Exception err) {
        // Datenzeile nicht korrekt
        giveErrorMess(lineNo,line,err.toString());
      }
    }

    // Eine generelle Mitteilung an die Listener, dass sich die Registry
    // geaendert hat
    registry.setEventFiringEnabled(true);
    registry.fireGeneralEvent();

    input.close();
    return registry;
  }

  /**
   * Erstellt eine neue XuluRegistry und befuellt diese.
   * @param appl     Instanz der Xulu-Applikation fuer die die Registry eingelesen
   *                 werden soll
   * @param input    Eingabe-Stream aus dem die Registry-Daten gelesen werden
   * @return Instanz der XuluRegistry, die erstellt wurde
   * @throws java.io.IOException falls ein Fehler beim Einlesen auftritt
   */
  public XuluRegistry readRegistry(XuluModellingPlatform appl, InputStream input) throws IOException {
    return readRegistry(appl,input,null);
  }

  /**
   * Laed eine Klasse. Ggf. wird dazu ein eigener eigener {@link ClassLoader}
   * erzeugt, so dass die Klasse dynamisch neu geladen werden kann.
   * Als Suchpfad fuer diesen ClassLoader wird der fuer die Xulu-Applikation
   * definierte Pfad fuer dynamische Klassen verwendet (siehe
   * {@link XuluModellingPlatform#getDynamicClassRootDirectory()}).
   * @param name  Name der zu ladenen Klasse
   * @param newCL Flag, ob zum Laden der Klasse ein eigener {@link ClassLoader}
   *              instanziiert werden soll.
   */
  private Class loadClass(String name, boolean newCL) throws ClassNotFoundException, MalformedURLException {
//    URL searchPath = newCL ? appl.getDynamicClassRootDirectory().toURL() : null;
//    return LangUtil.loadClass(name,searchPath);
      if ( newCL && appl != null )
        return LangUtil.loadPreferredClass(name,appl.getDynamicClassRootDirectory().toURL());
      return Class.forName(name);
  }
}
