/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.data.grid;


import schmitzm.data.WritableGrid;
import schmitzm.data.property.ListProperty;
import schmitzm.data.property.ScalarProperty;
import edu.bonn.xulu.data.AbstractXuluObject;

/**
 * Dieser Datentyp stellt eine Liste von Rastern ({@link WritableGrid} da.
 * In der Liste koennen beliebige Raster (auch unterschiedlicher Groesse und
 * Georeferenz) gespeichert werden. Deshalb hat das Objekt auch nur zwei
 * Eigenschaften (Properties):<br>
 * <ol>
 * <li>{@linkplain #PROP_GRIDS Grids}: Die Liste der Raster (ListProperty[WritableGrid])</li>
 * <li>{@linkplain #PROP_STYLE Style}: Darstellungs-Style fuer die Raster (ScalarProperty[Object])</li>
 * <li>{@linkplain #PROP_COUNT Count}: Die Anzahl der Raster in der Liste (ScalarProperty[Integer])</li>
 * </ol>
 * Die Eigenschaft 3 ist direkt von der zu grunde liegenden Liste
 * abhaengig (Eigenschaft 1) und kann deshalb nicht veraendert werden. Ein schreibender
 * Zugriff auf diese Property hat keine Auswirkung.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class GridList extends AbstractXuluObject {
  /** Name der Eigenschaft, die die Liste der {@link WritableGrid} beinhaltet. */
  public static final String PROP_GRIDS = "Grids";
  /** Name der Eigenschaft, die die aktuelle Listengroesse beinhaltet. */
  public static final String PROP_COUNT = "Count";

  /**
   * Erzeugt eine neue (leere) Raster-Liste.
   */
  public GridList() {
    this( new WritableGrid[0] );
  }

  /**
   * Erzeugt eine neue Raster-Liste.
   * @param gc erstes Raster fuer die Liste
   */
  public GridList(WritableGrid gc) {
    this( new WritableGrid[] { gc } );
  }

  /**
   * Erzeugt eine neue Raster-Liste.
   * @param gc Raster fuer die Liste
   */
  public GridList(WritableGrid[] gc) {
    // Beim Freigeben der Property sollen auch die Grids zerstoert werden
    propSet.addProperty( new ListProperty(PROP_GRIDS,WritableGrid.class) {
      public void dispose() {
        for (int i=0; i<elements.size(); i++)
          ((WritableGrid)elements.elementAt(i)).dispose();
        // die Elements selber werden ueber die Super-Methode geloescht!
        super.dispose();
      }
    } );
    // Property fuer den Style der Grids
    propSet.addProperty( new ScalarProperty(PROP_STYLE,Object.class) );
    // Property fuer die Anzahl der Grids
    propSet.addProperty( new ScalarProperty(PROP_COUNT,Integer.class) {
      protected Object getValue(int... coords) {
        super.getValue(coords);
        return getGridCount();
      }
    });
    // Raster hinzufuegen
    for (WritableGrid wg : gc)
      addGrid(wg);
  }

  /**
   * Liefert die Anzahl an Rastern in der Liste.
   */
  public int getGridCount() {
    return ((ListProperty)getProperty(PROP_GRIDS)).getOneTimeReadAccess().getCount();
  }

  /**
   * Liefert ein Raster aus der Liste.
   * @param idx Listen-Index
   */
  public WritableGrid getGrid(int idx) {
    return (WritableGrid) ((ListProperty)getProperty(PROP_GRIDS)).getOneTimeReadAccess().getValue(idx);
  }

  /**
   * Fuegt der Liste ein Raster hinzu.
   */
  public void addGrid(WritableGrid grid) {
    ( (ListProperty)getProperty(PROP_GRIDS) ).getOneTimeWriteAccess().addValue(grid);
  }

  /**
   * Entfernt ein Raster aus der Liste. Ist das Raster nicht in der Liste
   * vorhanden geschieht nichts.
   * @param grid zu entfernendes Raster
   */
  public void removeGrid(WritableGrid grid) {
    ( (ListProperty)getProperty(PROP_GRIDS) ).getOneTimeWriteAccess().removeValue(grid);
  }

  /**
   * Entfernt ein Raster aus der Liste.
   * @param idx Listen-Index
   */
  public void removeGrid(int idx) {
    ( (ListProperty)getProperty(PROP_GRIDS) ).getOneTimeWriteAccess().removeValue(idx);
  }

  /**
   * Ersetzt ein Raster in der Liste.
   * @param idx Listen-Index
   * @param grid zu entfernendes Raster
   */
  public void replaceGrid(int idx, WritableGrid grid) {
    ( (ListProperty)getProperty(PROP_GRIDS) ).getOneTimeWriteAccess().setValue(grid,idx);
  }

  /**
   * Liefert das erste Grid der Liste, das nicht <code>null</code> ist.
   * @return <code>null</code> falls es kein solches Grid gibt
   */
  public WritableGrid getFirstNotNullGrid() {
    return getFirstNotNullGrid(0);
  }

  /**
   * Liefert das erste Grid der Liste ab einem bestimmten Index, das
   * nicht <code>null</code> ist.
   * @return <code>null</code> falls es kein solches Grid gibt
   */
  public WritableGrid getFirstNotNullGrid(int firstIdx) {
    int idx = determineFirstNotNullIndex(firstIdx);
    if (idx >= 0)
      return getGrid(idx);
    return null;
  }

  /**
   * Liefert den ersten Listen-Index, der kein null-Element enthaelt.
   * @param firstIdx Index bei dem zu suchen begonnen wird
   */
  private int determineFirstNotNullIndex(int firstIdx) {
    for (int i=firstIdx; i<getGridCount(); i++)
      if ( getGrid(i)!=null )
        return i;
    return -1;
  }
}
