/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.data.grid;

import org.opengis.referencing.crs.CoordinateReferenceSystem;

import schmitzm.data.WritableGrid;
import schmitzm.data.AbstractWritableGrid;
import schmitzm.data.property.ScalarProperty;
import edu.bonn.xulu.plugin.io.grid.WritableGridFactory;

// nur fuer Doku
import java.awt.image.DataBuffer;

/**
 * Dieser Datentyp stellt eine Liste von Rastern ({@link WritableGrid} da.
 * In der Liste koennen nur gleichartige Raster (gleiche Groesse und gleiche
 * Georeferenz) gespeichert werden. Deshalb hat das Objekt neben der Raster-Liste
 * noch weitere Eigenschaften (Properties):<br>
 * <ol>
 * <li>{@linkplain GridList#PROP_GRIDS GridListe}: Die Liste der Raster (ListProperty[WritableGrid])</li>
 * <li>{@linkplain XuluObject#PROP_STYLE Style}: Darstellungs-Style fuer die Raster (ScalarProperty[Object])</li>
 * <li>{@linkplain GridList#PROP_COUNT Count}: Die Anzahl der Raster in der Liste (ScalarProperty[Integer])</li>
 * <li>{@linkplain #PROP_COORDX X-Coordinate}: Georeferenz Longitude (ScalarProperty[Double])</li>
 * <li>{@linkplain #PROP_COORDY Y-Coordinate}: Georeferenz Latitude (ScalarProperty[Double])</li>
 * <li>{@linkplain #PROP_WIDTH  Width}: Reale Breite der Raster (ScalarProperty[Double])</li>
 * <li>{@linkplain #PROP_HEIGHT Height}: Reale Hoehe der Raster (ScalarProperty[Double])</li>
 * <li>{@linkplain #PROP_WIDTHC Width in Cells}: Breite der Raster in Zellen (ScalarProperty[Integer])</li>
 * <li>{@linkplain #PROP_HEIGHTC Height in Cells}: Hoehe der Raster in Zellen (ScalarProperty[Integer])</li>
 * <li>{@linkplain #PROP_CELLWIDTH Cell Width}: Breite einer Rasterzelle (ScalarProperty[Double])</li>
 * <li>{@linkplain #PROP_CELLHEIGHT Cell Height}: Hoehe einer Rasterzelle (ScalarProperty[Double])</li>
 * <li>{@linkplain #PROP_TYPE Data Type}: {@linkplain DataBuffer#dataType Datentyp} den eine Rasterzelle aufnehmen kann (ScalarProperty[Integer])</li>
 * </ol>
 * <br>
 * Die Eigenschaften 4 bis 12 sind komplett von den zu grunde liegenden Rastern
 * abhaengig (Eigenschaft 1) und koennen nicht veraendert werden. Ein schreibender
 * Zugriff auf diese Properties hat keine Auswirkung.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class MultiGrid extends GridList {
  /** Name der Eigenschaft, die die X-Koordinate des Rasters beinhaltet. */
  public static final String PROP_COORDX     = "X-Coordinate";
  /** Name der Eigenschaft, die die Y-Koordinate des Rasters beinhaltet. */
  public static final String PROP_COORDY     = "Y-Coordinate";
  /** Name der Eigenschaft, die die reale Breite des Rasters beinhaltet. */
  public static final String PROP_WIDTH      = "Width";
  /** Name der Eigenschaft, die die reale Hoehe des Rasters beinhaltet. */
  public static final String PROP_HEIGHT     = "Height";
  /** Name der Eigenschaft, die die Breite des Rasters in Zellen beinhaltet. */
  public static final String PROP_WIDTHC     = "Width (in cells)";
  /** Name der Eigenschaft, die die Hoehe des Rasters in Zellen beinhaltet. */
  public static final String PROP_HEIGHTC    = "Height (in cells)";
  /** Name der Eigenschaft, die die Zellen-Breite des Rasters beinhaltet. */
  public static final String PROP_CELLWIDTH  = "Cell Width";
  /** Name der Eigenschaft, die die Zellen-Hoehe des Rasters beinhaltet. */
  public static final String PROP_CELLHEIGHT = "Cell Height";
  /** Name der Eigenschaft, die die Zellen-Typ des Rasters beinhaltet. */
  public static final String PROP_TYPE       = "Data Type";

  /** Speichern die Vorlagen-Informationen */
  private boolean templateExists     = false;
  private int     templateType       = 0;
  private double  templateX          = 0;
  private double  templateY          = 0;
  private int     templateMinX       = 0;
  private int     templateMinY       = 0;
  private int     templateWidth   	 = 0;
  private int     templateHeight     = 0;
  private double  templateRealWidth  = 0;
  private double  templateRealHeight = 0;
  private CoordinateReferenceSystem templateCRS = null;
  /** Speichert eine Factory, mit der neue Grids erzeugt werden koennen */
  private WritableGridFactory gridFactory = null;

  /**
   * Erzeugt eine neue (leere) Raster-Liste.
   * @param template Vorlage fuer das Raster-Format, das in der Liste gespeichert
   *                 werden kann
   * @param fac      Factory mit der weitere Instanzen erzeugt werden koennen
   */
  public MultiGrid(WritableGrid template, WritableGridFactory fac) {
    this( template == null ? -1   : template.getSampleType(),
          template == null ? 0    : template.getWidth(),
          template == null ? 0    : template.getHeight(),
          template == null ? 0    : template.getMinX(),
          template == null ? 0    : template.getMinY(),
          template == null ? 0    : template.getX(),
          template == null ? 0    : template.getY(),
          template == null ? 0    : template.getRealWidth(),
          template == null ? 0    : template.getRealHeight(),
          template == null ? null : template.getCoordinateReferenceSystem(),
          fac );
  }

  /**
   * Erzeugt eine neue (leere) Raster-Liste. Das Format fuer die Raster
   * wird von dem ersten Element uebernommen, welches in die Liste eingefuegt
   * wird.
   * @param fac      Factory mit der weitere Instanzen erzeugt werden koennen
   */
  public MultiGrid(WritableGridFactory fac) {
    this( (WritableGrid)null, fac );
  }
  
  /**
   * Erzeugt eine neue Raster-Liste.
   * @param gc  Raster fuer die Liste (das erste wird als Vorlage verwendet!)
   * @param fac Factory mit der weitere Instanzen erzeugt werden koennen
   * @exception ArrayIndexOutOfBoundsException falls der uebergebene Array
   *            kein Element enthaelt
   */
  public MultiGrid(WritableGrid[] gc, WritableGridFactory fac) {
    this(gc[0],fac);
    for (int i=0; i<gc.length; i++)
      addGrid(gc[i]);
  }

  /**
   * Erzeugt eine neue Raster-Liste. Die uebergebenen Raster-Parameter bilden
   * die Vorlage fuer die Raster-Art, die in der Liste gespeichert werden
   * koennen.
   * @param widthc  Breite in Zellen
   * @param heightc Hoehe in Zellen
   * @param x       Georeferenz Longitute (Suedliche/Untere Kante)
   * @param y       Georeferenz Latitude (Westliche/Linke Kante)
   * @param width   Breite
   * @param height  Hoehe
   * @param crs     Referenzsystem
   * @param fac     Factory mit der weitere Instanzen erzeugt werden koennen
   */
  public MultiGrid(int type, int widthc, int heightc, int minX, int minY, double x, double y, double width, double height, CoordinateReferenceSystem crs, WritableGridFactory fac) {
    super();
    this.gridFactory = fac;
    setTemplateDimension(type, widthc, heightc, minX, minY, x, y, width, height, crs);
    
    // Die weiteren Properties werden Wertmaessig auf die Methoden des
    // Grids umgeleitet
    propSet.addProperty( new ScalarProperty(PROP_COORDX,Double.class) {
      protected Object getValue(int... coords) {
        super.getValue(coords);
        return templateX;
      }
    });
    propSet.addProperty( new ScalarProperty(PROP_COORDY,Double.class) {
      protected Object getValue(int... coords) {
        super.getValue(coords);
        return templateY;
      }
    });
    propSet.addProperty( new ScalarProperty(PROP_WIDTH,Double.class) {
      protected Object getValue(int... coords) {
        super.getValue(coords);
        return templateRealWidth;
      }
    });
    propSet.addProperty( new ScalarProperty(PROP_HEIGHT,Double.class) {
      protected Object getValue(int... coords) {
        super.getValue(coords);
        return templateRealHeight;
      }
    });
    propSet.addProperty( new ScalarProperty(PROP_WIDTHC,Integer.class) {
      protected Object getValue(int... coords) {
        super.getValue(coords);
        return templateWidth;
      }
    });
    propSet.addProperty( new ScalarProperty(PROP_HEIGHTC,Integer.class) {
      protected Object getValue(int... coords) {
        super.getValue(coords);
        return templateHeight;
      }
    });
    propSet.addProperty( new ScalarProperty(PROP_CELLWIDTH,Double.class) {
      protected Object getValue(int... coords) {
        super.getValue(coords);
        return templateRealWidth / templateWidth;
      }
    });
    propSet.addProperty( new ScalarProperty(PROP_CELLHEIGHT,Double.class) {
      protected Object getValue(int... coords) {
        super.getValue(coords);
        return templateRealHeight / templateHeight;
      }
    });
    propSet.addProperty( new ScalarProperty(PROP_TYPE,Integer.class) {
      protected Object getValue(int... coords) {
        super.getValue(coords);
        return templateType;
      }
    });
  }
  
  /**
   * Belegt die Variablen der Raster-Vorlage, welche fuer jedes Raster gelten
   * muss, welches in die Liste ergefuegt werden soll.
   */
  public void setTemplateDimension(int type, int widthc, int heightc, int minX, int minY, double x, double y, double width, double height, CoordinateReferenceSystem crs) {
    if ( getGridCount() > 0 )
      throw new UnsupportedOperationException("Template dimension can only be changed if MultiGrid is empty!");
    this.templateExists     = type >= 0;
    this.templateType       = type;
    this.templateX          = x;
    this.templateY          = y;
    this.templateMinX       = minX;
    this.templateMinY       = minY;
    this.templateWidth      = widthc;
    this.templateHeight     = heightc;
    this.templateRealWidth  = width;
    this.templateRealHeight = height;
    this.templateCRS        = crs;
  }

  /**
   * Belegt die Variablen der Raster-Vorlage, welche fuer jedes Raster gelten
   * muss, welches in die Liste ergefuegt werden soll.
   */
  public void setTemplateDimension(WritableGrid template) {
    setTemplateDimension(
        template == null ? -1   : template.getSampleType(),
        template == null ? 0    : template.getWidth(),
        template == null ? 0    : template.getHeight(),
        template == null ? 0    : template.getMinX(),
        template == null ? 0    : template.getMinY(),
        template == null ? 0    : template.getX(),
        template == null ? 0    : template.getY(),
        template == null ? 0    : template.getRealWidth(),
        template == null ? 0    : template.getRealHeight(),
        template == null ? null : template.getCoordinateReferenceSystem()
    );
  }
  
  /**
   * Prueft, ob fuer das {@code MultiGrid} eine Vorlage gesetzt ist.
   */
  public boolean isTemplateDimensionSet() {
    return this.templateExists;
  }

  /**
   * Fuegt der Liste ein leeres Raster hinzu.
   * @param initVal initialer Wert aller Raster-Zellen
   */
  public void addGrid(Object initVal) {
    addGrid();
    WritableGrid newGrid = getGrid( getGridCount()-1 );
    int minX = newGrid.getMinX();
    int minY = newGrid.getMinY();
    int maxX = minX + newGrid.getWidth();
    int maxY = minY + newGrid.getHeight();
    for (int y=minY; y<maxY; y++)
      for (int x=minX; x<maxX; x++)
        newGrid.setRasterSample(initVal, x, y);
  }

  /**
   * Fuegt der Liste ein leeres Raster hinzu.
   */
  public void addGrid() {
    if ( !templateExists )
      throw new UnsupportedOperationException("New grid can not be added until a template grid is specified...");
    addGrid( gridFactory.newInstance(
                  this.templateType,
                  this.templateWidth,
                  this.templateHeight,
                  this.templateMinX,
                  this.templateMinY,
                  this.templateX,
                  this.templateY,
                  this.templateRealWidth,
                  this.templateRealHeight,
                  this.templateCRS
    ));
  }

  /**
   * Fuegt der Liste ein Raster hinzu. Dieses muss von der selben Groesse und
   * Georeferenz sein, wie die bereits in der Liste befindlichen Raster.
   * @exception IllegalArgumentException falls das neue Raster in Groesse oder
   *            Georeferenz nicht zum Rest der Liste passt
   */
  public void addGrid(WritableGrid grid) {
    if ( !templateExists )
      setTemplateDimension(
          grid.getSampleType(),
          grid.getWidth(),
          grid.getHeight(),
          grid.getMinX(),
          grid.getMinY(),
          grid.getX(),
          grid.getY(),
          grid.getRealWidth(),
          grid.getRealHeight(),
          grid.getCoordinateReferenceSystem()
      );
    else
      checkStructure(grid);
    super.addGrid(grid);
  }

  /**
   * Ersetzt ein Raster in der Liste. Das neue Raster muss von der selben
   * Groesse und Georeferenz sein, wie die restlichen in der Liste befindlichen Raster.
   * @param idx Listen-Index
   * @param grid zu entfernendes Raster
   * @exception IllegalArgumentException falls das neue Raster in Groesse oder
   *            Georeferenz nicht zum Rest der Liste passt
   */
  public void replaceGrid(int idx, WritableGrid grid) {
    checkStructure(grid);
    super.replaceGrid(idx,grid);
  }

  /**
   * Prueft, ob die Struktur eines Rasters mit der des Multi-Grid vertraeglich ist.
   * @param gc ein WritableGrid
   * @exception IllegalArgumentException falls sich das angegebene Raster in Groesse
   *            Aufloesung oder Geo-Referenz vom Template-Grid des Multi-Grid
   *            unterscheidet
   */
  protected void checkStructure(WritableGrid gc) {
    if ( gc == null ||
         gc.getX()          != templateX ||
         gc.getY()          != templateY ||
         gc.getMinX()       != templateMinX ||
         gc.getMinY()       != templateMinY ||
         gc.getWidth()      != templateWidth ||
         gc.getHeight()     != templateHeight ||
         gc.getRealWidth()  != templateRealWidth ||
         gc.getRealHeight() != templateRealHeight )
      throw new IllegalArgumentException("WritableGrid does not fit to the rest of the list!");
  }

  /**
   * Liefert ein Beispiel-Raster, das die Raster-Art repraesentiert, die in
   * der Liste gespeichert werden koennen. Dieses liefert fuer alle Zellen
   * den Wert 0 und kann nicht veraendert werden.
   * @return {@code null} wenn noch kein Beispiel-Raster gesetzt ist
   */
  public WritableGrid getTemplateGrid() {
    if ( !templateExists )
      return null;
    return new AbstractWritableGrid(templateCRS) {
      public Object getRasterSample(int... c) { return 0; };
      public void setRasterSample(Object o, int... c) {};
      public void dispose() {};
      public int getSampleType() { return templateType; }
      public int getWidth() { return templateWidth; }
      public int getHeight() { return templateHeight; }
      public int getMinX() { return templateMinX; }
      public int getMinY() { return templateMinY; }
      public double getX() { return templateX; }
      public double getY() { return templateY; }
      public double getRealWidth() { return templateRealWidth; }
      public double getRealHeight() { return templateRealHeight; }
    };
  }

  /**
   * Liefert die Werte aller Rasters.
   * @param coords Georeferenz
   */
  public Object[] getGridSample(double... coords) {
    Object[] sample = new Object[getGridCount()];
    for (int i=0; i<sample.length; i++)
      sample[i] = getGrid(i).getGridSample(coords);
    return sample;
  }

  /**
   * Liefert die Werte aller Rasters.
   * @param coords Raster-Position
   */
  public Object[] getRasterSample(int... coords) {
    Object[] sample = new Object[getGridCount()];
    for (int i=0; i<sample.length; i++)
      sample[i] = getGrid(i).getRasterSample(coords);
    return sample;
  }

  /**
   * @return an array of all grids in the {@link MultiGrid}
   * @author Dominik Appl
   */
public WritableGrid[] toArray(){
	  WritableGrid[] array = new WritableGrid[this.getGridCount()];
	  for (int i = 0; i < array.length; i++) {
		array[i] = this.getGrid(i);
	}
	  return array;
  }

}
