/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.data.grid;


import schmitzm.data.WritableGrid;
import schmitzm.data.property.ScalarProperty;
import schmitzm.data.property.ValuePropertyAccessParameters;
import edu.bonn.xulu.data.AbstractXuluObject;

/**
 * Dieser Datentyp stellt ein einzelnes Raster dar. Dieses baut auf
 * der Klasse {@link WritableGrid schmitzm.data.WritableGrid} auf und
 * hat folgende Eigenschaften:<br>
 * <ol>
 * <li>{@linkplain #PROP_GRID Grid}: Das gesamte Raster (ScalarProperty[WritableGrid])</li>
 * <li>{@linkplain #PROP_STYLE Style}: Darstellungs-Style fuer das Raster (ScalarProperty[Object])</li>
 * <li>{@linkplain #PROP_COORDX X-Coordinate}: Georeferenz Longitude der S�dwestlichen Ecke (ScalarProperty[Double])</li>
 * <li>{@linkplain #PROP_COORDY Y-Coordinate}: Georeferenz Latitude der S�dwestlichen Ecke (ScalarProperty[Double])</li>
 * <li>{@linkplain #PROP_WIDTH  Width}: Reale Breite des Rasters (ScalarProperty[Double])</li>
 * <li>{@linkplain #PROP_HEIGHT Height}: Reale Hoehe des Rasters (ScalarProperty[Double])</li>
 * <li>{@linkplain #PROP_WIDTHC Width in Cells}: Breite des Rasters in Zellen (ScalarProperty[Integer])</li>
 * <li>{@linkplain #PROP_HEIGHTC Height in Cells}: Hoehe des Rasters in Zellen (ScalarProperty[Integer])</li>
 * <li>{@linkplain #PROP_CELLWIDTH Cell Width}: Breite einer Rasterzelle (ScalarProperty[Double])</li>
 * <li>{@linkplain #PROP_CELLHEIGHT Cell Height}: Hoehe einer Rasterzelle (ScalarProperty[Double])</li>
 * </ol>
 * <br>
 * Die Eigenschaften 3 bis 10 sind komplett von dem zu grunde liegenden Raster
 * abhaengig (Eigenschaft 1) und koennen nicht veraendert werden. Ein schreibender
 * Zugriff auf diese Properties hat keine Auswirkung.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class SingleGrid extends AbstractXuluObject {
  /** Name der Eigenschaft, die die {@link WritableGrid} beinhaltet. */
  public static final String PROP_GRID       = "Grid";
  /** Name der Eigenschaft, die die X-Koordinate des Rasters beinhaltet. */
  public static final String PROP_COORDX     = "X-Coordinate";
  /** Name der Eigenschaft, die die Y-Koordinate des Rasters beinhaltet. */
  public static final String PROP_COORDY     = "Y-Coordinate";
  /** Name der Eigenschaft, die die reale Breite des Rasters beinhaltet. */
  public static final String PROP_WIDTH      = "Width";
  /** Name der Eigenschaft, die die reale Hoehe des Rasters beinhaltet. */
  public static final String PROP_HEIGHT     = "Height";
  /** Name der Eigenschaft, die die Breite des Rasters in Zellen beinhaltet. */
  public static final String PROP_WIDTHC     = "Width (in cells)";
  /** Name der Eigenschaft, die die Hoehe des Rasters in Zellen beinhaltet. */
  public static final String PROP_HEIGHTC    = "Height (in cells)";
  /** Name der Eigenschaft, die die Zellen-Breite des Rasters beinhaltet. */
  public static final String PROP_CELLWIDTH  = "Cell Width";
  /** Name der Eigenschaft, die die Zellen-Hoehe des Rasters beinhaltet. */
  public static final String PROP_CELLHEIGHT = "Cell Height";

  /**
   * Erzeugt ein neues Raster. Saemtliche Properties sind noch unbelegt.
   */
  public SingleGrid() {
    // Beim Freigeben der Property soll auch das Grid zerstoert werden
//    propSet.addProperty( new ScalarProperty(PROP_GRID,WritableGrid.class,new ValuePropertyAccessParameters(1,1)) {
    propSet.addProperty( new ScalarProperty(PROP_GRID,WritableGrid.class ) {
      public void dispose() {
        if ( value != null )
          ((WritableGrid)value).dispose();
        super.dispose();
      }
    });
    // Property fuer den Style
    propSet.addProperty( new ScalarProperty(PROP_STYLE,Object.class) );
    // Die weiteren Properties werden Wertmaessig auf die Methoden des
    // Grids umgeleitet
    propSet.addProperty( new ScalarProperty(PROP_COORDX,Double.class) {
      protected Object getValue(int... coords) {
        super.getValue(coords);
        return getGrid().getX();
      }
    });
    propSet.addProperty( new ScalarProperty(PROP_COORDY,Double.class) {
      protected Object getValue(int... coords) {
        super.getValue(coords);
        return getGrid().getY();
      }
    });
    propSet.addProperty( new ScalarProperty(PROP_WIDTH,Double.class) {
      protected Object getValue(int... coords) {
        super.getValue(coords);
        return getGrid().getRealWidth();
      }
    });
    propSet.addProperty( new ScalarProperty(PROP_HEIGHT,Double.class) {
      protected Object getValue(int... coords) {
        super.getValue(coords);
        return getGrid().getRealHeight();
      }
    });
    propSet.addProperty( new ScalarProperty(PROP_WIDTHC,Integer.class) {
      protected Object getValue(int... coords) {
        super.getValue(coords);
        return getGrid().getWidth();
      }
    });
    propSet.addProperty( new ScalarProperty(PROP_HEIGHTC,Integer.class) {
      protected Object getValue(int... coords) {
        super.getValue(coords);
        return getGrid().getHeight();
      }
    });
    propSet.addProperty( new ScalarProperty(PROP_CELLWIDTH,Double.class) {
      protected Object getValue(int... coords) {
        super.getValue(coords);
        WritableGrid g = getGrid();
        return g.getRealWidth() / g.getWidth();
      }
    });
    propSet.addProperty( new ScalarProperty(PROP_CELLHEIGHT,Double.class) {
      protected Object getValue(int... coords) {
        super.getValue(coords);
        WritableGrid g = getGrid();
        return g.getRealHeight() / g.getHeight();
      }
    });
  }

  /**
   * Erzeugt ein neues Raster.
   * @param gc zugrunde liegendes Raster
   */
  public SingleGrid(WritableGrid gc) {
    this();
    setGrid(gc);
  }

  /**
   * Liefert das zugrunde liegende Raster. Aenderungen an diesem Raster
   * wirken sich automatisch auf die anderen Properties aus!
   */
  public WritableGrid getGrid() {
    return (WritableGrid) ((ScalarProperty)getProperty(PROP_GRID)).getOneTimeReadAccess().getValue();
  }

  /**
   * Setzt das zugrunde liegende Raster. Hierdurch werden automatisch auch
   * die anderen Properties beeinflusst.
   */
  public void setGrid(WritableGrid grid) {
    ( (ScalarProperty)getProperty(PROP_GRID) ).getOneTimeWriteAccess().setValue(grid);
  }

  /**
   * Liefert einen Wert des Rasters.
   * @param coords Georeferenz
   */
  public Object getGridSample(double... coords) {
    return getGrid().getGridSample(coords);
  }

  /**
   * Liefert einen Wert des Rasters.
   * @param coords Raster-Position
   */
  public Object getRasterSample(int... coords) {
    return getGrid().getRasterSample(coords);
  }

}
