/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.data.misc;

import java.awt.Point;
import java.awt.geom.Point2D;
import java.util.Vector;

/**
 * Dieser Datentyp stellt eine Region (z.B. Siedlung) in einem
 * {@linkplain edu.bonn.xulu.plugin.model.ca.MultiCellularAutomaton multiplen zellularen Automaten}
 * dar. Eine <code>CAArea</code> hat folgende Eigenschaften:
 * <ul>
 *   <li>ein Name</li>
 *   <li>ein Start-Position (lat/lon-Georeferenz)</li>
 *   <li>eine aktuelle Bedeckung, wie viele qm der Region zugewiesen wurden</li>
 *   <li>eine Liste an Randpunkten die fuer die Expansion im naechsten Schritt in Frage kommen</li>
 *   <li>eine Flag, das signalisiert, ob der Bedarf fuer einen Zeitschritt gedeckt ist</li>
 *   <li>eine Flag, das signalisiert, ob sich die Region in einem Deadlock befindet</li>
 * </ul>
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class CAArea {
  /** Speichert den Namen der Area. */
  protected String          name         = "";
  /** Speichert die Start-Position der Area. */
  protected Point2D         startLoc     = null;
  /** Speichert die aktuelle Flaeche (in qm), die der CA der Area zugewiesen hat. */
  protected double          cover        = 0.0;
  /** Speichert die Randzellen der Region. */
  protected Vector<Point>   border       = null;
  /** Flag, ob sich die Area in einem Deadlock befindet. */
  protected boolean         deadlock     = false;

  /**
   * Erzeugt eine neue Area.
   */
  public CAArea() {
    this("",0,0);
  }

  /**
   * Erzeugt eine neue Area.
   * @param name Name der Area
   * @param x    X-Koodinate (Longitude)
   * @param y    Y-Koodinate (Latitude)
   */
  public CAArea(String name, double x, double y) {
    setName(name);
    this.startLoc       = new Point2D.Double(x,y);
    this.border         = new Vector<Point>();
    this.deadlock       = false;
  }


  /**
   * Initialisiert die Area, in dem die aktuelle Flaeche und die Liste der
   * Randzellen zurueckgesetzt werden.
   */
  public void init() {
    this.cover = 0.0;
    this.border.clear();
    this.deadlock = false;
  }

  /**
   * Liefert den Namen der Area.
   */
  public String getName() {
    return name;
  }

  /**
   * Setzt den Namen der Area.
   * @param name neuer Name
   */
  public void setName(String name) {
    this.name = name;
  }

  /**
   * Liefert die Startkoordinaten der Area.
   * @return loc Lat/Lon-Georeferenz
   */
  public Point2D getStartLocation() {
    return startLoc;
  }

  /**
   * Setzt die Startkoordinaten der Area.
   * @param loc Lat/Lon-Georeferenz
   */
  public void setStartLocation(Point2D loc) {
    this.startLoc = loc;
  }

  /**
   * Liefert die aktuelle qm-Flaeche, die der Area zugeordnet ist.
   */
  public double getCurrentCover() {
    return this.cover;
  }

  /**
   * Erhoeht die Flaeche, die der Area aktuell zugeordnet ist. Wird vom
   * CA verwendet.
   * @param value Wert um den Gesamt-Flaeche erhoeht wird (in qm)
   */
  public void incCurrentCover(double value) {
    this.cover += value;
  }

  /**
   * Reduziert die Flaeche, die der Area aktuell zugeordnet ist. Wird vom
   * CA verwendet.
   * @param value Wert um den Gesamt-Flaeche reduziert wird (in qm)
   */
  public void decCurrentCover(double value) {
    this.cover -= value;
  }

  /**
   * Liefert die Liste der Randpunkte der Area. Wird vom CA verwaltet und
   * durch {@link #init()} initialisiert.
   */
  public Vector<Point> getBorder() {
    return this.border;
  }

  /**
   * Checkt, ob sich der CA fuer diese Area in einem Deadlock befindet.
   */
  public boolean isDeadlock() {
    return this.deadlock;
  }

  /**
   * Setzt ein Flag, dass sich diese Area in einem Deadlock befindet.
   * @param deadlock <code>true</code> wenn der CA fuer diese Area nicht mehr
   *                 weiter expandieren kann
   */
  public void setDeadlock(boolean deadlock) {
    this.deadlock = deadlock;
  }

  /**
   * Prueft, ob die aktuell belegte Flaeche den Bedarf deckt.
   * @param step Schrittnummer (beginnend bei 0!)
   */
  public abstract boolean isDemandSatisfied(int step);

}
