/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.gui;

import java.awt.Container;
import java.awt.Dimension;
import java.awt.BorderLayout;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import javax.swing.JFileChooser;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeNode;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreePath;
import javax.swing.JTable;
import javax.swing.table.AbstractTableModel;
import java.util.Vector;
import java.util.Enumeration;
import java.io.File;

import schmitzm.io.FileInputStream;
import schmitzm.io.FileOutputStream;
import schmitzm.lang.AlreadyHandledException;
import schmitzm.swing.MultiSplitPane;
import schmitzm.swing.ObjectDisplayContainer;
import schmitzm.swing.event.PopupMenuListener;
import schmitzm.swing.menu.ObjectMenuItem;
import schmitzm.swing.tree.EmptyNode;
import schmitzm.swing.tree.ContentNode;
import schmitzm.swing.tree.EditableNode;

import schmitzm.data.ObjectStructure;
import schmitzm.data.event.ObjectListener;
import schmitzm.data.event.ObjectEvent;
import schmitzm.data.property.Property;
import schmitzm.data.property.PropertyType;
import schmitzm.data.property.Properties;
import schmitzm.data.property.ValueProperty;
import schmitzm.data.property.ValuePropertyType;
import schmitzm.data.property.ScalarProperty;
import schmitzm.data.property.ListProperty;
import schmitzm.data.property.MatrixProperty;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.PropertyWriteAccess;
import schmitzm.data.property.AccessViolationException;

import edu.bonn.xulu.io.InstantiationFactory;
import edu.bonn.xulu.io.ImportFactory;
import edu.bonn.xulu.io.ExportFactory;
import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.appl.DataPool;
import edu.bonn.xulu.appl.VisualisationTool;
import edu.bonn.xulu.data.XuluObject;
import edu.bonn.xulu.data.XuluDataException;
import edu.bonn.xulu.gui.XuluGUIMessages;
import edu.bonn.xulu.gui.XuluMainFrame;
import edu.bonn.xulu.gui.DataPoolMenu;
import edu.bonn.xulu.gui.DataPoolFrame;
import edu.bonn.xulu.gui.DataPoolTreeModel;

import schmitzm.temp.BaseTypeUtil;

import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeCellEditor;
import javax.swing.event.TreeSelectionEvent;
import java.util.EventObject;

/**
 * Diese Klasse stellt den Xulu-Datenpool als Baum dar, in der die Datenstrukturen
 * der einzelnen Objekte aufgeschluesselt werden.<br>
 * Die eigentlichen Werte der Eigenschaften werden im Baum nicht dargestellt.
 * Statt befindet sich im unteren Fensterbereich ein Detail-Fenster, in dem
 * zur Zeit{@link Properties}-, {@link ListProperty} und
 * {@link MatrixProperty}-Instanzen tabellarisch dargestellt werden, sowie
 * Raster/Geometrie-Daten grafisch.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class DataPoolFrame_Tree extends DataPoolFrame {
  private final Container                       DC_EMPTY          = new Container();
  private final DisplayContainer_Properties     DC_PROPERTIES     = new DisplayContainer_Properties();
  private final DisplayContainer_ListProperty   DC_LISTPROPERTY   = new DisplayContainer_ListProperty();
  private final DisplayContainer_MatrixProperty DC_MATRIXPROPERTY = new DisplayContainer_MatrixProperty();
  private final DisplayContainer_GeomAndGrid    DC_GEOMGRID       = new DisplayContainer_GeomAndGrid();

  private XuluModellingPlatform  appl              = null;
  private MultiSplitPane         dataPoolSplitPane = new MultiSplitPane(2,javax.swing.JSplitPane.VERTICAL_SPLIT);
  private Container              contentPane       = null;
  private JTree                  dataPoolTree      = null;
  private ObjectDisplayContainer displayCont       = null;
  private Object[]               selectedObjects   = new Object[0];
  private DataPoolMenu           dataPoolMenu      = null;

  /**
   * Erzeugt einen neue Instanz des grafischen Datenpools. Die Groesse ist
   * auf 300x350 Pixel voreingestellt.
   * @param appl Xulu-Applikation fuer die der Datenpool-Frame instanziiert wird
   */
  public DataPoolFrame_Tree(XuluModellingPlatform appl) {
    this(appl, new Dimension(300,350));
  }

  /**
   * Erzeugt einen neue Instanz des grafischen Datenpools.
   * @param appl   Xulu-Applikation fuer die der Datenpool-Frame instanziiert wird
   * @param size   Groesse des Fensters.
   */
  public DataPoolFrame_Tree(XuluModellingPlatform appl, Dimension size) {
    super();
    this.appl = appl;
    this.setSize(size);
    this.contentPane = this.getContentPane();
    contentPane.setLayout( new BorderLayout() );

    // Tabelle mit Xulu-Objekten
    this.dataPoolTree = new JTree( new DataPoolTreeModel(appl.getDataPool(),true,"ROOT") );
    this.dataPoolTree.setRootVisible(false);
    this.dataPoolTree.setEditable(true);
    this.dataPoolTree.setDragEnabled(true);
    this.dataPoolTree.setCellEditor( new DefaultTreeCellEditor(dataPoolTree, (DefaultTreeCellRenderer)dataPoolTree.getCellRenderer()) {
      private Object lastSelected = null;
      // Diese Methode wird aufgerufen, um zu entscheiden, ob ein
      // Knoten editierbar ist
      public boolean isCellEditable(EventObject event) {
        return super.isCellEditable(event) &&
               lastSelected != null &&
               lastSelected instanceof EditableNode &&
               ((EditableNode)lastSelected).isEditable();
      }
      // Diese Methode wird aufgerufen, sobald ein neuer Knoten im Baum
      // selektiert wurde --> merken, welcher das ist
      public void valueChanged(TreeSelectionEvent e) {
        if ( e.getNewLeadSelectionPath() != null )
          lastSelected = e.getNewLeadSelectionPath().getLastPathComponent();
        else
          lastSelected = null;
        super.valueChanged(e);
      }
    });

    this.dataPoolTree.addMouseListener(new MouseAdapter() {
      public void mouseClicked(MouseEvent e) {
        TreePath selPath = dataPoolTree.getSelectionPath();
        Object selObj = (selPath!=null && selPath.getLastPathComponent() instanceof ContentNode) ? ((ContentNode)selPath.getLastPathComponent()).getUserObject() : null;

        Object selVal = selObj; // darzustellender Wert
        try {
          // Handelt es sich um eine skalare Property, wird deren Wert
          // dargestellt
          if (selObj instanceof ScalarProperty)
            selVal = ( (ScalarProperty) selObj).getOneTimeReadAccess().getValue();
          // ggf. neuen DisplayContainer setzen
          setDisplayContainerForObject(selVal);
          // falls es fuer den Wert keinen Container gibt, es sich aber um
          // Property handelt, diese selbst darstellen
          if (displayCont == null && selObj instanceof Property) {
            setDisplayContainerForObject(selObj);
            selVal = selObj;
          }
          // falls es einen passenden Container gibt, Objekt setzen
          if (displayCont != null)
            displayCont.setObject(selVal);
        } catch ( AccessViolationException err ) {
          XuluGUIMessages.showAccessError(contentPane,err);
        } catch ( Exception err ) {
          XuluGUIMessages.showError(contentPane,err,"Unerwarteter Fehler:");
        }
      }
    });

    // SplitPane organisieren
    dataPoolSplitPane.setResizeWeigth( new double[] {0.5,0.5} );
    dataPoolSplitPane.setDividerSize(2);
    dataPoolSplitPane.setContainer(0,new JScrollPane(dataPoolTree));
    dataPoolSplitPane.setContainer(1,DC_EMPTY);
    contentPane.add(dataPoolSplitPane);

    // Popup-Menu anbinden
    dataPoolMenu = new DataPoolMenu(appl,this);
    PopupMenuListener listener = new PopupMenuListener(dataPoolMenu.getPopupMenu()) {
      public void mouseReleased(MouseEvent e) {
        // Wenn noch keine Baum-Element selektiert war, wird automatisch
        // das Element selektiert, das die geklickt wurde
        if ( e.isPopupTrigger() && dataPoolTree.getSelectionCount() <= 1 ) {
          TreePath path = dataPoolTree.getPathForLocation(e.getX(),e.getY());
          dataPoolTree.setSelectionPath(path);
        }
        super.mouseReleased(e);
      }
    };
    dataPoolTree.addMouseListener( listener );
    dataPoolSplitPane.getContainer(0).addMouseListener( listener );

  }

  /**
   * Liefert alle in der Datentabelle <b>und</b> in der Eigenschaftstabelle
   * selektierten Objekte. Dies koennen sowohl <code>XuluObject</code>s sein,
   * als auch einzelne <code>Property</code>s.<br>
   * <b>Bemerke:</b><br>
   * Zur Zeit werden nur die Xulu-Objekte zurueckgegeben!!
   * @return einen leeren Array wenn keine Zeilen selektiert sind
   */
  public Object[] getSelectedObjects() {
    // selektierte Objekte im Datenpool und DisplayContainer
    TreePath[] selXuluDataPath = dataPoolTree.getSelectionPaths();
    if ( selXuluDataPath == null )
      selXuluDataPath = new TreePath[0];
    Object[] selDisplayObjects = getSelectedObjectsFromDisplayContainer();

    // alle selektierten Objekte zusammen
    Object[]   selectedObjects = new Object[ selXuluDataPath.length + selDisplayObjects.length ];
    for (int i=0; i<selXuluDataPath.length; i++)
      selectedObjects[i] = ((ContentNode)selXuluDataPath[i].getLastPathComponent()).getUserObject();
    for (int i=0; i<selDisplayObjects.length; i++)
      selectedObjects[i+selXuluDataPath.length] = selDisplayObjects[i];
    return selectedObjects;
  }

  /**
   * Liefert das letzte in der Datentabelle selektierte Objekt.
   * @return <code>null</code> wenn keine Zeile selektiert ist
   */
  public Object getSelectedObject() {
    // selektierte Objekte im Datenpool und DisplayContainer
    TreePath selXuluDataPath = dataPoolTree.getSelectionPath();
    if ( selXuluDataPath == null )
      return null;
    return ((ContentNode)selXuluDataPath.getLastPathComponent()).getUserObject();
  }

  /**
   * Aktualisiert die Visualisierung des Datenpools, indem ein
   * <code>reload()</code>-Event auf das TreeModel des Datenpools
   * ausgeloest wird.
   */
  public void refresh() {
    ((DataPoolTreeModel)dataPoolTree.getModel()).reload();
    setDisplayContainerForObject(null);
  };


  /**
   * Setzt den passenden Display-Container fuer ein Objekt.
   */
  private void setDisplayContainerForObject(Object obj) {
    // wenn noch kein DisplayContainer besteht oder der bestehende
    // das Objekt nicht darstellen kann, wird ein neuer gewaehlt
    if ( displayCont == null || !displayCont.canDisplay(obj) ||
         // Da ListProperty und MatrixProperty auch Property implementieren,
         // koennen sie von DC_PROPERTIES dargestellt werden, dieser soll
         // jedoch nicht fuer Liste/Matrix verwendet werden!
         displayCont == DC_PROPERTIES && (obj instanceof ListProperty || obj instanceof MatrixProperty) ) {
      if ( DC_LISTPROPERTY.canDisplay(obj) )
        displayCont = DC_LISTPROPERTY;
      else if ( DC_MATRIXPROPERTY.canDisplay(obj) )
        displayCont = DC_MATRIXPROPERTY;
      else if ( DC_PROPERTIES.canDisplay(obj) )
        displayCont = DC_PROPERTIES;
      else if ( DC_GEOMGRID.canDisplay(obj) )
        displayCont = DC_GEOMGRID;
      else
        // fuer das Object gibt es keinen Container
        displayCont = null;
      // SplitPane aktualisieren
      dataPoolSplitPane.setContainer(1,displayCont!=null ? displayCont : DC_EMPTY);
    }
  }

  /**
   * Liefert alle im Display-Container selektierten Objekte.
   */
  private Object[] getSelectedObjectsFromDisplayContainer() {
    Object[] selDisplayObjects = new Object[0];
    if ( displayCont instanceof DisplayContainer_Table )
      selDisplayObjects = ((DisplayContainer_Table)displayCont).getSelectedElements();
    return selDisplayObjects;
  }

// Es gibt nun eine eigene Klasse fuer
//  ///////////////////////////////////////////////////////////////////
//  // TreeModel fuer die Datenpool-Tabelle
//  ///////////////////////////////////////////////////////////////////
//  /**
//   * Diese Klasse stellt ein TreeModel fuer den {@link schmitzm.data.DataPool} dar.
//   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
//   * @version 1.0
//   */
//  private class DataPoolTreeModel extends DefaultTreeModel {
//    private DataPool dataPool = null;
//
//    /**
//     * Erzeugt ein neues Daten-Modell fuer einen Datenpool.
//     * @param dp Instanz eines Xulu-Datenpools
//     */
//    public DataPoolTreeModel(DataPool dp) {
//      super( new EmptyNode("ROOT") );
//      this.dataPool = dp;
//
//      dp.addObjectListener( new ObjectListener() {
//        public void performObjectEvent(ObjectEvent e) {
//          DataPool.DataPoolChangeEvent dpEvent = (DataPool.DataPoolChangeEvent)e;
//          switch (dpEvent.getType()) {
//            // Objekt wurde hinzugefuegt
//            case DataPool.DataPoolChangeEvent.OBJECT_ADDED:
//              add( (XuluObject)dpEvent.getSource().getRoot() );
//              break;
//            // Objekt wurde entfernt
//            case DataPool.DataPoolChangeEvent.OBJECT_REMOVED:
//              remove( (XuluObject)dpEvent.getSource().getRoot() );
//              break;
//          }
//          reload();
//        }
//      });
//    }
//
//    public void reorganize() {
//      ((EmptyNode)getRoot()).removeAllChildren();
//      // Fuer alle Xulu-Objekte einen Knoten einfuegen
//      for ( Enumeration e = dataPool.asEnumeration(); e.hasMoreElements(); )
//        add( (XuluObject)e.nextElement() );
////      this.nodeStructureChanged( (TreeNode)getRoot() );
//      reload();
//    }
//
//    public void add(XuluObject obj) {
//      ObjectStructureNode node = new ObjectStructureNode(obj,true) {
//        // bei Aenderung aendert sich der Name des Objekts
//        public void setUserObject(Object newValue) {
//          if ( newValue instanceof String && getUserObject() instanceof XuluObject )
//            ((XuluObject)getUserObject()).setDescription( (String)newValue );
//        }
//      };
//      ((EmptyNode)getRoot()).add(node);
//    }
//
//    public void remove(XuluObject obj) {
//      // Knoten des Objekts suchen und entfernen
//      for (int i=0; i<((TreeNode)getRoot()).getChildCount();i++) {
//        ContentNode child = (ContentNode) ((EmptyNode)getRoot()).getChildAt(i);
//        if ( child.getUserObject() == obj )
//          ((EmptyNode)getRoot()).remove( i-- );
//      }
//    }
//
//    private class ObjectStructureNode extends ContentNode {
//      public ObjectStructureNode(Object obj) {
//        this(obj,false);
//      }
//
//      public ObjectStructureNode(Object obj, boolean editable) {
//        super(obj,editable);
//        reorganize();
//      }
//
//      public void reorganize() {
//        // Wenn kein strukturiertes Objekt, dann auch keine Kindknoten
//        if ( !(getUserObject() instanceof ObjectStructure) )
//          return;
//
//        ObjectStructure struct = (ObjectStructure)getUserObject();
//        removeAllChildren();
//
//        // Zwischenknoten?
//        DefaultMutableTreeNode node = this;
//        if ( struct.containsMultipleValues() )
//          node = new EmptyNode(struct.getClass().getSimpleName());
//
//        // Kindknoten
//        for (Enumeration attr = struct.getAttrTypes(); attr.hasMoreElements();)
//          node.add( new ObjectStructureNode(attr.nextElement()) );
//
//        // wenn Zwischenknoten, dann diesen einfuegen
//        if ( node != this )
//          add( node );
//      }
//
//      public String toString() {
//        if ( getUserObject() instanceof ObjectStructure &&
//             ((ObjectStructure)getUserObject()).isStructureNamed() )
//          return ((ObjectStructure)getUserObject()).getStructureName();
//        if ( getUserObject() instanceof Class )
//          return ((Class)getUserObject()).getSimpleName();
//
//        return getUserObject().getClass().getSimpleName();
//      }
//    }
//
////    private class ObjectStructureNode extends ContentNode {
////      public ObjectStructureNode(Object obj, String desc) {
////        super(obj,desc,false);
////        reorganize();
////      }
////
////      public void reorganize() {
////        removeAllChildren();
////
////        // Xulu-Objekte >> Jede Property erhaelt einen Knoten
////        if ( getUserObject() instanceof XuluObject ) {
////          Property[] prop = ((Properties)getUserObject()).getProperties();
////          for (int i=0; i<prop.length; i++) {
////            PropertyType type = prop[i].getPropertyType();
////            add(new ObjectStructureNode(prop[i],prop[i].getName()));
////          }
////        }
////        // andere Property-Mengen >> erst eigener "Menge"-Knoten, darunter
////        //                           ein Unterkoten fuer Property
////        else if ( getUserObject() instanceof Properties ) {
////          Property[] prop = ((Properties)getUserObject()).getProperties();
////          ContentNode inNode = new ContentNode(getUserObject(),"PropertySet",false);
////          for (int i=0; i<prop.length; i++) {
////            PropertyType type = prop[i].getPropertyType();
////            inNode.add(new ObjectStructureNode(prop[i],prop[i].getName()));
////          }
////          add(inNode);
////        }
//// /*
////        // ListProperty >> erst eigener "Liste"-Knoten, darunter
////        //                 ein Unterkoten den Elementtyp
////        else if ( getUserObject() instanceof ListProperty ) {
////          ContentNode inNode = new ContentNode(getUserObject(),"List",false);
////          Object sample = ((ListProperty)getUserObject()).getPropertyType().getSampleValue();
////          if ( sample == null )
////            inNode.add( new EmptyNode( ((ListProperty)getUserObject()).getType().getName() ) );
////          else
////            inNode.add(new ObjectStructureNode( ((ListProperty)getUserObject()).getPropertyType().getSampleValue(),"??" ));
////          add(inNode);
////        }
////        // ScalarProperty >> erst eigener "Liste"-Knoten, darunter
////        //                   ein Unterkoten den Elementtyp
////        else if ( getUserObject() instanceof ScalarProperty ) {
////          Object sample = ((ScalarProperty)getUserObject()).getPropertyType().getSampleValue();
////          if ( sample == null )
////            add( new EmptyNode( ((ScalarProperty)getUserObject()).getType().getName() ) );
////          else if ( sample instanceof Property )
////            add( new ObjectStructureNode( sample, ((Property)sample).getName() ) );
////          else
////            add( new EmptyNode(sample.getClass().getName()) );
////        }
////*/
////        else if ( getUserObject() instanceof ValueProperty ) {
////          Object sample = ((ValueProperty)getUserObject()).getPropertyType().getSampleValue();
////          if ( sample == null )
////            add( new EmptyNode( ((ValueProperty)getUserObject()).getType().getName() ) );
////          else
////            add(new ObjectStructureNode( ((ValueProperty)getUserObject()).getPropertyType().getSampleValue(),getUserObject().getClass().getSimpleName() ));
////        }
////
////        else {
////          add( new EmptyNode(getUserObject().getClass().getName()) );
////        }
////      }
////    }
//  }

}
