/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.gui;

import java.awt.BorderLayout;
import javax.swing.table.AbstractTableModel;
import javax.swing.JScrollPane;
import javax.swing.JTable;

import schmitzm.swing.ObjectDisplayContainer;
import schmitzm.data.property.Properties;
import schmitzm.data.property.Property;
import schmitzm.data.property.ListProperty;
import schmitzm.data.property.ListPropertyWriteAccess;
import schmitzm.data.property.AccessViolationException;

import edu.bonn.xulu.gui.XuluGUIMessages;

import schmitzm.temp.BaseTypeUtil;

/**
 * Diese Klasse stellt eine GUI-Komponente dar, in der eine
 * {@link ListProperty}-Instanz angezeigt werden kann. Die einzelnen Listen-Elemente
 * werden tabellarisch in 3 Spalten aufgelistet:
 * <ol>
 * <li>Listenindex</li>
 * <li>Typ des Listenelements</li>
 * <li>Typ des Listenelements (sofern darstellbar)</li>
 * </ol>
 * Darstellbar sind Elemente, die einen Basis-Datentyps (z.B. Integer) repraesentieren.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class DisplayContainer_ListProperty extends DisplayContainer_Table {
  /**
   * Erzeugt eine neue Darstellungskomponente.
   */
  public DisplayContainer_ListProperty() {
    this(null);
  }

  /**
   * Erzeugt eine neue Darstellungskomponente und zeigt sofort ein
   * Objekt an.
   * @param data anzuzeigendes Objekt
   */
  public DisplayContainer_ListProperty(Object data) {
    super(data);
  }

  /**
   * Erzeugt das TableModel fuer die Tabelle
   * @param o darzustellendes <code>ListProperty</code>-Objekt (kann <code>null</code> sein!)
   */
  protected ListPropertyTableModel createTableModel(Object o) {
    return new ListPropertyTableModel((ListProperty)o);
  }

  /**
   * Liefert die aktuell dargestellte {@link ListProperty}-Instanz.
   * @return <code>null</code> falls aktuell kein Objekt dargestellt wird
   */
  public ListProperty getObject() {
    return (ListProperty)super.getObject();
  }

  /**
   * Prueft, ob ein Objekt darstellbar ist.
   * @return <code>true</code> gdw. das angegebene Objekt eine Instanz von
   *         {@link ListProperty} ist.
   */
  public boolean canDisplay(Object data) {
    return data instanceof ListProperty;
  }

  /**
   * Liefert alle in der Tabelle selektierten Listenelemente.
   */
  public Object[] getSelectedElements() {
    int[] selIdx = table.getSelectedRows();
    Object[] selElements = new Object[selIdx.length];
    for (int i=0; i<selElements.length; i++)
      selElements[i] = ((ListPropertyTableModel)tableModel).getListElement( selIdx[i] );
    return selElements;
  }

  ///////////////////////////////////////////////////////////////////
  // TableModel fuer die ListProperty-Tabelle
  ///////////////////////////////////////////////////////////////////
  /**
   * Diese Klasse stellt ein TableModel fuer ListProperty-Instanzen dar.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  private class ListPropertyTableModel extends DisplayContainerTableModel {
    protected final String[] COLUMN_NAME = {"Index","Type","Value"};
    private ListProperty listProp = null;

    /**
     * Erzeugt ein neues Daten-Modell fuer ein <code>ListProperty</code>-Objekt.
     * @param data listProp Listen-Eigenschaft
     * @see schmitzm.data.property.ListProperty
     */
    public ListPropertyTableModel(ListProperty listProp) {
      setData(listProp);
    }

    /**
     * Erzeugt ein leeres Daten-Modell fuer ein <code>ListProperty</code>-Objekt.
     * @see schmitzm.data.property.ListProperty
     */
    public ListPropertyTableModel() {
      this(null);
    }

    /**
     * Setzt die Liste, deren Elemente dargestellt werden sollen.
     * @param listProp Listen-Eigenschaft (kann <code>null</code> sein!)
     */
    public void setData(Object listProp) {
      this.listProp = (ListProperty)listProp;
      fireTableDataChanged();
    }

    /**
     * Liefert ein Listen-Element.
     * @param row Zeile des Datenmodells
     */
    public Object getListElement(int row) {
      return this.listProp.getOneTimeReadAccess().getValue(row);
    }

    /**
     * Liefert einen Eintrag der Tabelle. Fuer Spalte 0 wird der Listenindex
     * zurueckgegeben, fuer Spalte 1 der Datentyp und fuer Spalte 2
     * der aktuelle Wert des Listenelements. Letzter wird nur angezeigt, wenn es
     * sich bei dem Listenelement um einen Basisdatentyp handelt.
     * @param row Tabellenzeile
     * @param col Tabellenspalte
     */
    public Object getValueAt(int row, int col) {
      switch( col ) {
        case 0: // Name der Property
                return row;
        case 1: // Typ des Listenelements
                String typeStr = getListElement(row).getClass().getSimpleName();
                if ( getListElement(row) instanceof Property )
                  typeStr += "["+((Property)getListElement(row)).getType().getSimpleName()+"]";
                return typeStr;
        case 2: // aktueller Wert des Listen-Elements
                try {
                  return getListElement(row);
                } catch( AccessViolationException err ) {
                  return "<no access!!>";
                }
      }
      return null;
    }

    /**
     * Liefert die Anzahl an Tabellen-Zeilen. Diese entspricht der Anzahl an
     * Elementen in der ListProperty.
     */
    public int getRowCount() {
      return (listProp!=null) ? listProp.getOneTimeReadAccess().getCount() :0;
    }

    /**
     * Liefert die Anzahl an Tabellen-Spalten. Diese ist konstant 3 (je eine
     * Spalte fuer Listenindex, Elementtyp und Elementwert).
     */
    public int getColumnCount() {
      return COLUMN_NAME.length;
    }

    /**
     * Liefert die Spaltennamen.
     * @see #COLUMN_NAME
     */
    public String getColumnName(int col) {
      return COLUMN_NAME[col];
    }

    /**
     * Prueft, ob eine Tabellen-Zelle editierbar ist. Dies ist nur fuer die
     * Spalte mit dem Element-Wert der Fall und nur wenn es sich um einen
     * Build-In-Type beinhaltet.
     * @param row Tabellen-Zeile
     * @param col Tabellen-Spalte
     * @see schmitzm.temp.BaseTypeUtil#isBaseType(Class)
     */
    public boolean isCellEditable(int row, int col) {
      return col == 2 &&
             BaseTypeUtil.isBaseType( getListElement(row) );
    }

    /**
     * Setzt den jeweiligen Property-Wert neu.
     * @param obj neue Objektbeschreibung (<code>String</code>)
     * @param row Tabellen-Zeile
     * @param col Tabellen-Spalte
     * @see #isCellEditable(int,int)
     * @see schmitzm.temp.BaseTypeUtil#convertFromString(String,Class)
     */
    public void setValueAt(Object obj, int row, int col) {
      if ( isCellEditable(row,col) ) {
        ListPropertyWriteAccess w = null;
        try {
          w = listProp.getWriteAccess(this);
          w.setValue(BaseTypeUtil.convertFromString((String)obj, listProp.getType()),row);
        } catch (AccessViolationException err) {
          XuluGUIMessages.showAccessError(GUI_COMPONENT,err);
        } catch (NumberFormatException err) {
          XuluGUIMessages.showBaseTypeError(GUI_COMPONENT,err);
        } catch (Exception err) {
          XuluGUIMessages.showError(GUI_COMPONENT,err);
        } finally {
          // Schreibrecht muss unbedingt wieder freigegeben werden!!
          if ( w!=null ) w.release();
        }

      }
    }
  }

}
