/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.gui;

import java.awt.Container;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import java.util.Vector;

import schmitzm.lang.WorkingThread;
import schmitzm.lang.WorkingThreadListener;
import schmitzm.lang.WorkingThreadAdapter;
import edu.bonn.xulu.model.XuluModel;
import edu.bonn.xulu.model.StepModel;
import edu.bonn.xulu.model.event.ModelAdapter;
import edu.bonn.xulu.model.XuluModelThread;
import edu.bonn.xulu.model.StepModelThread;

import edu.bonn.xulu.gui.XuluGUIMessages;

// fuer doku
import appl.plugin.multimodelcontrol.MultiModelControlPlugin;

/**
 * Diese Klasse stellt einen Container dar, ueber den der Modellablauf gesteuert
 * werden kann. Hierzu enthaelt er 5 Buttons
 * <ul>
 * <li><b>Init</b><br>
 *     erzeugt einen neuen Thread fuer das Modell (entweder {@link StepModelThread}
 *     falls das Modell {@link StepModel} implementiert, sonst {@link XuluModelThread})
 *     und fuehrt {@link XuluModel#initModel()} aus</li>
 * <li><b>Start/Resume</b><br>
 *     fuehrt {@link XuluModelThread#start()} aus</li>
 * <li><b>Pause</b><br>
 *     fuehrt {@link XuluModelThread#pause()} aus</li>
 * <li><b>Step</b><br>
 *     fuehrt {@link StepModelThread#step()} aus</li>
 * <li><b>Stop</b><br>
 *     fuehrt {@link XuluModelThread#terminate()} aus</li>
 * </ul>
 * Je nach Modell-Status sind die Button aktivier oder deaktiviert.
 * Uber die Methode {@link #addButtonActionListener(int,ActionListener)} koennen den
 * Button zusaetzliche Aktionen zugewiesen werden.
 *
 * <br><b>29.04.07, Dominik Appl</b><br>
 *  it is now possible to access start, stop, init, step and pause from outside (e.g. for controlling the
 *  model execution from the {@link MultiModelControlPlugin})
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ModelControlContainer extends Container {
  /** Identifiziert den Init-Button. */
  public static final int BUTTON_INIT  = 1;
  /** Identifiziert den Start/Resume-Button. */
  public static final int BUTTON_START = 2;
  /** Identifiziert den Pause-Button. */
  public static final int BUTTON_PAUSE = 3;
  /** Identifiziert den Step-Button. */
  public static final int BUTTON_STEP  = 4;
  /** Identifiziert den Stop-Button. */
  public static final int BUTTON_STOP  = 5;

  protected XuluModel model        = null;
  private JButton   initButton   = new JButton("Init");
  private JButton   startButton  = new JButton("Start/Resume");
  private JButton   pauseButton  = new JButton("Pause");
  private JButton   stepButton   = new JButton("Step");
  private JButton   stopButton   = new JButton("Stop");

  private XuluModelThread modelThread     = null;
  private Vector          threadListeners = new Vector<WorkingThreadListener>();
  private ModelAdapter    modelListener   = null;


  /**
   * Erzeugt eine neue Modell-Steuerungs-Komponente.
   * @param model zu steuerndes XuluModel
   */
  public ModelControlContainer(XuluModel model) {
    super();
    this.setLayout( new BoxLayout( this,BoxLayout.X_AXIS ) );
    this.add(initButton);
    this.add(startButton);
    //Pause funktioniert zur Zeit noch nicht richtig
    //this.add(pauseButton);
    this.add(stepButton);
    this.add(stopButton);
    setButtonsEnabled(true,false,false,false,false);
    initActions();
    // Auf Thread-Aktionen lauschen, um Button zu (de)aktivieren
    addThreadListener( new WorkingThreadAdapter() {
      public void threadStarted(WorkingThread thread) {
        setButtonsEnabled(false,false,true,true,true);
      }
      public void threadResumed(WorkingThread thread) {
        setButtonsEnabled(false,false,true,true,true);
      }
      public void threadPaused(WorkingThread thread) {
        setButtonsEnabled(false,true,true,false,true);
      }
      public void threadStopped(WorkingThread thread) {
        setButtonsEnabled(true,false,false,false,false);
      }
    });
    // Auf Model-Aktionen lauschen, um Button zu (de)aktivieren
    // Bemerke: Kann nicht ueber ThreadListener erfolgen, da Thread
    //          erst beim Start initialisiert wird!
    this.modelListener = new ModelAdapter() {
      public void modelInitialised(XuluModel model) {
        setButtonsEnabled(true, true, true, false, false);
      }
    };
    // Model setzen
    setModel(model);
  }

  /**
   * Setzt das Modell, das durch den Container kontrolliert wird.
   * @param model ein Modell
   */
  public void setModel(XuluModel model) {
    if ( model != null ) {
      if ( modelThread != null && modelThread.getMode() != modelThread.MODE_TERMINATE )
        perfomModelAction( BUTTON_STOP );
      model.removeModelListener( modelListener );
    }

    this.model = model;
    model.addModelListener( modelListener );
  }

  /**
   * Liefert das Modell, das durch den Container kontrolliert wird.
   */
  public XuluModel getModel() {
    return this.model;
  }

  /**
   * Belegt die Button mit Modell-Aktionen.
   */
  private void initActions() {
    addButtonActionListener( BUTTON_INIT, new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        perfomModelAction(BUTTON_INIT);
      }
    });
    addButtonActionListener( BUTTON_START, new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        perfomModelAction(BUTTON_START);
      }
    });
    addButtonActionListener( BUTTON_PAUSE, new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        perfomModelAction(BUTTON_PAUSE);
      }
    });
    addButtonActionListener( BUTTON_STEP, new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        perfomModelAction(BUTTON_STEP);
      }
    });
    addButtonActionListener( BUTTON_STOP, new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        perfomModelAction(BUTTON_STOP);
      }
    });
  }

//  /**
//   * Liefert den Thread, der durch die Komponente kontrolliert wird.
//   */
//  public XuluModelThread getModelThread() {
//    return modelThread;
//  }

  /**
   * Liefert einen Button des Containers.
   */
  private JButton getButton(int button) {
    switch( button ) {
      case BUTTON_INIT:  return initButton;
      case BUTTON_START: return startButton;
      case BUTTON_PAUSE: return pauseButton;
      case BUTTON_STEP:  return stepButton;
      case BUTTON_STOP:  return stopButton;
    }
    return null;
  }

  /**
   * Belegt einen Button mit einer Aktion.
   * @param button   Code fuer den Button
   * @param listener Aktion
   * @see #BUTTON_INIT
   * @see #BUTTON_START
   * @see #BUTTON_PAUSE
   * @see #BUTTON_STEP
   * @see #BUTTON_STOP
   */
  public void addButtonActionListener(int button, ActionListener listener) {
    JButton b = getButton(button);
    if (b != null) b.addActionListener(listener);
  }

//  /**
//   * Liefert alle bisher einem Button zugewiesenen Aktionen.
//   * @param button Code fuer den Button
//   * @see #BUTTON_INIT
//   * @see #BUTTON_START
//   * @see #BUTTON_PAUSE
//   * @see #BUTTON_STEP
//   * @see #BUTTON_STOP
//   */
//  public ActionListener[] getButtonActionListeners(int button) {
//    JButton b = getButton(button);
//    if (b != null) return b.getActionListeners().clone();
//    return new ActionListener[0];
//  }

  /**
   * Fuegt dem Modell-Thread einen Listener hinzu.
   * @param listener Modell-Listener
   */
  public void addThreadListener(WorkingThreadListener listener) {
    threadListeners.add(listener);
    if ( modelThread != null )
      modelThread.addThreadListener(listener);
  }

  /**
   * (De)Aktiviert die Steuerungsbutton.
   */
  private void setButtonsEnabled(boolean initEn, boolean startEn, boolean stepEn, boolean pauseEn, boolean stopEn) {
    initButton.setEnabled(initEn);
    startButton.setEnabled(startEn);
    stepButton.setEnabled( (model instanceof StepModel) && stepEn);
    pauseButton.setEnabled(pauseEn);
    stopButton.setEnabled(stopEn);
  }

  /**
   * Fuehrt eine Modell-Aktion aus. Diese koennten auch einzeln direkt in die
   * ActionListener programmiert werden. So kann jedoch das Error-Handling
   * einheitlich erfolgen.
   */
  private void perfomModelAction(int action) {
    try {
      switch( action ) {
        case BUTTON_INIT:  init(); break;
        case BUTTON_START: start(); break;
        case BUTTON_PAUSE: pause(); break;
        case BUTTON_STEP:  step();break;
        case BUTTON_STOP:  stop(); break;
      }
    } catch( Exception err ) {
      XuluGUIMessages.showError(this,err);
    }
  }

  /**
	 * starts the model
	 */
	public void start() {
		modelThread.start();
	}

	/**
	 * inits the model
	 */
	public void init() {
		createNewThread();
	}

	/**
	 * stops the model
	 */
	public void stop() {
		modelThread.terminate();
	}

	/**
	 * makes one model step
	 */
	public void step() {
		if (modelThread instanceof StepModelThread)
			((StepModelThread) modelThread).step();
	}

	/**
	 * pauses the model
	 */
	public void pause() {
		modelThread.pause();
	}

	public XuluModelThread getModelThread(){
		return modelThread;
	}


  /**
   * Erzeugt einen neuen Thread fuer das Modell. Handelt es sich um eine
   * StepModel-Modell, wird ein StepModelThread erzeugt, ansonsten ein
   * XuluModelThread.
   */
  protected void createNewThread() {
    // Wenn die Ressourcen noch nicht richtig zugeordnet sind, wird auch kein
    // Thread erzeugt.
    // Bemerke: Meldungen zu fehlerhaften Ressourcen sollten bereits bei
    //          beim Befuellen des ResourceManagers mit den GUI-Komponenten
    //          gemeldet werden (z.B. ModelControlFrame_Basic.setResources())
    if ( !model.getContentManager().check() )
      return;

    // noch aktiven Thread anhalten
    if ( modelThread != null && modelThread.isAlive() )
      modelThread.terminate();

    // neuen Thread erzeugen
    if ( model instanceof StepModel )
      modelThread = new StepModelThread((StepModel)model);
    else
      modelThread = new XuluModelThread(model);

    // dem neuen Thread die Listener hinzufuegen
    for (int i=0; i<threadListeners.size(); i++)
      modelThread.addThreadListener( (WorkingThreadListener)threadListeners.get(i) );

    // Modell initialisieren
    model.initModel();
  }

}
