/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.io.feature.gt;

import java.io.File;

import schmitzm.geotools.io.GeoImportUtil;
import schmitzm.geotools.io.GeoExportUtil;

import edu.bonn.xulu.io.ImportFactory;
import edu.bonn.xulu.io.ExportFactory;
import edu.bonn.xulu.io.InstantiationFactory;
import edu.bonn.xulu.io.AbstractFactory;
import edu.bonn.xulu.plugin.io.IOUtil;
import edu.bonn.xulu.plugin.data.feature.SingleFeatureCollection;
import edu.bonn.xulu.appl.XuluRegistry;

/**
 * Diese Factory importiert und exportiert Instanzen des Datentyps
 * {@link SingleFeatureCollection} aus/in Shape-Files.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class SingleFeatureCollectionFactory_ShapeFile extends AbstractFactory implements ImportFactory, ExportFactory {
  ////////////////////////////////////////////////////////////////////////
  ///////////////     Implementierung von ImportFactory     //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Liefert den Datentyp, den die Factory erzeugt.
   * @return immer {@link SingleFeatureCollection SingleFeatureCollection.class}
   */
  public Class getImportType() {
    return SingleFeatureCollection.class;
  }

  /**
   * Liefert den Datentyp, den die Factory als Import-Quelle benoetigt.
   * @return immer {@link File File.class}
   */
  public Class getImportSourceType() {
    return File.class;
  }

  /**
   * Importiert eine Instanz von {@link SingleFeatureCollection} aus einer Datei im
   * ShapeFile-Format (.shp).
   * @param input Eingabe-Quelle (muss ein {@link File} sein!)
   * @param reg   Instanz der Xulu-Registry, ueber die eine Factory ermittelt
   *              wird, die eine Standard-Instanz von {@link SingleFeatureCollection}
   *              erzeugt, in die das Objekt importiert wird
   * @exception UnsupportedOperationException falls als Eingabe-Quelle keine
   *            Datei angegeben wird
   * @see GeoImportUtil#readFeaturesFromShapeFile(File)
   */
  public SingleFeatureCollection importObject(Object input, XuluRegistry reg) throws Exception {
    AbstractFactory.checkImportSourceObject(this,input,getImportSourceType(),true);
    // Eine SingleFeatureCollection instanziieren und ShapeFile importieren
    InstantiationFactory instFac = AbstractFactory.getInstantiationFactoryFromRegistry(reg,getImportType(),true);
    SingleFeatureCollection sfc = (SingleFeatureCollection)instFac.newInstance(false);
    sfc.setFeatureCollection( GeoImportUtil.readFeaturesFromShapeFile((File)input) );

    // Style importieren und falls vorhanden in Property einfuegen
    IOUtil.importFeatureStyle( sfc, ((File)input).toURI().toURL() );

    // Dateinamen in die Bezeichnung aufnehmen
    sfc.setDescription(sfc.getDescription().concat(" [").concat(((File)input).getName()).concat("]"));
    return sfc;

  }

  ////////////////////////////////////////////////////////////////////////
  ///////////////     Implementierung von ExportFactory     //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Prueft, ob ein Objekt exportiert werden kann. Dies ist der
   * Fall, wenn es sich um ein {@link SingleFeatureCollection} handelt.
   * @param object zu pruefendes Objekt
   */
  public boolean isExportable(Object object) {
    return object!=null && object instanceof SingleFeatureCollection;
  }

  /**
   * Prueft, ob ein Objekttyp exportiert werden kann. Dies ist der
   * Fall, wenn es sich um {@link SingleFeatureCollection} (oder eine Unterklasse)
   * handelt.
   * @param c zu pruefender Objekttyp
   */
  public boolean isExportable(Class c) {
    return SingleFeatureCollection.class.isAssignableFrom(c);
  }

  /**
   * Liefert den Objekt-Typ, den die Factory als Ziel zum Exportieren
   * benoetigt.
   * @return immer {@link File File.class}
   */
  public Class getExportDestinationType() {
    return File.class;
  }

  /**
   * Exportiert eine Instanz von {@link SingleFeatureCollection} in das
   * ShapeFile-Format.
   * @param fc     zu exportiertendes Objekt (muss ein {@link SingleFeatureCollection} sein
   * @param output Export-Ziel (muss ein {@link File} sein!)
   * @exception UnsupportedOperationException falls als Export-Ziel keine
   *            Datei angegeben wird
   * @exception IllegalArgumentException falls es sich bei <code>fc</code>
   *            nicht um ein {@link SingleFeatureCollection} handelt
   */
  public void exportObject(Object fc, Object output) throws Exception {
    AbstractFactory.checkExportDestinationObject(this,output,getExportDestinationType(),true);
    if ( !isExportable(fc) )
      throw new IllegalArgumentException(getClass().getName().concat(" can not export instances of ").concat(fc.getClass().getName()));
    SingleFeatureCollection sfc = (SingleFeatureCollection)fc;
    GeoExportUtil.writeFeaturesToShapeFile(sfc.getFeatureCollection(),(File)output);
    IOUtil.exportStyle( sfc, ((File)output).toURI().toURL() );
  }
}
