/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.io.grid;

import java.awt.image.DataBuffer;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import schmitzm.data.WritableGrid;
import schmitzm.swing.InputOption;
import schmitzm.swing.SelectionInputOption;
import schmitzm.swing.ManualInputOption;
import schmitzm.swing.MultipleOptionPane;
import schmitzm.geotools.io.GeoImportUtil;
import edu.bonn.xulu.io.InstantiationFactory;
import edu.bonn.xulu.io.AbstractFactory;
import edu.bonn.xulu.io.FactoryCanceledException;
import edu.bonn.xulu.appl.XuluConstants;

// nur fuer Doku
import edu.bonn.xulu.data.XuluObject;
import edu.bonn.xulu.plugin.data.grid.SingleGrid;

/**
 * Diese Klasse bildet eine Oberklasse fuer alle Factorys, die Standard-Instanzen
 * des Datentyps {@link WritableGrid} erzeugen.<br>
 * <b>Bemerke:</b><br>
 * Dieser Datentyp stellt <b>kein</b> {@linkplain XuluObject Xulu-Objekt} dar,
 * kann also nicht direkt im Xulu-Datenpool gespeichert werden. Hierzu ist der Datentyp
 * {@link SingleGrid} (bzw. eine entsprechende Factory) zu verwenden!
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class WritableGridFactory extends AbstractFactory implements InstantiationFactory {
  /**
   * Liefert den Datentyp, den die Factory erzeugt.
   * @return immer {@link WritableGrid WritableGrid.class}
   */
  public Class getInstanceType() {
    return WritableGrid.class;
  }

  /**
   * Erzeugt ein neues Raster. Die dazu benoetigten Angaben werden von einem
   * Vorlagen-Raster uebernommen, koennen aber ueber einen Anwender-Dialog
   * abgeaendert werden.
   * @param sample Vorlage
   * @param dialog gibt an, ob die Factory einen Dialog mit dem Anwender
   *               fuehren darf, um die Instanz zu initialisieren (bei
   *               <code>false</code> wird die Vorlage ohne Dialog kopiert)
   * @exception IllegalArgumentException falls es sich bei der Vorlage nicht
   *            um ein <code>WritableGridCoverage</code> handelt
   * @exception FactoryCanceledException falls der Anwender-Dialog ueber
   *            "Abbrechen" verlassen wird
   */
  public WritableGrid newInstance(boolean dialog, Object sample) {
    if ( sample == null )
      throw new IllegalArgumentException(this.getClass().getSimpleName().concat(" can not create an instance of ").concat(getInstanceType().getSimpleName()).concat(" from null") );
    if ( !getInstanceType().isInstance(sample) )
      throw new IllegalArgumentException(this.getClass().getSimpleName().concat(" can not create an instance of ").concat(getInstanceType().getSimpleName()).concat(" from ").concat(sample.getClass().getName()) );

    WritableGrid wg = (WritableGrid)sample;
    return newInstance( dialog,
                        wg.getSampleType(),
                        wg.getWidth(),
                        wg.getHeight(),
                        wg.getX(),
                        wg.getY(),
                        wg.getRealWidth(),
                        wg.getRealHeight(),
                        wg.getCoordinateReferenceSystem()
    );
  }

  /**
   * Erzeugt ein neues Raster. Die dazu benoetigten Angaben
   * <ul>
   * <li>Datentyp</li>
   * <li>Rasterbreite (in Zellen)</li>
   * <li>Rasterhoehe (in Zellen)</li>
   * <li>GeoReferenz-Koordinaten (lat/lon)</li>
   * <li>Reale Groesse des Rasters</li>
   * </ul>
   * werden durch einen Anwender-Dialog abgefragt.
   * @param dialog gibt an, ob die Factory einen Dialog mit dem Anwender
   *               fuehren darf, um die Instanz zu initialisieren (bei
   *               <code>false</code> wird eine Standard-Instanz erzeugt)
   * @exception FactoryCanceledException falls der Anwender-Dialog ueber
   *            "Abbrechen" verlassen wird
   */
  public WritableGrid newInstance(boolean dialog) {
    return newInstance(dialog, DataBuffer.TYPE_FLOAT, 500, 500, 0.0, 0.0, 500.0, 500.0, null);
  }

  /**
   * Erzeugt ein neues Raster aus einer Vorlage. Trotzdem wird ein Dialog
   * angezeigt, in dem die Vorlagen-Daten nacheditiert werden koennen.
   * Falls die Vorlage <code>null</code> ist, werden Default-Angaben vorgeblendet.
   * @param dialog bestimmt, ob die Vorlagen-Daten sofort uebernommen werden, oder
   *               ueber einen Anwender-Dialog abgeaendert werden koennen
   * @param type Datentyp der gespeicherten Objekte
   * @param wCells Breite in Zellen
   * @param hCells Hoehe in Zellen
   * @param xCoord Georeferenz Longitute (Westliche/Linke Kante)
   * @param yCoord Georeferenz Latitude (Suedliche/Untere Kante)
   * @param wReal  Breite
   * @param hReal  Hoehe
   * @param crs Referenzsystem fuer das Raster (wenn {@code null}, wird
   *            {@link GeoImportUtil#DEFAULT_CRS} verwendet
   */
  public WritableGrid newInstance(boolean dialog, int type, int wCells, int hCells, double xCoord, double yCoord, double wReal, double hReal, CoordinateReferenceSystem crs) {
    if ( crs == null )
      crs = GeoImportUtil.DEFAULT_CRS;

    // Sample-Angaben ohne Dialog uebernehmen
    if ( !dialog )
      return newInstance(type,wCells,hCells,0,0,xCoord,yCoord,wReal,hReal,crs);

    InputOption[] inpOpt = new InputOption[] {
        new SelectionInputOption.Combo("Typ", true,
                        new Integer[] {/*DataBuffer.TYPE_BYTE, DataBuffer.TYPE_USHORT, DataBuffer.TYPE_INT,*/ DataBuffer.TYPE_FLOAT},
                        (Object)type,
                        new String[] {/*"Byte", "Unsigned Short", "Integer",*/"Float"}),
        new ManualInputOption.Integer("Breite in Zellen", true, wCells),
        new ManualInputOption.Integer("Hoehe in Zellen", true, hCells),
        new ManualInputOption.Double("Geo-Referenz Lon/X (Südwestliche Ecke)", true, xCoord),
        new ManualInputOption.Double("Geo-Referenz Lat/Y (Südwestliche Ecke)", true, yCoord),
        new ManualInputOption.Double("Geo-Referenz Breite", true, wReal),
        new ManualInputOption.Double("Geo-Referenz Hoehe", true, hReal)
    };

    Object[] inpValue = MultipleOptionPane.showMultipleInputDialog(
      null,"Grid erzeugen", inpOpt);
    if (inpValue == null)
      throw new FactoryCanceledException(this);

    return newInstance(
      (Integer)inpValue[0], // Datentyp
      (Integer)inpValue[1], // Breite in Zellen
      (Integer)inpValue[2], // Hoehe in Zellen
      0,
      0,
      (Double)inpValue[3],  // Lat
      (Double)inpValue[4],  // Lon
      (Double)inpValue[5],  // Breite
      (Double)inpValue[6],  // Hoehe
      crs
    );
  }

  /**
   * Erzeugt ein neues Grid.
   * @param type Datentyp der gespeicherten Objekte
   * @param widthc Breite in Zellen
   * @param heightc Hoehe in Zellen
   * @param minX    Index der ersten Zelle in X-Richtung
   * @param minY    Index der ersten Zelle in Y-Richtung
   * @param x Georeferenz Longitute (Westliche/Linke Kante)
   * @param y Georeferenz Latitude (Suedliche/Untere Kante)
   * @param width Breite
   * @param height Hoehe
   * @param crs Referenzsystem fuer das Raster
   */
  public abstract WritableGrid newInstance(int type, int widthc, int heightc, int minX, int minY, double x, double y, double width, double height, CoordinateReferenceSystem crs);
}
