/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.io.grid.awt;

import schmitzm.data.WritableGrid;
import edu.bonn.xulu.io.AbstractFactory;
import edu.bonn.xulu.io.InstantiationFactory;
import edu.bonn.xulu.io.FactoryCanceledException;
import edu.bonn.xulu.plugin.data.grid.GridList;
import schmitzm.swing.MultipleOptionPane;
import schmitzm.swing.InputOption;
import schmitzm.swing.ManualInputOption;

/**
 * Diese Factory erzeugt Standard-Instanzen von {@link GridList} mit
 * durch die Factory {@link WritableGridRasterFactory} erzeugten (auf Standard-AWT
 * basierenden) {@link WritableGrid}-Instanzen als Inhalt.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class GridListFactory extends AbstractFactory implements InstantiationFactory {
  /**
   * Liefert den Datentyp, der von der Factory erzeugt wird.
   * @return immer {@link GridList GridList.class}
   */
  public Class getInstanceType() {
    return GridList.class;
  }

  /**
   * Erzeugt eine neue Instanz von {@link GridList}. Hierzu wird ueber einen
   * Anwenderdialog abgefragt wie viele Raster die Liste enthalten soll und
   * die von {@link #getWritableGridFactory()} gelieferte Factory herangezogen,
   * um die Raster zu erzeugen.
   * @param sample Vorlage
   * @param dialog gibt an, ob die Factory einen Dialog mit dem Anwender
   *               fuehren darf, um die Instanz zu initialisieren (bei
   *               <code>false</code> wird die Vorlage ohne Dialog kopiert)
   * @exception IllegalArgumentException falls es sich bei der Vorlage nicht
   *            um eine <code>GridList</code> handelt
   * @exception FactoryCanceledException falls der Anwender-Dialog ueber
  *            "Abbrechen" verlassen wird
  */
  public GridList newInstance(boolean dialog, Object sample) throws Exception {
    if ( sample == null )
      throw new IllegalArgumentException(this.getClass().getSimpleName().concat(" can not create an instance of ").concat(getInstanceType().getSimpleName()).concat(" from null") );
    if ( !getInstanceType().isInstance(sample) )
      throw new IllegalArgumentException(this.getClass().getSimpleName().concat(" can not create an instance of ").concat(getInstanceType().getSimpleName()).concat(" from ").concat(sample.getClass().getName()) );

    return createFromSample(dialog,(GridList)sample);
  }

  /**
   * Erzeugt eine neue Instanz von {@link GridList}. Hierzu wird ueber einen
   * Anwenderdialog abgefragt wie viele Raster die Liste enthalten soll und
   * die von {@link #getWritableGridFactory()} gelieferte Factory herangezogen,
   * um die Raster zu erzeugen.
   * @param dialog gibt an, ob die Factory einen Dialog mit dem Anwender
   *               fuehren darf, um die Instanz zu initialisieren (bei
   *               <code>false</code> wird eine Standard-Instanz erzeugt)
   * @exception FactoryCanceledException falls der Anwender-Dialog ueber
   *            "Abbrechen" verlassen wird
   */
  public GridList newInstance(boolean dialog) throws Exception {
    return createFromSample(dialog, null);
  }

  /**
   * Erzeugt eine neue GridList aus einer Vorlage. Trotzdem wird ein Dialog
   * angezeigt, in dem die Vorlagen-Daten nacheditiert werden koennen.
   * Falls die Vorlage <code>null</code> ist, werden Default-Angaben vorgeblendet.
   * @param sample Vorlage
   * @param dialog bestimmt, ob die Vorlagen-Daten sofort uebernommen werden, oder
   *               ueber einen Anwender-Dialog abgeaendert werden koennen
   */
  public GridList createFromSample(boolean dialog, GridList sample) throws Exception {
    // Anzahl an Grids aus dem Sample uebernehmen
    int count = (sample == null) ? 0 : sample.getGridCount();
    // Anzahl zu erzeugender Grids abfragen
    if (dialog) {
      Object[] inpValue = MultipleOptionPane.showMultipleInputDialog(
                               null,
                               this.getInstanceType().getSimpleName().concat(" erzeugen"),
                               new InputOption[] { new ManualInputOption.Integer("Anzahl an Grids in der Liste",true,count) }
      );
      if ( inpValue == null )
        throw new FactoryCanceledException(this);
      count = (Integer)inpValue[0];
    }

    // nacheinander die Grids erzeugen
    InstantiationFactory fac = getWritableGridFactory();
    if ( !WritableGrid.class.isAssignableFrom(fac.getInstanceType()) )
      throw new UnsupportedOperationException(getClass().getSimpleName().concat(".getWritableGridFactory() must provide a factory that creates a WritableGrid!"));

    WritableGrid[] grid = new WritableGrid[count];
    for (int i=0; i<grid.length; i++) {
      // falls vorhanden, die Werte der Grids aus der Vorlage vorblenden
      WritableGrid sampleGrid = (sample!=null && i<sample.getGridCount()) ? sample.getGrid(i) : null;
      if ( sampleGrid == null )
        grid[i] = (WritableGrid)fac.newInstance(dialog);
      else
        grid[i] = (WritableGrid)fac.newInstance(dialog,sampleGrid);
    }
    GridList gl = new GridList(grid);

    // Style von der Vorlage uebernehmen
    edu.bonn.xulu.plugin.io.IOUtil.setStyleToXuluObject(
      gl,
      edu.bonn.xulu.plugin.io.IOUtil.getStyleFromXuluObject( sample )
    );

    return gl;
  }

  /**
   * Liefert eine {@link InstantiationFactory}, die Instanzen von {@link WritableGrid}
   * erzeugt. Mit diesen Instanzen wird die Liste gefuellt.<br>
   * Standardmaessig wird eine Instanz von {@link WritableGridRasterFactory}
   * zurueckgegeben. Sub-Klassen koennen diese Methode ueberschreiben, um
   * andere Arten von WritableGrid zu erzeugen.
   */
  protected InstantiationFactory getWritableGridFactory() {
    return new WritableGridRasterFactory();
  }
}
