/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.io.grid.awt;

import schmitzm.data.WritableGrid;
import edu.bonn.xulu.io.AbstractFactory;
import edu.bonn.xulu.io.InstantiationFactory;
import edu.bonn.xulu.plugin.data.grid.SingleGrid;
import edu.bonn.xulu.plugin.io.grid.awt.WritableGridRasterFactory;
import schmitzm.data.property.ScalarProperty;

/**
 * Diese Factory erzeugt Standard-Instanzen von {@link SingleGrid} mit einem
 * durch die Factory {@link WritableGridRasterFactory} erzeugten (auf Standard-AWT
 * basierenden) {@link WritableGrid}.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class SingleGridFactory extends AbstractFactory implements InstantiationFactory {
  /**
   * Liefert den Datentyp, der von der Factory erzeugt wird.
   * @return immer {@link SingleGrid SingleGrid.class}
   */
  public Class getInstanceType() {
    return SingleGrid.class;
  }

  /**
   * Erzeugt eine neue Instanz von {@link SingleGrid}. Hierzu wird die
   * Factory {@link WritableGridRasterFactory} herangezogen.
   * @param dialog gibt an, ob die Factory einen Dialog mit dem Anwender
   *               fuehren darf, um die Instanz zu initialisieren (bei
   *               <code>false</code> wird die Vorlage ohne Dialog kopiert)
   * @param sample Vorlage
   * @exception IllegalArgumentException falls es sich bei der Vorlage nicht
   *            um ein <code>SingleGrid</code> oder um ein <code>WritableGrid</code>
   *            handelt
   */
  public SingleGrid newInstance(boolean dialog, Object sample) throws Exception {
    if ( sample == null )
      throw new IllegalArgumentException(this.getClass().getSimpleName().concat(" can not create an instance of ").concat(getInstanceType().getSimpleName()).concat(" from null") );
    if ( !getInstanceType().isInstance(sample) && !WritableGrid.class.isInstance(sample) )
      throw new IllegalArgumentException(this.getClass().getSimpleName().concat(" can not create an instance of ").concat(getInstanceType().getSimpleName()).concat(" from ").concat(sample.getClass().getName()) );

    InstantiationFactory fac = getWritableGridFactory();
    if ( !WritableGrid.class.isAssignableFrom(fac.getInstanceType()) )
      throw new UnsupportedOperationException(getClass().getSimpleName().concat(".getWritableGridFactory() must provide a factory that creates a WritableGrid!"));

    WritableGrid sampleGrid = ( sample instanceof WritableGrid ) ? (WritableGrid)sample : ((SingleGrid)sample).getGrid();
    SingleGrid sg = new SingleGrid( (WritableGrid)fac.newInstance( dialog, sampleGrid ) );
    // Style von der Vorlage uebernehmen
    if ( sample instanceof SingleGrid )
      edu.bonn.xulu.plugin.io.IOUtil.setStyleToXuluObject(
        sg,
        edu.bonn.xulu.plugin.io.IOUtil.getStyleFromXuluObject( (SingleGrid) sample )
      );
    return sg;
  }

  /**
   * Erzeugt eine neue Instanz von {@link SingleGrid}. Fuer den Inhalt des
   * Grids wird die von der Methode {@link #getWritableGridFactory()} gelieferte
   * Factory herangezogen.
   * @param dialog gibt an, ob die Factory einen Dialog mit dem Anwender
   *               fuehren darf, um die Instanz zu initialisieren (bei
   *               <code>false</code> wird eine Standard-Instanz erzeugt)
   */
  public SingleGrid newInstance(boolean dialog) throws Exception {
    InstantiationFactory fac = getWritableGridFactory();
    if ( !WritableGrid.class.isAssignableFrom(fac.getInstanceType()) )
      throw new UnsupportedOperationException(getClass().getSimpleName().concat(".getWritableGridFactory() must provide a factory that creates a WritableGrid!"));

    return new SingleGrid( (WritableGrid)fac.newInstance(dialog) );
  }

  /**
   * Liefert eine {@link InstantiationFactory}, die Instanzen von {@link WritableGrid}
   * erzeugt. Mit diesen Instanzen wird das SingleGrid gefuellt.<br>
   * Standardmaessig wird eine Instanz von {@link WritableGridRasterFactory}
   * zurueckgegeben. Sub-Klassen koennen diese Methode ueberschreiben, um
   * andere Arten von WritableGrid zu erzeugen.
   */
  protected InstantiationFactory getWritableGridFactory() {
    return new WritableGridRasterFactory();
  }
}
