/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.io.grid.gt;

import java.io.File;

import org.geotools.coverage.grid.GridCoverage2D;

import schmitzm.geotools.grid.WritableGridCoverage;
import schmitzm.geotools.io.GeoImportUtil;
import schmitzm.geotools.io.GeoExportUtil;

import edu.bonn.xulu.io.AbstractFactory;
import edu.bonn.xulu.io.ImportFactory;
import edu.bonn.xulu.io.ExportFactory;
import edu.bonn.xulu.appl.XuluRegistry;

// nur fuer Doku
import edu.bonn.xulu.data.XuluObject;
import edu.bonn.xulu.plugin.data.grid.SingleGrid;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

/**
 * Diese Factory importiert und exportiert Instanzen des Datentyps
 * {@link WritableGridCoverage} aus/in das GeoTiff-Format.<br>
 * <b>Bemerke:</b><br>
 * Die erzeugten Instanzen stellen <b>keine</b> {@linkplain XuluObject Xulu-Objekte} dar,
 * koennen also nicht direkt im Xulu-Datenpool gespeichert werden. Hierzu ist der Datentyp
 * {@link SingleGrid} (bzw. eine entsprechende Factory) zu verwenden!
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class GridCoverageFactory_GeoTiff extends AbstractFactory implements ImportFactory, ExportFactory {
  ////////////////////////////////////////////////////////////////////////
  ///////////////     Implementierung von ImportFactory     //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Liefert den Datentyp, den die Factory erzeugt.
   * @return immer {@link WritableGridCoverage WritableGridCoverage.class}
   */
  public Class getImportType() {
    return WritableGridCoverage.class;
  }

  /**
   * Liefert den Datentyp, den die Factory als Import-Quelle benoetigt.
   * @return immer {@link File File.class}
   */
  public Class getImportSourceType() {
    return File.class;
  }

  /**
   * Importiert eine Instanz von {@link WritableGridCoverage} aus einer
   * Datei.
   * @param input Eingabe-Quelle (muss ein {@link File} sein!)
   * @exception UnsupportedOperationException falls als Eingabe-Quelle keine
   *            Datei angegeben wird
   * @param reg wird nicht beruecksichtigt, da immer Instanzen von
   *            {@link WritableGridCoverage} verwendet werden!
   * @see GeoImportUtil#readGridFromGeoTiff(File,CoordinateReferenceSystem)
   */
  public WritableGridCoverage importObject(Object input, XuluRegistry reg) throws Exception {
    AbstractFactory.checkImportSourceObject(this,input,getImportSourceType(),true);
    return (WritableGridCoverage)GeoImportUtil.readGridFromGeoTiff((File)input);
  }

  ////////////////////////////////////////////////////////////////////////
  ///////////////     Implementierung von ExportFactory     //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Prueft, ob ein Objekt exportiert werden kann. Dies ist der
   * Fall, wenn es sich um ein {@link GridCoverage2D} handelt.
   * @param obj zu pruefendes Objekt
   */
  public boolean isExportable(Object obj) {
    return obj!=null & obj instanceof GridCoverage2D;
  }

  /**
   * Prueft, ob ein Objekttyp exportiert werden kann. Dies ist der
   * Fall, wenn es sich um {@link GridCoverage2D} (oder eine Unterklasse)
   * handelt.
   * @param c zu pruefender Objekttyp
   */
  public boolean isExportable(Class c) {
    return GridCoverage2D.class.isAssignableFrom(c);
  }

  /**
   * Liefert den Objekt-Typ, den die Factory als Ziel zum Exportieren
   * benoetigt.
   * @return immer {@link File File.class}
   */
  public Class getExportDestinationType() {
    return File.class;
  }

  /**
   * Exportiert eine Instanz von {@link GridCoverage2D} in eine Datei.
   * @param grid zu exportiertendes Raster (muss ein {@link GridCoverage2D} sein
   * @param output Export-Ziel (muss ein {@link File} sein!)
   * @exception UnsupportedOperationException falls als Export-Ziel keine
   *            Datei angegeben wird
   * @exception IllegalArgumentException falls es sich bei dem angegeben Raster
   *            nicht um ein {@link GridCoverage2D} handelt
   * @see GeoExportUtil#writeGridToGeoTiff(GridCoverage2D,File)
   */
  public void exportObject(Object grid, Object output) throws Exception {
    AbstractFactory.checkExportDestinationObject(this,output,getExportDestinationType(),true);
    if ( !isExportable(grid) )
      throw new IllegalArgumentException(getClass().getName().concat(" can not export instances of ").concat(grid.getClass().getName()));
    GeoExportUtil.writeGridToGeoTiff((GridCoverage2D)grid,(File)output);
  }
}
