/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.io.misc;

import java.io.File;
import java.awt.geom.Point2D;

import org.geotools.feature.Feature;
import org.geotools.feature.FeatureIterator;
import com.vividsolutions.jts.geom.Point;

import schmitzm.geotools.io.GeoImportUtil;

import schmitzm.data.property.ListProperty;
import schmitzm.data.property.ListPropertyWriteAccess;

import edu.bonn.xulu.io.ImportFactory;
import edu.bonn.xulu.io.InstantiationFactory;
import edu.bonn.xulu.io.AbstractFactory;
import edu.bonn.xulu.plugin.data.misc.CAArea;
import edu.bonn.xulu.appl.XuluRegistry;
import edu.bonn.xulu.data.XuluObject;
import edu.bonn.xulu.data.DynamicXuluObject;

/**
 * Diese Factory importiert Instanzen des Datentyps {@link CAArea} aus einem
 * Shape-File. Dessen Features muessen folgende Attribute besitzen:
 * <table width="70%">
 *   <tr><td><b>Attribut</b></td>
 *       <td><b>Datentyp</b></td>
 *       <td><b>Beschreibung</b></td>
 *       </tr>
 *   <tr><td><i>Default-Geometry</i></td>
 *       <td>Punkt</td>
 *       <td>Lat/Lon-Koordinate der Area</td>
 *       </tr>
 *   <tr><td><code>Name</code></td>
 *       <td><code>String</code></td>
 *       <td>Name der Area</td>
 *       </tr>
 * </table>
 * Aus den importierten {@link CAArea}-Objekten wird anschliessend ein
 * {@link XuluObject} erstellt, das aus einer {@link ListProperty} besteht,
 * in der die Areas gespeichert werden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class CAAreaListFactory_ShapeFile extends AbstractFactory implements ImportFactory {
  /** (Unter-)Typ von {@link CAArea} fuer den die {@link ListProperty} des
   *  {@link XuluObject} erzeugt wird. Unterklassen muessen diese Variable
   *  neu belegen, um Spezialisierungen von {@link CAArea} zu verwenden.
   */
  protected Class<? extends CAArea> caAreaType = CAArea.class;

  ////////////////////////////////////////////////////////////////////////
  ///////////////     Implementierung von ImportFactory     //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Liefert den Datentyp, den die Factory erzeugt.
   * @return immer {@link XuluObject XuluObject.class}
   */
  public Class getImportType() {
    return XuluObject.class;
  }

  /**
   * Liefert den Datentyp, den die Factory als Import-Quelle benoetigt.
   * @return immer {@link File File.class}
   */
  public Class getImportSourceType() {
    return File.class;
  }

  /**
   * Importiert eine Instanz von {@link XuluObject} aus einer Datei im
   * oben beschriebenen ShapeFile-Format (.shp).
   * @param input Eingabe-Quelle (muss ein {@link File} sein!)
   * @param reg   Instanz der Xulu-Registry (wird nicht verwendet!)
   * @exception UnsupportedOperationException falls als Eingabe-Quelle keine
   *            Datei angegeben wird
   * @see GeoImportUtil#readFeaturesFromShapeFile(File)
   */
  public XuluObject importObject(Object input, XuluRegistry reg) throws Exception {
    AbstractFactory.checkImportSourceObject(this,input,getImportSourceType(),true);

    // Xulu-Objekt mit einer ListProperty[CAArea] erzeugen
    DynamicXuluObject xuluObject = new DynamicXuluObject();
    ListProperty      prop       = new ListProperty("Areas",caAreaType);
    xuluObject.addProperty( prop );
    // Features einlesen
    FeatureIterator featIter = GeoImportUtil.readFeaturesFromShapeFile((File)input).features();

    // CA-Areas erzeugen und der ListProperty hinzufuegen
    ListPropertyWriteAccess wa = prop.getWriteAccess(this);
    for (int i=1; featIter.hasNext(); i++ ) {
      Feature f = featIter.next();
      try {
        // Area der Liste hinzufuegen
        wa.addValue( interpreteFeature(i,f) );
      } catch (Exception err) {
        System.err.println("Error in feature "+i+": "+err);
        System.err.println("    > Feature ignored!");
      }
    }
    wa.release();

    // Dateinamen in die Bezeichnung aufnehmen
    xuluObject.setDescription(xuluObject.getDescription().concat(" [").concat(((File)input).getName()).concat("]"));
    return xuluObject;

  }

  /**
   * Erzeugt eine Instanz vom {@link CAArea} und befuellt diese mit Daten aus
   * dem Feature.
   * @param featNo Nummer des Features (beginnend bei 0)
   * @param f      zu interpretierendes Feature
   */
  protected CAArea interpreteFeature(int featNo, Feature f) {
    return interpreteFeature(featNo,f,null);
  }

  /**
   * Befuellt eine Instanz vom {@link CAArea} mit Daten aus dem Feature.
   * @param featNo Nummer des Features (beginnend bei 0)
   * @param f      zu interpretierendes Feature
   * @param area   Area in die die Informationen hineingeschrieben werden (wenn
   *               {@code null}, wird eine neue {@link CAArea}-Instanz erzeugt)
   */
  protected CAArea interpreteFeature(int featNo, Feature f, CAArea area) {
    Object attr = null;
    String attrName = "";

    // Startpunkt in Default-Geometry
    attr = f.getDefaultGeometry();
    if ( attr != null && attr instanceof Point ) {
      Point p = (Point)attr;
      area.setStartLocation( new Point2D.Double(p.getX(),p.getY()) );
    } else {
      System.err.println("Feature "+featNo+": No point found in default geometry");
      System.err.println("    > Area location set to (0,0)");
      area.setStartLocation( new Point2D.Double(0,0) );
    }

    // Name der Area
    attrName = "Name";
    attr = f.getAttribute(attrName);
    if ( attr != null && attr instanceof String )
      area.setName( (String)attr );
    else {
      System.err.println("Feature "+featNo+": No string-attribute '"+attrName+"' found");
      System.err.println("    > Area name set to 'Area "+featNo+"'");
      area.setName("Area "+featNo);
    }

    return area;
  }
}
