/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.io.misc;

import java.io.File;
import java.awt.geom.Point2D;

import org.geotools.feature.Feature;
import org.geotools.feature.FeatureIterator;
import com.vividsolutions.jts.geom.Point;

import schmitzm.geotools.io.GeoImportUtil;

import schmitzm.data.property.ListProperty;
import schmitzm.data.property.ListPropertyWriteAccess;

import edu.bonn.xulu.io.ImportFactory;
import edu.bonn.xulu.io.InstantiationFactory;
import edu.bonn.xulu.io.AbstractFactory;
import edu.bonn.xulu.plugin.data.misc.CASettlementArea;
import edu.bonn.xulu.appl.XuluRegistry;
import edu.bonn.xulu.data.XuluObject;
import edu.bonn.xulu.data.DynamicXuluObject;
import edu.bonn.xulu.plugin.data.misc.CAArea;

/**
 * Diese Factory importiert Instanzen des Datentyps {@link CASettlementArea} aus einem
 * Shape-File. Dessen Features muessen folgende Attribute besitzen:
 * <table width="70%">
 *   <tr><td><b>Attribut</b></td>
 *       <td><b>Datentyp</b></td>
 *       <td><b>Beschreibung</b></td>
 *       </tr>
 *   <tr><td><i>Default-Geometry</i></td>
 *       <td>Punkt</td>
 *       <td>Lat/Lon-Koordinate der Area</td>
 *       </tr>
 *   <tr><td><code>Name</code></td>
 *       <td><code>String</code></td>
 *       <td>Name der Area</td>
 *       </tr>
 *   <tr><td><code>Inhabits</code></td>
 *       <td><code>double</code></td>
 *       <td>Einwohneranzahl zum Zeitpunkt 0 (kann auch direkt fuer Bedarf genutzt werden)</td>
 *       </tr>
 *   <tr><td><code>growthrate</code></td>
 *       <td><code>double</code></td>
 *       <td>Wachstumsrate der Einwohner (in %)</td>
 *       </tr>
 *   <tr><td><code>demand</code></td>
 *       <td><code>double</code></td>
 *       <td>Flaechen-Bedarf pro Einwohner (in qm)</td>
 *       </tr>
 * </table>
 * Aus den importierten {@link CASettlementArea}-Objekten wird anschliessend ein
 * {@link XuluObject} erstellt, das aus einer {@link ListProperty} besteht,
 * in der die Areas gespeichert werden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class CASettlementAreaListFactory_ShapeFile extends CAAreaListFactory_ShapeFile {
  /**
   * Erzeugt einen neue Factory.
   */
  public CASettlementAreaListFactory_ShapeFile() {
    // CAArea-Typ fuer den die List-Property im XuluObject erzeugt wird
    // neu belegen
    caAreaType = CASettlementArea.class;
  }

  ////////////////////////////////////////////////////////////////////////
  ///////////////     Implementierung von ImportFactory     //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Erzeugt eine Instanz vom {@link CASettlementArea} und befuellt diese mit Daten aus
   * dem Feature.
   * @param featNo Nummer des Features (beginnend bei 0)
   * @param f      zu interpretierendes Feature
   */
  protected CASettlementArea interpreteFeature(int featNo, Feature f) {
    return interpreteFeature(featNo,f,null);
  }

  /**
   * Erzeugt eine Instanz vom {@link CASettlementArea} und befuellt diese mit Daten aus
   * dem Feature.
   * @param featNo Nummer des Features (beginnend bei 0)
   * @param f      zu interpretierendes Feature
   * @param settArea Area in die die Informationen hineingeschrieben werden (wenn
   *                 {@code null}, wird eine neue {@link CASettlementArea}-Instanz erzeugt)
   * @throws IllegalArgumentException wenn {@code settArea} keine Instanz von
   *         {@link CASettlementArea} ist
   */
  protected CASettlementArea interpreteFeature(int featNo, Feature f, CAArea settArea) {
    if ( settArea == null )
      settArea = new CASettlementArea();
    if ( !(settArea instanceof CASettlementArea) )
      throw new IllegalArgumentException(getClass().getSimpleName()+".interpreteFeature(.) can only be applied on CASettlementArea-objects!");
    CASettlementArea area = (CASettlementArea)settArea;

    // Attribute von CAArea befuellen
    super.interpreteFeature(featNo,f,settArea);

    // Einwohner pro Zeitschritt
    String attrName = "Inhabits";
    Object attr     = f.getAttribute(attrName);
    if ( attr != null && attr instanceof Number )
      area.setInhabitants( ((Number)attr).doubleValue() );
    else {
      System.err.println("Feature "+featNo+": No double-attribute '"+attrName+"' found");
      System.err.println("    > Initial inhabitants set to 1.0");
      area.setInhabitants(1.0);
    }

    // Wachstumsrate der Einwohner (in %)
    attrName = "growthrate";
    attr = f.getAttribute(attrName);
    if ( attr != null && attr instanceof Number )
      area.setGrowthRate( ((Number)attr).doubleValue()/100.0 );
    else {
      System.err.println("Feature "+featNo+": No double-attribute '"+attrName+"' found");
      System.err.println("    > Growth rate set to 0.0");
      area.setGrowthRate(0.0);
    }

    // Bedarf pro Einwohner
    attrName = "demand";
    attr = f.getAttribute(attrName);
    if ( attr != null && attr instanceof Number ) {
      area.setDemandPerInhabitant( ((Number)attr).doubleValue() );
    } else {
      System.err.println("Feature "+featNo+": No number-attribute '"+attrName+"' found");
      System.err.println("    > Demand per habitant set to 1");
      area.setDemandPerInhabitant(1);
    }

    return area;
  }
}
