/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.io.misc;

import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.BufferedWriter;
import java.util.Vector;
import java.util.Collection;
import java.util.StringTokenizer;

import schmitzm.lang.LangUtil;
import schmitzm.data.property.*;
import schmitzm.temp.BaseTypeUtil;

import edu.bonn.xulu.io.AbstractFactory;
import edu.bonn.xulu.io.ImportFactory;
import edu.bonn.xulu.io.InstantiationFactory;
import edu.bonn.xulu.io.ExportFactory;
import edu.bonn.xulu.data.XuluObject;
import edu.bonn.xulu.data.DynamicXuluObject;
import edu.bonn.xulu.data.XuluDataException;
import edu.bonn.xulu.appl.XuluRegistry;

import org.apache.log4j.Logger;
import java.io.OutputStreamWriter;
import java.util.Arrays;

/**
 * Diese Factory importiert und exportiert Instanzen des Datentyps
 * {@link DynamicXuluObject} aus einem (in ein) einfaches ASCII-Format.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class DynamicXuluObjectFactory_BasicStructure extends AbstractFactory implements ImportFactory, ExportFactory {
  /** Logger fuer Debug-Ausgaben und Warnings. */
  protected final Logger LOGGER = LangUtil.createLogger(this);

  private static final String TAG_OBJNAME  = "XuluObjectName";
  private static final String TAG_PROPTYPE = "Type";
  private static final String TAG_DATATYPE = "DataType";
  private static final String TAG_VALUE    = "Value";

  // Objekt-Parameter
  private String objectName   = null;
  // Property-Parameter
  private String propName     = null;
  private Class  propClass    = null;
  private Class  propDataType = null;
  private Vector propValues   = new Vector<String>();
  private Vector params       = new Vector();


  ////////////////////////////////////////////////////////////////////////
  ///////////////     Implementierung von ImportFactory     //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Liefert den Datentyp, den die Factory erzeugt.
   * @return immer {@link DynamicXuluObject DynamicXuluObject.class}
   */
  public Class getImportType() {
    return DynamicXuluObject.class;
  }

  /**
   * Liefert den Datentyp, den die Factory als Import-Quelle benoetigt.
   * @return immer {@link InputStream InputStream.class}
   */
  public Class getImportSourceType() {
    return InputStream.class;
  }

  /**
   * Importiert eine Instanz von {@link DynamicXuluObject} aus einem InputStream.
   * Das Format des Streams ist der Klassendokumentation (siehe oben) zu
   * entnehmen.
   * @param input Eingabe-Quelle (muss ein {@link InputStream} sein!)
   * @param reg bietet Zugriff auf andere registrierte Factories, um eine
   *            Standard-Instanz von {@link DynamicXuluObject} zu erzeugen,
   *            in die das Objekt importiert wird
   * @exception UnsupportedOperationException falls als Eingabe-Quelle keine
   *            InputStream angegeben wird
   */
  public DynamicXuluObject importObject(Object input, XuluRegistry reg) throws Exception {
    // Factory benoetigt Input-Stream als Eingabe
    AbstractFactory.checkImportSourceObject(this,input,getImportSourceType(),true);
    // Factory benoetigt eine InstantiationFactory, die ein DynamicXuluObject
    // erzeugt
    InstantiationFactory instFac = AbstractFactory.getInstantiationFactoryFromRegistry(reg,getImportType(),true);

    // Instanz erzeugen und nach und nach mit Propertys fuellen
    DynamicXuluObject xuluObject = (DynamicXuluObject)instFac.newInstance(false);

    // Datei zeilenweise einlesen
    BufferedReader in   = new BufferedReader( new InputStreamReader((InputStream)input) );
    String  line        = null;
    boolean propSection = false; // Flag, ob generelle Paramater gelesen werden,
                                 // oder eine Propertry
    initPropertyVariables();
    for (int lineNo=1; in.ready() && (line=in.readLine().trim()) != null; lineNo++ ) {
      try {
          // Kommentare und Leerzeilen ignorieren
          if ( line.startsWith("//") || line.startsWith("#") || line.equals("") )
            continue;

          // Es beginnt eine neue Property-Definition
          if ( line.startsWith("[") ) {
            // wenn bereits eine Property eingelesen wurde, diese
            // erzeugen und dem Objekt hinzufuegen
            if ( propSection )
              try {
                xuluObject.addProperty(createProperty());
              } catch ( Exception err ) {
                System.err.println( getClass().getSimpleName().concat(" --> ").concat(err.toString()) );
                System.err.println( "     Construction of property '".concat(propName).concat("'") );
              }

            // Nachdem die Property erzeugt wurde, werden die Variablen
            // fuer die naechste Property initialisiert
            initPropertyVariables();

            // Namen einlesen
            propName = line.substring(1, line.indexOf("]"));
            if ( propName.equals("") )
              throw new XuluDataException("Property name must be specified!");

            // Flag setzen, dass (von nun an) Propertys gelesen werden
            propSection = true;
            continue;
          }

          Line l = new Line( (Object)line );
          if ( !propSection ) {
            // Zeile als globalen Parameter interpretieren
            if ( l.getTagName().equalsIgnoreCase(TAG_OBJNAME) )
              // Beschreibung des Xulu-Objekts
              xuluObject.setDescription( l.getFirstValue() );
            else
              // unbekannter Parameter
              throw new IOException( "Unknown Tag for XuluObject: ".concat(l.getTagName()) );
          } else {
            // Zeile als Property-Parameter interpretieren
            if ( l.getTagName().equalsIgnoreCase(TAG_PROPTYPE) ) {
              // Typ der Property
              propClass = Class.forName(l.getFirstValue());
              // nur Skalar-, List- oder Matrix-Property zulaessig
              if ( !ScalarProperty.class.equals(propClass) &&
                   !ListProperty.class.equals(propClass) &&
                   !MatrixProperty.class.equals(propClass) )
                throw new IOException( "Only ScalarProperty, ListProperty and MatrixProperty are allowed!" );
              // Alle Zeilen-Elemente NACH dem Klassennamen sind Parameter
              params = new Vector( l.values.subList(1,l.values.size()) );
            }
            else if ( l.getTagName().equalsIgnoreCase(TAG_DATATYPE) )
              // Daten-Typ der in der Property gespeichert wird
              propDataType = Class.forName( l.getFirstValue() );
            else if ( l.getTagName().equalsIgnoreCase(TAG_VALUE) )
              // Initiale Werte der Property
              propValues.addAll( l.getValues() );
            else
              // unbekannter Parameter
              throw new IOException( "Unknown Tag for XuluObject: ".concat(l.getTagName()) );
          }

      } catch ( Exception err ) {
        System.err.println( getClass().getSimpleName().concat(" --> ").concat(err.toString()) );
        System.err.println( "     Line ".concat(String.valueOf(lineNo)).concat(": ").concat(line) );
      }
    }

    // wenn bereits eine Property eingelesen wurde, diese
    // erzeugen und dem Objekt hinzufuegen
    if ( propSection )
      try {
        xuluObject.addProperty(createProperty());
      } catch ( Exception err ) {
        System.err.println( getClass().getSimpleName().concat(" --> ").concat(err.toString()) );
        System.err.println( "     Construction of property '".concat(propName).concat("'") );
      }
    // Eingabe-Stream schliessen
//    in.close();
    return xuluObject;
  }


  private void initPropertyVariables() {
    propName     = null;
    propClass    = null;
    propDataType = null;
    propValues.clear();
    params.clear();
  }

  private void checkProperty() throws Exception {
      // Property muss einen Namen haben
      if ( propName == null )
        throw new XuluDataException("Property name must be specified!");

      // Wurde kein Property-Typ angegeben, wird eine Liste angenommen,
      // wenn mehrere Werte angegeben wurden, sonst ein skalarer Wert
      if ( propClass == null ) {
        if ( propValues.size() > 1 )
          propClass = ListProperty.class;
        else
          propClass = ScalarProperty.class;
      }
      // nur Skalar-, List- oder Matrix-Property zulaessig
      if ( !ScalarProperty.class.equals(propClass) &&
           !ListProperty.class.equals(propClass) &&
           !MatrixProperty.class.equals(propClass) )
        throw new IOException( "Only ScalarProperty, ListProperty and MatrixProperty are allowed for Property '".concat(propName).concat("'") );

      // Property muss einen Datentyp haben oder (mind.) einen Wert
      if ( propDataType == null && propValues.isEmpty() )
        throw new XuluDataException("Data type to store in property must be specified or at least one value!");

      // Wurde kein Daten-Typ angegeben, wird der "optimale" gewaehlt
      if ( propDataType == null )
        propDataType = BaseTypeUtil.convertFromString((String)propValues.elementAt(0)).getClass();

      // Wenn fuer eine Matrix Werte angegeben werden, sollten alle Werte
      // angegeben werden
      if ( MatrixProperty.class.isAssignableFrom(propClass) ) {
        int    valueCountNeeded = 1;
        String sizeDesc = "";
        for (int i = 0; i < params.size(); i++) {
          int size = Integer.parseInt((String)params.elementAt(i));
          valueCountNeeded *= size;
          // Beschreibung zusammensetzen
          if ( i > 0 )
            sizeDesc += ",";
          sizeDesc.concat(String.valueOf(size));
        }
        if ( propValues.size() > valueCountNeeded )
          System.err.println( "Warning: To much values specified for matix(".concat(sizeDesc).concat(") property '").concat(propName).concat("'") );
        if ( propValues.size() < valueCountNeeded )
          System.err.println( "Warning: Less values specified for matix(".concat(sizeDesc).concat(") property '").concat(propName).concat("'") );
      }
  }

  private Property createProperty() throws Exception {
    // Pruefen, ob alle Angaben korrekt sind
    checkProperty();
    ValueProperty prop          = null;
    ValuePropertyType  propType = ValuePropertyType.createValuePropertyType(propDataType);

    // SKALARE PROPERTY
    if ( ScalarProperty.class.equals(propClass) ) {
      // Property erzeugen
      prop = new ScalarProperty(propName, propType);
      // Property mit initialem Wert belegen
      if ( propValues.size() > 0 )
        prop.getOneTimeWriteAccess().setValue( BaseTypeUtil.convertFromString((String)propValues.firstElement(), propDataType) );
    }

    // LIST PROPERTY
    if ( ListProperty.class.equals(propClass) ) {
      // Property erzeugen
      prop = new ListProperty(propName, propType);
      // Propertry mit initialen Werten belegen
      ListPropertyWriteAccess wa = (ListPropertyWriteAccess)prop.getWriteAccess(this);
      for (int i=0; i<propValues.size(); i++)
        wa.addValue( BaseTypeUtil.convertFromString((String)propValues.elementAt(i),propDataType) );
      wa.release();
    }

    // MATRIX PROPERTY
    if ( MatrixProperty.class.equals(propClass) ) {
      // Dimension der Matrix aus Parametern lesen
      int[] size        = new int[params.size()];
      int   maxElements = 1; // max. Anzahl an Elementen fuer die Matrix
      for ( int i=0; i<size.length; i++ ) {
        size[i] = Integer.parseInt( (String)params.elementAt(i) );
        maxElements *= size[i];
      }
      // Property erzeugen
      prop = new MatrixProperty(propName,size,propType);
      // Propertry mit initialen Werten belegen
      int[] c = new int[size.length]; // Koordinate zum Listenindex
      PropertyWriteAccess wa = prop.getWriteAccess(this);
      for (int i=0; i<propValues.size() && i<maxElements; i++, c[0]++) {
        // Koordinaten-Uebertrag checken
        for (int j=0; j<size.length && c[j]==size[j]; j++) {
          c[j] = 0;
          c[j+1]++;
        }
        wa.setValue( BaseTypeUtil.convertFromString((String)propValues.elementAt(i),propDataType), c );
      }
      wa.release();
    }

    return prop;
  }

  private class Line {
    private String tagName = "";
    private Vector values  = new Vector<String>();

    private void initFromString(String line) {
      StringTokenizer st = new StringTokenizer(line);
      // erstes Token (bis zum Gleichheitszeichen) = Name des Tags
      tagName = st.nextToken("=\n").trim();
      while ( st.hasMoreTokens() )
        values.add( st.nextToken("=;\n").trim() );
    }

    public Line(Object line) {
      initFromString((String)line);
    }

    public Line(String tagName) {
      this.tagName = tagName;
    }

    public String getTagName() {
      return this.tagName;
    }

    public int getValueCount() {
      return values.size();
    }

    public String getValue(int i) {
      return (String)values.elementAt(i);
    }

    public String getFirstValue() {
      return (String)values.elementAt(0);
    }

    public void addValue(Object o) {
      values.add(o);
    }

    public Collection getValues() {
      return values;
    }
  }

  ////////////////////////////////////////////////////////////////////////
  ///////////////     Implementierung von ExportFactory     //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Prueft, ob ein Objekt exportiert werden kann. Dies ist der
   * Fall, wenn es sich um ein {@link DynamicXuluObject} handelt.
   * @param obj zu pruefendes Objekt
   */
  public boolean isExportable(Object obj) {
    return obj!=null && obj instanceof DynamicXuluObject;
  }

  /**
   * Prueft, ob ein Objekttyp exportiert werden kann. Dies ist der
   * Fall, wenn es sich um {@link Properties} (oder eine Unterklasse)
   * handelt.
   * @param c zu pruefender Objekttyp
   */
  public boolean isExportable(Class c) {
    return c != null && Properties.class.isAssignableFrom(c);
  }

  /**
   * Liefert den Objekt-Typ, den die Factory als Ziel zum Exportieren
   * benoetigt.
   * @return immer {@link OutputStream OutputStream.class}
   */
  public Class getExportDestinationType() {
    return OutputStream.class;
  }

  /**
   * Exportiert eine Instanz von {@link Properties} in ein einfaches
   * ASCII-Format.
   * @param obj    zu exportiertendes Objekt (muss ein {@link DynamicXuluObject} sein
   * @param output Export-Ziel (muss ein {@link OutputStream} sein!)
   * @exception UnsupportedOperationException falls als Export-Ziel kein
   *            OutputStream angegeben wird
   * @exception IllegalArgumentException falls es sich bei <code>obj</code>
   *            nicht um ein {@link Properties}-Objekt handelt
   */
  public void exportObject(Object obj, Object output) throws IOException {
    AbstractFactory.checkExportDestinationObject(this,output,getExportDestinationType(),true);
    if ( !isExportable(obj) )
      throw new IllegalArgumentException(getClass().getName().concat(" can not export instances of ").concat(obj.getClass().getName()));

    Properties   propObj = (Properties)obj;
    BufferedWriter  out     = new BufferedWriter( new OutputStreamWriter((OutputStream)output) );
    // Name des Objekts schreiben
    if ( propObj instanceof XuluObject )
      out.write(TAG_OBJNAME + " = " + ((XuluObject)propObj).getDescription()+"\n\n");
    // Properties schreiben
    for (Property prop : propObj.getProperties()) {
      // Es koennen nur Instanzen von ValueProperty in die Datei
      // geschrieben werden
      if ( !(prop instanceof ValueProperty) ) {
        LOGGER.error("Property type not supported: "+prop.getClass().getSimpleName()+". Property ignored!");
        continue;
      }
      // Als Property-Wert koennen nur Stings, numerische oder boolesche
      // Werte in die Datei geschrieben werden
      Class dataType = prop.getType();
      if ( !BaseTypeUtil.isNumeric( dataType ) &&
           !BaseTypeUtil.isBoolean( dataType ) &&
           !BaseTypeUtil.isString( dataType ) &&
           !BaseTypeUtil.isCharacter( dataType ) ) {
        LOGGER.error("Property data type not supported: "+prop.getType().getSimpleName()+". Property ignored!");
        continue;
      }

      //**** Name ****
      out.write("["+prop.getName().trim()+"]\n");

      //**** Property-Typ ****
      out.write(TAG_PROPTYPE+" = "+prop.getClass().getName());
      // Parameter hinzufuegen
      if ( prop instanceof MatrixProperty)
        for (int dimSize : ((MatrixProperty)prop).getSize())
          out.write("; "+dimSize);
      // Zeile beenden
      out.write("\n");

      //**** Daten-Typ ****
      out.write(TAG_DATATYPE+" = "+prop.getType().getName()+"\n");

      //**** Werte ****
      if ( prop instanceof ScalarProperty ) {
        writeValueToOutput( (ScalarProperty)prop, out );
      } else if ( prop instanceof ListProperty) {
        writeValueToOutput( (ListProperty)prop, out );
      } else if ( prop instanceof MatrixProperty) {
        writeValueToOutput( (MatrixProperty)prop, out );
      } else
        LOGGER.error("ValueProperty type not supported: "+prop.getType().getSimpleName()+". Values could not be stored!");

      // Leerzeile nach jeder Property
      out.write("\n");
    }

    out.flush();
    out.close();
  }

  /**
   * Exportiert den Wert-Block einer {@link ScalarProperty}.
   * @param prop zu exportierende {@link ScalarProperty}
   * @param out  Ausgabe
   */
  private void writeValueToOutput(ScalarProperty prop, BufferedWriter out) throws IOException {
    out.write(TAG_VALUE+" =\t"+prop.getOneTimeReadAccess().getValue()+"\n" );
  }

  /**
   * Exportiert den Wert-Block einer {@link ListProperty}.
   * @param prop zu exportierende {@link ListProperty}
   * @param out  Ausgabe
   */
  private void writeValueToOutput(ListProperty prop, BufferedWriter out) throws IOException {
    ListPropertyReadAccess RA_prop = prop.getReadAccess(this);
    for (int i=0; i<RA_prop.getCount(); i++) {
      if ( i==0 )
        out.write(TAG_VALUE+" =");
      out.write("\t"+RA_prop.getValue(i)+(i<RA_prop.getCount()-1 ? ";" : "\n") );
    }
    RA_prop.release();
  }

  /**
   * Exportiert den Wert-Block einer {@link MatrixProperty}.
   * @param prop zu exportierende {@link MatrixProperty}
   * @param out  Ausgabe
   */
  private void writeValueToOutput(MatrixProperty prop, BufferedWriter out) throws IOException {
    PropertyReadAccess RA_prop = prop.getReadAccess(this);
    int[] dimSize   = prop.getSize();  // Matrix-Groesse in jeder Dimension
    int[] coord     = new int[dimSize.length]; // Var. fuer eine Koordinate
    int   lineCount = 1; // Gesamt-Anzahl an VALUE-Zeilen
    for (int d=1; d<dimSize.length; d++)
      lineCount *= dimSize[d];
    // Zeilen erzeugen
    for (int l=0; l<lineCount; l++) {
      // Koordinate erhoehen (ab dem 2. Durchlauf, da initial 'coord'
      // bereits auf (0,0,...,0) steht)
      for (int d = 1; l > 0 && d < dimSize.length; d++)
        if (++coord[d] < dimSize[d])
          break; // Koord. in Dimensionen > d nicht beeinflusst
        else {
          coord[d] = 0; // Koord. in Dimension d initialisieren und in d+1 erhoehen
          out.write("\n");
        }
      // Wenn Matrix mehr als 2 Dimensionen hat, vor jedem
      // Matrix-Block die darin enthaltenen Koordinaten als Kommentar
      // darstellen
      if ( dimSize.length > 2 && coord[1] == 0 )
        out.write("# "  + Arrays.toString( LangUtil.cloneArray(coord,0,0)) +
                  " - " + Arrays.toString( LangUtil.cloneArray(coord,dimSize[0]-1,dimSize[1]-1) ) + "\n");

      // Zeile erzeugen (= Werte in Dimension 0)
      out.write(TAG_VALUE+" =");
      for (coord[0]=0; coord[0]<dimSize[0]; coord[0]++)
        out.write( "\t"+RA_prop.getValue(coord)+(coord[0]<dimSize[0]-1 ? ";" : "\n") );
    }
    RA_prop.release();
  }

}
