/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.model.ca.agric;

import java.awt.Point;
import java.awt.geom.Point2D;
import java.util.Vector;

import schmitzm.data.WritableGrid;
import schmitzm.data.AbstractWritableGrid;

import edu.bonn.xulu.plugin.data.misc.CAArea;
import edu.bonn.xulu.plugin.data.misc.CASettlementArea;
import edu.bonn.xulu.plugin.model.ca.MultiCellularAutomaton;
import edu.bonn.xulu.model.ValuePropertyResource;
import schmitzm.data.property.ScalarProperty;
import schmitzm.data.property.ListProperty;
import edu.bonn.xulu.data.XuluDataException;
import schmitzm.data.property.PropertyReadAccess;

// nur fuer Doku
import edu.bonn.xulu.model.ModelContentManager;

/**
 * Dieser zellulare Automat modelliert die Ausbreitung von Feldflaechen (Ackerland)
 * im Impetus-Untersuchungsgebiet.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ImpetusCellularAutomaton extends MultiCellularAutomaton {
  // Zugriffsrechte auf Ressourcen
  private PropertyReadAccess nutz_RA = null;
  private PropertyReadAccess brach_RA = null;
  private PropertyReadAccess dhm_RA = null;
  private PropertyReadAccess dsm_RA  = null;
  private PropertyReadAccess lucHist_WA  = null;

  /** Raster mit Gelaende-Hoehe (DHM).*/
  protected WritableGrid dhm = null;
  /** Raster mit Gelaende-Neigung (DSM).*/
  protected WritableGrid dsm = null;
  /** Raster fuer LUC-Historie.*/
  protected WritableGrid lucHist = null;

  /**
   * Erzeugt einen neuen CA.
   */
  public ImpetusCellularAutomaton() {
    super( new CAContentManager() );
  }

  /**
   * Initialisiert das Modell.
   */
  public void performModelInit() {
    super.performModelInit();
    //########## Lese/Schreib-Rechte auf Ressourcen ##########
    // Leserecht auf Nutzungsdauer
    nutz_RA = ((ScalarProperty)contManager.getResource(CAContentManager.NUTZSTEPS).getData()).getReadAccess(this);
    // Leserecht auf Nutzungsdauer
    brach_RA = ((ScalarProperty)contManager.getResource(CAContentManager.BRACHSTEPS).getData()).getReadAccess(this);
    // Leserecht auf Hoehen-Modell
    dhm_RA = ((ScalarProperty)contManager.getResource(CAContentManager.DHM).getData()).getReadAccess(this);
    // Leserecht auf Neigungs-Modell
    dsm_RA = ((ScalarProperty)contManager.getResource(CAContentManager.DSM).getData()).getReadAccess(this);
    // Schreibrecht auf LUC-Historie
    lucHist_WA = ((ScalarProperty)contManager.getResource(CAContentManager.LUCHISTORY).getData()).getReadAccess(this);

    //########## Ressourcen in lokalen Variablen verlinken ##########
    dhm = (WritableGrid)dhm_RA.getValue();
    dsm = (WritableGrid)dsm_RA.getValue();
    lucHist = (WritableGrid)lucHist_WA.getValue();

    //########## Pro-Kopf-Bedarf ist (noch) nicht im Shape-File ##########
//    double dpi = ((ScalarProperty)contManager.getResource(CAContentManager.QMDPI).getData()).getOneTimeReadAccess().getValueAsDouble();
    // Bedarf pro Einwohner im Shape-File ist in ha angegeben
    for (int a=0; a<areas.getCount(); a++)
      ((CASettlementArea)areas.getValue(a)).setDemandToQMFactor(10000);

    //########## Nutzungszeiten initialisieren ##########
    for (int y=lucHist.getMinY(); y<lucHist.getMinY()+lucHist.getHeight(); y++)
      for (int x=lucHist.getMinX(); x<lucHist.getMinX()+lucHist.getWidth(); x++)
        lucHist.setRasterSample(0,x,y);
  }

  /**
   * Zerstoert das Modell. Alle gehaltenen Zugriffsrechte auf Ressourcen
   * werden wieder freigegeben.
   */
  public void performModelDispose() {
    super.performModelDispose();
    releaseAccess(nutz_RA);
    releaseAccess(brach_RA);
    releaseAccess(dhm_RA);
    releaseAccess(dsm_RA);
    releaseAccess(lucHist_WA);
  }

  /**
   * Modelliert genau einen Zeitschritt.
   * <ol>
   *   <li>Zu Beginn wird fuer jede Zellen des Untersuchungsgebiets geprueft,
   *       ob die maximale Nutzungsdauer erreicht ist. Ggf. faellt die Zelle
   *       brach ({@link #checkBrache()}).</li>
   *   <li>Es wird ein CA-Schritt modelliert ({@link MultiCellularAutomaton#performModelStep(int)}</li>
   *   <li>Die Nutzungsdauer (LUC-Historie) wird fuer alle Zellen des
   *       Untersuchungsgebiets um 1 erhoeht ({@link #incLUCHistory()})</li>
   * </ol>
   * @param stepNo Nummer des Zeitschritts
   */
  public void performModelStep(int stepNo) {
    // Deadlock-Flag von allen Gebieten entfernt, da durch das Ablaufen der Brachezeiten
    // u.U. ein Weiterlaufen moeglich wird
    for (int a=0; a<areaCount; a++)
      ((CASettlementArea)areas.getValue(a)).setDeadlock(false);

    // Flaechen brach fallen lassen.
    checkBrache();

    // Status der Gebiete fuer diesesn Zeitschritt ausgeben
    statusOut.print("Step "+stepNo+":\t");
    for (int a=0; a<areaCount; a++) {
      CASettlementArea area = (CASettlementArea)areas.getValue(a);
      double demandHA = area.getDemandPerInhabitant()*area.getInhabitants(stepNo-1);
      double coverHA  = area.getCurrentCover()/area.getDemandToQMFactor();
      statusOut.print("("+a+") "+coverHA+"/"+demandHA+"ha \t");
    }
    statusOut.println();

    // CA ausfuehren
    super.performModelStep(stepNo);

    // Nutzungszeiten erhoehen
    incLUCHistory();
  }

  /**
   * Setzt die LUC-Historie fuer eine Zelle auf 0 zurueck.
   * @param p       Raster-Koordinaten der umgewandelten Zelle
   * @param area    Gebiet dem die Zelle zugeordnet wurde
   * @param newLUC  LUC-Typ, der der Zelle zugeordnet wurde
   * @param prevLUC LUC-Typ, den die Zelle zuvor hatte
   */
  public void cellChanged(Point p, CASettlementArea area, int newLUC, int prevLUC) {
    lucHist.setRasterSample(0, p.x, p.y);
  }

  /**
   * Bewertet eine Zelle des Untersuchungsgebiets hinsichtlich des LUC-Typs
   * 'Ackerland'. Folgende LUC-Typen stellen ein Ausschluss-Kriterium dar:
   * <ul>
   *   <li>0: Forest Dense</li>
   *   <li>5: Inselberg mit Vegetation</li>
   *   <li>8: Fluss</li>
   *   <li>9: Flussbett</li>
   *   <li>10: Dichte Siedlung</li>
   *   <li>11: Lockere Siedlung</li>
   *   <li>13: Inselberg</li>
   * </ul>
   * Daneben entstehen keine Acker-Flaechen in Zellen, die weiter als 5km
   * vom Startpunkt der Siedlung (Area) entfernt sind oder deren Hangneigung
   * 4� ueberschreitet.<br>
   * Ein weiteres Ausschuluss-Kriterium stellen <i>Brache</i>-Zellen (14) dar,
   * deren Brachezeit ({@link CAContentManager#BRACHSTEPS}) noch nicht
   * erreicht wurde.<br>
   * Alle anderen Zellen werden bewertet mit<br>
   * <center><code>
   *   value = 2 - distToStart/5000
   * </code></center>
   * wobei <code>distToStart</code> die Distanz (in m) zum Siedlungskern
   * darstellt. Dies bewirkt, dass Zellen nahe des Siedlungskern generell
   * bevorzugt werden (da 5000m Entfernung ein generelles Ausschlusskriterium
   * darstellt, ist gewaehrleistet, dass <code>value</code> immer >= 1 ist!).
   * @param p  zu bewertende Zelle des Untersuchungsgebiets.
   * @param luc LUC-Typ hinsichtlich dessen die Bewertung durchgefuehrt wird
   *            (wird nicht genutzt! Bewertung immer hinsichtlich Ackerland)
   * @param area CA-Gebiet fuer das die Bewertung durchgefuehrt wird
   * @param areaNo Nummer des Gebiets (wird nicht genutzt!)
   */
  public double getPreferenceValue(Point p, int luc, CAArea area, int areaNo) {
    // Ausschuluss-Kriterium: aktuelle LUC
    int currLUC  = outLUC.getRasterSampleAsInt(p.x,p.y);
    int currHist = lucHist.getRasterSampleAsInt(p.x,p.y);

    switch (currLUC) {
      case 0: // Forest Dense
      case 5: // Inselberg mit Vegetation
      case 8: // Fluss
      case 9: // Flussbett
      case 10: // Dichte Siedlung
      case 11: // Lockere Siedlung
      case 13: // Inselberg
               return 0;
      case 14: // Brache (kommt nur in Frage, wenn Brachezeit bereits erreicht)
               if ( currHist < brach_RA.getValueAsInt() )
                 return 0;
    }

    // Ausschuss-Kriterium: > 4.0� Gelaendeneigung
    float degree = dsm.getRasterSampleAsFloat(p.x,p.y);
    if ( degree > 4.0 )
      return 0;

    // Ausschluss-Kriterium: 5km Distanz vom Siedlungskern
    // -> Mittelpunkt der betrachteten Zelle betrachten
    double realX = baseLUC.convertRasterToReal(p.x,0);
    double realY = baseLUC.convertRasterToReal(p.y,1);
    double distToStart = Math.abs(area.getStartLocation().distance(realX,realY));
//    int startX = baseLUC.convertRealToRaster(area.getStartLocation().getX(),0);
//    int startY = baseLUC.convertRealToRaster(area.getStartLocation().getY(),1);
//    double distToStart = Math.sqrt(Math.pow(startX-p.x,2)+Math.pow(startY-p.y,2))*baseLUC.getCellWidth();
    if ( distToStart > 5000 )
      return 0.0;

    // alle anderen Zellen geeignet
    return 2 - distToStart/5000;
  }


  /**
   * Prueft fuer jede Zelle des Untersuchungsgebiets, ob die maximale
   * Nutzungsdauer erreicht ist. Ist dies der Fall faellt die Zelle brach,
   * d.h. sie wird in den Typ 14 (Brache) umgewandelt.
   */
  protected void checkBrache() {
    Vector<Point> brachCells = new Vector<Point>();

    //###########################################################
    // Folgender Abschnitt fuegt nur Brach-Zellen in die Border-
    // Mengen ein, die (NACH dem Brachfallen) noch in der
    // Nachbarschaft einer Area liegen.
    //###########################################################
    // Alle Zellen identifizieren, die brach fallen muessen. Diese werden in
    // einer Liste zwischen gespeichert, um anschliessend wieder in die
    // Border-Menge der Areas eingefuegt zu werden
    // Bemerke: Dies kann nicht on-the-fly geschehen, da die Nachbarzellen
    //          der brach-gefallenen Zellen auch noch brach fallen koennen!
    for (int y=lucHist.getMinY(); y<lucHist.getMinY()+lucHist.getHeight(); y++)
      for (int x=lucHist.getMinX(); x<lucHist.getMinX()+lucHist.getWidth(); x++) {
        checkBreakingCommands();
        int currLUC  = outLUC.getRasterSampleAsInt(x,y);
        int currArea = areaMapping.getRasterSampleAsByte(x,y);
        int steps    = lucHist.getRasterSampleAsInt(x,y);
        // wenn Zelle noch nicht brach ist und Nutzungsdauer erreichen, faellt
        // sie brach (Flaeche des "Besitzer"-Gebiets wird reduziert)
        if ( currLUC == luccType && // evt. hier noch den anderen Feld-Typ 14 beruecksichtigen
             steps >= nutz_RA.getValueAsInt() ) {
          outLUC.setRasterSample(14,x,y);
          lucHist.setRasterSample(0,x,y);
          areaMapping.setRasterSample(-1,x,y);
          ((CASettlementArea)areas.getValue(currArea)).decCurrentCover(cellSize);
          // alle brach gefallenen Zellen merken
          brachCells.add( new Point(x,y) );
        }
      }

    // Brach-gefallene Zellen in die Border-Menge der angrenzenden Areas einfuegen
    for (int i=0; i<brachCells.size(); i++) {
      Point p = brachCells.elementAt(i);
      // Umgebung der Zelle betrachten
      // > Wenn eine Area angrenzt, wird die Zelle deren Border-Menge hinzugefuegt
      for (int n = 0; n<NEIGHBORHOOD_AREA.length; n++) {
        checkBreakingCommands();
        Point np = new Point(p.x+NEIGHBORHOOD_AREA[n][0],p.y+NEIGHBORHOOD_AREA[n][1]);
        int areaIdx = areaMapping.getRasterSampleAsByte(np.x,np.y);
        if ( areaIdx >= 0 ) {
          CASettlementArea area = (CASettlementArea)areas.getValue(areaIdx);
          if ( !area.getBorder().contains(p) )
            area.getBorder().add(p);
        }
      }
    }

//    //###########################################################
//    // Folgender Abschnitt fuegt ALLE Brach-Zellen in die Border-
//    // Mengen ein.
//    //###########################################################
//    // Alle Zellen identifizieren, die brach fallen muessen. Diese werden in
//    // einer Liste zwischen gespeichert, um anschliessend die Area-Zuordnung
//    // zu loeschen.
//    // Bemerke: Dies kann nicht on-the-fly geschehen, da die bestehende
//    //          Nachbarschaftsbeziehung fuer die Border-Zuordnung noch benoetigt
//    //          wird.
//    for (int y=lucHist.getMinY(); y<lucHist.getMinY()+lucHist.getHeight(); y++)
//      for (int x=lucHist.getMinX(); x<lucHist.getMinX()+lucHist.getWidth(); x++) {
//        int currLUC  = outLUC.getRasterSampleAsInt(x,y);
//        int currArea = areaMapping.getRasterSampleAsByte(x,y);
//        int steps    = lucHist.getRasterSampleAsInt(x,y);
//        // wenn Zelle noch nicht brach ist und Nutzungsdauer erreichen, faellt
//        // sie brach (Flaeche des "Besitzer"-Gebiets wird reduziert)
//        if ( currLUC == luccType && // evt. hier noch den anderen Feld-Typ 14 beruecksichtigen
//             steps >= nutz_RA.getValueAsInt() ) {
//          Point p = new Point(x,y);
//          brachCells.add( p ); // alle brach gefallenen Zellen merken
//          outLUC.setRasterSample(14,x,y);
//          lucHist.setRasterSample(0,x,y);
//          ((CASettlementArea)areas.getValue(currArea)).decCurrentCover(cellSize);
//          // Zelle in Border der angrenzenden Areas einfuegen
//          for (int n = 0; n<NEIGHBORHOOD_AREA.length; n++) {
//            checkBreakingCommands();
//            Point np = new Point(p.x+NEIGHBORHOOD_AREA[n][0],p.y+NEIGHBORHOOD_AREA[n][1]);
//            int areaIdx = areaMapping.getRasterSampleAsByte(np.x,np.y);
//            if ( areaIdx >= 0 ) {
//              CASettlementArea area = (CASettlementArea)areas.getValue(areaIdx);
//              if ( !area.getBorder().contains(p) )
//                area.getBorder().add(p);
//            }
//          }
//      }
//    }
//
//    // Area-Zuordnung der brach-gefallene Zellen loeschen
//    for (int i=0; i<brachCells.size(); i++) {
//      Point p = brachCells.elementAt(i);
//      areaMapping.setRasterSample(-1,p.x,p.y);
//    }

  }

  /**
   * Erhoeht die Nutzungsdauer aller Zellen im Untersuchungsgebiet um 1.
   */
  protected void incLUCHistory() {
    for (int y=lucHist.getMinY(); y<lucHist.getMinY()+lucHist.getHeight(); y++)
      for (int x=lucHist.getMinX(); x<lucHist.getMinX()+lucHist.getWidth(); x++) {
        checkBreakingCommands();
        lucHist.setRasterSample( lucHist.getRasterSampleAsInt(x,y)+1 ,x,y);
      }
  }


  /**
   * Diese Klasse definiert einen {@link ModelContentManager} fuer den
   * {@linkplain ImpetusCellularAutomaton Impetus-Zellularen-Automat}. Neben den
   * durch die Oberklasse spezifizierten Ressourcen, definiert dieser Content-Manager
   * folgende:
   * <ol>
   *  <li><b>Nutzungsdauer (<code>int</code>)</b><br>
   *      Bestimmt nach wie vielen Modellschritten der Nutzung eine modellierte
   *      Zelle wieder brach faellt.</li>
   *  <li><b>Brachezeit (<code>int</code>)</b><br>
   *      Bestimmt nach wie vielen Modellschritten eine brach-gefallene Zelle
   *      wieder genutzt werden kann.</li>
   *  <li><b>Hoehen-Modell ({@link WritableGrid})</b><br>
   *      Hoehen-Modell (DHM) des Untersuchungsgebiets.</li>
   *  <li><b>Neigungs-Modell ({@link WritableGrid})</b><br>
   *      Gelaendeneigung des Untersuchungsgebiets.</li>
   *  <li><b>LUC-Historie ({@link WritableGrid})</b><br>
   *      Speichert (temporaer) die Dauer, wie lange sich eine jede Zelle
   *      unter der aktuellen LUC befindet.</li>
   * </ol>
   * Fuer Sub-Klassen von {@link MultiCellularAutomaton} kann dieser Content-Manager
   * um weitere Ressourcen erweitert werden.
   */
  public static class CAContentManager extends MultiCellularAutomaton.CAContentManager {
//    /** ID der <code>int</code>-Ressource, die den Bedarf pro Einwohner
//     *  (in qm) definiert. */
//    public static final int    QMDPI                 = BASE_RESCOUNT + 0;
//    /** Beschreibung fuer die {@link #QMDPI}-Ressource. */
//    public static final String QMDPI_DESC            = "qm-Demand per Inhabitant (int)";

    /** ID der {@link Long}-Ressource, in der die Nutzungsdauer des modellierten
     *  LUC-Typs hinterlegt ist (nach der angegebenen Anzahl an Modellschritten
     *  faellt das Land brach).
     */
    public static final int NUTZSTEPS = BASE_RESCOUNT + 0;
    /** Beschreibung fuer die {@link #NUTZSTEPS}-Ressource. */
    public static final String NUTZSTEPS_DESC = "Nutzungsdauer (int)";

    /** ID der {@link Long}-Ressource, in der die Brachezeit des modellierten
     *  LUC-Typs hinterlegt ist. Nachdem eine Zelle brach gefallen ist, kann sie
     *  erst nach dieser Anzahl an Modellschritten wiederverwendet werden. */
    public static final int BRACHSTEPS = BASE_RESCOUNT + 1;
    /** Beschreibung fuer die {@link #BRACHSTEPS}-Ressource. */
    public static final String BRACHSTEPS_DESC = "Brachezeit (int)";

      /** ID der {@link WritableGrid}-Ressource fuer das Raster, in dem
     *  das digitale Gelaendemodell hinterlegt ist. */
    public static final int    DHM = BASE_RESCOUNT + 2;
    /** Beschreibung fuer die {@link #DHM}-Ressource. */
    public static final String DHM_DESC = "Height raster (WritableGrid)";

    /** ID der {@link WritableGrid}-Ressource fuer das Raster, in dem
     *  die Gelaendeneigung hinterlegt ist. */
    public static final int    DSM = BASE_RESCOUNT + 3;
    /** Beschreibung fuer die {@link #DSM}-Ressource. */
    public static final String DSM_DESC = "Slope raster (WritableGrid)";

    /** ID der {@link WritableGrid}-Ressource fuer das Raster, in dem
     *  die aktuelle Nutzungsdauer der Zellen hinterlegt wird. */
    public static final int    LUCHISTORY = BASE_RESCOUNT + 4;
    /** Beschreibung fuer die {@link #LUCHISTORY}-Ressource. */
    public static final String LUCHISTORY_DESC              = "Temp: LUC history (WritableGrid)";

    public CAContentManager() {
      super(5);
//      // Bedarf pro Einwohner
//      resource[QMDPI] = new ValuePropertyResource(ValuePropertyResource.CATEGORY_INPUT,QMDPI_DESC,ScalarProperty.class,Double.class,false);

      // Area-Konfiguration der Oberklasse verfeinern (muss CASettlementArea enthalten!)
      resource[AREACONF] = new ValuePropertyResource(ValuePropertyResource.CATEGORY_INPUT | ValuePropertyResource.CATEGORY_OUTPUT,AREACONF_DESC,ListProperty.class,CASettlementArea.class,false);
      // Nutzungsdauer (in Modellschritten)
      resource[NUTZSTEPS] = new ValuePropertyResource(ValuePropertyResource.CATEGORY_INPUT,NUTZSTEPS_DESC,ScalarProperty.class,Integer.class,false);
      // Brachezeit (in Modellschritten)
      resource[BRACHSTEPS] = new ValuePropertyResource(ValuePropertyResource.CATEGORY_INPUT,BRACHSTEPS_DESC,ScalarProperty.class,Integer.class,false);
      // Gelaendemodell
      resource[DHM] = new ValuePropertyResource(ValuePropertyResource.CATEGORY_INPUT,DHM_DESC,ScalarProperty.class,WritableGrid.class,false);
      // Gelaendeneigung
      resource[DSM] = new ValuePropertyResource(ValuePropertyResource.CATEGORY_INPUT,DSM_DESC,ScalarProperty.class,WritableGrid.class,false);
      // LUC-History
      resource[LUCHISTORY] = new ValuePropertyResource(ValuePropertyResource.CATEGORY_TEMP,LUCHISTORY_DESC,ScalarProperty.class,WritableGrid.class,false);
    }

    /**
     * Prueft die Ressourcen-Zuordnung auf Korrektheit. Neben den Checks der
     * Oberklasse werden folgende Sachverhalte geprueft:
     * <ol>
     *   <li>Das Raster mit dem Gelaendemodell muss das gleiche Format haben, wie das Ausgangsszenario.</li>
     *   <li>Das Raster mit der Gelaendeneigung muss das gleiche Format haben, wie das Ausgangsszenario.</li>
     * </ol>
     */
    public void checkAndError() {
      super.checkAndError();

//      // Einwohner-Bedarf muss > 0 sein
//      double dpi = ((ScalarProperty)resource[QMDPI].getData()).getOneTimeReadAccess().getValueAsDouble();
//      if ( dpi <= 0 )
//        throw new XuluDataException("Demand per inhabitant must be greater zero!");

      // Nutzungsdauer muss > 0 sein
      long nutz = ((ScalarProperty)resource[NUTZSTEPS].getData()).getOneTimeReadAccess().getValueAsInt();
      if ( nutz <= 0 )
        throw new XuluDataException("'Nutzungsdauer' must be greater zero!");

      // Nutzungsdauer muss > 0 sein
      long brach = ((ScalarProperty)resource[BRACHSTEPS].getData()).getOneTimeReadAccess().getValueAsInt();
      if ( brach <= 0 )
        throw new XuluDataException("'Brachezeit' must be greater zero!");

      // Raster fuer das Gelaendemodell muss das gleiche Format haben, wie
      // das Ausgangs-Szenarion
      AbstractWritableGrid.compareStructure(
                      (WritableGrid)((ScalarProperty)resource[DHM].getData()).getOneTimeReadAccess().getValue(),
                      DHM_DESC,
                      (WritableGrid)((ScalarProperty)resource[BASELUC].getData()).getOneTimeReadAccess().getValue(),
                      BASELUC_DESC,
                      false );

      // Raster fuer die Gelaendeneigung muss das gleiche Format haben, wie
      // das Ausgangs-Szenarion
      AbstractWritableGrid.compareStructure(
                      (WritableGrid)((ScalarProperty)resource[DSM].getData()).getOneTimeReadAccess().getValue(),
                      DSM_DESC,
                      (WritableGrid)((ScalarProperty)resource[BASELUC].getData()).getOneTimeReadAccess().getValue(),
                      BASELUC_DESC,
                      false );

      // Raster fuer die LUC-History muss das gleiche Format haben, wie
      // das Ausgangs-Szenarion
      AbstractWritableGrid.compareStructure(
                      (WritableGrid)((ScalarProperty)resource[LUCHISTORY].getData()).getOneTimeReadAccess().getValue(),
                      LUCHISTORY_DESC,
                      (WritableGrid)((ScalarProperty)resource[BASELUC].getData()).getOneTimeReadAccess().getValue(),
                      BASELUC_DESC,
                      false );
    }
  }
}
