/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.model.ca.fire;

import java.util.Map;
import java.util.Vector;
import java.util.HashMap;
import java.util.Random;
import java.util.GregorianCalendar;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.awt.Component;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Dimension;
import java.awt.GridBagLayout;
import java.awt.BorderLayout;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import javax.swing.JLabel;
import javax.swing.JSlider;
import javax.swing.JTable;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;

import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureIterator;
import org.geotools.feature.Feature;

import com.vividsolutions.jts.geom.Envelope;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.MultiPoint;

import edu.bonn.xulu.data.DynamicXuluObject;
import edu.bonn.xulu.model.AbstractStepModel;
import edu.bonn.xulu.plugin.data.feature.SingleFeatureCollection;

import schmitzm.data.WritableGrid;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.PropertyWriteAccess;
import schmitzm.data.property.ListPropertyReadAccess;
import schmitzm.data.property.ScalarProperty;
import schmitzm.data.property.ListProperty;
import schmitzm.data.property.MatrixProperty;
import schmitzm.lang.LangUtil;
import schmitzm.swing.CaptionsChangeable;
import schmitzm.swing.JPanel;
import schmitzm.swing.InputCompass;
import schmitzm.swing.SliderSpinnerPanel;

import edu.bonn.xulu.model.event.ModelIterationStepStartedEvent;
import edu.bonn.xulu.model.event.ModelIterationStepFinishedEvent;

import org.apache.log4j.Logger;
import schmitzm.lang.ResourceProvider;
import edu.bonn.xulu.XuluModellingPlatform;
import java.util.Locale;

/**
 * Diese Klasse stellt eine Implementierung des Modells {@linkplain ImpetusFireCA}
 * dar. Dieses benoetigt 7 Ressourcen, welche durch den
 * {@linkplain ImpetusFireCAContentManager} spezifiziert werden:
 * <ol>
 * <li><b>Allgemeine Modell-Parameter ({@code modelParam}):</b>
 *     Dynamisches Xulu-Objekt mit folgenden Propertys:
 *     <ul>
 *     <li><b>Modeltime</b> (int): Anzahl der zu modellierenden Stunden.</li>
 *     <li><b>LUC_Burning</b> (int): Nummer des LUC-Typs, mit dem die brennenden Zellen
 *            im Ausgabe-Raster markiert werden.</li>
 *     <li><b>LUC_BurnedOut</b> (int): Nummer des LUC-Typs, mit dem die ausgebrannten Zellen
 *            im Ausgabe-Raster markiert werden.</li>
 *     <li><b>BurnSuitability</b> (double-Liste): Brenn-Wahrscheinlichkeit pro LUC-Typ (globale
 *            Brenneignung).</li>
 *     <li><b>BurnTime</b> (double-Liste): Zeit in Stunden pro LUC-Typ, die eine Zelle brennt,
 *            bevor sie als "burned out" markiert wird.</li>
 *     <li><b>WindDirection</b> (double): Windrichtung in Grad (Nord = 0Grad, West = 90Grad)</li>
 *     <li><b>WindPower</b> (double): Windstaerke in m/s. Zur Zeit wird die
 *            Windstaerke als Ausbreitungsgeschwindigkeit des Feuers verwendet,
 *            was nicht realistisch ist! Eine Umrechnung von Wind- in Ausbreitungsgeschwindigkeit
 *            ist noch nicht implementiert.</li>
 *     <li><b>WindDirectionDev</b> (double): Max. Abweichung der Windrichtung in Grad.</li>
 *     <li><b>WindPowerDev</b> (double): Max. Abweichung der Windstaerke in m/s (Boeigkeit).</li>
 *     <li><b>WindPowerMaxBack</b> (double): Windstaerke in m/s ab der kein Rueckschreiten des
 *            Feuers (entgegen der Windrichtung) mehr moeglich ist.</li>
 *     </ul>
 *     </li>
 * <li><b>Brandherde ({@code startPoints}):</b>
 *     FeatureCollection (z.B. aus Shape-File) mit ein oder mehreren georeferenzierten
 *     Punkten, die als Startpunkte fuer die Feuer-Ausbreitung verwendet werden.
 *     </li>
 * <li><b>LUC Classification ({@code baseLUC}):</b>
 *     Raster mit LUC-Klassifikation (Ausgangsszenario).
 *     </li>
 * <li><b>Brenn-Eignung ({@code burnFit}, optional):</b>
 *     Raster mit Brenn-Wahrscheinlichkeiten pro Zelle (lokale Brenneignung)..
 *     </li>
 * <li><b>Ausgabe-Raster fuer modellierte LUC ({@code outLUC}):</b>
 *     Raster, in dem die modellierte LUC ("burning" und "burned out") hinterlegt
 *     wird.
 *     </li>
 * <li><b>Ausgabe-Raster fuer verbrannte Biomasse ({@code outBurnMass}, optional):</b>
 *     Raster, in dem die verbrannte Biomasse pro Zelle hinterlegt wird.
 *     Wird zur Zeit noch nicht genutzt!
 *     </li>
 * <li><b>Temp-Raster fuer Brenn-Zeit ({@code burnTime}):</b>
 *     Raster, in dem waehrend des Modellablaufs die Zeiten hinterlegt werden,
 *     die eine Zelle brennt, um die Zellen "irgendwann" als ausgebrannt zu
 *     markieren.
 *     </li>
 * </ol>
 * @see ImpetusFireCAContentManager
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ImpetusFireCA extends AbstractStepModel {
  /** {@link ResourceProvider}, der die Uebersetzungen fuer die GUI
   *  bereitstellt: {@code locales.XuluModel_ImpetusFireCA} */
  protected static ResourceProvider RESOURCE = XuluModellingPlatform.registerResourceProvider("XuluModel_ImpetusFireCA",Locale.ENGLISH);

  /** Logger fuer Debug-Ausgaben und Warnings. */
  public static final Logger LOGGER = LangUtil.createLogger(ImpetusFireCA.class);

  public static final String PARAM_PROP_MODELTIME  = "Modeltime";
  public static final String PARAM_PROP_BURNLUC    = "LUC_Burning";
  public static final String PARAM_PROP_BURNOUTLUC = "LUC_BurnedOut";
  public static final String PARAM_PROP_BURNSUIT   = "BurnSuitability";
  public static final String PARAM_PROP_BURNTIME   = "BurnTime";
//  public static final String PARAM_PROP_WINDMATRIX = "WindMatrix";
  public static final String PARAM_PROP_WINDDIRECTION     = "WindDirection";
  public static final String PARAM_PROP_WINDPOWER         = "WindPower";
  public static final String PARAM_PROP_WINDDIRECTIONDEV  = "WindDirectionDev";
  public static final String PARAM_PROP_WINDPOWERDEV      = "WindPowerDev";
  public static final String PARAM_PROP_WINDPOWERMAXBW    = "WindPowerMaxBack";
  public static final String PARAM_PROP_WINDMATRIX_PCONST = "WindMatrix_PowerConst";
  public static final String PARAM_PROP_WINDMATRIX_PFACT  = "WindMatrix_PowerFact";

  /** Speichert den ContentManager fuer das Modell.
   *  @see ImpetusFireCAContentManager */
  protected ImpetusFireCAContentManager contManager;

  //********** Lese/Schreibrechte, die fuer den gesamten ***********
  //********** Modellanlauf gehalten werden muessen      ***********
  private PropertyReadAccess   RA_startPoints = null;  // Brandherde
  private PropertyReadAccess   RA_baseLUC = null;  // LUC Classification
  private PropertyReadAccess   RA_burnSuitability = null;  // Brenn-Eignung
  private PropertyWriteAccess  WA_outLUC = null;  // Ausgabe-Raster für modellierte LUC
  private PropertyWriteAccess  WA_outBurnMass = null;  // Ausgabe-Raster für verbrannte Biomasse
  private PropertyWriteAccess  WA_burnTime = null;  // Temp-Raster für Brenn-Zeit

  //**************** Variablen mit denen gearbeitet wird *******************
  protected  DynamicXuluObject  modelParam = null; // Allgemeine Modell-Parameter
  protected  FeatureCollection  startPoints = null; // Bandherde
  protected  WritableGrid       baseLUC = null; // LUC Classification
  protected  WritableGrid       burnSuitability = null; // Brenn-Eignung
  protected  WritableGrid       outLUC = null; // Ausgabe-Raster für modellierte LUC
  protected  WritableGrid       outBurnMass = null; // Ausgabe-Raster für verbrannte Biomasse
  protected  WritableGrid       burnTime = null; // Temp-Raster für Brenn-Zeit

  protected  int                areaMinX = 0;
  protected  int                areaMaxX = 0;
  protected  int                areaMinY = 0;
  protected  int                areaMaxY = 0;
  protected  double             cellSize = 0;
  protected  int                lucCount = 0;
  protected  int                burningLUC = 0;
  protected  int                burnedOutLUC = 0;
  protected  double[]           lucBurningSuitability = null;
  protected  double[]           lucBurningTime = null;
  protected  double[]           windMatrix = new double[8]; // Wind-Matrix als Liste (ohne Zentrum)
  protected  double             windMatrix_powerConst = 0;
  protected  double             windMatrix_powerFact  = 0;
  protected  double             windDir = 0; // Windrichtung
  protected  double             windDirDev = 0; // max. Abw. Windrichtung
  protected  double             windPow = 0; // Windstaerke
  protected  double             windPowDev = 0; // max. Abw. Windstaerke
  protected  double             windPowMaxBW = 0; // Windstaerke ab der kein Rueckschreiten mehr moeglich ist
  protected  int                iterCount = 0;
  protected  double             iterSeconds = 0;

  protected double              MAX_lucBurningSuitability = 0;
  protected double              MAX_windMatrix = 0;
  protected double              MAX_burningSuitability = 0;

  // Nachbar-Zellen relativ zur Zentrum-Zelle
  private  static final int[][] NEIGHBOR = new int[][] {
      {-1,-1}, {0,-1}, {1,-1},
      {-1, 0},         {1, 0},
      {-1, 1}, {0, 1}, {1, 1}
  };

  protected  Vector<int[]>         burningCells = new Vector<int[]>();
  protected  Vector<int[]>         uselessBurningCells = new Vector<int[]>();
  protected  HashMap<int[],Object> newBurningCells = new HashMap<int[],Object>();

  // Modell-Spezifische GUI
  protected  Component modelGUI = null;

  /**
   * Erzeugt eine neue Instanz des Modells.
   */
  public ImpetusFireCA() {
    super( new ImpetusFireCAContentManager() );
    contManager = (ImpetusFireCAContentManager)super.contManager;
    modelResourceProvider = RESOURCE;

    // Iterations-Ereignisse der Event-Liste hinzufuegen
    events.add(ModelIterationStepStartedEvent.class);
    events.add(ModelIterationStepFinishedEvent.class);
    // GUI erzeugen
    modelGUI = new GUI();
    modelGUI.setEnabled(false);
  }

  ////////////////////////////////////////////////////////////////////////
  ///////////////   Implementierung von AbstractXuluModel   //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Initialisiert das Model, indem die Ressourcen aus dem ContentManager
   * geladen werden und mit Lese/Schreib-Rechten versehen werden. Zudem
   * werden Referenzen auf die Ressourcen in lokalen Variablen
   * gespeichert, um waehrend des Modellablaufs effizienter darauf zugreifen
   * zu koennen.
   */
  public void performModelInit() {
    if ( contManager == null )
      return;
    // Zugriffsrechte aus Ressourcen/Propertys holen
    RA_startPoints = null;
    if ( contManager.getResource(1).getData() != null )
      RA_startPoints = ((ScalarProperty)((SingleFeatureCollection)contManager.getResource(1).getData()).getProperty("FEATURES")).getReadAccess(this);
    RA_baseLUC = null;
    if ( contManager.getResource(2).getData() != null )
      RA_baseLUC = ((ScalarProperty)contManager.getResource(2).getData()).getReadAccess(this);
    RA_burnSuitability = null;
    if ( contManager.getResource(3).getData() != null )
      RA_burnSuitability = ((ScalarProperty)contManager.getResource(3).getData()).getReadAccess(this);
    WA_outLUC = null;
    if ( contManager.getResource(4).getData() != null )
      WA_outLUC = ((ScalarProperty)contManager.getResource(4).getData()).getWriteAccess(this);
    WA_outBurnMass = null;
    if ( contManager.getResource(5).getData() != null )
      WA_outBurnMass = ((ScalarProperty)contManager.getResource(5).getData()).getWriteAccess(this);
    WA_burnTime = null;
    if ( contManager.getResource(6).getData() != null )
      WA_burnTime = ((ScalarProperty)contManager.getResource(6).getData()).getWriteAccess(this);

    // Variablen belegen mit denen gearbeitet wird
    modelParam      = (DynamicXuluObject)contManager.getResource(0).getData();
    startPoints     = (FeatureCollection)RA_startPoints.getValue();
    baseLUC         = (WritableGrid)RA_baseLUC.getValue();
    areaMinX        = baseLUC.getMinX();
    areaMaxX        = baseLUC.getMinX() + baseLUC.getWidth() - 1;
    areaMinY        = baseLUC.getMinY();
    areaMaxY        = baseLUC.getMinY() + baseLUC.getHeight() - 1;
    cellSize        = baseLUC.getCellWidth();
    burnSuitability = RA_burnSuitability != null ? (WritableGrid)RA_burnSuitability.getValue() : null;
    outLUC          = (WritableGrid)WA_outLUC.getValue();
    outBurnMass     = WA_outBurnMass != null ? (WritableGrid)WA_outBurnMass.getValue() : null;
    burnTime        = (WritableGrid)WA_burnTime.getValue();
    // Der Effizienz halber Parameter aus dyn. Xulu-Objekt heraus holen
    burningLUC      = ((ScalarProperty)modelParam.getProperty(PARAM_PROP_BURNLUC)).getOneTimeReadAccess().getValueAsInt();
    burnedOutLUC    = ((ScalarProperty)modelParam.getProperty(PARAM_PROP_BURNOUTLUC)).getOneTimeReadAccess().getValueAsInt();
    lucCount        = ((ListProperty)modelParam.getProperty(PARAM_PROP_BURNSUIT)).getOneTimeReadAccess().getCount();
    lucBurningSuitability = new double[lucCount];
    lucBurningTime        = new double[lucCount];
    ListPropertyReadAccess RA_lucBurningSuitability = ((ListProperty)modelParam.getProperty(PARAM_PROP_BURNSUIT)).getReadAccess(this);
    ListPropertyReadAccess RA_lucBurningTime        = ((ListProperty)modelParam.getProperty(PARAM_PROP_BURNTIME)).getReadAccess(this);
    for (int luc=0; luc<lucCount; luc++) {
      lucBurningSuitability[luc] = RA_lucBurningSuitability.getValueAsDouble(luc);
      lucBurningTime[luc]        = RA_lucBurningTime.getValueAsDouble(luc);
      // maximale Brenn-Eignung ermitteln (fuer Normierung)
      if ( luc==0 || lucBurningSuitability[luc] > MAX_lucBurningSuitability )
        MAX_lucBurningSuitability = lucBurningSuitability[luc];
    }
    RA_lucBurningSuitability.release();
    RA_lucBurningTime.release();
    windDir      = ((ScalarProperty)modelParam.getProperty(PARAM_PROP_WINDDIRECTION)).getOneTimeReadAccess().getValueAsDouble();
    windDirDev   = ((ScalarProperty)modelParam.getProperty(PARAM_PROP_WINDDIRECTIONDEV)).getOneTimeReadAccess().getValueAsDouble();
    windPow      = ((ScalarProperty)modelParam.getProperty(PARAM_PROP_WINDPOWER)).getOneTimeReadAccess().getValueAsDouble();
    windPowDev   = ((ScalarProperty)modelParam.getProperty(PARAM_PROP_WINDPOWERDEV)).getOneTimeReadAccess().getValueAsDouble();
    windPowMaxBW = ((ScalarProperty)modelParam.getProperty(PARAM_PROP_WINDPOWERMAXBW)).getOneTimeReadAccess().getValueAsDouble();
    // Faktoren zur Gewichtung der Windgeschwindigkeit bei
    // der Berechnung der Wind-Matrix
    if ( modelParam.containsProperty(PARAM_PROP_WINDMATRIX_PCONST) )
    	windMatrix_powerConst = ((ScalarProperty)modelParam.getProperty(PARAM_PROP_WINDMATRIX_PCONST)).getOneTimeReadAccess().getValueAsDouble();
    else
    	windMatrix_powerConst = 0.0;
    if ( modelParam.containsProperty(PARAM_PROP_WINDMATRIX_PFACT) )
    	windMatrix_powerFact  = ((ScalarProperty)modelParam.getProperty(PARAM_PROP_WINDMATRIX_PFACT)).getOneTimeReadAccess().getValueAsDouble();
    else
    	windMatrix_powerConst = 1.0;
    // Wind-Matrix berechnen
    calculateWindMatrix(windDir,windPow,windDirDev,windPowDev,windPowMaxBW,windMatrix_powerConst,windMatrix_powerFact);
//    PropertyReadAccess RA_windMatrix = ((MatrixProperty)modelParam.getProperty(PARAM_PROP_WINDMATRIX)).getReadAccess(this);
//    int offset = 0;
//    for (int y=0; y<3; y++)
//      for (int x=0; x<3; x++) {
//        // Zelle in der Mitte ausnehmen
//        if ( y==1 && x==1 ) {
//          offset = 1; // um Position im Array zu reduzieren
//          continue;
//        }
//        double windValue = RA_windMatrix.getValueAsDouble(x,y);
//        windMatrix[y*3+x-offset] = windValue;
//        // maximalen Wert der Wind-Matrix ermitteln (fuer Normierung)
//        if ( (y==0 && x==0) || windValue > MAX_windMatrix )
//          MAX_windMatrix = windValue;
//      }
//    RA_windMatrix.release();

    // Startzellen initialisieren
    initBurningCells( startPoints );

    // Ausgabe-LUC mit NoData initialisieren
    // Ausgaberaster fuer BioMasse initialisieren
    // Raster fuer Brennzeit initialisieren
    for (int y=areaMinY; y<=areaMaxY; y++)
      for (int x=areaMinX; x<=areaMaxX; x++) {
        outLUC.setRasterSample( Float.NaN, x, y );
        if ( outBurnMass != null )
          outBurnMass.setRasterSample( 0, x, y );
        burnTime.setRasterSample( 0, x, y );
        // maximalen Wert des Eignungsrasters ermitteln (fuer Normierung)
        if ( burnSuitability != null ) {
          double value = burnSuitability.getRasterSampleAsDouble(x,y);
          if ( (y == areaMinY && x == areaMinX) || value > MAX_burningSuitability )
            MAX_burningSuitability = value;
        }
      }

    // WICHTIG:
    // An dieser Stelle die Anzahl an Modellschritten (aus einer
    // Ressource) zuweisen, die das Modell laufen soll!
    // Sonst laeuft das Modell nicht an!!
    this.stepCount = ((ScalarProperty)modelParam.getProperty(PARAM_PROP_MODELTIME)).getOneTimeReadAccess().getValueAsInt();

    // 1 Modellschritt = 1 Stunde
    // -> Anzahl an Iterationen pro Schritt richtet sich nach der Windgeschwindigkeit
    // -> ein CA-Schritt (Iteration) modelliert die Ausbreitung um eine Zelle,
    //    also S Meter, was bei Windstaerke P (in m/s) genau S/P Sekunden
    //    (= S/(3600*P) Stunden) entspricht
    // -> Fuer eine Stunde (ein Modellschritt) muss 3600*P/S mal iteriert werden
    this.iterCount = (int)Math.round( 3600 * windPow / baseLUC.getCellWidth() );
    this.iterCount = Math.max(1,this.iterCount);
    // Anzahl an Sekunden pro Iteration
    this.iterSeconds = baseLUC.getCellWidth() / windPow;
    statusOut.println( RESOURCE.getString("Model.ModelTime.Step") );
    statusOut.println( RESOURCE.getString("Model.ModelTime.Iter",iterSeconds) );
    statusOut.println("-----------------------");

    // GUI initialisieren
    if ( modelGUI != null ) {
      ((GUI)modelGUI).setWindDirection(windDir);
      ((GUI)modelGUI).setWindDirectionDeviation(windDirDev);
      ((GUI)modelGUI).setWindPower(windPow);
      ((GUI)modelGUI).setWindPowerDeviation(windPowDev);
      modelGUI.setEnabled(true);
    }
  }

  /**
   * Gibt die Lese/Schreibrechte auf das Ein- und Ausgaberaster wieder frei.
   * Interne Ressourcen brauchen keine aus dem Speicher entfernt werden.
   */
  public void performModelDispose() {
    // Ressourcen wieder freigeben
    releaseAccess(RA_startPoints);
    releaseAccess(RA_baseLUC);
    releaseAccess(RA_burnSuitability);
    releaseAccess(WA_outLUC);
    releaseAccess(WA_outBurnMass);
    releaseAccess(WA_burnTime);
  }

  /**
   * Fuehrt einen Schritt des Modellablaufs durch.
   * @param stepNo zu modellierender Schritt (beginnend bei 1!)
   */
  public void performModelStep(int stepNo) {
    Random RANDOM = new Random();

    GregorianCalendar calendar = new GregorianCalendar();
    DateFormat        format   = new SimpleDateFormat("HH:mm:ss.SSS");
    for (int iterNo=1; iterNo<=iterCount; iterNo++) {
      this.fireModelEvent( new ModelIterationStepStartedEvent(this,stepNo,iterNo) );

      // Wind-Matrix neu berechnen
      calculateWindMatrix(windDir,windPow,windDirDev,windPowDev,windPowMaxBW,windMatrix_powerConst,windMatrix_powerFact);
      ((GUI)modelGUI).repaintWindMatrx();

      long millis = Math.round( ((stepNo-2)*3600+iterNo*iterSeconds) * 1000 );
      calendar.setTimeInMillis( millis );
      statusOut.println("Iteration "+stepNo+"."+iterNo+": "+format.format(calendar.getTime()));
//      LOGGER.debug( "Burning cells: "+burningCells.size() + "   Useless burning: " + uselessBurningCells.size() );

      // neu brennende Zellen muessen zunaecht in eine eigene Menge
      // aufgenommen werden
      newBurningCells.clear();

      // Alle brennenden Zellen betrachten
      int[] bCell = null;
      int[] nbCell = null;
      int x   = 0; // X-Koordinate einer brennenden Zelle
      int y   = 0; // Y-Koordinate einer brennenden Zelle
      int nbX = 0; // X-Koordinate einer Nachbarzelle
      int nbY = 0; // Y-Koordinate einer Nachbarzelle
      float luc = 0;
      for (int bCellIdx=0; bCellIdx<burningCells.size(); bCellIdx++) {
        bCell = burningCells.elementAt(bCellIdx);

        x   = bCell[0];
        y   = bCell[1];

        // Alle Nachbarzellen betrachten
        boolean useless = true;
        for (int nbCellIdx=0; nbCellIdx<NEIGHBOR.length; nbCellIdx++) {
          checkBreakingCommands();
          nbX = x+NEIGHBOR[nbCellIdx][0];
          nbY = y+NEIGHBOR[nbCellIdx][1];
          nbCell = new int[] {nbX,nbY};
          // Zelle nicht verarbeiten, wenn
          // - Zelle ausserhalb des Untersuchungsgebiets liegt
          // - Zelle keinen Wert hat
          // - Zelle bereits brennt
          // - Zelle bereits fuer diesen Schritt als brennend klassifiziert
          if ( nbX < areaMinX || nbY < areaMinY || nbX > areaMaxX || nbY > areaMaxY ||
               newBurningCells.containsKey(nbCell) ||
               Float.isNaN(baseLUC.getRasterSampleAsFloat(nbX,nbY)) )
            continue;
          luc = outLUC.getRasterSampleAsFloat(nbX,nbY);
          if ( luc == burningLUC || luc == burnedOutLUC )
            continue;
          // Brenn-Wahrscheinlichkeit erreichen
          double burnProb = calculateBurnProbability(nbX,nbY,windMatrix[nbCellIdx]);
          double random   = RANDOM.nextDouble(); // Zufalls-Wert aus [0..1[
          if ( random < burnProb ) {
            // Zelle als brennend markieren und Brennzeit initialisieren
            outLUC.setRasterSample(burningLUC,nbX,nbY);
            burnTime.setRasterSample(0,nbX,nbY);
            // Zelle in Liste der neuen brennenden Zellen aufnehmen
            newBurningCells.put( nbCell, null );
          } else {
            // Laufzeit-Optimierung:
            // Wenn Nachbar potentiell brennbar ist, ist die betrachtete Zelle
            // nicht "nutzlos", sondern muss im naechsten Schritt nochmals
            // betrachtet werden
            if ( burnProb > 0 )
              useless = false;
          }
        }

        // Laufzeit-Optimierung:
        // Wenn Zelle nicht mehr zur Feuerausbreitung beitragen kann, braucht
        // sie im naechsten Schritt nicht mehr betrachtet werden
        if ( useless ) {
          burningCells.remove( bCell );
          uselessBurningCells.add( bCell );
        }
      }

      // Verbrannte Biomasse und Brennzeit erhoehen
      performBurning(burningCells);
      performBurning(uselessBurningCells);

      // neue brennende Zellen in Liste aufnehmen
      for ( int[] cell : newBurningCells.keySet() ) {
        checkBreakingCommands();
        burningCells.add(cell);
      }
      this.fireModelEvent( new ModelIterationStepFinishedEvent(this,stepNo,iterNo) );
    }

  }

  /**
   * Erhoeht die Menge der verbrannten Biomasse einer Zelle, sowie die
   * Brennzeit. Ist die maximale Brennzeit ueberschritten wird die Zelle
   * aus der Menge der Brennenden Zellen entfernt.
   * @param bCells Zellen
   */
  protected void performBurning(Vector<int[]> bCells) {
    int[] bCell = null;
    int x = 0;
    int y = 0;
    for (int bCellIdx=0;  bCellIdx<bCells.size(); bCellIdx++) {
      checkBreakingCommands();
      bCell = bCells.elementAt(bCellIdx);
      x     = bCell[0];
      y     = bCell[1];

      // Brennzeit der betrachteten Zelle erhoehen
      float bTime = burnTime.getRasterSampleAsFloat(x, y);
      bTime += iterSeconds;
      burnTime.setRasterSample(bTime, x, y);

      // TODO: Menge der verbrannten Biomasse erhoehen
      if (outBurnMass != null) {
        float bMass = outBurnMass.getRasterSampleAsFloat(x, y);
//        bMass += ...;
        outBurnMass.setRasterSample(bMass, x, y);
      }

      // Wenn Zelle ausgebrannt, als solche markieren und aus Rand
      // entfernen
      int luc = baseLUC.getRasterSampleAsInt(x, y);
      if (bTime > lucBurningTime[ (int) luc] * 3600) {
        outLUC.setRasterSample(burnedOutLUC, x, y);
        bCells.remove(bCellIdx--);
      }
    }
  }

  /**
   * Initialisiert die Liste der Brandherd-Zellen ueber die georeferenzierten
   * Punkte einer {@link FeatureCollection}.
   * @param burnSource FeatureCollection von {@link Point} und {@link MultiPoint}
   */
  protected void initBurningCells(FeatureCollection burnSource) {
    // brennende Zellen
    if ( burningCells == null )
      burningCells = new Vector<int[]>();
    else
      burningCells.clear();
    // brennende Zellen, die keinen Einfluss mehr auf Ihre Nachbarzellen
    // haben (also nicht mehr fuer die Expansion betrachtet werden muessen)
    if ( uselessBurningCells == null )
      uselessBurningCells = new Vector<int[]>();
    else
      uselessBurningCells.clear();

    FeatureIterator fi = burnSource.features();
    for (; fi.hasNext();) {
      Geometry g = fi.next().getDefaultGeometry();
      if ( g == null )
        continue;
      if ( g instanceof Point ) {
        Point p = (Point)g;
        burningCells.add( new int[] {
          baseLUC.convertRealToRaster(p.getX(),0),
          baseLUC.convertRealToRaster(p.getY(),1),
        } );
      }
      if ( g instanceof MultiPoint ) {
        MultiPoint mp = (MultiPoint)g;
        for (int i=0; i<mp.getNumPoints(); i++)
          burningCells.add( new int[] {
            baseLUC.convertRealToRaster(((Point)mp.getGeometryN(i)).getX(),0),
            baseLUC.convertRealToRaster(((Point)mp.getGeometryN(i)).getY(),1),
          } );
      }
    }

    for ( int[] p : burningCells )
      LOGGER.debug("initial burning cell: "+p[0]+" / "+p[1]);
  }

  /**
   * Berechnet die Wahrscheinlichkeit, dass eine Zelle brennt.
   * @param x    horizontale Raster-Koordinate
   * @param y    vertikale Raster-Koordinate
   * @param wind Wert der Wind-Matrix (fuer Windrichtung)
   * @return einen Wert zwischen 0 und 1
   */
  protected double calculateBurnProbability(int x, int y, double wind) {
    int luc = baseLUC.getRasterSampleAsByte(x,y);

    double burnProb = 1;
    // lokale Eignung einrechnen (pro Zelle)
    if ( burnSuitability != null )
      burnProb *= burnSuitability.getRasterSampleAsDouble(x,y) / MAX_burningSuitability;
    // globale Eignung einrechnen (pro LUC-Typ)
    burnProb *= lucBurningSuitability[luc] / MAX_lucBurningSuitability;
    // Wind einrechnen
    burnProb *= wind;

    return burnProb;
  }

  ////////////////////////////////////////////////////////////////////////
  /////////////////////   Berechnung der Wind-Matrix   ///////////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Berechnet die Windmatrix fuer gegebene Windrichtung und -staerke.
   * @param direction Windrichtung in Grad (Nord = 0, West = 90)
   * @param power Windstaerke in m/s
   * @param directionDev Max. Abweichung der Windrichtung in Grad
   * @param powerDev Max. Abweichung der Windstaerke in m/s
   * @param powerMaxBW Windstaerke ab der kein Rueckschreiten des Feuers
   *                   (entgegen der Windrichtung) mehr moeglich ist
   */
  private void calculateWindMatrix(double direction, double power, double directionDev, double powerDev, double powerMaxBW, double powerConst, double powerFact) {
  	// Die Windgeschwindigkeit fuer die Berechnung der Matrix
  	// wie folgt gewichten: power_for_matrix = powerConst + powerFact * power
  	power = powerConst + powerFact * power;
  	
  	// Zufaellige Abweichung in Windrichtung und -staerke einrechnen
    Random random = new Random();
    direction += directionDev * (random.nextDouble()*2 - 1);
    power     += powerDev     * (random.nextDouble()*2 - 1);

    // Mittelpunkt der 3x3-Matrix (>>> Ursprung links-oben)
    double mX = 1.5 * cellSize;
    double mY = 1.5 * cellSize;

    // Referenz-Quadrat, welches fuer die Ueberlageung mit den "Matrix"-Zellen
    // verwendet wird
    //== Durchmesser ==
    double windEnvW = calculateReferenceSize(cellSize, power, powerMaxBW);
    //== Flaeche ==
    double windArea = windEnvW * windEnvW;
    //== Mittelpunkt ==
    double windEnvMX = mX - cellSize*Math.sin(direction/180 * Math.PI);
    double windEnvMY = mY - cellSize*Math.cos(direction/180 * Math.PI);
    //== Referenz-Quadrat ==
    Envelope windEnv = new Envelope(
      windEnvMX-windEnvW/2, // X1
      windEnvMX+windEnvW/2, // X2
      windEnvMY-windEnvW/2, // Y1
      windEnvMY+windEnvW/2  // Y2
    );
    // Ueberlagerung jeder "Matrix"-Zelle mit dem Referenz-Quadrat berechnen
    Envelope cellEnv = new Envelope();
    double   totalIntersectionArea = 0.0;
    for (int i=0; i<windMatrix.length; i++) {
      initEnvelopeForMatrixCell(cellEnv, i<4 ? i : i+1,cellSize);
      windMatrix[i] = calculateIntersection(cellEnv,windEnv);
      totalIntersectionArea += windMatrix[i];
    }

    // Eintrag in Wind-Matrix entspricht dem Anteil am Referenz-Quadrat
    for (int i=0; i<windMatrix.length; i++)
      windMatrix[i] = Math.round(windMatrix[i]/totalIntersectionArea*10000)/10000.0;

  }

  /**
   * Berechnet den Durchmesser des Referenzkreises (oder Quadrats), welches
   * fuer die Ueberlageung mit den "Matrix"-Zellen verwendet wird.
   * @param cellSize Groesse (Breite) einer Zelle
   * @param power    Windstaerke
   * @param powerMaxBW Windstaerke ab der kein Rueckschreiten des Feuers
   *                   (entgegen der Windrichtung) mehr moeglich ist
   */
  private double calculateReferenceSize(double cellSize, double power, double powerMaxBW) {
    // ????
    return 3*cellSize * powerMaxBW / power;
  }

  /**
   * Bestimmt das Rechteck einer Matrix-Zelle (3x3).
   * @param env      Rechteck, das mit den Eckpunkten der Matrix-Zelle initialisiert wird
   * @param i        Index der Matrix-Zelle (0..8)
   * @param cellSize Groesse (Breite) einer Raster-Zelle
   */
  private void initEnvelopeForMatrixCell(Envelope env, int i, double cellSize) {
    int col = i % 3;
    int row = i / 3;

    double x1 = col*cellSize;
    double x2 = x1 + cellSize;
    double y1 = row*cellSize;
    double y2 = y1 + cellSize;

    env.init(x1,x2,y1,y2);
  }

  /**
   * Berechnet die Groesse der Schnittflaeche zwischen zwei Rechtecken.
   * @param env1 Rechteck 1
   * @param env2 Rechteck 2
   */
  private double calculateIntersection(Envelope env1, Envelope env2) {
    Envelope intersectionEnv  = env1.intersection( env2 );
    return intersectionEnv.getWidth() * intersectionEnv.getHeight();
  }

  ////////////////////////////////////////////////////////////////////////
  ////////////////   Implementierung der modellspez. GUI   ///////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Liefert die modellspezifische GUI.
   */
  public Component getModelSpecificGUI() {
    return this.modelGUI;
  }

  private class GUI extends JPanel implements CaptionsChangeable {
    /** GUI-Komponente fuer die Windrichtung (Kompass). */
    protected InputCompass windDirection = null;
    /** Slider fuer die Abweichung der Windrichtung. */
    protected SliderSpinnerPanel windDirectionDev = null;
    /** Slider fuer die Windstaerke. */
    protected SliderSpinnerPanel windPower = null;
    /** Slider fuer die Abweichung der Windstaerke. */
    protected SliderSpinnerPanel windPowerDev = null;
    /** Beschriftung der Wind-Matrix. */
    protected JLabel windMatrixLabel = null;
    /** Tabelle fuer Windmatrix. */
    protected JTable windMatrixTable = null;
    /** Grafische Darstellung der Windmatrix. */
    protected JPanel windMatrixGraphic = null;
    /**
     * Erzeugt einen neue GUI fuer das Feuer-Modell.
     */
    public GUI() {
      super();
      setLayout( new GridBagLayout() );

      // Kompass fuer Windrichtung
      windDirection = new InputCompass(0);
      windDirection.setPreferredSize( new Dimension(250,150) );
      // Slider fuer Abweichung der Windrichtung
      windDirectionDev = new SliderSpinnerPanel(JSlider.HORIZONTAL,0,360,0.1,10," ");
      initSlider( windDirectionDev.getSlider(), 45 );
      // Slider fuer Windstaeke
      windPower = new SliderSpinnerPanel(JSlider.HORIZONTAL,1,40,0.1,10," ");
      initSlider( windPower.getSlider(), 5 );
      // Slider fuer Boeigkeit
      windPowerDev = new SliderSpinnerPanel(JSlider.HORIZONTAL,0,100,0.1,2," ");
      initSlider( windPowerDev.getSlider(), 20 );
      this.add( windDirection, new GridBagConstraints(
          0, 0, 1, 5, 0.2, 0.5, GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(5, 5, 5, 0), 0, 0) );
      this.add( windPower, new GridBagConstraints(
          1, 0, 1, 1, 1.0, 0.0, GridBagConstraints.NORTH, GridBagConstraints.HORIZONTAL, new Insets(5, 5, 5, 5), 0, 0) );
      this.add( windDirectionDev, new GridBagConstraints(
          1, 1, 1, 1, 1.0, 0.0, GridBagConstraints.NORTH, GridBagConstraints.HORIZONTAL, new Insets(5, 5, 5, 5), 0, 0) );
      this.add( windPowerDev, new GridBagConstraints(
          1, 2, 1, 1, 1.0, 0.0, GridBagConstraints.NORTH, GridBagConstraints.HORIZONTAL, new Insets(5, 5, 5, 5), 0, 0) );

      // Anzeige der Wind-Matrix
      windMatrixTable = new JTable( new AbstractTableModel() {
        public Object getValueAt(int row, int col) {
          // Zentrum der Matrix enthaelt keinen Wert
          if ( windMatrix == null || (row == 1 && col == 1) )
            return "";
          int idx = row*3+col;
          if ( idx > 4 )
            idx--;
          return windMatrix[idx];
        }
        public int getRowCount() {
          return 3;
        }
        public int getColumnCount() {
          return 3;
        }
        public boolean isCellEditable(int row, int col) {
          return false;
        }
      });
      windMatrixTable.setCellSelectionEnabled(false);
      // Eintraege horizontal zentrieren
      windMatrixTable.setDefaultRenderer(Object.class, new DefaultTableCellRenderer() {
        public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
          JLabel comp = (JLabel)super.getTableCellRendererComponent(table,value,isSelected,hasFocus,row,column);
          comp.setHorizontalAlignment(JLabel.CENTER);
          return comp;
        }
      });
      windMatrixLabel = new JLabel(" ");
      windMatrixLabel.setFont( windDirectionDev.getHeaderLabel().getFont() );
      this.add( windMatrixLabel, new GridBagConstraints(
         1, 3, 1, 1, 0.0, 0.0, GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE, new Insets(5, 5, 0, 5), 0, 0) );
      this.add( windMatrixTable, new GridBagConstraints(
         1, 4, 1, 1, 0.0, 0.0, GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH, new Insets(0, 5, 5, 5), 0, 0) );

      ChangeListener l = new ChangeListener() {
        public void stateChanged(ChangeEvent e) {
          windDir = windDirection.getValue();
          windPow = windPower.getValue();
          windDirDev = windDirectionDev.getValue();
          windPowDev = windPowerDev.getValue();
          // Waehrend des Modellablaufs werden nur die Modell-Variablen neu gesetzt,
          // die Neuberechnung  (und Anzeige) der Matrix erfolgt zu Beginn der
          // naechsten Iteration
          if ( !isRunning() ) {
            // Wind-Matrix ohne Abweichung berechnen und anzeigen
            calculateWindMatrix(
              windDir,
              windPow,
              0,
              0,
              windPowMaxBW,
              windMatrix_powerConst,
              windMatrix_powerFact
            );
            repaintWindMatrx();
            if ( windMatrixGraphic != null )
              windMatrixGraphic.repaint();
          }
        }
      };
      windDirection.addChangeListener(l);
      windPower.getSpinner().addChangeListener(l);

      // grafische Darstellung der Wind-Matrix
      windMatrixGraphic = new JPanel() {
        public void paint(Graphics g) {
          super.paint(g);
          int cellSize = Math.min(getWidth(), getHeight()) / 3;
          // Mittelpunkt der Komponente
          int xM       = getWidth() / 2;
          int yM       = getHeight() / 2;
          // Ursprung der Matrix (links-oben)
          int x0       = (int) (xM - 1.5 * cellSize);
          int y0       = (int) (yM - 1.5 * cellSize);
          // Matrix
          g.setColor( Color.BLACK );
          g.drawRect(x0,          y0,          3*cellSize, 3*cellSize);
          g.drawRect(x0,          y0+cellSize, 3*cellSize, cellSize);
          g.drawRect(x0+cellSize, y0,          cellSize,   3*cellSize);
          // Basis-Kreis (Radius 1 Zelle)
          g.setColor( Color.BLUE );
          g.drawArc(xM-cellSize, yM-cellSize,2*cellSize,2*cellSize,0,360);
          // Referenzquadrat, welches fuer die Ueberlageung mit den
          // "Matrix"-Zellen verwendet wird
          //== Durchmesser ==
          int windEnvW = (int)Math.round( calculateReferenceSize(cellSize,windPow,windPowMaxBW) );
          //== Mittelpunkt ==
          int windEnvMX = (int)Math.round( xM - cellSize*Math.sin(windDir/180 * Math.PI) );
          int windEnvMY = (int)Math.round( yM - cellSize*Math.cos(windDir/180 * Math.PI) );
          //== Referenz-Quadrat ==
          g.setColor( Color.RED );
          g.drawRect(windEnvMX-windEnvW/2,windEnvMY-windEnvW/2,windEnvW,windEnvW);
        }
      };
      windMatrixGraphic.setPreferredSize( new Dimension(100,100) );
      this.add( windMatrixGraphic, new GridBagConstraints(
         1, 5, 1, 1, 1.0, 1.0, GridBagConstraints.NORTH, GridBagConstraints.BOTH, new Insets(5, 5, 5, 5), 0, 0) );
      
      // Labels belegen
      resetCaptions(null);
    }

    /**
     * Belegt die Labels in der GUI neu. Sind in der uebergebenen Map fuer die Keys
     * keine Werte hinterlegt (oder wenn {@code captionMap = null}, wird auf den
     * {@linkplain ImpetusFireCA#RESOURCE Resource-Provider} des Modells zurueckgegriffen.
     * @param captionMap Map in der nach den Label-Bezeichnungen gesucht wird (kann
     *                   {@code null} sein)
     */
    public void resetCaptions(Map<String,Object> captionMap) {
      windDirectionDev.getHeaderLabel().setText( determineCaption(captionMap, RESOURCE, "Model.GUI.WindDir.Dev") );
      windPower.getHeaderLabel().setText( determineCaption(captionMap, RESOURCE, "Model.GUI.WindPower") );
      windPowerDev.getHeaderLabel().setText( determineCaption(captionMap, RESOURCE, "Model.GUI.WindPower.Dev") );
      windMatrixLabel.setText( determineCaption(captionMap, RESOURCE, "Model.GUI.WindMatrix") );
    } 
  
    /**
     * Aktualisiert die Anzeige der Wind-Matrix.
     */
    public void repaintWindMatrx() {
      ((AbstractTableModel)windMatrixTable.getModel()).fireTableDataChanged();
    }

    /**
     * Initialisiert das Slider-Layer fuer die GUI.
     * @param slider zu initialisierender Slider
     * @param tickSize Abstand fuer die Ticks/Labels
     */
    private void initSlider(JSlider slider, int tickSize) {
      slider.setMajorTickSpacing(tickSize);
      slider.setPaintLabels(true);
      slider.setPaintTicks(true);
    }

    /**
     * Setzt die eingestellte Windrichtung.
     * @param degree Windrichtung in Grad (Nord = 0°, West = 90°)
     */
    public void setWindDirection(double degree) {
      windDirection.setValue(degree);
    }

    /**
     * Liefert die eingestellte Windrichtung.
     */
    public double getWindDirection() {
      return windDirection.getValue();
    }

    /**
     * Setzt die eingestellte Abweichung der Windrichtung.
     * @param degree Abweichung der Windrichtung in Grad
     */
    public void setWindDirectionDeviation(double degree) {
      windDirectionDev.setValue((int)degree);
    }

    /**
     * Liefert die eingestellte Abweichung der Windrichtung.
     */
    public double getWindDirectionDeviation() {
      return windDirectionDev.getValue();
    }

    /**
     * Setzt die eingestellte Windstaerke.
     * @param power Windstaerke in m/s
     */
    public void setWindPower(double power) {
      windPower.setValue((int)power);
    }

    /**
     * Liefert die eingestellte Windstaeke.
     */
    public double getWindPower() {
      return windPower.getValue();
    }

    /**
     * Setzt die eingestellte Abweichung der Windstaerke.
     * @param power Abweichung der Windstaerke in m/s
     */
    public void setWindPowerDeviation(double power) {
      windPowerDev.setValue((int)power);
    }

    /**
     * Liefert die eingestellte Abweichung der Windstaerke.
     */
    public double getWindPowerDeviation() {
      return windPowerDev.getValue();
    }
  }
}
