/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.model.ca.fire;

import java.util.Map;

import edu.bonn.xulu.data.DynamicXuluObject;
import edu.bonn.xulu.data.XuluDataException;
import edu.bonn.xulu.model.AbstractModelContentManager;
import edu.bonn.xulu.model.ModelResource;
import edu.bonn.xulu.model.PropertiesResource;
import edu.bonn.xulu.model.ValuePropertyResource;
import edu.bonn.xulu.plugin.data.feature.SingleFeatureCollection;

import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureIterator;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.MultiPoint;

import schmitzm.data.WritableGrid;
import schmitzm.data.AbstractWritableGrid;
import schmitzm.data.property.ScalarProperty;
import schmitzm.data.property.ListProperty;
import schmitzm.data.property.MatrixProperty;
import schmitzm.lang.ResourceProvider;


/**
 * Diese Klasse definiert und verwaltet die 7
 * Ressourcen fuer das Modell {@linkplain ImpetusFireCA}
 * und prueft diese auf Korrektheit.<br>
 * Die Art der benoetigten Ressourcen ist der {@linkplain ImpetusFireCA}-Doku
 * zu entnehmen.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ImpetusFireCAContentManager extends AbstractModelContentManager {

  /**
   * Erzeugt einen neuen ContentManager fuer das Modell {@linkplain ImpetusFireCA}.
   */
  public ImpetusFireCAContentManager() {
    super(7);
    // ===== Ressourcen festlegen =====
    // Allgemeine Modell-Parameter
    resource[0] = new PropertiesResource(ModelResource.CATEGORY_INPUT,"",DynamicXuluObject.class,false);
    // Bandherde
    resource[1] = new PropertiesResource(ModelResource.CATEGORY_INPUT,"",SingleFeatureCollection.class,false);
    // LUC Classification
    resource[2] = new ValuePropertyResource(ModelResource.CATEGORY_INPUT,"",ScalarProperty.class,WritableGrid.class,false);
    // Brenn-Eignung
    resource[3] = new ValuePropertyResource(ModelResource.CATEGORY_INPUT,"",ScalarProperty.class,WritableGrid.class,true);
    // Ausgabe-Raster fuer modellierte LUC
    resource[4] = new ValuePropertyResource(ModelResource.CATEGORY_OUTPUT,"",ScalarProperty.class,WritableGrid.class,false);
    // Ausgabe-Raster fuer verbrannte Biomasse
    resource[5] = new ValuePropertyResource(ModelResource.CATEGORY_OUTPUT,"",ScalarProperty.class,WritableGrid.class,true);
    // Temp-Raster fuer Brenn-Zeit
    resource[6] = new ValuePropertyResource(ModelResource.CATEGORY_TEMP,"",ScalarProperty.class,WritableGrid.class,false);
  
    // Ressourcen-Beschreibungen neu belegen
    resetCaptions(null);
  }
  
  /**
   * Belegt die Ressourcen-Beschreibungen neu. Sind in der uebergebenen Map fuer die Keys
   * keine Werte hinterlegt (oder wenn {@code captionMap = null}), wird auf den
   * {@linkplain ImpetusFireCA#RESOURCE Resource-Provider} des Modells zurueckgegriffen.
   * @param captionMap Map in der nach den Ressourcen-Beschreibungen gesucht wird (kann
   *                   {@code null} sein)
   */
  public void resetCaptions(Map<String,Object> captionMap) {
    resetResourceDescription(0, captionMap, ImpetusFireCA.RESOURCE, "MRDesc.In.ModelParams");
    resetResourceDescription(1, captionMap, ImpetusFireCA.RESOURCE, "MRDesc.In.BurnSource");
    resetResourceDescription(2, captionMap, ImpetusFireCA.RESOURCE, "MRDesc.In.BaseLUC");
    resetResourceDescription(3, captionMap, ImpetusFireCA.RESOURCE, "MRDesc.In.BurnSuit");
    resetResourceDescription(4, captionMap, ImpetusFireCA.RESOURCE, "MRDesc.Out.LUs");
    resetResourceDescription(5, captionMap, ImpetusFireCA.RESOURCE, "MRDesc.Out.BioMass");
    resetResourceDescription(6, captionMap, ImpetusFireCA.RESOURCE, "MRDesc.Temp.BurnTime");
  }
  
  public void checkAndError() {
    super.checkAndError();

    // Pruefen, ob Modell-Parameter alle erforderlichen Properties enthaelt
    DynamicXuluObject modelParam = (DynamicXuluObject)getResource(0).getData();
    if ( !modelParam.containsProperty(ImpetusFireCA.PARAM_PROP_MODELTIME) ||
         !(modelParam.getProperty(ImpetusFireCA.PARAM_PROP_MODELTIME) instanceof ScalarProperty) ||
         ((ScalarProperty)modelParam.getProperty(ImpetusFireCA.PARAM_PROP_MODELTIME)).getOneTimeReadAccess().getValueAsInt() <= 0 )
      throwResourceException(0, ImpetusFireCA.RESOURCE.getString("Error.In.ModelParams.ModelTime",ImpetusFireCA.PARAM_PROP_MODELTIME));
    if ( !modelParam.containsProperty(ImpetusFireCA.PARAM_PROP_BURNLUC) ||
         !(modelParam.getProperty(ImpetusFireCA.PARAM_PROP_BURNLUC) instanceof ScalarProperty) ||
         ((ScalarProperty)modelParam.getProperty(ImpetusFireCA.PARAM_PROP_BURNLUC)).getOneTimeReadAccess().getValueAsInt() <= 0 )
      throwResourceException(0, ImpetusFireCA.RESOURCE.getString("Error.In.ModelParams.BurnLuc",ImpetusFireCA.PARAM_PROP_BURNLUC));
    if ( !modelParam.containsProperty(ImpetusFireCA.PARAM_PROP_BURNOUTLUC) ||
         !(modelParam.getProperty(ImpetusFireCA.PARAM_PROP_BURNOUTLUC) instanceof ScalarProperty) ||
         ((ScalarProperty)modelParam.getProperty(ImpetusFireCA.PARAM_PROP_BURNOUTLUC)).getOneTimeReadAccess().getValueAsInt() <= 0 )
      throwResourceException(0, ImpetusFireCA.RESOURCE.getString("Error.In.ModelParams.BurnOutLuc",ImpetusFireCA.PARAM_PROP_BURNOUTLUC));
    if ( !modelParam.containsProperty(ImpetusFireCA.PARAM_PROP_BURNSUIT) ||
         !(modelParam.getProperty(ImpetusFireCA.PARAM_PROP_BURNSUIT) instanceof ListProperty) ||
         !Number.class.isAssignableFrom( modelParam.getProperty(ImpetusFireCA.PARAM_PROP_BURNSUIT).getPropertyType().getType() ) )
      throwResourceException(0, ImpetusFireCA.RESOURCE.getString("Error.In.ModelParams.BurnSuit",ImpetusFireCA.PARAM_PROP_BURNSUIT));
    if ( !modelParam.containsProperty(ImpetusFireCA.PARAM_PROP_BURNTIME) ||
         !(modelParam.getProperty(ImpetusFireCA.PARAM_PROP_BURNTIME) instanceof ListProperty) ||
         !Number.class.isAssignableFrom( modelParam.getProperty(ImpetusFireCA.PARAM_PROP_BURNTIME).getPropertyType().getType() ) )
      throwResourceException(0, ImpetusFireCA.RESOURCE.getString("Error.In.ModelParams.BurnTime",ImpetusFireCA.PARAM_PROP_BURNTIME));
    if ( !modelParam.containsProperty(ImpetusFireCA.PARAM_PROP_WINDDIRECTION) ||
         !(modelParam.getProperty(ImpetusFireCA.PARAM_PROP_WINDDIRECTION) instanceof ScalarProperty) ||
         ((ScalarProperty)modelParam.getProperty(ImpetusFireCA.PARAM_PROP_WINDDIRECTION)).getOneTimeReadAccess().getValueAsDouble() < 0 )
      throwResourceException(0, ImpetusFireCA.RESOURCE.getString("Error.In.ModelParams.WindDir",ImpetusFireCA.PARAM_PROP_WINDDIRECTION));
    if ( !modelParam.containsProperty(ImpetusFireCA.PARAM_PROP_WINDPOWER) ||
         !(modelParam.getProperty(ImpetusFireCA.PARAM_PROP_WINDPOWER) instanceof ScalarProperty) ||
         ((ScalarProperty)modelParam.getProperty(ImpetusFireCA.PARAM_PROP_WINDPOWER)).getOneTimeReadAccess().getValueAsDouble() <= 0 )
      throwResourceException(0, ImpetusFireCA.RESOURCE.getString("Error.In.ModelParams.WindPower",ImpetusFireCA.PARAM_PROP_WINDPOWER));
    if ( !modelParam.containsProperty(ImpetusFireCA.PARAM_PROP_WINDDIRECTIONDEV) ||
         !(modelParam.getProperty(ImpetusFireCA.PARAM_PROP_WINDDIRECTIONDEV) instanceof ScalarProperty) ||
         ((ScalarProperty)modelParam.getProperty(ImpetusFireCA.PARAM_PROP_WINDDIRECTIONDEV)).getOneTimeReadAccess().getValueAsDouble() < 0 )
      throwResourceException(0, ImpetusFireCA.RESOURCE.getString("Error.In.ModelParams.WindDirDev",ImpetusFireCA.PARAM_PROP_WINDDIRECTIONDEV));
    if ( !modelParam.containsProperty(ImpetusFireCA.PARAM_PROP_WINDPOWERDEV) ||
         !(modelParam.getProperty(ImpetusFireCA.PARAM_PROP_WINDPOWERDEV) instanceof ScalarProperty) ||
         ((ScalarProperty)modelParam.getProperty(ImpetusFireCA.PARAM_PROP_WINDPOWERDEV)).getOneTimeReadAccess().getValueAsDouble() < 0 )
      throwResourceException(0, ImpetusFireCA.RESOURCE.getString("Error.In.ModelParams.WindPowerDev",ImpetusFireCA.PARAM_PROP_WINDPOWERDEV));
    if ( !modelParam.containsProperty(ImpetusFireCA.PARAM_PROP_WINDPOWERMAXBW) ||
         !(modelParam.getProperty(ImpetusFireCA.PARAM_PROP_WINDPOWERMAXBW) instanceof ScalarProperty) ||
         ((ScalarProperty)modelParam.getProperty(ImpetusFireCA.PARAM_PROP_WINDPOWERMAXBW)).getOneTimeReadAccess().getValueAsDouble() <= 0 )
      throwResourceException(0, ImpetusFireCA.RESOURCE.getString("Error.In.ModelParams.WindPowerMaxBW",ImpetusFireCA.PARAM_PROP_WINDPOWERMAXBW));
//    if ( !modelParam.containsProperty(ImpetusFireCA.PARAM_PROP_WINDMATRIX) ||
//         !(modelParam.getProperty(ImpetusFireCA.PARAM_PROP_WINDMATRIX) instanceof MatrixProperty) ||
//         ((MatrixProperty)modelParam.getProperty(ImpetusFireCA.PARAM_PROP_WINDMATRIX)).getDimension() < 2 ||
//         ((MatrixProperty)modelParam.getProperty(ImpetusFireCA.PARAM_PROP_WINDMATRIX)).getSize(0) < 3 ||
//         ((MatrixProperty)modelParam.getProperty(ImpetusFireCA.PARAM_PROP_WINDMATRIX)).getSize(1) < 3 ||
//         !Number.class.isAssignableFrom( modelParam.getProperty(ImpetusFireCA.PARAM_PROP_WINDMATRIX).getPropertyType().getType() ) )
//      throwResourceException(0, ImpetusFireCA.RESOURCE.getString("Error.In.ModelParams.WindMatrix",ImpetusFireCA.PARAM_PROP_WINDMATRIX));
    // Konstante fuer Gewichtung der Windgeschwindigkeit in Windmatrix-
    // Berechnung ist optional; Wenn angegeben, dann muss es ein Double-Wert sein
    if ( !modelParam.containsProperty(ImpetusFireCA.PARAM_PROP_WINDMATRIX_PCONST) )
    	LOGGER.warn( ImpetusFireCA.RESOURCE.getString("Warn.In.ModelParams.Missing",ImpetusFireCA.PARAM_PROP_WINDMATRIX_PCONST,0) );
    else
    	if ( !(modelParam.getProperty(ImpetusFireCA.PARAM_PROP_WINDMATRIX_PCONST) instanceof ScalarProperty) )
    		throwResourceException(0, ImpetusFireCA.RESOURCE.getString("Error.In.ModelParams.WindMatrix_PowerConst",ImpetusFireCA.PARAM_PROP_WINDMATRIX_PCONST));
    // Faktor fuer Gewichtung der Windgeschwindigkeit in Windmatrix-
    // Berechnung ist optional; Wenn angegeben, dann muss es ein Double-Wert
    // <> 0 sein
    if ( !modelParam.containsProperty(ImpetusFireCA.PARAM_PROP_WINDMATRIX_PFACT) )
    	LOGGER.warn( ImpetusFireCA.RESOURCE.getString("Warn.In.ModelParams.Missing",ImpetusFireCA.PARAM_PROP_WINDMATRIX_PFACT,1) );
    else
    	if ( !(modelParam.getProperty(ImpetusFireCA.PARAM_PROP_WINDMATRIX_PFACT) instanceof ScalarProperty) ||
    			 ((ScalarProperty)modelParam.getProperty(ImpetusFireCA.PARAM_PROP_WINDMATRIX_PFACT)).getOneTimeReadAccess().getValueAsDouble() == 0 )
    		throwResourceException(0, ImpetusFireCA.RESOURCE.getString("Error.In.ModelParams.WindMatrix_PowerFact",ImpetusFireCA.PARAM_PROP_WINDMATRIX_PCONST));

    // Pruefen, ob FeatureCollection "Punkte" enthaelt
    FeatureCollection fc = ((SingleFeatureCollection)getResource(1).getData()).getFeatureCollection();
    FeatureIterator   fi = fc.features();
    Geometry          g  = null;
    if ( !fi.hasNext() ||
         (g = fi.next().getDefaultGeometry()) == null ||
         !( g instanceof Point || g instanceof MultiPoint ) )
      throwResourceException(1,ImpetusFireCA.RESOURCE.getString("Error.In.BurnSource.Data"));

    WritableGrid baseLUC = (WritableGrid)((ScalarProperty)getResource(2).getData()).getOneTimeReadAccess().getValue();
    // wenn angegeben, muss Brenneignungs-Raster mit LUC-Szenario uebereinstimmen
    if ( getResource(3).getData() != null ) {
      WritableGrid burnSuit = (WritableGrid)((ScalarProperty)getResource(3).getData()).getOneTimeReadAccess().getValue();
      AbstractWritableGrid.compareStructure(
          baseLUC,  ImpetusFireCA.RESOURCE.getString("Error.In.BaseLUC.Grid"),
          burnSuit, ImpetusFireCA.RESOURCE.getString("Error.In.BurnSuit.Grid"),
          false
      );
    }
    // Ausgabe-Raster fuer LUC muss mit LUC-Szenario uebereinstimmen
    WritableGrid outLUC = (WritableGrid)((ScalarProperty)getResource(4).getData()).getOneTimeReadAccess().getValue();
    AbstractWritableGrid.compareStructure(
        baseLUC,  ImpetusFireCA.RESOURCE.getString("Error.In.BaseLUC.Grid"),
        outLUC,   ImpetusFireCA.RESOURCE.getString("Error.Out.LUC.Grid"),
        false
    );

    if ( getResource(5).getData() != null ) {
      WritableGrid burnedMass = (WritableGrid)((ScalarProperty)getResource(5).getData()).getOneTimeReadAccess().getValue();
      AbstractWritableGrid.compareStructure(
          baseLUC,    ImpetusFireCA.RESOURCE.getString("Error.In.BaseLUC.Grid"),
          burnedMass, ImpetusFireCA.RESOURCE.getString("Error.Out.BioMass.Grid"),
          false
      );
    }
    // Temp-Raster fuer Brennzeit muss mit LUC-Szenario uebereinstimmen
    WritableGrid burnTime = (WritableGrid)((ScalarProperty)getResource(6).getData()).getOneTimeReadAccess().getValue();
    AbstractWritableGrid.compareStructure(
        baseLUC,  ImpetusFireCA.RESOURCE.getString("Error.In.BaseLUC.Grid"),
        burnTime, ImpetusFireCA.RESOURCE.getString("Error.Temp.BurnTime.Grid"),
        false
    );
  }

  private void throwResourceException(int resIdx, String errDesc) {
    throw new XuluDataException(
      ImpetusFireCA.RESOURCE.getString(
        "Error.Global.Mess",
        getResource(resIdx).getDesc(),
        errDesc
      )
    );
  }
}
