/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.model.clue;

import java.awt.Component;
import java.awt.Font;
import java.awt.geom.AffineTransform;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.Locale;
import java.lang.reflect.Array;

import javax.swing.BorderFactory;
import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JSlider;
import javax.swing.SpringLayout;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import schmitzm.data.WritableGrid;
import schmitzm.data.property.ListProperty;
import schmitzm.data.property.ListPropertyReadAccess;
import schmitzm.data.property.ListPropertyWriteAccess;
import schmitzm.data.property.MatrixProperty;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.PropertyWriteAccess;
import schmitzm.lang.LangUtil;
import schmitzm.lang.ResourceProvider;
import schmitzm.swing.ExpansionBar;
import schmitzm.swing.SpringUtilities;
import schmitzm.swing.SwingUtil;
import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.model.AbstractStepModel;
import edu.bonn.xulu.model.event.ModelIterationStepFinishedEvent;
import edu.bonn.xulu.model.event.ModelIterationStepStartedEvent;
import edu.bonn.xulu.plugin.data.grid.MultiGrid;
import edu.bonn.xulu.plugin.model.ca.fire.ImpetusFireCA;

/**
 * Diese Klasse stellt eine Nachbildung des
 * <a href="http://www.cluemodel.nl/" target="_blank">CLUE</a>-Modells dar
 * (CLUE = The <b>C</b>onversion of <b>L</b>and <b>U</b>se and its <b>E</b>ffects).
 * Es wird nicht garantiert, dass diese Klasse identisch zum CLUE-Modell arbeitet.
 * Alle Rechte am CLUE-Algrithmus liegen beim CLUE-Team um
 * <a href="mailto:peter.verburg@wur.nl">Peter Verburg</a>.<br>
 * <br>
 * Das CLUE-Modell arbeitet mit Rasterdaten und berechnet die zeitliche
 * Landnutzungsveraenderung (LUCC) in einem Untersuchungsgebiet iterativ auf
 * Basis von statistischen Eingabeparametern und vorgegebenen Landnutzungsbedarfen.<br>
 * Das CLUE-Modell benoetigt folgende Ressourcen, welche durch den
 * {@link ClueModelContentManager} spezifiziert werden:
 * <ol>
 * <li><b>Modellparameter:</b>
 *     <ul>
 *     <li><i>n</i> = Anzahl modellierter LU-Typen (<b>implizit</b> durch die LU-Conversion-Matrix gegeben!)</li>
 *     <li><i>m</i> = Anzahl an Driving Forces (<b>implizit</b> durch die Anzahl der Driving-Force-Raster gegeben!))</li>
 *     <li><i>o</i> = Anzahl an dynamischen Driving Forces (<b>implizit</b> durch die Liste der dynamischen DF gegeben!))</li>
 *     <li><i>t</i> = Anzahl zu simulierende Zeitschritte (Integer)</li>
 *     <li><i>x</i> = Breite des Untersuchungsgebiets (<b>implizit</b> durch das Ausgangsszenario gegeben!)</li>
 *     <li><i>y</i> = Anzahl zu simulierende Zeitschritte (<b>implizit</b> durch das Ausgangsszenario gegeben!)</li>
 *     <li><i>a</i> = Erlaubte Durchschnitts-Abweichung zu den vorgegebenen LU-Bedarfen</li>
 *     <li><i>s</i> = Erlaubte Abweichung eines jeden LU-Typs zum vorgegebenen LU-Bedarf</li>
 *     <li><i>d</i> = Art der Abweichungsangaben (0=prozentual, 1=absolut in Bedarf-Einheit)</li>
 *     </ul></li>
 * <li><b>Ein Ausgangsszenario ({@link WritableGrid})</b><br>
 *     Stellt die Ausgangs-Landnutzung dar und muss Werte zwischen 0 und <i>n</i>
 *     (sowie {@link Float#NaN} = NoData) enthalten.<br>
 *     Dieses Raster bestimmt die Groesse (<i>x</i>, <i>y</i>) des
 *     betrachteten Untersuchungsgebiets. Alle anderen Ein/Ausgabe-Raster muessen
 *     dieser Dimension entsprechen.</li>
 * <li><b>LU-History ({@link WritableGrid})</b><br>
 *     Stellt die Vergangenheit des Ausgangsszenarios dar. Fuer jede Zelle
 *     wird angegeben, wie lange die aktuelle Landnutzung bereits besteht.<br>
 *     Das Raster muss dieselbe Dimension besitzen, wie das Ausgangsszenario.</li>
 * <li><b>Eine LU-Conversion-Matrix ({@link MatrixProperty MatrixProperty[Integer]})</b><br>
 *     Beschreibt ob und wie ein direkterWechsel zwischen den LU-Typen erlaubt
 *     ist. Stellt eine <i>n</i>x<i>n</i>-Matrix dar und bestimmt somit,
 *     wie viele LU-Typen modelliert werden.</li>
 * <li><b>Eine LU-Conversion-Deviation-Matrix ({@link MatrixProperty MatrixProperty[Integer]})</b><br>
 *     Stellt eine <i>n</i>x<i>n</i>-Matrix dar und beschreibt eine Abweichungen
 *     fuer die LU-Conversion-Matrix-Eintraege > 100 und < -100, damit nicht
 *     alle Zellen gleichzeitig welchseln, sobald die Grenze erreicht ist.</li>
 * <li><b>Conversion Elasticity ({@link ListProperty ListProperty[Float]})</b><br>
 *     Beschreibt die Bestaendigkeit der LU-Typen und stellt so eine weitere
 *     Einschraenkung an den LU-Wechsel dar. Die Liste muss fuer jeden LU-Typ
 *     einen Wert zwischen 0 und 1 enthalten (also <i>n</i> Float-Werte).<br>
 *     <ul>
 *     <li><i>ELAS[U] = 1</i> bedeutet, dass wenn an einer Stelle eine U-Zelle "verschwindet",
 *     an anderer Stelle im gleichen Zeitschritt keine Zelle vom Typ U entstehen darf.</li>
 *     <li><i>ELAS[U] = 0</i> bedeutet, dass uneingeschraenkt an einer Stelle ein Wechsel von U
 *         nach A und gleichzeitig an anderer Stelle von B nach U moeglich ist.</li>
 *     <li><i>0 < ELAS[U] < 1</i> drueckt eine Tendenz zu einem der beiden Extremwerte 0 oder
 *         1 aus.</li>
 *     </ul></li>
 * <li><b>Area Restrictions, Spatial policies ({@link WritableGrid})</b><br>
 *     Beschreiben, an welchen Stellen des Untersuchungsgebiets ein LU-Wechsel generell
 *     verboten ist (z.B. Schutzgebiete oder Bereiche, die nicht zum Untersuchungsgebiet
 *     gehoeren).<br>
 *     Das Raster muss dieselbe Dimension besitzen, wie das Ausgangsszenario.
 *     Der Rasterwert NoData ({@link Float#NaN}) bedeutet, dass ein LU-Wechsel
 *     verboten ist, jeder andere Wert erlaubt den Wechsel.</li>
 * <li><b>Demand Scenario ({@link MatrixProperty MatrixProperty[Integer]})</b><br>
 *     Beschreibt den LU-Bedarf (in Zellen!) fuer jeden LU-Typ zu jedem
 *     modellierten Zeitschritt. Muss eine <i>n</i>x(<i>t</i>+1)-Matrix darstellen,
 *     die Integer-Werte enthaelt.<br>
 *     <u>Bemerke:</u><br>
 *     Der "Bedarf" in der Zeile 0 stellt die Situation
 *     des Ausgangsszenarios dar. Deshalb sind <i>t</i>+1 Zeilen anzugeben!</li>
 * <li><b>Static Driving Forces ({@link MultiGrid})</b><br>
 *     Beschreiben die Auspraegung der statischen Einflussfaktoren an jeder Stelle des
 *     Untersuchungsgebiets. Zusammen mit den Regressionsparametern werden
 *     hieraus die LU-Wahrscheinlichkeiten errechnet.<br>
 *     Das MultiGrid muss mind. <i>m</i> numerische Raster enthalten, die alle dieselbe
 *     Dimension besitzen, wie das Ausgangsszenario. Werden dynamische DF verwendet
 *     werden die Elemente an deren Index <b>nicht</b> verwendet, sie muessen
 *     jedoch vorhanden sein!</li>
 * <li><b>Opt: Dynamic Driving Force Types ({@link ListProperty ListProperty[Integer]})</b><br>
 *     Diese Liste enthaelt die Nummern der dynamischen Driving Forces. Die
 *     Angabe ist optional, muss aber - wenn angegeben - Werte zwischen
 *     0 und <i>m</i> enthalten.</li>
 * <li><b>Opt: Dynamic Driving Forces ({@link MultiGrid})</b><br>
 *     Beschreiben die Auspraegung der dynamischen Einflussfaktoren an jeder Stelle des
 *     Untersuchungsgebiets ueber die Zeit. Zusammen mit den Regressionsparametern werden
 *     hieraus die LU-Wahrscheinlichkeiten errechnet.<br>
 *     Das MultiGrid muss mind. <i>t</i>*<i>o</i> numerische Raster enthalten,
 *     die alle dieselbe Dimension besitzen, wie das Ausgangsszenario. Es koennen
 *     aber auch mehr Raster angegeben werden. Bedingung ist aber, dass fuer alle
 *     dyn. DF gleich viele Raster (=Zeitpunkte) angegeben werden! Die Grids
 *     sind in der Liste gruppiert nach DF anzugeben, d.h. erst alle Grids
 *     fuer den ersten dyn. DF, dann alle Grids fuer den zweiten, usw.</li>
 * <li><b>Regression Results ({@link MatrixProperty MatrixProperty[Float]})</b><br>
 *     Beschreiben die Einfluss-Staerke (Betas) der Driving Forces auf jeden einzelnen
 *     LU-Typ. Muss eine (<i>m</i>+1)x<i>n</i>-Matrix darstellen,
 *     die Float-Werte enthaelt.</li>
 * <li><b>Opt: Neigborhood Weights ({@link ListProperty ListProperty[Float]})</b><br>
 *     Diese (optionale) Liste enthaelt die Gewichte, wie stark der Nachbarschaftsfaktor
 *     in die Berechnung der LU-Wahrscheinlichkeit fuer jeden LU-Typ eingeht.
 *     Sie muss <i>n</i> Werte enthalten.</li>
 * <li><b>Opt: Neighborhood Settings ({@link MatrixProperty MatrixProperty[Float]}})</b><br>
 *     Beschreibt die Nachbarschaft, die fuer jeden LU-Typ betrachtet wird.
 *     Muss eine <i>w</i>x<i>h</i>x<i>n</i>-Matrix darstellen, die Float-Werte enthaelt.
 *     Dabei bestimmten <i>w</i> und <i>h</i> die Hoehe und Breit der betrachteten
 *     Nachbarschaft. Werden Neighborhodd Weights angegeben, so werden auch die
 *     Settings zu einer Pflicht-Angabe.
 * <li><b>Opt: Neighborhood Regression Results ({@link MatrixProperty MatrixProperty[Float]}})</b><br>
 *     Beschreibt die Einfluss-Staerke (Betas) eines jeden LU-Typs fuer den
 *     Fall, dass er in der Nachbarschaft eines anderen LU-Typs auftritt.
 *     Muss eine (<i>n</i>+1)x<i>n</i>-Matrix darstellen, die Float-Werte enthaelt.
 *     Werden Neighborhodd Weights angegeben, so werden auch die Regression Results
 *     zu einer Pflicht-Angabe.
 * <li><b>Opt: Location specific preference addition (LSPA) ({@link MultiGrid})</b><br>
 *     Enthaelt Faktoren, die (gewichtet) der Gesamt-Wahrscheinlichkeit
 *     hinzuaddiert werden. Sie dienen dazu, bestimmte Regionen fuer bestimmte
 *     LUC-Typen zu bevorzugen. Das MultiGrid muss mind. <i>n</i> numerische Raster
 *     enthalten, die alle dieselbe Dimension besitzen, wie das Ausgangsszenario.</li>
 * <li><b>Opt: Location specific preference addition factors ({@link ListProperty ListProperty[Float]})</b><br>
 *     Enthaelt Gewichte, mit denen die LSPA-W'keiten gewichtet werden. Die Liste
 *     muss min. <i>n</i> Eintraege haben.
 * <li><b>Ein Ausgaberaster fuer die aktuelle LU ({@link WritableGrid})</b><br>
 *     Hierin wird die aktuell modellierte LU-Konfiguration gespeichert.
 *     Das Raster muss dieselbe Dimension besitzen, wie das Ausgangsszenario.</li>
 * <li><b>Ein Ausgaberaster fuer die aktuelle LU Historie ({@link WritableGrid})</b><br>
 *     Hierin wird die aktuelle Dauer gespeichert, fuer die eine Zelle
 *     unter der aktuellen LU befindet.<br>
 *     Das Raster muss dieselbe Dimension besitzen, wie das Ausgangsszenario.</li>
 * <li><b>Ein temp. Ausgaberaster fuer die LU-Wahrscheinlichkeiten ({@link MultiGrid})</b><br>
 *     Hierin werden die ueber die Driving Forces und Regression Results errechneten
 *     LU-Wahrscheinlichkeiten gespeichert.
 *     Das Raster muss dieselbe Dimension besitzen, wie das Ausgangsszenario.</li>
 * <li><b>Ein temp. Ausgaberaster fuer die Nachbarschafts-Wahrscheinlichkeiten ({@link MultiGrid})</b><br>
 *     Hierin werden die ueber die Nachbarschaftszellen und Nachbarschaftsregression errechneten
 *     Wahrscheinlichkeiten gespeichert.
 *     Das Raster muss dieselbe Dimension besitzen, wie das Ausgangsszenario.</li>
 * <li><b>Ausgabeliste fuer die Iterations-Variablen ({@link ListProperty ListProperty[Float]})</b><br>
 *     Speichert die aktuellen Werte der Iterationsvariablen ITER(U).</li>
 * <li><b>Ausgaberaster fuer die LU zu jedem Zeitpunkt ({@link MultiGrid})</b><br>
 *     Hierin wird die fuer jeden Zeitschritt modellierte LU-Konfiguration gespeichert.<br>
 *     Das MultiGrid muss dieselbe Dimension besitzen, wie das Ausgangsszenario.</li>
 * </ol>
 * Die LU-Wahrscheinlichkeiten P1 (fuer Driving Forces) und P2 (fuer Nachbarschaft)
 * werden jeweils ueber ein <i>Binary Logit Model</i> errechnet und gehen anteilig
 * in die Gesamtwahrscheinlichkeit TPROB ein:
 * <center>
 *    TPROB(i,U) = P1(i,U) * (1-NBH_WEIGHT(U)) + P2(i,U) * NBH_WEIGHT + ELAS(U) - ITER(U)
 * </center><br>
 * Die Elastizitaet ELAS(U) wird nur eingerechnet, wenn sich die Zelle i im vorangegebenen
 * Zeitschritt bereits unter der Landnutzung U befindet.
 * In jedem Zeitschritt wird solgange iteriert, bis die vorgegebenen Bedarfe fuer
 * jeden LU-Typ (bis auf die Abweichungen <i>a</i> und <i>s</i>) erfuellt sind.
 * Nach jedem Iterationsschritt werden die ITER(U)-Variablen (einzige variable
 * Groesse innerhalb einer Zeitschritt-Iteration!) auf Basis der
 * Bedarfsabweichnung <i>DEMAND(U) - COVER(U)</i> angepasst, um die
 * Gesamtwahrscheinlicheiten "in eine geeignete Richtung" zu lenken.
 * <br><br>
 * <b><u>BEACHTE</u>:</b><br>
 * Zur Zeit werden folgende in CLUE integriert Regeln noch nicht unterstuetzt:
 * <ul>
 * <li>Area Restrictions fuer bestimmte Wechsel LU1 -> LU2 (Angabe ueber LU-Conversion
 *     Matrix zusammen mit speziellen Eingabe-Rastern)</li>
 * <li>Aufteilung des Untersuchungsgebiets in mehrere Regionen</li>
 * </ul>
 * @see ClueModelContentManager
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ClueModel extends AbstractStepModel {
  /** {@link ResourceProvider}, der die Uebersetzungen fuer die Clue-GUI
   *  bereitstellt: {@code locales.XuluModel_Clue} */
  protected static ResourceProvider CLUE_RES = XuluModellingPlatform.registerResourceProvider("XuluModel_Clue",Locale.ENGLISH);

  /** Konstante fuer die Toleranz-Art "prozentual". */
  public static final int TOLTYPE_PCT = 0;
  /** Konstante fuer die Toleranz-Art "absolut in Bedarf-Einheit". */
  public static final int TOLTYPE_ABS = 1;

  /** Speichert den ContentManager fuer das Modell.
   *  @see ClueModelContentManager */
  protected ClueModelContentManager contManager;

  private final ClueModelGUI  modelGUI = new ClueModelGUI();
  private final Logger        LOG      = LogManager.getLogger(this.getClass().getName());

  //********** Lese/Schreibrechte, die fuer den gesamten ***********
  //********** Modellanlauf gehalten werden muessen      ***********
  private PropertyReadAccess      RA_baseLU       = null; // Base Scenario
  private PropertyReadAccess      RA_baseLUHist   = null; // Base LU History
  private PropertyReadAccess      RA_luConv       = null; // LU-Conversion-Matrix
  private PropertyReadAccess      RA_luConvDev    = null; // Deviation of LU-Conversion-Matrix
  private PropertyReadAccess      RA_luElas       = null; // LU-Conversion-Elasticity
  private PropertyReadAccess      RA_areaRestr    = null; // Area Restrictions
  private PropertyReadAccess      RA_luDemand     = null; // LU-Bedarfe
  private PropertyReadAccess      RA_sdfGrids     = null; // Static Driving Forces
  private PropertyReadAccess      RA_ddfList      = null; // List of dynmic Driving Forces
  private PropertyReadAccess      RA_ddfGrids     = null; // Dynamic Driving Forces
  private PropertyReadAccess      RA_regrRes      = null; // Regression Results
  private PropertyReadAccess      RA_nbhWeights   = null; // Gewichtung der Nachbarschaft pro LU-Typ
  private PropertyReadAccess      RA_nbhSettings  = null; // Definition der Umgebung pro LU-Typ
  private PropertyReadAccess      RA_nbhRegrRes   = null; // Regression Results fuer Nachbarschaft
  private PropertyReadAccess      RA_lspaProb     = null; // Regression Results fuer Nachbarschaft
  private PropertyWriteAccess     WA_outActLU     = null; // speichert die aktuell sim. LU
  private PropertyWriteAccess     WA_outActLUHist = null; // speichert die aktuell sim. LU Historie
  private PropertyWriteAccess     WA_outLUProb    = null; // speichert die LU-Wahrscheinlichkeiten
  private PropertyWriteAccess     WA_outNbhProb   = null; // speichert die Nachbarschafts-Wahrscheinlichkeiten
  private ListPropertyWriteAccess WA_outTotProb   = null; // speichert die TPROB fuer jeden LUC-Typ
  private PropertyWriteAccess     WA_outStepLU    = null; // speichert die LU fuer jeden Zeitpunkt

  //**************** Variablen mit denen gearbeitet wird *******************
  // Modell-Parameter
  private int    luCount    = 0;    // Anzahl an LU-Typen
  private int    dfCount    = 0;    // Anzahl an Driving Forces
  private int    ddfCount   = 0;    // Anzahl an dynamischen Driving Forces
  private int    areaWidth  = 0;    // Breite des Untersuchungsgebiets (Rasterzellen)
  private int    areaHeight = 0;    // Hoehe des Untersuchungsgebiets (Rasterzellen)
  private int    areaMinX   = 0;    // Start-Index des Rasters in X-Richtung
  private int    areaMinY   = 0;    // Start-Index des Rasters in Y-Richtung
  private int    areaMaxX   = 0;    // End-Index des Rasters in X-Richtung (EXKLUSIVE!)
  private int    areaMaxY   = 0;    // End-Index des Rasters in Y-Richtung (EXKLUSIVE!)
  private int    demandTolType    = 0;  // Art der Abweichungsangaben
  private float  demandTolSingle  = 0;  // erlaubte Abweichung fuer jeden einzelnen LU-Typ vom Bedarf
  private float  demandTolAverage = 0;  // erlaubte Durchschnitts-Abweichung fuer das Untersuchungsgebiet
  private float  demandConv = 0;    // Umrechnungsfaktor Bedarfseingeit -> qm
  private float  cellSize   = 0;    // Flaeche einer Rasterzelle (in Bedarfseinheit)
  // Eingabe-Daten, die ueber den gesamten Modellablauf benoetigt werden
  private WritableGrid            baseLU        = null; // Base Scenario
  private WritableGrid            baseLUHist    = null; // Base LU History
  private int[][]                 luConv        = null; // LU-Conversion-Matrix
  private boolean                 CALC_WITH_DEV = false; // Flag, ob mit Abweichungen in LUC-Conversion-Matrix gerechnet wird
  private int[][]                 luConvDev     = null; // Deviation of LU-Conversion-Matrix
  private float[]                 luElas        = null; // LU-Conversion-Elasticity
  private WritableGrid            areaRestr     = null; // Area Restrictions
  private float[][]               luDemand      = null; // LU-Bedarfe
  private WritableGrid[]          sdfGrids      = null; // Static Driving Forces
  private boolean                 CALC_WITH_DDF = false; // Flag, ob dynamische DF beruecksichtigt werden
  private WritableGrid[]          ddfGrids      = null; // Dynamic Driving Forces
  private int[]                   ddfStartIdx   = null; // Listenindex an dem ein dyn. DF beginnt
  private float[][]               regrRes       = null; // Regression Results fuer LUCC
  private boolean                 CALC_WITH_NBH = false; // Flag, ob Nachbarschaftsw'keit beruecksichtigt wird
  private float[]                 nbhWeights    = null; // Gewichtung der Nachbarschaft pro LU-Typ
  private float[][][]             nbhSettings   = null; // Definition der Umgebung pro LU-Typ
  private float[][]               nbhRegrRes    = null; // Regression Results fuer Nachbarschaft
  private int                     nbhRadiusX    = 0;    // horizontaler Radius der betrachteten Nachbarschaft
  private int                     nbhRadiusY    = 0;    // vertikaler Radius der betrachteten Nachbarschaft
  private boolean                 CALC_WITH_LSPA = false; // Flag, ob LSPA beruecksichtigt wird
  private WritableGrid[]          lspaProb      = null; // Location specific preference addition
  private float[]                 lspaFact      = null; // Faktoren der Location specific preference addition
  private float[][]               luCover       = null; // LU-Bedeckung der Typen ueber die Zeit in Bedarfseinheit
  private float                   totCover      = 0;    // Gesamte Landbedeckungsflaeche (ohne NoData!)
  private Random                  rand          = null; // Zufallsgenerator
  private float[]                 iterSpeed     = null; // Geschwindigkeit fuer die ITER-Variablen
  // Ausgabe-Daten, die ueber den gesamten Modellablauf benoetigt werden
  private WritableGrid            outActLU      = null; // speichert die aktuell sim. LU
  private WritableGrid            outActLUHist  = null; // speichert die aktuell sim. LU History
  private WritableGrid[]          outLUProb     = null; // speichert die LU-Wahrscheinlichkeiten
  private WritableGrid[]          outNbhProb    = null; // speichert die Nachbarschafts-Wahrscheinlichkeiten
  private WritableGrid[]          outTotProb    = null; // speichert die TPROB fuer jeden LUC-Typ
  private float[]                 outIter       = null; // speichert die Iterationsvariablen
  private MultiGrid               outStepLU     = null; // speichert die LU fuer jeden Zeitpunkt
  private Map<Float,Float>        luMap         = new HashMap<Float,Float>(); // speichert das Mapping von LUCC-Typen im BAseScenario zu 0,1,2,...
  private Map<Float,Float>        luMapReverse  = new HashMap<Float,Float>(); // speichert das Mapping von LUCC-Typen im BAseScenario zu 0,1,2,...


  /**
   * Erzeugt eine neue Instanz des CLUE-Modells.
   */
  public ClueModel() {
    super( new ClueModelContentManager() );
    contManager = (ClueModelContentManager)super.contManager;
    modelResourceProvider = CLUE_RES;
    
    // In dieser optimierten Clue-Implementierung, werden die Iterationsvariablen
    // lokal verwaltet und brauchen deshalb nicht als Ressource zur Verfuegung
    // gestellt werden
    contManager.getResource( contManager.ITERVAR ).setNullable(true);
    // Iterations-Ereignisse der Event-Liste hinzufuegen
    events.add(ModelIterationStepStartedEvent.class);
    events.add(ModelIterationStepFinishedEvent.class);
  }


  ////////////////////////////////////////////////////////////////////////
  ///////////////   Implementierung von AbstractXuluModel   //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Initialisiert das Model, indem die Ressourcen aus dem ContentManager
   * geladen werden und mit Lese/Schreib-Rechten versehen werden. Zudem
   * werden Referenzen auf einen Teil der Ressourcen in lokalen Variablen
   * gespeichert, um waehrend des Modellablaufs effizienter darauf zugreifen
   * zu koennen.
   */
  public void performModelInit() {
    if ( contManager == null )
      return;
    //######### Zugriffsrechte aus Ressourcen/Propertys holen #########
    RA_baseLU       = contManager.getBaseScenario().getReadAccess(this);
    RA_baseLUHist   = contManager.getLUHistory().getReadAccess(this);
    MatrixProperty conversionMatrix = contManager.getLUConversionMatrix();
    RA_luConv       = conversionMatrix.getReadAccess(this);
    // Abweichungen in der LUC-Matrix sind optional
    CALC_WITH_DEV   = contManager.getLUConversionDeviationMatrix() != null;
    RA_luConvDev    = CALC_WITH_DEV ? contManager.getLUConversionDeviationMatrix().getReadAccess(this) : null;
    RA_luElas       = contManager.getConversionElasticity().getReadAccess(this);
    RA_areaRestr    = contManager.getAreaRestrictions().getReadAccess(this);
    RA_luDemand     = contManager.getDemandScenario().getReadAccess(this);
    RA_regrRes      = contManager.getRegressionResults().getReadAccess(this);
    RA_sdfGrids     = ((ListProperty)contManager.getStaticDrivingForces().getProperty(MultiGrid.PROP_GRIDS)).getReadAccess(this);
    // Dynamische Driving Forces sind optional
    CALC_WITH_DDF   = contManager.getDynamicDrivingForceTypes() != null;
    if ( CALC_WITH_DDF ) {
      RA_ddfList    = contManager.getDynamicDrivingForceTypes().getReadAccess(this);
      RA_ddfGrids   = ((ListProperty)contManager.getDynamicDrivingForces().getProperty(MultiGrid.PROP_GRIDS)).getReadAccess(this);
    } else {
      RA_ddfList    = null;
      RA_ddfGrids   = null;
    }
    // Nachbarschafts-W'keiten sind optional
    CALC_WITH_NBH   = (contManager.getNeighborhoodWeights() != null);
    if ( CALC_WITH_NBH ) {
      RA_nbhWeights  = contManager.getNeighborhoodWeights().getReadAccess(this);
      RA_nbhSettings = contManager.getNeighborhoodSettings().getReadAccess(this);
      RA_nbhRegrRes  = contManager.getNeighborhoodRegression().getReadAccess(this);
      WA_outNbhProb  = ((ListProperty)contManager.getNeighborhoodProbabilities().getProperty(MultiGrid.PROP_GRIDS)).getWriteAccess(this);
    } else {
      RA_nbhWeights  = null;
      RA_nbhSettings = null;
      RA_nbhRegrRes  = null;
      WA_outNbhProb  = null;
    }
    // Location specific preference addition ist optional
    CALC_WITH_LSPA = (contManager.getLSPAProbabilities() != null);
    if ( CALC_WITH_LSPA )
      RA_lspaProb = ((ListProperty)contManager.getLSPAProbabilities().getProperty(MultiGrid.PROP_GRIDS)).getReadAccess(this);
    else
      RA_lspaProb = null;

    WA_outActLU     = contManager.getActLUResult().getWriteAccess(this);
    WA_outActLUHist = contManager.getActLUHistory().getWriteAccess(this);
    WA_outLUProb    = ((ListProperty)contManager.getLUProbabilities().getProperty(MultiGrid.PROP_GRIDS)).getWriteAccess(this);
    WA_outStepLU    = ((ListProperty)contManager.getStepLUResults().getProperty(MultiGrid.PROP_GRIDS)).getWriteAccess(this);
    // Speichern der Gesamt-W'keiten ist optional
    if ( contManager.getTotalProbabilities() != null )
      WA_outTotProb = ((ListProperty)contManager.getTotalProbabilities().getProperty( MultiGrid.PROP_GRIDS )).getWriteAccess(this);
    else
      WA_outTotProb = null;

    //#########  Variablen belegen mit denen gearbeitet wird  ###########
    this.stepCount    = contManager.getStepCount().getOneTimeReadAccess().getValueAsInt(); // WICHTIG!!
    this.dfCount      = contManager.getStaticDrivingForces().getGridCount();
    this.demandTolType    = contManager.getDemandToleranceType().getOneTimeReadAccess().getValueAsInt();
    this.demandTolSingle  = contManager.getSingleDemandTolerance().getOneTimeReadAccess().getValueAsFloat();
    this.demandTolAverage = contManager.getAverageDemandTolerance().getOneTimeReadAccess().getValueAsFloat();
    this.demandConv   = contManager.getConversionFactor().getOneTimeReadAccess().getValueAsFloat();
    this.luCount      = conversionMatrix.getSize(0);

    this.baseLU       = (WritableGrid)RA_baseLU.getValue();
    this.areaWidth    = baseLU.getWidth();
    this.areaHeight   = baseLU.getHeight();
    this.areaMinX     = baseLU.getMinX();
    this.areaMinY     = baseLU.getMinY();
    this.areaMaxX     = baseLU.getMinX() + areaWidth;
    this.areaMaxY     = baseLU.getMinY() + areaHeight;
    this.cellSize     = ((float)(baseLU.getCellWidth() * baseLU.getCellHeight())) / demandConv;
    this.baseLUHist   = (WritableGrid)RA_baseLUHist.getValue();
    this.outActLU     = (WritableGrid)WA_outActLU.getValue();
    this.outActLUHist = (WritableGrid)WA_outActLUHist.getValue();
    this.outStepLU    = contManager.getStepLUResults();
    this.outIter      = new float[luCount];

    // LU-Conversion-Matrix und Abweichung in Arrays kopieren (PERFORMANZ!)
    this.luConv    = new int[conversionMatrix.getSize(0)][conversionMatrix.getSize(1)];
    this.luConvDev = new int[conversionMatrix.getSize(0)][conversionMatrix.getSize(1)];;
    for (int i=0; i<luConv.length; i++)
      for (int j=0; j<luConv[i].length; j++) {
        luConv[i][j]    = RA_luConv.getValueAsInt(i,j);
        luConvDev[i][j] = CALC_WITH_DEV ? RA_luConvDev.getValueAsInt(i,j) : 0;
      }
    // LU-Demand-Matrix in Array kopieren (PERFORMANZ!)
    this.luDemand    = new float[contManager.getDemandScenario().getSize(0)][contManager.getDemandScenario().getSize(1)];
    for (int i=0; i<luDemand.length; i++)
      for (int j=0; j<luDemand[i].length; j++)
        luDemand[i][j] = RA_luDemand.getValueAsFloat(i,j);
    // LUCC Elastizitaet in Array kopieren (PERFORMANZ!)
    this.luElas      = new float[((ListPropertyReadAccess)RA_luElas).getCount()];
    for (int i=0; i<luElas.length; i++)
      this.luElas[i] = RA_luElas.getValueAsFloat(i);
    this.areaRestr   = (WritableGrid)RA_areaRestr.getValue();
    // Statische DF in Array kopieren (PERFORMANZ!)
    this.sdfGrids    = new WritableGrid[contManager.getStaticDrivingForces().getGridCount()];
    for (int i=0; i<sdfGrids.length; i++)
      sdfGrids[i] = contManager.getStaticDrivingForces().getGrid(i);

    // Die Liste der dyn. DF ist blockweise nach DF grupperiert
    // Es koennen mehr Zeitschritt-Raster angegeben werden, als simuliert
    // werden.
    // Im Array 'ddfStartIdx' wird fuer die dynamischen DF gespeichert, an
    // welcher Stelle in der Liste der Block beginnt.
    // -1 bedeutet, dass der DF ein statische DF ist!
    this.ddfStartIdx = new int[dfCount];
    for (int i=0; i<dfCount; i++)
      ddfStartIdx[i] = - 1;

    // Variablen fuer dynamische Driving Forces nur belegen, wenn auch mit
    // diesen gerechnet wird.
    if ( CALC_WITH_DDF ) {
      ListPropertyReadAccess ddfList = (ListPropertyReadAccess)RA_ddfList;
      this.ddfCount    = ddfList.getCount();
      // Raster mit dyn. DF in Array kopieren (PERFORMANZ!)
      this.ddfGrids    = new WritableGrid[contManager.getDynamicDrivingForces().getGridCount()];
      for (int i=0; i<ddfGrids.length; i++)
        ddfGrids[i] = contManager.getDynamicDrivingForces().getGrid(i);
      // Start-Indizes der DDF in der Liste der DDF-Grids
      for (int i=0; i<ddfCount; i++)
        ddfStartIdx[ddfList.getValueAsInt(i)] = i * ddfGrids.length / ddfCount;
    } else {
      this.ddfCount = 0;
      this.ddfGrids = null;
    }

    // Regression Results fuer LUCC-W'keiten in Array kopieren (PERFORMANZ!)
    this.regrRes = new float[contManager.getRegressionResults().getSize(0)][contManager.getRegressionResults().getSize(1)];
    for (int i=0; i<regrRes.length; i++)
      for (int j=0; j<regrRes[i].length; j++)
        this.regrRes[i][j] = RA_regrRes.getValueAsFloat(i,j);

    // Ausgabe-Raster-Liste fuer LUC-W'keiten aus Ressource holen,
    // auf benoetigte Groesse erweitern und in Array kopieren (PERFORMANZ!)
    MultiGrid outLUProbTemp = contManager.getLUProbabilities();
    for (int i=0; outLUProbTemp.getGridCount()<luCount; i++)
      outLUProbTemp.addGrid(); // Liste muss ein Raster fuer jeden LU-Typ enthalten
    this.outLUProb  = new WritableGrid[luCount];
    for (int i=0; i<luCount; i++)
      this.outLUProb[i]  = outLUProbTemp.getGrid(i);

    // Variablen fuer Nachbarschaftsw'keiten nur belegen, wenn auch mit diesen
    // gerechnet wird
    if ( CALC_WITH_NBH ) {
      this.nbhRadiusX  = (contManager.getNeighborhoodSettings().getSize(0)-1)/2;
      this.nbhRadiusY  = (contManager.getNeighborhoodSettings().getSize(1)-1)/2;
      // Nbh.-Gewichtungen in Array kopieren (PERFORMANZ!)
      this.nbhWeights  = new float[((ListPropertyReadAccess)RA_nbhWeights).getCount()];
      for (int i=0; i<nbhWeights.length; i++)
        this.nbhWeights[i] = RA_nbhWeights.getValueAsFloat(i);
      // Nbh.-Settings in Array kopieren (PERFORMANZ!)
      this.nbhSettings = new float[contManager.getNeighborhoodSettings().getSize(0)][contManager.getNeighborhoodSettings().getSize(1)][contManager.getNeighborhoodSettings().getSize(2)];
      for (int i=0; i<nbhSettings.length; i++)
        for (int j=0; j<nbhSettings[i].length; j++)
          for (int k=0; k<nbhSettings[i][j].length; k++)
            this.nbhSettings[i][j][k] = RA_nbhSettings.getValueAsFloat(i,j,k);
      // Regression Results fuer Nbh. in Array kopieren (PERFORMANZ!)
      this.nbhRegrRes  = new float[contManager.getNeighborhoodRegression().getSize(0)][contManager.getNeighborhoodRegression().getSize(1)];
      for (int i=0; i<nbhRegrRes.length; i++)
        for (int j=0; j<nbhRegrRes[i].length; j++)
          this.nbhRegrRes[i][j] = RA_nbhRegrRes.getValueAsFloat(i,j);
      // Ausgabe-Raster-Liste fuer Nachbarschafts-W'keiten aus Ressource holen,
      // auf benoetigte Groesse erweitern und in Array kopieren (PERFORMANZ)
      MultiGrid outNbhProbTemp = contManager.getNeighborhoodProbabilities();
      for (int i=0; outNbhProbTemp.getGridCount()<luCount; i++)
        outNbhProbTemp.addGrid(); // Liste muss ein Raster fuer jeden LU-Typ enthalten
      this.outNbhProb = new WritableGrid[luCount];
      for (int i=0; i<luCount; i++)
        this.outNbhProb[i] = outNbhProbTemp.getGrid(i);
    } else {
      this.nbhRadiusX  = 0;
      this.nbhRadiusY  = 0;
      this.nbhWeights  = null;
      this.nbhSettings = null;
      this.nbhRegrRes  = null;
      this.outNbhProb  = null;
    }

    // Bei Toleranz-Art "prozentual" werden die Werte in Prozent
    // angegeben (intern wird aber mit Abweichungen zwischen 0 und 1 gerechnet!)
    if ( demandTolType == TOLTYPE_PCT ) {
      this.demandTolSingle  /= 100.0;
      this.demandTolAverage /= 100.0;
    }

    // "Geschwindigkeit" fuer die Iterationsvariablen
    this.iterSpeed = new float[luCount];

    // Speichert die Flaechendeckung (in der Bedarfseinheit, z.B. ha)
    // fuer die einzelnen LU-Typen zu jedem Zeitpunkt
    // --> Die initialen Werte stehen in der 0-Zeile der Demand-Matrix
    this.luCover  = new float[luCount][stepCount+1];
    this.totCover = 0;
    statusOut.print(CLUE_RES.getString("Model.InitCover")+": ");
    for (int lu=0; lu<luCount; lu++) {
      luCover[lu][0] = luDemand[lu][0];
      totCover += luCover[lu][0]; // Gesamt-Flaeche des Untersuchungsgebiets
      statusOut.print(luCover[lu][0]+"   ");
    }
    statusOut.println("   >> "+CLUE_RES.getString("Model.TotalCover")+" "+totCover);

    // Ausgabe-Raster leeren
    for (int y=outActLU.getMinY(); y<outActLU.getMinY()+outActLU.getHeight(); y++)
      for (int x=outActLU.getMinX(); x<outActLU.getMinX()+outActLU.getWidth(); x++)
        outActLU.setRasterSample(Float.NaN,x,y);
    // Liste der Schritt-Resultate leeren (damit bei vorzeitigem Modellende
    // nicht neue mit alten Daten vermischt werden)
    while ( outStepLU.getGridCount() > 0 )
      outStepLU.removeGrid(0);

    // (optionale) Ausgabe-Raster-Liste fuer Gesamt-W'keiten aus Ressource holen,
    // auf benoetigte Groesse erweitern und in Array kopieren (PERFORMANZ!)
    MultiGrid outTotProbTemp = contManager.getTotalProbabilities();
    if ( outTotProbTemp != null ) {
      for (; outTotProbTemp.getGridCount()<luCount;)
        outTotProbTemp.addGrid(); // Liste muss ein Raster fuer jeden LU-Typ enthalten
      this.outTotProb  = new WritableGrid[luCount];
      for (int i=0; i<luCount; i++)
        this.outTotProb[i] = outTotProbTemp.getGrid(i);
    } else
      outTotProb = null;

    // (optionale) LSPA-Raster und -Faktoren in Array kopieren (PERFORMANZ!)
    if ( CALC_WITH_LSPA ) {
      // Raster in Array kopieren
      this.lspaProb = new WritableGrid[luCount];
      for (int i=0; i<lspaProb.length; i++)
        this.lspaProb[i] = (WritableGrid)RA_lspaProb.getValue(i);
      // Wenn Faktoren angegeben, dann in Array kopieren
      this.lspaFact = new float[luCount];
      ListProperty lspaFactProp = contManager.getLSPAFactors();
      if ( lspaFactProp == null )
        // wenn Faktoren nicht angegeben, dann 1 annehmen
        Arrays.fill(lspaFact, 1.0f);
      else {
        ListPropertyReadAccess RA_lspaFact = lspaFactProp.getReadAccess(this);
        for (int i=0; i<lspaFact.length; i++)
          this.lspaFact[i] = RA_lspaFact.getValueAsFloat(i);
        RA_lspaFact.release();
      }
    } else {
      this.lspaProb = null;
      this.lspaFact = null;
    }

    // spezielle GUI initialisieren
    if (modelGUI != null)
      modelGUI.init(this);

    LOG.info("Scanning for all LU-Values in the Base-Scenario and mapping to 0...");
    // Alle Zellen des Untersuchungsgebiets...
    Set<Float> luccs = new HashSet<Float>();

    for (Integer y=0; y<this.areaHeight; y++)
      for (Integer x=0; x<this.areaWidth; x++){
    	  float rasterSampleAsFloat = (float)baseLU.getRasterSampleAsFloat(x,y);
    	  if (! Float.isNaN(rasterSampleAsFloat) )
    		 luccs.add( rasterSampleAsFloat );
  	}
    Object[] lmo = luccs.toArray() ;
	Arrays.sort(lmo);
	//statusOut.println(CLUE_RES.getString("Model.LucTypes")+": ");
	luMap.put(Float.NaN, Float.NaN );
	luMapReverse.put(Float.NaN, Float.NaN );
	for (Integer i = 0; i < lmo.length; i++ ){
		luMap.put((Float) lmo[i], new Float(i) );
		luMapReverse.put( new Float(i), (Float) lmo[i]);
	}

	LOG.info("========================================");
    LOG.info("MODEL INITIALISED (Log-Level "+LOG.getLevel()+")");
  }

  /**
   * Gibt die Lese/Schreibrechte auf das Ein- und Ausgaberaster wieder frei.
   * Interne Ressourcen brauchen keine aus dem Speicher entfernt werden.
   */
  public void performModelDispose() {
    // Ressourcen wieder freigeben
    releaseAccess(RA_baseLU);
    releaseAccess(RA_baseLUHist);
    releaseAccess(RA_luConv);
    releaseAccess(RA_luConvDev);
    releaseAccess(RA_luElas);
    releaseAccess(RA_areaRestr);
    releaseAccess(RA_luDemand);
    releaseAccess(RA_sdfGrids);
    releaseAccess(RA_ddfList);
    releaseAccess(RA_ddfGrids);
    releaseAccess(RA_regrRes);
    releaseAccess(RA_nbhWeights);
    releaseAccess(RA_nbhSettings);
    releaseAccess(RA_nbhRegrRes);
    releaseAccess(RA_lspaProb);
    releaseAccess(WA_outActLU);
    releaseAccess(WA_outActLUHist);
    releaseAccess(WA_outLUProb);
    releaseAccess(WA_outNbhProb);
    releaseAccess(WA_outStepLU);
    releaseAccess(WA_outTotProb);
    LOG.info("MODEL DISPOSED");
    LOG.info("========================================");
  }

  /**
   * Fuehrt einen Schritt des Modellablaufs durch.
   * @param stepNo zu modellierender Schritt (beginnend bei 1!)
   */
  public void performModelStep(int stepNo) {
    // Unloading no longer needed Data:
    if(stepNo>2)
      outStepLU.getGrid(stepNo-3).unloadData();

    // Raster festlegen, in dem die LU des vorangegangenen
    // Schritts zu finden ist
    WritableGrid lastLUGrid = (stepNo==1) ? baseLU : outStepLU.getGrid(stepNo-2);
    // Raster festlegen, in dem die aktuelle LU History
    // zu finden ist
    WritableGrid luHistGrid = (stepNo==1) ? baseLUHist : outActLUHist;

    // Merken, ob an einer Stelle bereits eine U-Zelle in einen anderen
    // Typ gewechselt hat
    boolean[] alreadyLUCCFromType = new boolean[luCount];
    // Merken, ob an einer Stelle bereits eine U-Zelle neu entstanden ist
    boolean[] alreadyLUCCToType = new boolean[luCount];

    // LU-Wahrscheinlichkeiten berechnen
    calculateLUProbabilities(stepNo);

    // Nachbarschafts-Wahrscheinlichkeiten berechnen
    calulateNeighborhoodProbabilities(stepNo,lastLUGrid);

    // Randomizer fuer diesen Zeitschritt erzeugen
    rand = new Random(System.currentTimeMillis());

    int   iterNo     = 0;
    int   x          = 0;
    int   y          = 0;
    int   lu         = 0;
    int   lastLU     = 0;
    int   bestLU     = 0;
    float bestLUProb = 0;
    float baseProb   = 0;
    float nbhProb    = 0;
    float nbhWeight  = 0;
    float elas       = 0;
    float lspa       = 0;
    float totalProb  = 0;
    float lastLUC    = 0;
    float actLUC     = 0;

    statusOut.print( CLUE_RES.getString("Model.Step.Calc.Iter", stepNo) );
    for (iterNo=1; !checkDemandCompliedAndCalculateIterVar(stepNo,iterNo); iterNo++) {
      this.fireModelEvent( new ModelIterationStepStartedEvent(this,stepNo,iterNo) );
      if ( modelGUI != null /*&& modelGUI.isDeviationUpdateEnabled()*/ )
        modelGUI.setIterationNumber(stepNo+"/"+iterNo);
      // Arrays fuer Iterationsschritt initialisieren
      for (lu = 0; lu < luCount; lu++) {
        // noch kein Wechsel von einem LU-Typ
        alreadyLUCCFromType[lu] = false;
        // noch kein Wechsel zu einem LU-Typ
        alreadyLUCCToType[lu] = false;
        // Flaechen der LU-Typen fuer diesen Zeitschritt
        luCover[lu][stepNo] = 0;
        // Speed fuer die Iterationsvariablen initialisieren
        // --> liegt die Nachfrage bei weniger als 5% des U'gebiets
        //     wird langsamer begonnen
        if ( luDemand[lu][stepNo]/totCover < 0.05f )
          iterSpeed[lu] = 0.01f;
        else
          iterSpeed[lu] = 0.05f;
      }

      // Fuer alle Zellen des Untersuchungsgebiets
      for (y = areaMinY; y < areaMaxY; y++)
        for (x = areaMinX; x < areaMaxX; x++) {
          checkBreakingCommands();

          // Rasterzellen ausserhalb des Untersuchungsgebiets komplett
          // ignorieren
          // Grund: NoData-Zellen (Float.NaN) werden beim Cast in int zu 0,
          //        also zur Landnutzung 0!!
          if ( checkNoData(x,y,areaRestr) || checkNoData(x,y,baseLU) )
            continue;

          // Nutzungsdauern erhoehen
          if (iterNo==1)
            outActLUHist.setRasterSample(luHistGrid.getRasterSampleAsFloat(x, y) + 1, x, y);

          // vorangegangene LU merken.
          // Diese steht im Raster nicht unbedingt 0-basiert
          lastLU = luMap.get( new Float (lastLUGrid.getRasterSampleAsInt(x, y)) ).intValue();

          // Beste gefundene LU wird mit aktueller LU initialisiert
          bestLU = lastLU;
          bestLUProb = 0;

          // Pruefen, ob Zelle die LU wechseln darf
          if (checkLUCCGeneralAllowed(x, y, stepNo, lastLU, alreadyLUCCToType[ lastLU ])) {
            // Alle LU-Typen auf Wechsel pruefen
            for (lu = 0; lu < luCount; lu++) {
              // Pruefen, ob Zelle in den betrachteten Typ wechseln darf
              if (checkLUCCAllowed(x, y, stepNo, lastLU, lu, alreadyLUCCFromType[lu])) {
                // Gesamt-Wahrscheinlichkeit berechnen
                // a) DF-W'keit und Nachbarschaft-W'keit gehen anteilig ein
                // b) Elastizitaet geht nur ein, wenn Zelle bereits fuer den
                //    LU-Typ genutzt wird
                // c) ITER geht negativ ein, da es auf Basis der Abweichung
                //    luCover-luDemand berechnet wird. Ein negatives Iter
                //    kennzeichnet also einen Mangel und muss die Gesamtw'keit
                //    erhoehen
                baseProb  = outLUProb[lu].getRasterSampleAsFloat(x, y);
                nbhProb   = !CALC_WITH_NBH ? 0 : outNbhProb[lu].getRasterSampleAsFloat(x, y);
                nbhWeight = !CALC_WITH_NBH ? 0 : nbhWeights[lu];
                elas      = (lu == lastLU && luElas[lu] < 1) ? luElas[lu] : 0;
                lspa      = !CALC_WITH_LSPA ? 0 : lspaProb[lu].getRasterSampleAsFloat(x, y) * lspaFact[lu];
                // LU-Wahrscheinlichkeit zusammensetzen
                totalProb = baseProb * (1-nbhWeight)
                          + nbhProb  * nbhWeight
                          + elas
                          + lspa
                          - outIter[lu];
                // Gesamt-Wahrscheinlichkeit speichern
                if ( outTotProb != null )
                  outTotProb[lu].setRasterSample(totalProb,x,y);

                // betrachteter Typ ist geeigneter als der bisher beste
                if (totalProb > bestLUProb) {
                  bestLU = lu;
                  bestLUProb = totalProb;
                }
              }
            }
            // Merken wenn wirklich ein Wechsel stattfindet
            if (bestLU != lastLU) {
              alreadyLUCCFromType[ lastLU ] = true;
              alreadyLUCCToType[ bestLU ] = true;
            }
          }

          // LU fuer diesen Iterationschritt setzen
          outActLU.setRasterSample( luMapReverse.get ( (float) bestLU) , x, y);
          // Flaechendeckung fuer LU-Typ erhoehen
          luCover[ bestLU ][stepNo] += cellSize;
        }
      this.fireModelEvent( new ModelIterationStepFinishedEvent(this,stepNo,iterNo) );
    }
    statusOut.println( CLUE_RES.getString("Model.Step.Calc.Iter.done",iterNo) );

    // Liste der Zeitschritt-Ausgaberaster vergroessern
    if ( outStepLU.getGridCount() <= stepNo-1 )
      outStepLU.addGrid();

    // LU in Ausgaberaster des Zeitschritts uebernehmen und ggf.
    // Nutzungsdauer initialisieren
    WritableGrid outStepLUGrid = outStepLU.getGrid(stepNo-1);
    for (y=areaMinY; y<areaMaxY; y++)
      for (x=areaMinX; x<areaMaxX; x++) {
        checkBreakingCommands();
        lastLUC = luMap.get( lastLUGrid.getRasterSampleAsFloat(x,y) );
        actLUC  = luMap.get( outActLU.getRasterSampleAsFloat(x,y) );
        // Wenn die Landnutzung gewechselt hat, wird die Historie zurueck
        // gesetzt
        if ( actLUC != lastLUC )
          outActLUHist.setRasterSample(0f,x,y);

        // LU in Ausgabe-Raster des simulierten Schritts uebernehmen
        // Die 0-basierten LU-Nummern werden wieder zurueckgemapped. (SK)
        outStepLUGrid.setRasterSample( luMapReverse.get( actLUC) ,x,y);
      }
  }

  /**
   * Liefert eine CLUE-spezifische GUI, in der die Abweichungen waehrend der
   * Iteration angezeigt werden und ueber die die Toleranzen eingestellt
   * werden koennen
   * @return Instannz von {@link ClueModelGUI}
   */
  public Component getModelSpecificGUI() {
    return modelGUI;
  }

  //////////////////////////////////////////////////////////////////////
  ////////////////////////   Clue-Methoden   ///////////////////////////
  //////////////////////////////////////////////////////////////////////
  /**
   * Prueft, ob eine Rasterzelle mit <code>NoData</code> belegt ist
   * @param x     Raster-Koordinate
   * @param y     Raster-Koordinate
   * @param grid  zu pruefendes Raster
   */
  protected static boolean checkNoData(int x, int y, WritableGrid grid) {
    return Float.isNaN(grid.getRasterSampleAsFloat(x,y));
  }

  /**
   * Berechnet die auf den Driving Forces basierenden
   * Landnutzungs-Wahrscheinlichkeiten fuer einen Modell-Schritt und speichert
   * diese in der Ressource "temp. Ausgaberaster fuer die LU-Wahrscheinlichkeiten".<br>
   * Werden keine dynamischen Driving Forces verwendet, geschieht die Berechnung
   * nur beim ersten Modellierungsschritt.
   * @param stepNo modellierter Schritt (beginnend bei 1)
   */
  protected void calculateLUProbabilities(int stepNo) {
    // Wenn keine dynamischen Driving Forces verwendet werden
    // braucht die Berechnung nur fuer den ersten Schritt
    // durchgefuehrt werden
    if ( ddfCount == 0 && stepNo > 1 )
      return;

    statusOut.print( CLUE_RES.getString("Model.Step.Calc.LUCProb", stepNo) );
    int x            = 0;
    int y            = 0;
    int lu           = 0;
    int df           = 0;
    float betaSum    = 0;
    float expBetaSum = 0;
    float prob       = 0;

    // Fuer Zellen des Untersuchungsgebiets
    for (y=areaMinY; y<areaMaxY; y++)
      for (x=areaMinX; x<areaMaxX; x++)
        // NotData-Bereich und Area Restrictions ignorieren
        if ( !checkNoData(x,y,baseLU) && areaRestr.getRasterSampleAsFloat(x,y)==0 ) {
          // Fuer alle LU-Typen
          for (lu = 0; lu < luCount; lu++) {
            betaSum = 0;
            // Fuer alle Driving Forces die Summe der Betas berechnen
            for (df = 0; df < dfCount; df++) {
              checkBreakingCommands();
              // DF-Grid festlegen:
              // Bei dynamischem DF das jew. Zeitschritt-Raster aus der Liste
              // ansonsten das statische Raster
              WritableGrid dfGrid = null;
              if ( ddfStartIdx[df] >= 0 )
                dfGrid = this.ddfGrids[ddfStartIdx[df]+stepNo-1];
              else
                dfGrid = this.sdfGrids[df];

///
//   if ( ddfStartIdx[df] >= 0 ) {
//     if ( x == dfGrid.convertRealToRaster(379755.2,0) &&
//          y == dfGrid.convertRealToRaster(1028214.7,1) ) {
//       LOG.debug("Step "+stepNo+", LUC "+lu+", "+x+"/"+y+":  "+dfGrid.getGridSampleAsFloat(379755.2, 1028214.7)+"    "+dfGrid);
//     }
//   }
///
              // Beim ersten Driving Force wird die Summe mit der
              // Regressionskonstante initialisiert
              if (df == 0)
                betaSum = regrRes[0][lu];
              // betaSum = BETA0(U) + Sum( BETA(df+1,U) * DF(df,x,y) )
              betaSum += regrRes[df + 1][lu]
                       * dfGrid.getRasterSampleAsFloat(x, y);
///
//  if ( x == 184 && y == 396 )
//    LOG.debug("Step "+stepNo+", LUC "+lu+", DF "+df+", "+x+"/"+y+":  beta = "+regrRes[df + 1][lu]+"     dfValue = "+dfGrid.getRasterSampleAsFloat(x, y)+"     mult = "+(regrRes[df + 1][lu] * dfGrid.getRasterSampleAsFloat(x, y))+"    BetaSum = "+betaSum);
///
            }
            expBetaSum = (float)Math.exp(betaSum);
            prob = expBetaSum / (1 + expBetaSum);
            // errechnete W'keit im Raster speichern
///
//  if ( x == 184 && y == 396 )
//    LOG.debug("Step "+stepNo+", LUC "+lu+", "+x+"/"+y+":  Prob = "+prob);
///
            this.outLUProb[lu].setRasterSample( (float) prob, x, y);
          }
        }
    //unloading no longer used ddfs to save memory (if supported by WritableGrid)
    for (df = 0; df < dfCount; df++) {
        WritableGrid grid = null;
        if ( ddfStartIdx[df] >= 0 )
          grid = this.ddfGrids[ddfStartIdx[df]+stepNo-1];
        if(grid!=null) grid.unloadData();
     }

    statusOut.println( CLUE_RES.getString("Model.Step.Calc.done") );
  }

  /**
   * Berechnet die auf den Nachbarschaftsbeziehungen basierenden
   * Landnutzungs-Wahrscheinlichkeiten fuer einen Modell-Schritt und speichert
   * diese in der Ressource "temp. Ausgaberaster fuer die Nachbarschafts-Wahrscheinlichkeiten".<br>
   * Da sich die Landnutzung mit jedem Zeitschritt aendert geschieht dies
   * vor jedem Zeitschritt neu.<br>
   * Diese Methode macht nichts, wenn ohne Nachbarschafts-W'keiten simuliert wird,
   * also wenn keine Angabe fuer die "Neighborhood-Weights" gemacht wird.
   * @param stepNo modellierter Schritt (beginnend bei 1)
   * @param luGrid LU-Raster aus dem die Umgebung der jeweiligen Zellen
   *               ermittelt werden
   */
  protected void calulateNeighborhoodProbabilities(int stepNo, WritableGrid luGrid) {
    // Wenn ohne Nachbarschaftsbeziehungen simuliert wird
    if ( !CALC_WITH_NBH )
      return;

    statusOut.print( CLUE_RES.getString("Model.Step.Calc.NbhProb", stepNo) );

    int      x            = 0;
    int      y            = 0;
    int      lu           = 0;
    int      i            = 0;
    int      dx           = 0;
    int      dy           = 0;
    int      totCellCount = 0;
    float[]  luCellCount  = new float[luCount];
    float    betaSum      = 0;
    float    beta         = 0;
    float    nbhFact      = 0;
    float    expBetaSum   = 0;
    float    lucc         = 0;
    float    setting      = 0;
    // Fuer Zellen des Untersuchungsgebiets
    for (y=areaMinY; y<areaMaxY; y++)
      for (x=areaMinX; x<areaMaxX; x++)
        // NoData-Bereich und Area Restrictions ignorieren
        if ( !checkNoData(x,y,baseLU) && areaRestr.getRasterSampleAsFloat(x,y)==0 )
          // Fuer alle LU-Typen
          for (lu = 0; lu < luCount; lu++) {
            checkBreakingCommands();

            // Anzahl der Zellen in der Umgebung initialisieren
            totCellCount = 0;
            for (i=0; i<luCount; i++)
              luCellCount[i] = 0;

            // Umgebung der Zelle betrachten
            for (dy=-nbhRadiusY; dy<=nbhRadiusY; dy++)
              if ( areaMinY <= y+dy && y+dy<areaMaxY )
                for (dx=-nbhRadiusX; dx<=nbhRadiusX; dx++)
                  if ( areaMinX <= x+dx && x+dx<areaMaxX ) {


                    lucc = luGrid.getRasterSampleAsFloat(x+dx,y+dy); // PERFORMANZ > nur einmal den Zellen-Wert ermitteln
                    if ( !Float.isNaN(lucc) /*!checkNoData(x+dx,y+dy,baseLU)*/ ) {
                    	// Die LU-Typen stehen im Raster nicht 0-basiert, deshalb mapping
                   		lucc = luMap.get( lucc );

                        // Alle Zellen in Umgebung zaehlen (jedoch nur diejenigen,
                        // die potentiell auch in die Bewertung eingehen koennen!)
                        setting = nbhSettings[dx+nbhRadiusX][dy+nbhRadiusY][lu];
                        if ( setting != 0.0 ) {
                          totCellCount++;
                          // Anzahl der Zellen eines bestimmten Typs aufsummieren
                          // (relativ zum in der Umgebung angegebenen Faktor)
//                          luCellCount[ luGrid.getRasterSampleAsInt(x+dx,y+dy) ] += setting;

                          //TODO Bricht hier ab, wenn LUCC-Klassen nicht 0,1,2 ... zaehlen...
//                         luCellCount[ (int)lucc ] += setting;

                          luCellCount[ (int)lucc ] += setting;

                        }
                      }
                    }
              // Summe der Betas berechnen
              betaSum = nbhRegrRes[0][lu]; // Regr.Konstante
              for (i=0; i<luCount; i++) {
                beta    = nbhRegrRes[i + 1][lu];
                nbhFact = (luCellCount[i]/totCellCount) / (luCover[i][stepNo-1]/totCover);
                betaSum += beta * nbhFact;
              }
              // Wahrscheinlichkeit berechnen
              expBetaSum = (float)Math.exp(betaSum);
              outNbhProb[lu].setRasterSample((float)(expBetaSum/(1 + expBetaSum)),x,y);
          }
    statusOut.println( CLUE_RES.getString("Model.Step.Calc.done") );
  }


  /**
   * Prueft, ob ein genereller LU-Wechsel fuer eine Zelle erlaubt ist.
   * <ul>
   * <li>Schliessen die Area Restrictions einen Wechsel aus?<br>
   *     > Rasterwert(x,y) != 0
   *     </li>
   * <li>Schliesst die Conversion Elasticity einen Wechsel aus (Elastizitaet 1
   *     und steigende Bedarf)?<br>
   *     > ELAS(lastLU) == 1 und DEMAND(lastLU,t) >= DEMAND(lastLU,t-1)
   *     </li>
   * </ul>
   * @param x      Raster-Koordinate
   * @param y      Raster-Koordinate
   * @param stepNo modellierter Schritt (beginnend bei 1)
   * @param lastLU Bedeckung von (x,y) zum Zeitpunkt stepNo-1
   * @param alreadyLUCCToActualType Flag gibt an, ob in der laufenden Iteration
   *                                bereits ein LU-Wechsel in den den aktuellen
   *                                Typ der Zelle (lastLU) vorgenommen wurde.<br>
   *                                <b>Dieses Flag wird zur Zeit nicht verwendet!</b>
   */
  protected boolean checkLUCCGeneralAllowed(int x, int y, int stepNo, int lastLU, boolean alreadyLUCCToActualType) {
    // Area Restrictions checken
    if ( areaRestr.getRasterSampleAsFloat(x,y) != 0 )
      return false;
// DER ERSTE TEIL IST KEINE ORIGINAL-CLUE-REGEL!
//    // Conversion Elasticity checken
//    // --> wenn Elasticity 1 und bereits an anderer Stelle eine Zelle des
//    //     bisherigen Typs entstanden ist, darf sie nicht wechseln
//    // --> wenn Elasticity 1 und Bedarf steigt, darf kein Wechsel
//    //     stattfinden, denn dann muessten (um den Bedarf zu decken)
//    //     an anderer Stelle neue entstehen
//    if (  luElas.getValueAsFloat(lastLU) == 1.0
//       && ( alreadyLUCCToActualType ||
//           luDemand.getValueAsFloat(lastLU,stepNo) >= luCover[lastLU][stepNo-1] ) )
//      return false;

    // Conversion Elasticity checken
    // --> wenn Elasticity 1 und Bedarf steigt, darf kein Wechsel
    //     stattfinden, denn dann muessten (um den Bedarf zu decken)
    //     an anderer Stelle neue entstehen
    if (  luElas[ lastLU ] == 1.0 &&
//          luDemand[lastLU][stepNo] >= luCover[lastLU][stepNo-1] )
          luDemand[ lastLU ][stepNo] >= luDemand[ lastLU ][stepNo-1] )
      return false;

    return true;
  }

  /**
   * Prueft, ob ein bestimmter LU-Wechsel fuer eine Zelle erlaubt ist.
   * <ul>
   * <li>Schliesst die Conversion Elasticity einen Wechsel aus (Elastizitaet des
   *     potentiellen Wechseltyps 1 und sinkender Bedarf)?<br>
   *     > lastLU != newLU und ELAS(newLU) == 1 und DEMAND(newLU,t) <= DEMAND(newLU,t-1)
   *     </li>
   * <li>Schliesst die Conversion Matrix einen Wechsel aus (generell oder ueber
   *     Historie)?<br>
   *     > CONV(lastLU,newLU) == 0 oder<br>
   *     > CONV(lastLU,newLU) > 100 und HIST(x,y) < CONV(lastLU,newLU)-101 oder<br>
   *     > CONV(lastLU,newLU) < -100 und HIST(x,y) > -CONV(lastLU,newLU)-101
   *     </li>
   * </ul>
   * @param x      Raster-Koordinate
   * @param y      Raster-Koordinate
   * @param stepNo modellierter Schritt (beginnend bei 1)
   * @param lastLU Bedeckung von (x,y) zum Zeitpunkt stepNo-1
   * @param newLU  potentielle neue Bedeckung von (x,y)
   * @param alreadyLUCCFromNewType  Flag gibt an, ob in der laufenden Iteration
   *                                bereits ein LU-Wechsel vom potentiellen neuen
   *                                Typ in einen anderen Typ vorgenommen wurde.<br>
   *                                <b>Dieses Flag wird zur Zeit nicht verwendet!</b>
   */
  protected boolean checkLUCCAllowed(int x, int y, int stepNo, int lastLU, int newLU, boolean alreadyLUCCFromNewType) {
// IST KEINE ORIGINAL-CLUE-REGEL!
//    // LU-Conversion-Elasticity des neuen Typs checken
//    // --> wenn Elasticity des neuen Typs 1 ist und an anderer Stelle bereits
//    //     eine Zelle des dieses Typs "verloren" ging, darf er hier
//    //     nicht entstehen
//    if (  alreadyLUCCFromNewType && luElas.getValueAsFloat(newLU) == 1.0 )
//      return false;

    // Conversion Elasticity checken
    // --> wenn Elasticity des neuen Typs "newLU" 1 und dessen Bedarf sinkt,
    //     darf kein Wechsel in diesen Typ stattfinden, denn dann muessten (um
    //     den Bedarf fuer "newLU" zu erreichen) an anderer Stelle newLU-Zellen
    //     entfernt werden
    if (  lastLU != newLU &&
          luElas[ newLU ] == 1.0 &&
          luDemand[ newLU ][stepNo] <= luCover[ newLU ][stepNo-1] )
//          luDemand.getValueAsFloat(newLU,stepNo) <= luDemand.getValueAsFloat(newLU,stepNo-1) )
      return false;

    // Wert aus LU-Conversion-Matrix ermitteln
    // Bemerke: "von" steht in den Zeilen (also Y, also Dim1)
    //          "nach" steht in den Spalten (also X, also Dim2)
    int conv    = luConv[ newLU ][ lastLU ];
    // 1 --> LUCC generell erlaubt
    if ( conv == 1 )
      return true;
    // 0 --> LUCC generell verboten
    if ( conv == 0 )
      return false;

    // Max. Abweichung der LU-Conversion
    int convDev = Math.abs( luConvDev[ newLU ][ lastLU ] );
    // Zufallswert zwischen -convDev und +convDev
    if ( convDev > 0 )
      convDev = rand.nextInt(2*convDev+1) - convDev;

    // 100+t --> Wechsel erst nach t Schritten der vorangegangenen
    //           LU moeglich
    if ( conv > 100 ) {
//      int minLUHist = conv-100;
//      int luHist    = outActLUHist.getRasterSampleAsInt(x,y);
//      return luHist >= minLUHist;
      int minLUHist = conv-101 + convDev;
      int luHist    = outActLUHist.getRasterSampleAsInt(x,y);
      return luHist >= minLUHist;
    }

    // -100-t --> Wechsel nicht mehr moeglich, wenn Zelle bereits
    //            t Schritte unter der vorangegangenen LU
    if ( conv < -100 ) {
//      int maxLUHist = -conv-100;
//      int luHist    = outActLUHist.getRasterSampleAsInt(x,y);
//      return luHist < maxLUHist;
      int maxLUHist = -conv-101 + convDev;
      int luHist    = outActLUHist.getRasterSampleAsInt(x,y);
      return luHist <= maxLUHist;
    }

    return true;
  }

  /**
   * Prueft, ob die aktuelle LU-Konfiguration die einzelenen Bedarfsanforderungen
   * (bis auf eine Abweichung) erfuellt und berechnet die Iterationsvariablen
   * neu.
   * @param stepNo modellierter Schritt (beginnend bei 1)
   * @param iterNo aktuelle Iteration (beginnend bei 1)
   */
  protected boolean checkDemandCompliedAndCalculateIterVar(int stepNo, int iterNo) {
    boolean updateEnabled  = (modelGUI!=null && modelGUI.isDeviationUpdateEnabled());
    boolean demandComplied = true;
    // Abweichnungen berechnen
    float[] singleDevAbs = new float[luCount];
    float[] singleDevPct = new float[luCount];
    float   averageDevAbs = 0;
    float   averageDevPct = 0;
    float   totalDevAbs = 0;
    float   totalDevPct = 0;
    for (int lu = 0; lu < luCount; lu++) {
      // ##### Einzelabweichungen #####
      // Absolute Abweichung der Landbedeckung vom Bedarf berechnen
      float demand = luDemand[lu][stepNo];
      singleDevAbs[lu]  = luCover[lu][stepNo] - demand;
      // Prozentuale Einzel-Abweichnung relativ zum Bedarf
      if (demand == 0)
        // ist der Bedarf 0, aber auch die Bedeckung, ist die Abw. 0%
        singleDevPct[lu] = luCover[lu][stepNo] == 0 ? 0 : 1;
      else
        singleDevPct[lu] = singleDevAbs[lu] / demand;

      // Abweichung ueber erlaubter Toleranz?
      float singleDev = (demandTolType == TOLTYPE_PCT) ? singleDevPct[lu] : singleDevAbs[lu];
      if ( updateEnabled ) {
        modelGUI.setDeviation(lu,singleDev);
        modelGUI.setIterationVariable(lu, outIter[lu]);
      }
      if (Math.abs(singleDev) > demandTolSingle)
        demandComplied = false;

//      // Abweichung des LU-Typs ausgeben
//      appendLog(singleDevAbs[lu],singleDevPct[lu],luDemand.getValueAsFloat(lu, stepNo),luCover[lu][stepNo]);

      // ##### Gesamtabweichungen aufsummieren #####
      totalDevAbs += Math.abs(singleDevAbs[lu]);
    }

    // Prozentuale Gesamt-Abweichnung relativ zum Gesamt-Untersuchungsgebiet
    totalDevPct = totalDevAbs / totCover;
    if ( updateEnabled )
      modelGUI.setDeviation(luCount,(demandTolType == TOLTYPE_PCT) ? totalDevPct : totalDevAbs);

    // Durchschnittliche Abweichnung pro LU-Typ
    averageDevAbs = totalDevAbs / luCount;

    // Durchschnittliche Abweichnung pro LU-Typ relativ zum Gesamt-Untersuchungsgebiet
    averageDevPct = averageDevAbs / totCover;

    // Abweichung ueber erlaubter Toleranz?
    float avgDev = (demandTolType == TOLTYPE_PCT) ? averageDevPct : averageDevAbs;
    if ( updateEnabled )
      modelGUI.setDeviation(luCount+1,avgDev);
    if ( Math.abs(avgDev) > demandTolAverage )
      demandComplied = false;

    // Iterationsvariablen neu berechnen
    for (int lu = 0; lu < luCount; lu++) {
      // Um Alternieren zwischen 2 Konfiguration zu verhindern
      // eine kleine Zufallszahl hinzuaddieren (zwischen -0,005 und 0,005)
      float randomValue = (rand.nextFloat()-0.5f)/100;

      // --> Im ersten Schritt erhalten alle Iterationsvariablen den Wert 0,5
      float iter = outIter[lu];
      if (iterNo == 1)
        iter = 0.5f+randomValue;
//        iter = 0.5 + new float[] {0.002f,0f,0.002f,0f,0f,-0.003f}[lu];
//        iter = 0.5 + new float[] {0.002f,0f,0.001f,0f,0f,-0.002f}[lu];
      else {
//        // Anteil der Einzelabweichnung an der Gesamtabweichnung
////        float dev = (totalDevAbs!=0) ? singleDevAbs[lu] / totalDevAbs : 0.0;
//        float dev  = singleDevPct[lu];
//        iter = outIter.getValueAsFloat(lu) // alter Iter-Wert
//             + dev // Anteil der Abweichung zum Bedarf
//             + randomValue;
        /////////////////////////////////////////
        // Original-Iter-Berechnung aus CLUE
        iterSpeed[lu] += 0.00005;
        iter += (singleDevAbs[lu]/totCover)/(iterSpeed[lu]*(rand.nextInt(101)+1));

        if (iter > 2f)
          iter = 0.5f;
        if ( Math.abs(iter) < 0.001 )
          iter = 0.01f * Math.signum(iter);
        if (iter < -0.5f)
          iter = -0.4f;
        /////////////////////////////////////////
      }
      // Iter speichern
      outIter[lu] = iter;
    }

    return demandComplied;
  }

  //////////////////////////////////////////////////////////////////////
  /////////////////////   Clue-spezifische GUI   ///////////////////////
  //////////////////////////////////////////////////////////////////////

  /**
   * Diese Klasse stellt eine modellspezifische GUI-Komponente fuer das
   * {@link ClueModel CLUE-Modell} dar.
   * In dieser wird fuer jeden modellierten LU-Typ ein {@linkplain ExpansionBar Status-Balken}
   * fuer die Abweichnung vom geforderten Bedarf angezeigt. Sowie je ein Balken
   * fuer die aktuelle Gesamt- und Durchschittsabweichung. Je nach gewaehlter
   * Abweichungsart (prozentual oder absolute Angaben), werden Abweichungen
   * von +/- 100% oder +/- 10000 Bedarfseinheiten angezeigt.<br>
   * Darueber hinaus enthaelt sie je einen Regler, ueber den die maximal
   * erlaubte Abweichung von den Bedarfen (Einzelabweichung der LU-Typen und
   * Durchschnittsabweichung) eingestellt werden kann.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class ClueModelGUI extends JComponent {
    private       float MIN = -1;
    private       float MAX = 1;
    private final float TOL = 0.2f;
    private       float SLIDER_FACT = 1000;
    // Format fuer Wert-Anzeige des Sliders
    private DecimalFormat sliderPctFormat = new DecimalFormat("0.0%");
    private DecimalFormat sliderDecFormat = new DecimalFormat("0.0");
    // Format fuer Wert-Anzeige im Balken
    private DecimalFormat statusPctFormat = new DecimalFormat("0.0000%");
    private DecimalFormat statusDecFormat = new DecimalFormat("0.0");
    // Format fuer Wert-Anzeige unterhalb der Balken (Max/Min/Middle)
    private DecimalFormat infoPctFormat   = new DecimalFormat("0.0%");
    private DecimalFormat infoDecFormat   = new DecimalFormat("0.0");

    private DeviationContainer  divCont          = null;
    private JSlider             singleDivSlider  = null;
    private JSlider             averageDivSlider = null;
    private ClueModel           model            = null;
    private DeviationLabel      singleDivLabel   = new DeviationLabel(" ",sliderPctFormat);
    private DeviationLabel      averageDivLabel  = new DeviationLabel(" ",sliderPctFormat);
    private Font                labelFont        = singleDivLabel.getFont().deriveFont(Font.BOLD | Font.CENTER_BASELINE);
    private boolean             alreadyPacked    = false;
    private JCheckBox           updateCheckBox   = null;
    private JLabel              iterationLabel   = null;
    private String              iterationText    = "";

    /**
     * Erzeugt eine neue Clue-GUI. Diese wird jedoch erst nach Aufruf
     * der {@link #init(ClueModel)}-Methode komplett erzeugt, da erst dann
     * die Anzahl der angezeigten Balken (= Anzahl LU-Typen) feststeht.
     */
    public ClueModelGUI() {
      super();
      this.setLayout( new SpringLayout() );
      // Label in dem die Iterationsnummer angezeigt wird
      iterationLabel = new JLabel(" ");
      iterationLabel.setBorder( BorderFactory.createLoweredBevelBorder() );
      // Slider/Label fuer Abweichnug der einzelnen Typen
      singleDivSlider  = new JSlider(JSlider.HORIZONTAL,0,(int)(SLIDER_FACT*MAX),(int)(SLIDER_FACT*TOL));
      singleDivSlider.setPaintTicks(true);
      singleDivSlider.setPaintLabels(true);
      singleDivSlider.setPaintTrack(true);
      singleDivSlider.setEnabled(false);
      singleDivSlider.addChangeListener( new ChangeListener() {
        public void stateChanged(ChangeEvent e) {
          float tol = singleDivSlider.getValue() / SLIDER_FACT;
          if ( model!=null )
            model.demandTolSingle = tol;
          if ( singleDivLabel!=null )
            singleDivLabel.setValue(tol);
          if ( divCont!=null )
            for (int i=0; i<divCont.getCount()-1; i++)
              divCont.setTolerance(i,tol);
        }
      });
      singleDivLabel.setFont(labelFont);
      singleDivLabel.setValue(singleDivSlider.getValue()/SLIDER_FACT);

      // Slider/Label fuer Abweichnug der einzelnen Typen
      averageDivSlider  = new JSlider(JSlider.HORIZONTAL,0,(int)(SLIDER_FACT*MAX),(int)(SLIDER_FACT*TOL));
      averageDivSlider.setPaintTicks(true);
      averageDivSlider.setPaintLabels(true);
      averageDivSlider.setPaintTrack(true);
      averageDivSlider.setEnabled(false);
      averageDivSlider.addChangeListener( new ChangeListener() {
        public void stateChanged(ChangeEvent e) {
          float tol = averageDivSlider.getValue() / SLIDER_FACT;
          if ( model!=null )
            model.demandTolAverage = tol;
          if ( averageDivLabel!=null )
            averageDivLabel.setValue(tol);
          if ( divCont!=null )
            divCont.setTolerance( divCont.getCount()-1, tol );
        }
      });
      averageDivLabel.setFont(labelFont);
      averageDivLabel.setValue(averageDivSlider.getValue()/SLIDER_FACT);

      // Checkbox zur (De)Aktivierung der Anzeige-Aktualisierung
      updateCheckBox = new JCheckBox(" ",true);

      // Labels lokalisieren
      resetCaptions(null);

      this.add( averageDivLabel );
      this.add( averageDivSlider );
      this.add( singleDivLabel );
      this.add( singleDivSlider );
      this.add( updateCheckBox );
      this.add( iterationLabel );
      SpringUtilities.makeCompactGrid(this,6,1,0,0,0,5);
    }

    /**
     * Initialisiert die GUI fuer eine <b>initialisierte</b> Modell-Instanz.
     * Wird zum Ende {@link ClueModel#performModelInit()} aufgerufen.
     */
    public void init(ClueModel model) {
      this.model = model;
      if ( divCont != null )
        remove(divCont);
      SLIDER_FACT = (model.demandTolType == TOLTYPE_PCT) ? 1000 : 1;
      MIN         = (model.demandTolType == TOLTYPE_PCT) ? -1 : ((float)-Math.pow(10,(int)Math.log10(model.totCover)))/1000;
      MAX         = (model.demandTolType == TOLTYPE_PCT) ?  1 : -MIN;
      averageDivSlider.setMaximum((int)(MAX*SLIDER_FACT));
      averageDivLabel.setNumberFormat( model.demandTolType == TOLTYPE_PCT ? sliderPctFormat : sliderDecFormat );
      singleDivSlider.setMaximum((int)(MAX*SLIDER_FACT));
      singleDivLabel.setNumberFormat( model.demandTolType == TOLTYPE_PCT ? sliderPctFormat : sliderDecFormat );

      String[] divDesc = new String[model.luCount+2];
      for (int i=0; i<model.luCount; i++)
        divDesc[i] = String.valueOf(i);
      divDesc[model.luCount]   = "";
      divDesc[model.luCount+1] = "";

      divCont = new DeviationContainer(
                         divDesc.length,
                         divDesc,
                         MIN,
                         MAX,
                         0,
                         model.demandTolSingle,
                         model.demandTolType == TOLTYPE_PCT ? statusPctFormat : statusDecFormat,
                         model.demandTolType == TOLTYPE_PCT ? infoPctFormat   : infoDecFormat
      );
      divCont.setBarDescHeader( CLUE_RES.getString("Model.GUI.LucType") );
      divCont.setBarHeader( CLUE_RES.getString("Model.GUI.CurrDemandDev") );
      this.add(divCont,4);

      // Wert des Balkens darf erst hier (nach Neu-Erstellung des
      // Abweichungs-Containers) neu gesetzt werden, da beim Setzen
      // des neuen Wertes automatisch die Toleranzen der Abweichungsbalken
      // aktualisiert werden
      setSingleTolerance(model.demandTolSingle);
      setAverageTolerance(model.demandTolAverage);

      // Labels lokalisieren
      resetCaptions(null);
      
      SpringUtilities.makeCompactGrid(this,7,1,0,0,0,5);
      if ( !alreadyPacked )
        SwingUtil.packParentWindow(this);
      else
        this.validate();
      alreadyPacked = true;
      singleDivSlider.setEnabled( true );
      averageDivSlider.setEnabled( true );
    }

    /**
     * Belegt die Labels in der GUI neu. Sind in der uebergebenen Map fuer die Keys
     * keine Werte hinterlegt (oder wenn {@code captionMap = null}, wird auf den
     * {@linkplain ClueModel#CLUE_RES Resource-Provider} des Modells zurueckgegriffen.
     * @param captionMap Map in der nach den Label-Bezeichnungen gesucht wird (kann
     *                   {@code null} sein)
     */
    public void resetCaptions(Map<String,Object> captionMap) {
      singleDivLabel.setText( determineCaption(captionMap, CLUE_RES, "Model.GUI.Single.Dev.Allowed") );
      averageDivLabel.setText( determineCaption(captionMap, CLUE_RES, "Model.GUI.Avg.Dev.Allowed") );
      iterationText = determineCaption(captionMap, CLUE_RES, "Model.GUI.Iteration");
      updateCheckBox.setText( determineCaption(captionMap, CLUE_RES, "Model.GUI.AutoUpdate") );
      
      if ( divCont != null ) {
        divCont.setBarDescHeader( determineCaption(captionMap, CLUE_RES, "Model.GUI.LucType") );
        divCont.setBarHeader( determineCaption(captionMap, CLUE_RES, "Model.GUI.CurrDemandDev") ); 
        divCont.setBarDesc( model.luCount,   determineCaption(captionMap, CLUE_RES, "Model.GUI.Total") );
        divCont.setBarDesc( model.luCount+1, determineCaption(captionMap, CLUE_RES, "Model.GUI.Avg") );
      } 
    }
    
    /**
     * Setzt die in der GUI angezeigte Interationsnummer. Diese wird
     * durch einen String repraesentiert, damit auch berliebige andere
     * Informationen (anstelle der Nummer) dargestellt werden koennen (z.B.
     * zusaetzlich noch die Schrittnummer).
     * @param iterString dargestellte Iterationsnummer
     */
    public void setIterationNumber(String iterString) {
      iterationLabel.setText( iterationText+": "+iterString );
    }

    /**
     * Stellt den Slider fuer die Einzelabweichung auf einen bestimmten Wert
     * ein. Hierdurch wird automatisch auch der Wert fuer das Modell neu
     * gesetzt.
     */
    public void setSingleTolerance(float tol) {
      singleDivSlider.setValue((int)(tol*SLIDER_FACT));
    }

    /**
     * Stellt den Slider fuer die Durchschnittsabweichung auf einen bestimmten Wert
     * ein. Hierdurch wird automatisch auch der Wert fuer das Modell neu
     * gesetzt.
     */
    public void setAverageTolerance(float tol) {
      averageDivSlider.setValue((int)(tol*SLIDER_FACT));
    }

    /**
     * Setzt die Abweichnungswerte fuer alle Statusbalken.
     * @param value Array mit den Abweichungen
     */
    public void setDeviation(float[] value) {
      if ( divCont!= null && isDeviationUpdateEnabled() )
        divCont.setExpansion(value);
    }

    /**
     * Setzt den Abweichnungswert fuer einen Statusbalken.
     * @param idx   Index eines Balkens
     * @param value Abweichung die im Balken dargestellt wird
     */
    public void setDeviation(int idx, float value) {
      if ( divCont!= null && isDeviationUpdateEnabled() )
        divCont.setExpansion(idx,value);
    }

    /**
     * Setzt die Anzeige Iterations-Variablen fuer alle Statusbalken (LUC-Typen).
     * @param value Array mit den Iterations-Variablen
     */
    public void setIterationVariable(float[] value) {
      if ( divCont!= null && isDeviationUpdateEnabled() )
        divCont.setIterVariable(value);
    }

    /**
     * Setzt die Anzeige Iterations-Variablen fuer einen Statusbalken (LUC-Typ).
     * @param idx   Index eines Balkens
     * @param value Wert der Iterations-Variablen
     */
    public void setIterationVariable(int idx, float value) {
      if ( divCont!= null && isDeviationUpdateEnabled() )
        divCont.setIterVariable(idx,value);
    }

    /**
     * Aktiviert oder Deaktiviert die Aktualisierung der Abweichungsbalken.
     */
    public void setDeviationUpdateEnabled(boolean enabled) {
      updateCheckBox.setSelected(enabled);
    }

    /**
     * Aktiviert oder Deaktiviert die Aktualisierung der Abweichungsbalken.
     */
    public boolean isDeviationUpdateEnabled() {
      return updateCheckBox.isSelected();
    }

    /**
     * Diese Klasse stellt ein Label dar, in dem (hinter einem Basistext)
     * ein numerischer Abweichungswert (z.B. erlaubte Maximalabweichung) angezeigt wird.
     * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
     * @version 1.0
     */
    private static class DeviationLabel extends JLabel {
      private String baseText        = "";
      private float dev              = 0;
      private NumberFormat numFormat = null;

      /**
       * Erzeugt ein neues leeres Label.
       */
      public DeviationLabel() {
        this("");
      }

      /**
       * Erzeugt ein neues Label. Der dargestellte Wert ist initial 0%.
       * @param text      Basistext
       */
      public DeviationLabel(String text) {
        this(text,new DecimalFormat("0.0%"));
      }

      /**
       * Erzeugt ein neues Label. Der dargestellte Wert ist initial 0.
       * @param text      Basistext
       * @param numFormat Anzeigeformat fuer den dargestellten Abweichungswert
       */
      public DeviationLabel(String text, NumberFormat numFormat) {
        this(text,numFormat,0);
      }

      /**
       * Erzeugt ein neues Label.
       * @param text      Basistext
       * @param numFormat Anzeigeformat fuer den dargestellten Abweichungswert
       * @param dev       initialer Wert fuer den Abweichungswert
       */
      public DeviationLabel(String text, NumberFormat numFormat, float dev) {
        super();
        this.baseText  = text;
        this.dev    = dev;
        setNumberFormat(numFormat);
        setText(baseText);
      }

      /**
       * Setzt den darzustellenden numerischen Abweichungswert.
       */
      public void setValue(float dev) {
        this.dev = dev;
        setText(baseText);
      }

      /**
       * Setzt den Basistext des Labels.
       */
      public void setText(String text) {
        this.baseText = text;
        String displayText = baseText;
        if ( numFormat != null )
          displayText = displayText+" "+numFormat.format(this.dev);
        super.setText(displayText);
      }

      /**
       * Setzt das Darstellungsformat fuer den Abweichungswert.
       */
      public void setNumberFormat(NumberFormat format) {
        this.numFormat = format;
      }
    }

    /**
     * Diese Klasse stellt einen Container dar, in dem vertikal eine Reihe von
     * {@linkplain ExpansionBar Status-Balken} angezeigt werden. Alle Balken
     * haben dieselben Extremwerte (Min, Max, Middle). Deshalb erhaelt nur
     * der unterste Balken eine Info-Leiste mit diesen Extremwerten.
     * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
     * @version 1.0
     */
    private static class DeviationContainer extends JComponent {
      private static final NumberFormat DEV_NUM_FORMAT = new DecimalFormat("0.0%");
      private static final NumberFormat ITER_NUM_FORMAT = new DecimalFormat("0.0000");
      /** Konstante fuer eine vertikale Ausrichtung der Status-Balken */
      public static final int VERTICAL   = ExpansionBar.VERTICAL;
      /** Konstante fuer eine horizontale Ausrichtung der Status-Balken */
      public static final int HORIZONTAL = ExpansionBar.HORIZONTAL;

      private ExpansionBar[] expansionBar   = null;
      private JLabel[]       expansionLabel = null;
      private JLabel[]       iterLabel      = null;
      private JLabel         typeHeader     = null;
      private JLabel         divHeader      = null;
      private JLabel         iterHeader     = null;
      private Font           sampleFont     = new JLabel().getFont();

      /**
       * Erzeugt einen neuen Container. Die Balken stellen eine prozentuale
       * Abweichung von +/- 100% um 0% dar. Die Balken erhalten keine Beschreibung,
       * sondern werden durchnummeriert.
       * @param count    Anzahl an Statusbalken im Conatiner
       * @param tol      Toleranzwert fuer alle Balken (siehe {@link ExpansionBar})
       */
      public DeviationContainer(int count, float tol) {
        this(count,-1,1,0,tol,DEV_NUM_FORMAT);
      }

      /**
       * Erzeugt einen neuen Container. Als Mitte-Wert fuer die Balken wird
       * 0 verwendet. Die Balken erhalten keine Beschreibung, sondern werden
       * durchnummeriert.
       * @param count    Anzahl an Statusbalken im Conatiner
       * @param min      Minimum-Wert fuer die Balken
       * @param max      Maximum-Wert fuer die Balken
       * @param tol      Toleranzwert fuer alle Balken (siehe {@link ExpansionBar})
       * @param form     Darstellungsformat fuer die Zahlen im Balken und in der
       *                 Info-Leiste des untersten Balken
       */
      public DeviationContainer(int count, float min, float max, float tol, NumberFormat form) {
        this(count,min,max,0,tol,form);
      }

      /**
       * Erzeugt einen neuen Container. Die Balken werden durchnummeriert.
       * @param count    Anzahl an Statusbalken im Conatiner
       * @param min      Minimum-Wert fuer die Balken
       * @param max      Maximum-Wert fuer die Balken
       * @param mid      Mitte-Wert fuer die Balken
       * @param tol      Toleranzwert fuer alle Balken (siehe {@link ExpansionBar})
       * @param form     Darstellungsformat fuer die Zahlen im Balken und in der
       *                 Info-Leiste des untersten Balken
       */
      public DeviationContainer(int count, float min, float max, float mid, float tol, NumberFormat form) {
        this(count,null,min,max,mid,tol,form,form);
      }

      /**
       * Erzeugt einen neuen Container.
       * @param count    Anzahl an Statusbalken im Conatiner
       * @param desc     Beschreibungen fuer die Balken (kann <code>null</code> sein!)
       * @param min      Minimum-Wert fuer die Balken
       * @param max      Maximum-Wert fuer die Balken
       * @param mid      Mitte-Wert fuer die Balken
       * @param tol      Toleranzwert fuer alle Balken (siehe {@link ExpansionBar})
       * @param barForm  Darstellungsformat fuer die Zahlen im Balken
       * @param infoForm Darstellungsformat fuer die Zahlen in der Info-Leiste
       *                 des untersten Balken
       */
      public DeviationContainer(int count, String[] desc, float min, float max, float mid, float tol, NumberFormat barForm, NumberFormat infoForm) {
        this(count,desc,min,max,mid,LangUtil.createArray(count,tol),barForm,infoForm);
      }

      /**
       * Erzeugt einen neuen Container.
       * @param count    Anzahl an Statusbalken im Conatiner
       * @param desc     Beschreibungen fuer die Balken (kann <code>null</code> sein!)
       * @param min      Minimum-Wert fuer die Balken
       * @param max      Maximum-Wert fuer die Balken
       * @param mid      Mitte-Wert fuer die Balken
       * @param tol      Toleranzwert fuer jeden einzelnen Balken (siehe {@link ExpansionBar})
       * @param barForm  Darstellungsformat fuer die Zahlen im Balken
       * @param infoForm Darstellungsformat fuer die Zahlen in der Info-Leiste
       *                 des untersten Balken
       */
      public DeviationContainer(int count, String[] desc, float min, float max, float mid, float[] tol, NumberFormat barForm, NumberFormat infoForm) {
        super();
        if ( count <= 0 )
          throw new IllegalArgumentException("At least one divergence bar must specified!");
        if ( desc!=null && desc.length < count )
          throw new IllegalArgumentException("Less describtions specified!");

        this.setLayout( new SpringLayout() );

        // PROBLEME mit den Labels bei VERTICAL
        // --> Label-Breite wird wohl zu 0
        int orient = HORIZONTAL;

        sampleFont = sampleFont.deriveFont(Font.BOLD);
        if (orient == VERTICAL)
          sampleFont = sampleFont.deriveFont( AffineTransform.getRotateInstance(Math.PI/2) );

        // Label erzeugen
        typeHeader = new JLabel( CLUE_RES.getString("Model.GUI.Type") );
        divHeader  = new JLabel( CLUE_RES.getString("Model.GUI.Divergence") );
        iterHeader = new JLabel( CLUE_RES.getString("Model.GUI.IterVar") );
        typeHeader.setFont(sampleFont);
        divHeader.setFont(sampleFont);
        iterHeader.setFont(sampleFont);

        // Statusbalken erzeugen
        this.expansionBar   = new ExpansionBar[count];
        this.expansionLabel = new JLabel[count];
        this.iterLabel      = new JLabel[count];
        for (int i=0; i<count; i++) {
          expansionBar[i]   = new ExpansionBar(orient,ExpansionBar.LINEAR,min,max,mid,-tol[i],tol[i],0,barForm,infoForm);
          expansionLabel[i] = new JLabel( desc!=null && i<desc.length ? desc[i] : String.valueOf(i));
          iterLabel[i]      = new JLabel("");
          if ( orient == HORIZONTAL ) {
            expansionLabel[i].setHorizontalAlignment(JLabel.CENTER);
            expansionLabel[i].setVerticalAlignment(JLabel.TOP);
          } else {
            expansionLabel[i].setFont( expansionLabel[i].getFont().deriveFont( AffineTransform.getRotateInstance(Math.PI/2) ) );
            expansionLabel[i].setHorizontalAlignment(JLabel.LEFT);
            expansionLabel[i].setVerticalAlignment(JLabel.CENTER);
          }
        }
        // beim untersten Balken sollen die Labels erscheinen
        expansionBar[count-1].setValueLabelsPainted(true);

        // Komponenten einfuegen und ausrichten
        if ( orient == HORIZONTAL ) {
          typeHeader.setHorizontalAlignment(JLabel.CENTER);
          typeHeader.setVerticalAlignment(JLabel.BOTTOM);
          divHeader.setHorizontalAlignment(JLabel.CENTER);
          divHeader.setVerticalAlignment(JLabel.BOTTOM);
          iterHeader.setHorizontalAlignment(JLabel.CENTER);
          iterHeader.setVerticalAlignment(JLabel.BOTTOM);
          add( typeHeader );
          add( divHeader );
          add( iterHeader );
          for (int i=0; i<count; i++) {
            add( expansionLabel[i] );
            add( expansionBar[i] );
            add( iterLabel[i] );
          }
          // Layout ausrichten
          SpringUtilities.makeCompactGrid(this,getComponentCount()/3,3,0,0,10,0);
        } else {
          typeHeader.setHorizontalAlignment(JLabel.CENTER);
          typeHeader.setVerticalAlignment(JLabel.CENTER);
          divHeader.setHorizontalAlignment(JLabel.CENTER);
          divHeader.setVerticalAlignment(JLabel.CENTER);
          iterHeader.setHorizontalAlignment(JLabel.CENTER);
          iterHeader.setVerticalAlignment(JLabel.CENTER);
          add( divHeader );
          for (int i=0; i<count; i++)
            add( expansionBar[i] );
          add( typeHeader );
          for (int i=0; i<count; i++)
            add( expansionLabel[i] );
          add( iterHeader );
          for (int i=0; i<count; i++)
            add( iterLabel[i] );
          // Layout ausrichten
          SpringUtilities.makeCompactGrid(this,3,getComponentCount()/3,0,0,10,0);
        }
      }

      /**
       * Liefert die Anzahl an Balken.
       */
      public int getCount() {
        return expansionBar.length;
      }

      /**
       * Setzt den Toleranzwert fuer alle Balken.
       */
      public void setTolerance(float tol) {
        for (int i=0; i<expansionBar.length; i++)
          setTolerance(i,tol);
      }

      /**
       * Setzt den Toleranzwert fuer einen Balken.
       */
      public void setTolerance(int i, float tol) {
        expansionBar[i].setMinimumTolerance( -tol );
        expansionBar[i].setMaximumTolerance(  tol );
      }

      /**
       * Setzt den Wert fuer alle Balken.
       */
      public void setExpansion(float[] value) {
        for (int i=0; i<value.length; i++)
          setExpansion(i,value[i]);
      }

      /**
       * Setzt den Wert fuer einen Balken.
       */
      public void setExpansion(int idx, float value) {
        expansionBar[idx].setValue(value);
      }

      /**
       * Setzt die Anzeige der Iterations-Variablen fuer alle Balken (LUC-Typ).
       */
      public void setIterVariable(float[] value) {
        for (int i=0; i<value.length; i++)
          setIterVariable(i,value[i]);
      }

      /**
       * Setzt die Anzeige der Iterations-Variablen fuer einen Balken (LUC-Typ).
       */
      public void setIterVariable(int idx, float value) {
        iterLabel[idx].setText( ITER_NUM_FORMAT.format( value ) );
      }

      /**
       * Setzt die Beschreibung fuer einen Balken.
       */
      public void setBarDesc(int idx, String text) {
        expansionLabel[idx].setText(text);
      }

      /**
       * Setzt die Beschreibung die oberhalb der Balken-Beschreibungen steht.
       */
      public void setBarDescHeader(String text) {
        typeHeader.setText(text);
      }

      /**
       * Setzt die Beschreibung die oberhalb der Balken steht.
       */
      public void setBarHeader(String text) {
        divHeader.setText(text);
      }

      /**
       * Setzt die Beschreibung die oberhalb der Iterations-Variablen steht.
       */
      public void setIterHeader(String text) {
        iterHeader.setText(text);
      }

    }
  }
}
