package edu.bonn.xulu.plugin.model.parallel.demo;

import java.awt.Rectangle;

import schmitzm.data.WritableGrid;
import appl.parallel.spmd.AbstractSPMDTask;
import appl.parallel.spmd.AdvancedSPMDServerInterface;
import appl.parallel.spmd.SPMDServerInterface;
import appl.parallel.spmd.SPMDTask;
import appl.parallel.spmd.SyncPoint;
import appl.util.RasterUtil;

/**
 * This is the {@link SPMDTask} which will be executed on server side by
 * the parallel model {@link AverageNeighborhoodParallelDemoModel}. There are
 * only very little changes to the serial code you have seen in
 * {@link AverageNeighborhoodSerialDemoModel}. Basically you have to retrieve
 * the partitions you have submitted to Xulu / V control on client side and use
 * the copy paste the sequential code. One important change has to be made
 * though: you now must iterate over the so called calculation area of the
 * partition. This becomes even more important when use multithreading to
 * calculate on many calculation areas at once.
 * <p/>
 * <strong>Multithreading is enabled!</strong>
 * <p>
 * Preloading is also used, but the effect is in this case very small (because data is 
 * only loaded once and not every step).
 * 
 * @see SPMDTask
 * @see AverageNeighborhoodParallelDemoModel
 * @see AVNTuned
 * @author Dominik Appl
 */
public class AVNTunedTask extends AbstractSPMDTask {

	private static final long serialVersionUID = 13L;

	AdvancedSPMDServerInterface controller;

	int neighborhoodRange;

	WritableGrid gridOutput;

	WritableGrid gridInput;

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.spmd.AbstractSPMDTask#init()
	 */
	@Override
	public void init() {
		//get the server controller for Xulu / V access
		controller = getAdvancedSPMDServerController();

		/***********************************************************************
		 * TUNING: we can load the grids concurrently using syncPoints. Of
		 * course in case of this simple algorithm with only two grids this will
		 * not make much difference. Normaly you would use SyncPoints when you
		 * have many grids or want to preload data for the NEXT step. 
		 * Look at the parallel Clue model for an example.
		 **********************************************************************/
		SyncPoint loadPoint = new SyncPoint(1, SyncPoint.Priority.HIGH);
		controller.preloadPartition("inputGrid", loadPoint);
		controller.preloadPartition("outputGrid", loadPoint);
		// wait for the loading to be finished
		controller.synchronizeToSyncPoint(loadPoint);

		// Now the data is loaded and the following 2 assignments are no-time
		// operations
		gridInput = (WritableGrid) controller.getPartition("inputGrid");
		gridOutput = (WritableGrid) controller.getPartition("outputGrid");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.spmd.SPMDTask#run()
	 */
	public Object run(Object... parameters) {

		//neighborhood range is given as a parameter
		neighborhoodRange = (Integer) parameters[0];

		/***********************************************************************
		 * START REMOTE ALGORITHM 
		 * Notice that only the outer loop variables have changed to the serial version!!!
		 **********************************************************************/
		float overallSum = 0;
		Rectangle partition = controller.getLocalBounds();
		//iterate over all cells
		for (int y = controller.getLocalCalcMinY(); y < controller
				.getLocalCalcMaxY() + 1; y++)
			for (int x = controller.getLocalCalcMinX(); x < controller
					.getLocalCalcMaxX() + 1; x++) {
				float tmp = gridInput.getRasterSampleAsFloat(x, y);
				if (!Float.isNaN(tmp))
					overallSum += tmp;
				// the local sum is simply the sum over all elements in the
				// neighborhood
				float localSum = 0;
				// number of cells over which the sum is calculated
				int noOfCells = 0;
				// for each cell: calculate the sum of all neighbors
				for (int y2 = y - neighborhoodRange; y2 <= y
						+ neighborhoodRange; y2++)
					for (int x2 = x - neighborhoodRange; x2 <= x
							+ neighborhoodRange; x2++)
						// check if the coordinates are valid (inside the grid
						// and not NaN)
						if (partition.contains(x2, y2))
							if (!Float.isNaN(gridInput.getRasterSampleAsFloat(
									x2, y2))) {
								localSum += gridInput.getRasterSampleAsFloat(
										x2, y2);
								noOfCells++;
							}
				gridOutput
						.setRasterSample(localSum / (float) (noOfCells), x, y);

			}
		// copy values into results from output to input (for next step)
		RasterUtil.copyInto(gridOutput, gridInput);
		return overallSum;
	}

	/* (non-Javadoc)
	 * @see appl.parallel.spmd.AbstractSPMDTask#supportsMultiThreading()
	 */
	@Override
	public boolean supportsMultiThreading() {
		return true;
	}

}