/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.vis;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import org.geotools.feature.FeatureCollection;
import org.geotools.map.MapLayer;
import org.geotools.styling.Style;

import schmitzm.data.property.ValueProperty;
import schmitzm.geotools.gui.GeoMapPane;
import schmitzm.geotools.gui.JEditorPane;
import schmitzm.geotools.gui.LayeredEditorFrame;
import schmitzm.geotools.gui.LayeredMapPane;
import schmitzm.geotools.map.event.JMapPaneEvent;
import schmitzm.geotools.map.event.JMapPaneListener;
import schmitzm.geotools.map.event.LayerEditFinishedEvent;
import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.appl.DataPool;
import edu.bonn.xulu.appl.XuluConstants;
import edu.bonn.xulu.data.XuluDataException;
import edu.bonn.xulu.data.XuluObject;
import edu.bonn.xulu.gui.XuluGUIMessages;
import edu.bonn.xulu.io.InstantiationFactory;
import edu.bonn.xulu.plugin.data.feature.SingleFeatureCollection;

/**
 * Diese Klasse stellt einen Editor fuer geografische Vektor-Daten dar.
 * Neben Vektor-Daten koennen auch Raster angezeigt werden.
 * Gesteuert wird der Editor durch eine {@linkplain MapEditorToolBar Button-Leiste}.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class GTEditorTool extends GTVisualisationTool implements JMapPaneListener {
  /** Xulu-Instanz fuer die der Editor erstellt wurde. */ 
  protected XuluModellingPlatform appl = null;
  
  /**
   * Erstellt ein neues Tool. Erweitert das Fenster um eine Button-Leiste
   * zur Steuerung des Editors. 
   */
  public GTEditorTool() {
    super( new LayeredEditorFrame(
                new LayeredMapPane(
                    new GeoMapPane(
                        new JEditorPane(),
                        null,
                        null,
                        null,
                        null
                    )
                )
           ),
           XuluConstants.XULU_VIS_RES.getString("GTEditorTool.FrameTitle")
    );
    // set listener to recognize new layers
    ((JEditorPane)this.layeredMapPane.getMapPane()).addMapPaneListener(this);
  }
  
  /**
   * Speichert die Xulu-Instanz, um in den Datenpool die neu erstellten
   * Objekte einzufuegen.
   * @param appl Xulu-Instanz
   */
  public void initialize(XuluModellingPlatform appl) {
    super.initialize(appl);
    this.appl = appl;
  }

  /**
   * Lauscht auf das {@link JEditorPane}, um nach Beendigung eines Layers,
   * dieses in den Xulu-Datenpool einzufuegen. 
   */
  public void performMapPaneEvent(JMapPaneEvent e) {
    try {
      if ( e instanceof LayerEditFinishedEvent ) {
        LayerEditFinishedEvent lefe   = (LayerEditFinishedEvent)e;
        JEditorPane            editor = (JEditorPane)lefe.getSource();
        JEditorPane.EditorMode mode   = editor.getEditorMode();
        // neues Objekt im Xulu-Datenpool nur, wenn komplett neues Layer
        // erstellt wurde
        if ( JEditorPane.EditorMode.New_Point.equals(mode) ||
             JEditorPane.EditorMode.New_Line.equals(mode) ||
             JEditorPane.EditorMode.New_Polygon.equals(mode) )
          createNewXuluObjectInDatapool(editor);
      }
    } catch (Exception err) {
      XuluGUIMessages.showError(layeredMapPane, err);
    }
  }
    
  /**
   * Erstellt eine {@link SingleFeatureCollection} fuer eine {@link FeatureCollection}
   * und fuegt es in den {@linkplain DataPool Xulu-Datenpool} ein.
   * @param name Bezeichnung fuer das neue Objekt im Datenpool
   * @param fc   neue {@link FeatureCollection}   
   */
  protected void createNewXuluObjectInDatapool(JEditorPane editor) throws Exception {
    FeatureCollection  fc    = editor.getEditorFeatureCollection();
    MapLayer           layer = editor.getEditorLayer();
    String             title = layer.getTitle();
    Style              style = layer.getStyle();
    
    // Xulu-Objekt erstellen und mit FeatureCollection befuellen
    InstantiationFactory fac = appl.getRegistry().getDefaultFactory( SingleFeatureCollection.class );
    if ( fac == null )
      throw new UnsupportedOperationException("No factory registered in XuluRegistry for type SingleFeatureCollection");
    SingleFeatureCollection sfc = (SingleFeatureCollection)fac.newInstance(false);
    sfc.setFeatureCollection(fc);
    sfc.setDescription( appl.getDataPool().createUniqueDescription(title) );
    ((ValueProperty)sfc.getProperty( SingleFeatureCollection.PROP_STYLE )).getOneTimeWriteAccess().setValue(style);

    // Objekt in den Datenpool einfuegen
    appl.getDataPool().addObject(sfc);
    
    // Damit Layer auf Datenpool-Aktionen (z.B. Loeschen)
    // reagiert, muessen diverse Listener gesetzt werden
    // -> Editor-Layer entfernen
    // -> neues Layer als Xulu-Objekt neu in GTEditorTool
    //    einfuegen
    editor.getContext().removeLayer( layer );
    this.add(sfc, sfc.getDescription());
  }
  
}
