/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.vis;

import java.util.Vector;
import java.awt.Color;
import java.awt.Point;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.awt.image.DataBufferFloat;
import java.awt.image.ComponentSampleModel;
import javax.media.jai.RasterFactory;

import schmitzm.lang.LangUtil;
import schmitzm.data.WritableGrid;
import schmitzm.data.WritableGridArray;
import schmitzm.data.property.Property;
import schmitzm.data.property.Properties;
import schmitzm.data.property.DynamicProperties;
import schmitzm.data.property.ValueProperty;
import schmitzm.data.property.ScalarProperty;
import schmitzm.data.property.ListProperty;
import schmitzm.data.property.ListPropertyReadAccess;
import schmitzm.data.event.ObjectTraceable;
import schmitzm.data.event.ObjectListener;
import schmitzm.data.event.ObjectEvent;
import schmitzm.data.event.ObjectChangeEvent;
import schmitzm.data.event.NameChangeEvent;
import schmitzm.geotools.GTUtil;
import schmitzm.geotools.gui.LayeredMapFrame;
import schmitzm.geotools.gui.LayeredMapPane;
import schmitzm.geotools.grid.WritableGridCoverage;
import schmitzm.geotools.grid.GridUtil;

import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.appl.VisualisationTool;
import edu.bonn.xulu.appl.VisualisationUpdateListener;
import edu.bonn.xulu.data.XuluObject;
import edu.bonn.xulu.appl.XuluConstants;

import org.geotools.map.MapContext;
import org.geotools.map.MapLayer;
import org.geotools.coverage.grid.GridCoverageFactory;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.feature.FeatureCollection;
import org.geotools.geometry.Envelope2D;
import org.geotools.geometry.GeneralEnvelope;
import org.geotools.coverage.GridSampleDimension;
import org.geotools.coverage.Category;
import org.geotools.styling.Style;

import appl.data.DataProxy;
import appl.data.LateLoadingProxy;
import appl.data.LoadingException;

// nur fuer Doku
import schmitzm.data.WritableGridRaster;
import skrueger.geotools.StyledLayerStyle;

/**
 * Diese Klasse stellt ein Visualisierungstool fuer geografische Raster- und
 * Vektor-Daten dar. Es basiert auf {@link LayeredMapFrame} und kann Layer fuer
 * folgende Datentypen darstellen:
 * <ul>
 *   <li>{@link GridCoverage2D org.geotools.coverage.grid.GridCoverage2D}</li>
 *   <li>{@link FeatureCollection org.geotools.feature.FeatureCollection}</li>
 *   <li><u>gleichzeitig</u> {@link WritableRaster java.awt.image.WritableRaster}
 *       und {@link WritableGrid schmitzm.data.WritableGrid} (z.B. {@link WritableGridRaster})</li>
 *   <li>{@linkplain Property Xulu-Properties}, die einen der obigen Typen
 *       beinhalten.</li>
 *   <li>{@linkplain XuluObject Xulu-Objekte}, die mindestens eine Property
 *       der obigen Typen besitzen.</li>
 * </ul>
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class GTVisualisationTool /*extends LayeredMapFrame*/ implements VisualisationTool {
  /** Zaehlt die Instanzen (nur fuer den Titel) */
  protected static int instanceCount = 1;
  /** Speichert die verteilten UpdateListener */
  protected Vector<VisualisationUpdateListener> updateListener = new Vector<VisualisationUpdateListener>();
  
  /** Fenster mit Kartenbereich und Layer-Liste */
  protected LayeredMapFrame mapFrame = null;
  /** Kartenbereich und Layer-Liste */
  protected LayeredMapPane  layeredMapPane = null;

  /**
   * Erzeugt ein neues <code>GTVisualisationTool</code>-Fenster.
   */
  protected GTVisualisationTool(LayeredMapFrame frame, String title) {
    super();
    this.mapFrame       = frame != null ? frame : new LayeredMapFrame();
    this.layeredMapPane = mapFrame.getLayeredMapPane();
    this.mapFrame.setIconImage( XuluConstants.XULU_ICON.getImage() );
    setTitle(title+" (".concat(String.valueOf(instanceCount++)).concat(")"));
  }

  /**
   * Erzeugt ein neues <code>GTVisualisationTool</code>-Fenster.
   */
  public GTVisualisationTool() {
    this( null, XuluConstants.XULU_VIS_RES.getString("GTVisualisationTool.FrameTitle") );
  }
  
  /**
   * Macht nichts, da die Xulu-Applikation nicht fuer die Visualisierung
   * benoetigt wird.
   */
  public void initialize(XuluModellingPlatform appl) {
  }

  /**
   * Verbirgt das <code>GTVisualisationTool</code>-Fenster, entfernt alle (durch
   * <code>UpdateListener</code>) damit verbundenen Layer und "vernichtet"
   * anschliessend das Fenster mit {@link #dispose()}.
   */
  public void close() {
    // Fenster unsichtbar machen
    setVisible(false);
    // Alle Layer schliessen
    for (int i=0; i<updateListener.size(); i++)
      ((VisualisationUpdateListener)updateListener.elementAt(i)).close();
    // Frame "vernichten"
    mapFrame.dispose();
  }
  
  /**
   * Zeigt oder verbirgt das Fenster.
   */
  public void setVisible(boolean visible) {
    mapFrame.setVisible( visible );
  }

  /**
   * Prueft, ob das Fenster angezeigt wird.
   */
  public boolean isVisible() {
    return mapFrame.isVisible();
  }
  
  /**
   * Setzt den Titel des Fensters.
   */
  public void setTitle(String title) {
    mapFrame.setTitle(title);
  }

  /**
   * Liefert den Titel des Fensters.
   */
  public String getTitle() {
    return mapFrame.getTitle();
  }
  
  /**
   * Liefert das Visualisierungsfenster.
   */
  public LayeredMapFrame getLayeredMapFrame() {
    return mapFrame;
  }

  /**
   * Prueft, ob Instanzen einer bestimmten Klasse visualisiert werden koennen.
   * Neben den direkt von der Oberklasse {@link LayeredMapPane#isVisualisable(Class) LayeredMapPane}
   * darstellbaren Klassen, kann das <code>GTVisualisationTool</code> Instanzen
   * folgende Klassen visualisieren:
   * <ul>
   *   <li><u>gleichzeitig</u> {@link WritableRaster} und {@link WritableGrid} (z.B. {@link WritableGridRaster})</li>
   *   <li>{@link WritableGridArray}</li>
   * </ul>
   */
  public boolean isVisualisable(Class c) {
    return layeredMapPane.isVisualisable( c ) ||
           // Aus einem Objekt, das WritableGrid und WritableRaster implementiert
           // kann ein GridCoverage2D erzeugt werden
           WritableRaster.class.isAssignableFrom( c ) && WritableGrid.class.isAssignableFrom( c ) ||
           // Aus einem Objekt, das WritableGridArray implementiert
           // kann ein GridCoverage2D erzeugt werden
           WritableGridArray.class.isAssignableFrom( c );
  }

  /**
   * Ermittelt alle darstellbaren Eigenschaften eines aus Properties
   * bestehenden Objekts.
   */
  private Vector<ValueProperty> determineVisualisableProperties(Properties obj) {
    Vector visProp = new Vector<ValueProperty> ();
    // Alle skalaren Propertys ermitteln
    Property[] allProp = (Property[]) obj.getProperties(ScalarProperty.class).getProperties();
    // Alle darstellbaren skalaren Propertys ermitteln
    for (int i = 0; i < allProp.length; i++)
      if ( /*super.*/isVisualisable( ( (ScalarProperty) allProp[i]).getOneTimeReadAccess().getValue()))
        visProp.add(allProp[i]);

    // Alle Listen ermitteln
    allProp = (Property[]) obj.getProperties(ListProperty.class).getProperties();
    // Alle darstellbaren Listen-Properties ermitteln
    for (int i = 0; i < allProp.length; i++)
// Bei Listen haengt die Darstellbarkeit vom jeweiligen Element ab. Sie kann
// nicht ueber den allgemeinen Listen-Typ geprueft werden (z.B. Listentyp
// WritableGrid nicht darstellbar, obwohl Instanzen von WritableGridCoverage
// in der Liste)!
// Deshalb wird dies erst beim Einfuegen (Layer-Erzeugen) gecheckt!
//    if ( super.isVisualisable( ((ListProperty)allProp[i]).getPropertyType().getType() ) )
      visProp.add(allProp[i]);

    return visProp;
  }

  /**
   * Prueft, ob ein Objekt visualisiert werden kann. Dies ist der Fall,
   * wenn es
   * <ul>
   *   <li>direkt von {@link LayeredMapPane#isVisualisable(Class) LayeredMapPane}
   *       dargestellt werden kann</li>
   *   <li><u>gleichzeitig</u> {@link WritableRaster} und {@link WritableGrid}
   *       implementiert (z.B. {@link WritableGridRaster})
   *   <li>sich um ein {@link XuluObject} handelt, das mind. eine darstellbare
   *       Property beinhaltet</li>
   * </ul>
   * @param obj ein Objekt
   */
  public boolean isVisualisable(Object obj) {
    obj = extractRealObject(obj);
    if ( obj==null )
      return false;
    if ( isVisualisable( obj.getClass() ) )
      return true;
    // Pruefen, ob Objekt mind. eine darstellbaren Property hat
    if ( obj instanceof Properties )
      return determineVisualisableProperties((Properties)obj).size() > 0;
    return false;
  }

  /**
   * Fuegt ein Layer (als oberstes Layer) ein. Instanzen,
   * <ul>
   *   <li>die <u>gleichzeitig</u> {@link WritableRaster} und {@link WritableGrid}
   *       implementieren (z.B. {@link WritableGridRaster})</li>
   *   <li>von {@link WritableGridArray}</li>
   * </ul>
   * werden umgewandelt in ein Geotools-{@link GridCoverage2D}.
   * @param obj ein (darstellbares) Objekt
   * @param desc Beschreibung fuer das Objekt
   * @param style Darstellungs-Style fuer das Layer
   * @exception UnsupportedOperationException falls ein nicht-darstellbares
   *            Objekt uebergeben wird
   * @see #isVisualisable(Object)
   * @see #isVisualisable(Class)
   */
  protected MapLayer addLayer(Object obj, String desc, Style style) {
    obj = extractRealObject(obj);
    try {
      obj = GridUtil.convertToGridCoverage2D(obj);
    } catch ( IllegalArgumentException err ) {
      // Objekt kann nicht in GridCoverage2D umgewandelt werden
    }

    return layeredMapPane.addLayer(obj,desc,style);
  }

  /**
   * Visualisiert ein Objekt. Handelt es sich bei dem Objekt um ein
   * darstellbares Xulu-Objekt mit mehreren darstellbaren Eigenschaften,
   * wird der Beschreibung der Property-Name hinzugefuegt.
   * @param obj  Objekt
   * @param desc Beschreibung zu dem Objekt, die in der Layer-Uebersicht
   *             angezeigt wird (kann <code>null</code> sein)
   * @exception java.lang.UnsupportedOperationException falls das Objekt nicht
   *            visualisiert werden kann
   */
  public boolean add(Object obj, String desc) {
    obj = extractRealObject(obj);

    VisualisationUpdateListener newListener = null;

    // Style aus XuluObject ermitteln
    Style style = determineStyleFromObject(obj);

    // Objekt ist direkt visualisierbar
    if ( layeredMapPane.isVisualisable(obj) )
      newListener = new UpdateListener(this,obj,this.addLayer(obj,desc,style));
    else
      // von einem Xulu-Objekt werden alle darstellbaren Propertys dargestellt
      // > Proxy-UpdateListener
      if ( obj instanceof Properties ) {
        newListener = new UpdateListenerProxy(this, obj, "");
        ((UpdateListenerProxy)newListener).reorganize();
      }
    // UpdateListener der Liste hinzufuegen
    if ( newListener != null ) {
      updateListener.add(newListener);
      return true;
    }

    // Objekt konnte nicht dargestellt werden
    if ( obj==null )
      throw new UnsupportedOperationException( getClass().getSimpleName().concat(" can not visualise null-Objects!") );
    throw new UnsupportedOperationException( getClass().getSimpleName().concat(" can not visualise instances of ").concat(obj.getClass().getName()) );
  }

  /**
   * Schliesst alle Visualisierungen eines Objekts.
   * @param obj zu entfernendes Objekt
   * @return <code>false</code> falls das Objekt nicht (mehr) visualisiert wird
   */
  public boolean remove(Object obj) {
    boolean objRemoved = false;
    for (int i=updateListener.size()-1; i>=0; i--) {
      VisualisationUpdateListener l = (VisualisationUpdateListener)updateListener.elementAt(i);
      if ( l.getSourceObject().equals(obj) ) {
        l.close();
        updateListener.remove(l);
        objRemoved = true;
      }
    }
    return objRemoved;
  }

  /**
   * Aktualisiert alle Visualisierungen eines Objekts.
   * @param obj zu aktualisierendes Objekt
   * @return <code>false</code> falls das Objekt nicht (mehr) visualisiert wird
   */
  public boolean update(Object obj) {
    boolean objUpdated = false;
    for (int i=updateListener.size()-1; i>=0; i--) {
      VisualisationUpdateListener l = updateListener.elementAt(i);
      if ( l.getSourceObject().equals(obj) ) {
        l.refresh();
        objUpdated = true;
      }
    }
    return objUpdated;
  }

  /**
   * If the given Object is a is an instance of <code>DataProxy</code> then
   * this method returns the real encapulated Object. If the Object implements
   * late loading functionality the Object is first loaded into memory.
   */
  public Object extractRealObject(Object obj) {
    if (obj != null)
      if (obj instanceof LateLoadingProxy) {
        LateLoadingProxy proxy = (LateLoadingProxy) obj;
        if (!proxy.isLoaded())
          try {
            proxy.loadData();
          }
          catch (LoadingException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
          }
      }
    if (obj instanceof DataProxy)
      obj = ( (DataProxy) obj).getProxiedObject();
    return obj;
  }

  protected static Style determineStyleFromObject(Object obj) {
    Style style = null;
    // Style aus XuluObject ermitteln
    if ( obj instanceof XuluObject )
      style = determineStyleFromObject( ((XuluObject)obj).getProperty( XuluObject.PROP_STYLE ) );
    else if ( obj instanceof ScalarProperty ) {
      Object styleObj = ((ScalarProperty)obj).getOneTimeReadAccess().getValue();
      if ( styleObj != null ) {
        if ( styleObj instanceof Style )
          style = (Style)styleObj;
        else if ( styleObj instanceof StyledLayerStyle )
          style = ((StyledLayerStyle)styleObj).getGeoObjectStyle();
      }
    }
    return style;
  }

  /**
   * Dieser UpdateListener verknuepft ein Objekt mit einem oder mehreren
   * Layern einer {@link LayeredMapFrame}-Instanz. Er kann unmittelbar als
   * {@link schmitzm.data.event.ObjectListener} eingesetzt werden, um
   * mit der Objekt-Aenderung automatisch die Visualisierung zu aktualisieren.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  protected class UpdateListener extends VisualisationUpdateListener {
    private MapLayer[] layer      = null;
    private MapContext mapContext = null;

    /**
     * Erzeugt einen neuen <code>UpdateListener</code>.
     * @param tool  Visualisierungstool, mit dem der Listener verbunden ist
     * @param obj   Objekt welches visualisiert wird
     * @param layer Layer des {@link LayeredMapFrame} in denen das Objekt dargestellt ist
     * @deprecated Ein UpdateListener soll nur noch fuer ein einziges Layer
     *             zustaendig sein. In einer zukuenftigen Version wird dieser
     *             Konstruktor entfernt!!
     */
    public UpdateListener(GTVisualisationTool tool, Object obj, MapLayer[] layer) {
      super(tool,obj);
      this.layer = layer;
      this.mapContext = tool.layeredMapPane.getMapPane().getContext();
    }

    /**
     * Erzeugt einen neuen <code>UpdateListener</code>.
     * @param tool  Visualisierungstool, mit dem der Listener verbunden ist
     * @param obj   Objekt welches visualisiert wird
     * @param layer Layer des {@link LayeredMapFrame} in dem das Objekt dargestellt ist
     */
    public UpdateListener(GTVisualisationTool tool, Object obj, MapLayer layer) {
      this(tool,obj,new MapLayer[] {layer});
    }

    /**
     * Entfernt alle Layer, die zum Objekts gehoeren aus dem {@link LayeredMapFrame}.
     */
    public void close() {
      for (int i=0; i<layer.length; i++)
        // Layer kann schon nicht mehr existieren, wenn es bereits manuell
        // geschlossen wurde!
        if (layer[i] != null)
          mapContext.removeLayer(layer[i]);
      ((GTVisualisationTool)this.visTool).layeredMapPane.getMapPane().refresh();
    }

    /**
     * Aktualisiert alle Layer im {@link LayeredMapFrame}.
     */
    public void refresh() {
      // Aktualierung eines einzelnen Layers nicht moeglich
      // --> alle Layer aktualisieren
      ((GTVisualisationTool)this.visTool).layeredMapPane.getMapPane().refresh();

//      for (int i=0; i<layer.length; i++)
//        // Layer kann schon nicht mehr existieren, wenn es bereits manuell
//        // geschlossen wurde!
//        if (layer[i] != null) {
//          // Update eines einzelnen Layers hier implementierten
//        }
    }
  }


  /**
   * Dieser UpdateListener verknuepft ein Objekt mit einem oder mehreren
   * anderen UpdateListenern und alle alle Funktionsaufrufe an diese weiter.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  protected class UpdateListenerProxy extends VisualisationUpdateListener {
    private Vector<VisualisationUpdateListener> listener = null;
    private String                              desc     = "";
    private Style                               style    = null;
    private boolean                             visObjectStructureChanged = true;
    private ObjectListener                      objectListener = new ObjectListener() {
      public void performObjectEvent(ObjectEvent e) {
        if ( !(e instanceof NameChangeEvent) )
          visObjectStructureChanged = true;
      }
    };

    /**
     * Erzeugt einen neuen <code>UpdateListenerProxy</code>, dem noch keine
     * Listener zugeordnet sind.
     * @param tool     Visualisierungstool, mit dem der Listener verbunden ist
     * @param obj      Objekt welches visualisiert wird
     * @param desc     Beschreibung fuer das Objekt (wird fuer die Bezeichnung
     *                 des Layers benoetigt)
     * @param style    Darstellungsstyle fuer neue Layer (wenn {@code null}, wird
     *                 versucht, den Style aus dem Object (z.B. {@link XuluObject})
     *                 zu ermitteln
     */
    public UpdateListenerProxy(GTVisualisationTool tool, Object obj, String desc, Style style) {
      super(tool,obj);
      this.listener = new Vector<VisualisationUpdateListener>();
      this.desc     = desc;
      // wenn kein Style angegeben wurde, versuchen, diesen aus dem Objekt
      // zu ermitteln
      if ( style == null )
        style = determineStyleFromObject(obj);
      this.style = style;
    }

    /**
     * Erzeugt einen neuen <code>UpdateListenerProxy</code>, dem noch keine
     * Listener zugeordnet sind. Der Style wird - sofern moeglich - aus dem Objekt
     * (z.B. {@link XuluObject}) ermittelt.
     * @param tool     Visualisierungstool, mit dem der Listener verbunden ist
     * @param obj      Objekt welches visualisiert wird
     * @param desc     Beschreibung fuer das Objekt (wird fuer die Bezeichnung
     *                 des Layers benoetigt)
     */
    public UpdateListenerProxy(GTVisualisationTool tool, Object obj, String desc) {
      this(tool, obj, desc, (Style)null);
    }

    /**
     * Erzeugt einen neuen <code>UpdateListenerProxy</code>.
     * @param tool     Visualisierungstool, mit dem der Listener verbunden ist
     * @param obj      Objekt welches visualisiert wird
     * @param desc     Beschreibung fuer das Objekt (wird fuer die Bezeichnung
     *                 des Layers benoetigt)
     * @param style    Darstellungsstyle fuer neue Layer (wenn {@code null}, wird
     *                 versucht, den Style aus dem Object (z.B. {@link XuluObject})
     *                 zu ermitteln
     * @param listener anderer VisualisationUpdateListener an den die Ereignisse
     *                 weitergegeben werden
     */
    public UpdateListenerProxy(GTVisualisationTool tool, Object obj, String desc, Style style, VisualisationUpdateListener listener) {
      this(tool,obj,desc,style);
      add( listener );
    }

    /**
     * Erzeugt einen neuen <code>UpdateListenerProxy</code>. Der Style wird - sofern
     * moeglich - aus dem Objekt (z.B. {@link XuluObject}) ermittelt.
     * @param tool     Visualisierungstool, mit dem der Listener verbunden ist
     * @param obj      Objekt welches visualisiert wird
     * @param desc     Beschreibung fuer das Objekt (wird fuer die Bezeichnung
     *                 des Layers benoetigt)
     * @param listener anderer VisualisationUpdateListener an den die Ereignisse
     *                 weitergegeben werden
     */
    public UpdateListenerProxy(GTVisualisationTool tool, Object obj, String desc, VisualisationUpdateListener listener) {
      this(tool, obj,desc, null, listener);
    }
    /**
     * Liefert die Anzahl der im Proxy enthaltenen Listener.
     */
    public int getSize() {
      return this.listener.size();
    }

    /**
     * Fuegt dem Proxy einen UpdateListener hinzu.
     * @param l neuer UpdateListener
     */
    public void add(VisualisationUpdateListener l) {
      this.listener.add( l );
    }

    /**
     * Entfernt einen UpdateListener hinzu vom Proxy.
     * @param l neuer UpdateListener
     */
    public void remove(VisualisationUpdateListener l) {
      this.listener.remove( l );
    }

    /**
     * Entfernt einen UpdateListener hinzu vom Proxy.
     * @param i Index des zu entfernenden UpdateListener
     * @return den entfernten UpdateListener
     */
    public VisualisationUpdateListener remove(int i) {
      return this.listener.remove( i );
    }

    /**
     * Ruft {@code close()} fuer alle Listener des Proxy auf.
     */
    public void close() {
      for (int i=0; i<listener.size(); i++) {
        VisualisationUpdateListener l = listener.elementAt(i);
        // Listener kann schon nicht mehr existieren
        if (l != null)
          l.close();
      }
    }

    /**
     * Reorganisiert die Listener. Wenn sich die Struktur des Basis-Objekts
     * geaendert hat (z.B. wenn einem Xulu-Objekt eine visualisierbare Property
     * hinzugefuegt wurde), werden entsprechende Listener hinzugefuegt.
     */
    public void reorganize() {
      // Flag zuruecksetzen, da Struktur nun aktualisiert wird
      visObjectStructureChanged = false;

      // falls es sich bei dem dargestellten Objekt um ein ObjectTraceable
      // handelt, wird ein Listener hinzugefuegt, damit Wert-Aenderungen
      // gemeldet werden
      if ( this.visObject instanceof ObjectTraceable && !((ObjectTraceable)visObject).containsObjectListener(objectListener) )
        ((ObjectTraceable)visObject).addObjectListener(objectListener);

      // ListenerProxy fuer eine ScalarProperty erzeugen
      if ( this.visObject instanceof ScalarProperty ) {
        ScalarProperty propObj = (ScalarProperty)this.visObject;
        Object         obj     = propObj.getOneTimeReadAccess().getValue();
        // Wenn Inhalt der ScalarProperty nun ein anderer, bisheriges Layer
        // ersetzen
        if ( listener.size() == 1 && listener.firstElement().getSourceObject() != obj ) {
          listener.firstElement().close();
          listener.clear();
        }
        // Listener erzeugen, wenn noch kein Listener (mehr) fuer den Inhalt der
        // ScalarProperty vorhanden
        if ( listener.isEmpty() ) {
          UpdateListenerProxy l = new UpdateListenerProxy((GTVisualisationTool)visTool, obj, desc, style);
          l.reorganize();
          listener.add( l );
        }
        return;
      }

      // ListenerProxy fuer die Elemente eine ListProperty erzeugen
      if ( this.visObject instanceof ListProperty) {
        ListProperty propObj = (ListProperty)this.visObject;
        ListPropertyReadAccess listAccess = propObj.getReadAccess(this);
        Vector listObjects = new Vector();
        for (int j = 0; j < listAccess.getCount(); j++) {
          Object listObject = listAccess.getValue(j);
          // Darstellbarkeit des Listenelements checken
          if (visTool.isVisualisable(listObject))
            listObjects.add(listObject);
        }

        // Pruefen, ob darstellbare Objekte hinzugekommen ist
        // --> aus 'listObjects' alle Objekte loeschen, fuer die bereits ein
        //     Listener besteht
        for (int i = 0; i < listener.size(); i++)
          while (listObjects.remove(listener.elementAt(i).getSourceObject()));

        // Fuer alle neuen (uebriggebliebenen) Objekte ein neues
        // Listener-Element hinzufuegen
        for (int i = 0; i < listObjects.size(); i++) {
          Object listObject = listObjects.elementAt(i);
          // Der Beschreibung den Listen-Index hinzufuegen
          String layerDesc = desc.trim() + "[" + listAccess.indexOf(listObject) + "]";
          UpdateListenerProxy l = new UpdateListenerProxy((GTVisualisationTool)this.visTool, listObject, layerDesc, style);
          listener.add(l);
          l.reorganize();
        }
        listAccess.release();
        return;
      }

      // von einem Xulu-Objekt werden alle darstellbaren Propertys dargestellt
      if ( this.visObject instanceof Properties ) {
        Properties propObj = (Properties)this.visObject;

        // wenn keine Beschreibung angegeben wurde, wird die
        // Objekt-Beschreibung genommen
        if ( (desc == null || desc.trim().equals("")) && propObj instanceof XuluObject)
          desc = ((XuluObject)propObj).getDescription();
        // alle visualisierbaren Propertys ermitteln
        Vector<ValueProperty> prop = ((GTVisualisationTool)visTool).determineVisualisableProperties(propObj);

        // Pruefen, ob darstellbare Property hinzugekommen ist
        // --> aus 'prop' alle Properies loeschen, fuer die bereits ein
        //     Listener besteht
        for (int i=0; i < listener.size(); i++)
          while ( prop.remove( listener.elementAt(i).getSourceObject() ) );

        // Fuer alle neuen (uebriggebliebenen) Properties ein neues
        // Listener-Element hinzufuegen
        for (int i=0; i<prop.size(); i++) {
          Property p = prop.elementAt(i);
          // Der Beschreibung den Property-Namen hinzufuegen
          String layerDesc = desc.trim() + "." + p.getName().trim();
          UpdateListenerProxy l = new UpdateListenerProxy((GTVisualisationTool)this.visTool, prop.elementAt(i), layerDesc, style);
          listener.add( l );
          l.reorganize();
        }
        return;
      }

      // fuer alle anderen Objekte einen "normalen" Listener und ein Layer
      // in der Map erzeugen
      if ( listener.isEmpty() ) {
        MapLayer layer = ((GTVisualisationTool)visTool).addLayer(visObject,desc,style);
        UpdateListener l = new UpdateListener((GTVisualisationTool)visTool, visObject, layer);
        listener.add( l );
      }
    }


    /**
     * Ruft {@code refresh()} fuer alle Listener des Proxy auf.
     */
    public void refresh() {
      // nur wenn sich die Struktur des Basis-Objekts geaendert hat, wird
      // die Listener-Struktur aktualisiert
      if ( visObjectStructureChanged )
        reorganize();

      for (int i=0; i<listener.size(); i++) {
        VisualisationUpdateListener l = listener.elementAt(i);
        // Listener kann schon nicht mehr existieren
        if (l != null)
          l.refresh();
      }
    }
  }

}
